/*
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

import berkeley.cs.dmc.numericalMethods.Integrator;
import berkeley.cs.dmc.numericalMethods.Derivative;

public class DiveBomberBurnIn implements Derivative {

  private static final double BURN_TIME = 20.0;
  private static final double OMEGA_DOT = DiveBomberSystem.OMEGA / BURN_TIME;

  private double t_start = 0.0;
  private double theta_start = 0.0;

  public final int getDimension() { return 2; }

  public final void derivs( double x, int n, double y[], double dydx[] ) {

    double t = x - t_start;
    double omega = t * OMEGA_DOT;
    double theta = theta_start + 0.5 * OMEGA_DOT * t * t;
    
    double sin_phi = Math.sin(y[0]);
    double cos_phi = Math.cos(y[0]);

    dydx[0] = y[1];
    dydx[1] = ( 9.81 * Math.sin(theta) * sin_phi
	      - omega * omega * ( DiveBomberSystem.RADIUS1
	                        + DiveBomberSystem.RADIUS2 * cos_phi ) * sin_phi
	      ) / DiveBomberSystem.RADIUS2 - DiveBomberSystem.DAMPING * y[1];
  }

  public double burnIn(double[] vars) {

    double theta_diff = Math.random() * 100.0;

    theta_start = -0.5 * OMEGA_DOT * BURN_TIME * BURN_TIME - theta_diff;
    t_start = -theta_diff / DiveBomberSystem.OMEGA - BURN_TIME;

    double offset = Math.IEEEremainder(theta_start, 2.0 * Math.PI);
    if ( offset < 0.0 )
	offset += Math.PI * 2.0;
    vars[0] = offset > Math.PI ? 0.0 : Math.PI;
    vars[1] = Math.random() * 0.01;

    Integrator integ = new Integrator(this);

    integ.ODEIntegrate(vars, t_start, t_start + BURN_TIME, 1.0e-4, 0.5, 1.0e-6);

    return t_start + BURN_TIME;
  }

}


