/*
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
import berkeley.cs.dmc.numericalMethods.Integrator;
import berkeley.cs.dmc.numericalMethods.Derivative;

import berkeley.cs.dmc.system.DynamicsSystem;

public class DiveBomberSystem implements DynamicsSystem, Derivative {

    public static final double OMEGA = Math.PI * 0.25;
    public static final double RADIUS1 = 5.0;
    public static final double RADIUS2 = 1.0;
    public static final double DAMPING = 0.1;

    private static final int techniques[] = { DynamicsSystem.ITERATIVE,
					      DynamicsSystem.FIRST_DERIVATIVE };
    private static final double period_time = 2.0 * Math.PI / OMEGA;
    private static final boolean space_wraps[] = { true, false };

    private double min_bounds[] = { -Math.PI, -10.0 };
    private double max_bounds[] = {  Math.PI,  10.0 };

    private double params[] = new double[2];

    private Integrator 		integ = null;

    public DiveBomberSystem() {
	integ = new Integrator( this );
    }

    public int numVariables() {	return 2; }
    public int numFields() { return 1; }
    public int subsetLength() { return 2; }
    public int evaluationTechnique(int index) {	return techniques[index]; }
    public double period() { return period_time; }
    public boolean isCyclic(int index) { return space_wraps[index]; }

    public double[] getBound(int index) {
	double[] result = new double[2];
	result[0] = min_bounds[index];
	result[1] = max_bounds[index];
	return result;
    }

    public boolean initialCondition( double p[] ) {
	DiveBomberBurnIn starter = new DiveBomberBurnIn();
	double t_start = starter.burnIn(p);

        integ.ODEIntegrate(p, t_start, 0.0, 1.0e-4, 0.5, 1.0e-5 );
        while ( p[0] > Math.PI )
	    p[0] -= 2.0 * Math.PI;
        while ( p[0] <= -Math.PI )
	    p[0] += 2.0 * Math.PI;

	return true;
    }

    public float evaluate(double t_last, double[] p_last,
		 	  double t_now, double[] p_now)
    {
	params[0] = p_last[0];
	params[1] = p_last[1];

	double t = t_last;
	double t_next = t + 5.0;
	if ( t_next > t_now )
	    t_next = t_now;

	while ( t < t_now - 0.0001 )
	{
	    integ.ODEIntegrate(params, t, t_next, 1.0e-4, 0.5, 1.0e-5 );
	    t = t_next;
	    t_next += 5.0;
	    if ( t_next > t_now )
		t_next = t_now;
	}


	p_now[0] = params[0];
	p_now[1] = params[1];
	return 0.1f;
    }

    public void evaluateFields(double t, double params[], float fields[] ) {
	fields[0] = (float)params[0];
    }

    public boolean checkState( double x, double p[], boolean b ) {
	return true;
    }

    public float bufferStep() { return 0.1f; }

    public int bufferSize() { return 1; } // 200; }

    public double[] resolutions() {
        double[] res = new double[2];
        res[0] = Math.PI / 15.0;
        res[1] = 0.3;
        return res;
    }

    public double mixDiff() { return 0.5; }

    public double mixTestMax() { return 0.5; }

    public double maxApproxError() {
	return 0.01;
    }


    public final int getDimension() { return 2; }

    public final void derivs( double x, int n, double y[], double dydx[] ) {
	double theta = ( x - Math.floor(x/period_time)*period_time ) * OMEGA;
    
	double sin_phi = Math.sin(y[0]);
	double cos_phi = Math.cos(y[0]);

	dydx[0] = y[1];
	dydx[1] = ( 9.81 * Math.sin(theta) * sin_phi
                  - OMEGA * OMEGA * ( RADIUS1 + RADIUS2 * cos_phi ) * sin_phi
		  ) / RADIUS2 - DAMPING * y[1];
    }

}



