/*
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
import berkeley.cs.dmc.numericalMethods.Integrator;
import berkeley.cs.dmc.numericalMethods.Derivative;

public class OctopusCarBurnIn implements Derivative {

  private static final double BURN_TIME = 20.0;
  private static final double OMEGA1_DOT = OctopusCarSystem.OMEGA1 / BURN_TIME;
  private static final double OMEGA2_DOT = OctopusCarSystem.OMEGA2 / BURN_TIME;

  private double t_start = 0.0;
  private double theta2_start = 0.0;

  public final int getDimension() { return 2; }

  public final void derivs( double x, int n, double y[], double dydx[] ) {

    double t = x - t_start;
    double omega1 = t * OMEGA1_DOT;
    double omega2 = t * OMEGA2_DOT;
    double theta2 = theta2_start + 0.5 * OMEGA2_DOT * t * t;
    
    double sin_x = Math.sin(y[0]);
    double cos_x = Math.cos(y[0]);

    double r1 = OctopusCarSystem.R1;
    double r2 = OctopusCarSystem.R2;
    double r3 = OctopusCarSystem.R3;

    dydx[0] = y[1];
    dydx[1] = ( - r1 * omega1 * omega1 * Math.sin(theta2 + y[0])
	      - r2 * ( omega1 + omega2 ) * ( omega1 + omega2 ) * Math.sin(y[0])
	      - r3 * ( OMEGA1_DOT + OMEGA2_DOT )
	      - r1 * OMEGA1_DOT * Math.cos(theta2 + y[0])
	      - r2 * ( OMEGA1_DOT + OMEGA2_DOT ) * Math.cos(y[0])
	      ) / r3 - OctopusCarSystem.DAMPING * y[1];
  }

  public double burnIn(double[] vars) {

    double time_diff = Math.random() * 500.0;

    theta2_start = -0.5 * OMEGA2_DOT * BURN_TIME * BURN_TIME
                   - time_diff * OctopusCarSystem.OMEGA2;
    t_start = -time_diff - BURN_TIME;

    vars[0] = Math.random() * Math.PI * 2.0;
    vars[1] = 0.0;

    Integrator integ = new Integrator(this);

    integ.ODEIntegrate(vars, t_start, -time_diff, 1.0e-4, 0.5, 1.0e-6);

    return -time_diff;
  }

}


