/*
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
import berkeley.cs.dmc.numericalMethods.Integrator;
import berkeley.cs.dmc.numericalMethods.Derivative;

import berkeley.cs.dmc.system.DynamicsSystem;

public class OctopusCarSystem implements DynamicsSystem, Derivative {

    public static final double OMEGA1 = (7.3 * Math.PI / 30.0);
    public static final double OMEGA2 = (4.0 * Math.PI / 30.0);
    public static final double R1 = 4.5;
    public static final double R2 = 1.8;
    public static final double R3 = 0.5;
    public static final double DAMPING = 0.1;

    private static final int techniques[] = { DynamicsSystem.ITERATIVE,
					      DynamicsSystem.FIRST_DERIVATIVE };
    private static final double period_time = 15.0;
    private static final boolean space_wraps[] = { true, false };

    private double min_bounds[] = { -Math.PI, -6.0 };
    private double max_bounds[] = {  Math.PI,  6.0 };

    private double params[] = new double[2];

    private Derivative 		deriv = null;
    private Integrator 		integ = null;

    public OctopusCarSystem()
    {
	integ = new Integrator( this );
    }

    public int numVariables()
    {
	return 2;
    }

    public int subsetLength()
    {
	return 2;
    }

    public int numFields()
    {
        return 1;
    }

    public int evaluationTechnique(int index)
    {
	return techniques[index];
    }

    public double period()
    {
	return period_time;
    }

    public boolean isCyclic(int index)
    {
	return space_wraps[index];
    }

    public double[] getBound(int index)
    {
	double[] result = new double[2];
	result[0] = min_bounds[index];
	result[1] = max_bounds[index];
	return result;
    }

    public boolean initialCondition(double[] p)
    {
	OctopusCarBurnIn starter = new OctopusCarBurnIn();
	double t_start = starter.burnIn(p);

        integ.ODEIntegrate(p, t_start, 0.0, 1.0e-4, 0.5, 1.0e-5 );
        while ( p[0] > Math.PI )
	    p[0] -= 2.0 * Math.PI;
        while ( p[0] <= -Math.PI )
	    p[0] += 2.0 * Math.PI;

	return true;
    }

    public float evaluate(double t_last, double[] p_last,
			  double t_now, double[] p_now)
    {
	params[0] = p_last[0];
	params[1] = p_last[1];

	double t = t_last;
	double t_next = t + 5.0;
	if ( t_next > t_now )
	    t_next = t_now;

	while ( t < t_now - 0.0001 )
	{
	    integ.ODEIntegrate(params, t, t_next, 1.0e-4, 0.5, 1.0e-5 );
	    t = t_next;
	    t_next += 5.0;
	    if ( t_next > t_now )
		t_next = t_now;
	}

	p_now[0] = params[0];
	p_now[1] = params[1];

	return 0.1f;
    }

    public void evaluateFields(double t, double[] p, float[] f)
    {
	f[0] = (float)p[0];
    }

    public boolean checkState(double t, double[] p, boolean discard)
    {
	return true;
    }

    public float bufferStep()
    {
	return 0.1f;
    }

    public int bufferSize()
    {
	return 200;
    }

    public double[] resolutions()
    {
	double[]    res = new double[2];

	res[0] = Math.PI / 15.0;
	res[1] = 0.3;

	return res;
    }

    public double mixTestMax()
    {
	return 0.8;
    }

    public double mixDiff()
    {
	return 0.8;
    }

    public double maxApproxError() {
        return 0.01;
    }

    public final int getDimension() { return 2; }

    public final void derivs( double x, int n, double y[], double dydx[] ) {
	double  theta2 = x * OMEGA2;

	dydx[0] = y[1];
	dydx[1] = ( - R1 * OMEGA1 * OMEGA1 * Math.sin(theta2 + y[0])
		  - R2 * ( OMEGA1 + OMEGA2 )
		       * ( OMEGA1 + OMEGA2 ) * Math.sin(y[0]) )
		  / R3 - DAMPING * y[1];
    }
}


