/*
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
import berkeley.cs.dmc.numericalMethods.Integrator;
import berkeley.cs.dmc.numericalMethods.Derivative;

public class PendulumBurnIn implements Derivative {

  private static final double OMEGA_DOT = PendulumSystem.OMEGA_RUN
                                        * PendulumSystem.OMEGA_RUN
					/ ( 32.0 * Math.PI );
  private static final double BURN_TIME = PendulumSystem.OMEGA_RUN / OMEGA_DOT;

  private static final double DAMPING = 0.1;

  private double t_start = 0.0;
  private double theta_start = 0.0;

  public final int getDimension() { return 2; }

  public final void derivs( double x, int n, double y[], double dydx[] ) {

    double t = x - t_start;
    double omega = t * OMEGA_DOT;
    double theta = theta_start + 0.5 * OMEGA_DOT * t * t;
    
    double sin_tx = Math.sin(theta - y[0]);
    double sin_x = Math.sin(y[0]);
    double cos_tx = Math.cos(theta - y[0]);

    dydx[0] = y[1];
    dydx[1] = ( PendulumSystem.R1 * omega * omega * sin_tx - 9.81 * sin_x
              - PendulumSystem.R1 * OMEGA_DOT * OMEGA_DOT * cos_tx )
              / PendulumSystem.R2
	    - DAMPING * ( y[1] - omega );
  }

  public double burnIn(double[] vars) {

    double theta_diff = Math.random() * 100.0;

    theta_start = -0.5 * OMEGA_DOT * BURN_TIME * BURN_TIME - theta_diff;
    t_start = -theta_diff / PendulumSystem.OMEGA_RUN - BURN_TIME;

    vars[0] = 0.0;
    vars[1] = 0.0;

    Integrator integ = new Integrator(this);

    integ.ODEIntegrate(vars, t_start, t_start + BURN_TIME, 1.0e-4, 0.5, 1.0e-6);

    return t_start + BURN_TIME;
  }

}


