/*
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
import berkeley.cs.dmc.numericalMethods.Integrator;
import berkeley.cs.dmc.numericalMethods.Derivative;

import berkeley.cs.dmc.system.*;

public class PendulumSystem implements DynamicsSystem, Derivative {

    private static final double DAMPING = 0.01;

    public static final double OMEGA_RUN = 3.264275;
    public static final double R1 = 1.0;
    public static final double R2 = 3.0;

    private static final int techniques[] = { DynamicsSystem.ITERATIVE,
					      DynamicsSystem.FIRST_DERIVATIVE };
    private static final double period_time = 2.0 * Math.PI / OMEGA_RUN;
    private static final boolean space_wraps[] = { true, false };

    private double min_bounds[] = { -Math.PI, -10.0 };
    private double max_bounds[] = {  Math.PI,  10.0 };

    private double params[] = new double[2];

    private Derivative 		deriv = null;
    private Integrator 		integ = null;

    public PendulumSystem() {
	integ = new Integrator( this );
    }

    public final void derivs( double x, int n, double y[], double dydx[] ) {
	double omega = PendulumSystem.OMEGA_RUN;
	double theta = PendulumSystem.OMEGA_RUN * x;
	
        double sin_tx = Math.sin(theta - y[0]);
	double sin_x = Math.sin(y[0]);

	dydx[0] = y[1];
	dydx[1] = ( PendulumSystem.R1 * omega * omega * sin_tx - 9.81 * sin_x )
		  / PendulumSystem.R2
		- DAMPING * ( y[1] - omega );
    }


    public final int getDimension() {
	return 2;
    }

    public int numVariables() {
	return 2;
    }

    public int numFields() {
	return 1;
    }

    public int subsetLength() {
	return 2;
    }

    public boolean initialCondition(double p[]) {
	PendulumBurnIn starter = new PendulumBurnIn();
	double t_start = starter.burnIn(p);

        integ.ODEIntegrate(p, t_start, 0.0, 1.0e-3, 0.1, 1.0e-6 );
        while ( p[0] > Math.PI )
	    p[0] -= 2.0 * Math.PI;
        while ( p[0] <= -Math.PI )
	    p[0] += 2.0 * Math.PI;
	return true;
    }

    public boolean checkState( double time, double p[], boolean discard ) {
	return true;
    }

    public float bufferStep() {
	return 0.1f;
    }

    public int bufferSize() {
	return 200;
    }

    public int evaluationTechnique(int index)
    {
	return techniques[index];
    }

    public double period()
    {
	return period_time;
    }

    public boolean isCyclic(int index)
    {
	return space_wraps[index];
    }

    public double[] getBound(int index)
    {
	double[] result = new double[2];
	result[0] = min_bounds[index];
	result[1] = max_bounds[index];
	return result;
    }

    public float evaluate(double t_last, double[] p_last,
		 	  double t_now, double[] p_now)
    {
	params[0] = p_last[0];
	params[1] = p_last[1];

	double t = t_last;
	double t_next = t + 1.0;
	if ( t_next > t_now )
	    t_next = t_now;

	while ( t < t_now - 0.0001 )
	{
	    integ.ODEIntegrate(params, t, t_next, 1.0e-4, 0.5, 1.0e-5 );
	    t = t_next;
	    t_next += 1.0;
	    if ( t_next > t_now )
		t_next = t_now;
	}

	p_now[0] = params[0];
	p_now[1] = params[1];
	return 0.1f;
    }

    public void evaluateFields(double t, double[] params, float[] fields) {
	fields[0] = (float)params[0];
    }

    public double mixDiff() {
	return 0.5;
    }

    public double mixTestMax() {
	return 0.5;
    }

    public double[] resolutions() {
	double[]  res = new double[2];
	res[0] = Math.PI / 15.0;
	res[1] = 0.4;
	return res;
    }
	
    public double maxApproxError() {
	return 0.005;
    }
}

