/*
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
import berkeley.cs.dmc.system.DynamicsSystem;
import berkeley.cs.dmc.numericalMethods.Integrator;
import berkeley.cs.dmc.numericalMethods.Derivative;
import berkeley.cs.dmc.bspline.BSpline;
public class RollerSystem implements DynamicsSystem {
  BSpline     pspline;
  BSpline     nspline;
  private Derivative deriv = null;
  private Integrator integ = null;
  private double[] params = new double[2];
  private double[] a = new double[3];
  private double[] v = new double[3];
  private float[] n = new float[3];
  private float[] t = new float[3];
  private float[] pt = new float[3];
  private static final double INIT_VEL = 2.0;
  private double energy = 0.0;

  static final double PERIOD = 16.0;
  private float[][]   t_control_pts = {
    { 21.518963f, 1.0f, 4.427049f},
    { 14.042409f, 1.0413438f, 5.579808f},
    { 14.0492325f, 1.0198106f, -1.4436634f},
    { 7.7339516f, 9.219274f, -0.8862512f},
    { 0.02485007f, 6.30905f, -1.0354645f},
    { -6.8167505f, 4.522277f, -0.6704053f},
    { -14.109677f, 2.068463f, -0.5064248f},
    { -20.632145f, 1.3742733f, 1.5510848f},
    { -19.381033f, 1.0f, 9.507294f},
    { -14.168044f, 1.2612991f, 12.104755f},
    { -5.849577f, 12.569652f, 11.43606f},
    { -1.3165593f, 0.99716485f, 11.393228f},
    { 2.6194885f, 1.1348124f, 11.329943f},
    { 7.791444f, 6.366364f, 11.002439f},
    { 14.794787f, 1.3413737f, 10.4128685f},
    { 21.290281f, 2.9845026f, 10.452518f},
    { 21.518963f, 1.0f, 4.427049f},
    { 14.042409f, 1.0413438f, 5.579808f},
    { 14.0492325f, 1.0198106f, -1.4436634f }
  };
  private float[][]   n_control_pts = {
    { -0.759105f, 0.44796634f, 0.47231948f},
    { 0.72161037f, 0.5824345f, -0.37423065f},
    { -0.39547473f, 0.43725294f, 0.80771875f},
    { 0.0f, 1.0f, 0.0f},
    { 0.07305053f, -0.03351159f, -0.9967651f},
    { 0.16007826f, -0.86539006f, 0.4748421f},
    { 0.24286665f, -0.015218149f, 0.9699403f},
    { 0.0f, 1.0f, 0.0f},
    { 0.0f, 1.0f, 0.0f},
    { 0.0f, 1.0f, 0.0f},
    { 0.0f, 1.0f, 0.0f},
    { 0.0f, 1.0f, 0.0f},
    { 0.0f, 1.0f, 0.0f},
    { 0.0f, 1.0f, 0.0f},
    { 0.0f, 1.0f, 0.0f},
    { -0.89251554f, -0.0041346587f, -0.45099756f},
    { -0.759105f, 0.44796634f, 0.47231948f},
    { 0.72161037f, 0.5824345f, -0.37423065f},
    { -0.39547473f, 0.43725294f, 0.80771875f }
  };
  public float[][] lifters = null;
  public RollerSystem() {
    pspline = new BSpline(3, 19, t_control_pts);
    float u = 9.0f;
    float[] x = new float[3];
    pspline.evaluate(u, x);
    energy = x[1] * 9.81 + 0.5 * INIT_VEL*INIT_VEL;
    nspline = new BSpline(3, 19, n_control_pts);
    deriv = new RollerDerivative(this);
    integ = new Integrator(deriv);
  }
  public int numVariables() { return 2; }
  public int numFields() { return 11; }
  public int subsetLength() { return 2; }
  public int evaluationTechnique(int i) {
    if ( i == 0 )
      return ITERATIVE;
    else
      return FIRST_DERIVATIVE;
  }
  public double period() { return 0.0; }
  public boolean isCyclic(int ind) {
    if ( ind == 0 )
      return true;
    else
      return false;
  }
  public double[] getBound(int ind) {
    double[] res = new double[2];
    if ( ind == 0 ) {
      res[0] = 0.0; res[1] = PERIOD;
    } else {
      res[0] = 0.0; res[1] = 1.0;
    }
    return res;
  }
  public boolean initialCondition(double[] p) {
    p[0] = 9.0;
    float[] dp = new float[3];
    pspline.evaluateDpDu((float)p[0], dp);
    p[1] = Math.sqrt(INIT_VEL*INIT_VEL/(dp[0]*dp[0]+dp[1]*dp[1]+dp[2]*dp[2]));
    return false;
  }
  public float evaluate(double t_last, double[] p_last,
                        double t_now, double[] p_now) {
    params[0] = p_last[0];
    params[1] = p_last[1];
    double  min_step = 1.0e-4;
    if ( t_now - t_last < min_step )
      min_step = t_now - t_last;
    integ.ODEIntegrate(params, t_last, t_now, 1.0e-3, 0.01,min_step);
    float   u = (float)Math.IEEEremainder(params[0], PERIOD);
    if ( u < 0.0f )
       u += PERIOD;
    if ( u >= PERIOD )
       u -= PERIOD;
    float[] x = new float[3];
    pspline.evaluate(u, x);
    float[] dp = new float[3];
    pspline.evaluateDpDu(u, dp);
    float t_dot = dp[0] * dp[0] + dp[1] * dp[1] + dp[2] * dp[2];
    double v_req_sq = 2.0 * ( energy - x[1] * 9.81 );
    if ( v_req_sq < 0.0 )
         params[1] = 0.0;
    else
         params[1] = Math.sqrt(v_req_sq / t_dot);
    p_now[0] = params[0];
    p_now[1] = params[1];
    return 1.0f;
  }
  public void evaluateFields(double time,
                     double[] p_in, float[] f_out) {
    float u = (float)Math.IEEEremainder(p_in[0], PERIOD);
    if ( u < 0.0f )
      u += PERIOD;
    if ( u >= PERIOD )
      u -= PERIOD;
    pspline.evaluate(u, pt);
    f_out[0] = (float)pt[0];
    f_out[1] = (float)pt[1];
    f_out[2] = (float)pt[2];
    pspline.evaluateDpDu(u, t);
    double l = Math.sqrt(t[0] * t[0] + t[1] * t[1] + t[2] * t[2]);
    a[0] = 0.0;
    a[1] = -t[2] / l;
    a[2] = t[1] / l;
    double sin_a = Math.sqrt(a[1] * a[1] + a[2] * a[2]);
    if ( sin_a < 1.0e-5 && sin_a > -1.0e-5 ) {
      f_out[7] = 1.0f;
      f_out[8] = 0.0f;
      f_out[9] = 0.0f;
      f_out[10] = 0.0f;
    } else {
      f_out[7] = 0.0f;
      f_out[8] = (float)(a[1] / sin_a);
      f_out[9] = (float)(a[2] / sin_a);
      if ( t[0] >= 0.0 )
        f_out[10] = (float)Math.asin(sin_a);
      else
        f_out[10] = (float)(Math.PI - Math.asin(sin_a));
    }
    nspline.evaluate(u, n);
    v[0] = n[1] * t[2] - n[2] * t[1];
    v[1] = n[2] * t[0] - n[0] * t[2];
    v[2] = n[0] * t[1] - n[1] * t[0];
    n[0] = (float)(t[1] * v[2] - t[2] * v[1]);
    n[1] = (float)(t[2] * v[0] - t[0] * v[2]);
    n[2] = (float)(t[0] * v[1] - t[1] * v[0]);
    l = Math.sqrt(n[0] * n[0] + n[1] * n[1] + n[2] * n[2]);
    n[0] /= l;
    n[1] /= l;
    n[2] /= l;
    double cos_a = ( 1.0 - Math.cos(f_out[10]) );
    v[0] = a[1];
    v[1] = f_out[8] * f_out[9] * cos_a;
    v[2] = 1.0 - f_out[8] * f_out[8] * cos_a;
    a[0] = v[1] * n[2] - n[1] * v[2];
    a[1] = v[2] * n[0] - n[2] * v[0];
    a[2] = v[0] * n[1] - n[0] * v[1];
    sin_a = Math.sqrt(a[0] * a[0] + a[1] * a[1] + a[2] * a[2]);
    if ( sin_a < 1.0e-5 && sin_a > -1.0e-5 ) {
      f_out[3] = 1.0f;
      f_out[4] = 0.0f;
      f_out[5] = 0.0f;
      f_out[6] = 0.0f;
    } else {
      f_out[3] = (float)(a[0] / sin_a);
      f_out[4] = (float)(a[1] / sin_a);
      f_out[5] = (float)(a[2] / sin_a);
      if ( v[0] * n[0] + v[1] * n[1] + v[2] * n[2] >= 0.0 )
        f_out[6] = (float)Math.asin(sin_a);
      else
        f_out[6] = (float)(Math.PI - Math.asin(sin_a));
    }
  }
  public boolean checkState(double t, double[] p, boolean discard) {
    return true;
  }
  public float bufferStep() { return 0.1f; }
  public int bufferSize() { return 100; }
  public double[] resolutions() {
    double[] res = new double[2];
    res[0] = 0.2;
    res[1] = 0.2;
    return res;
  }
  public double mixTestMax() { return 0.5; }
  public double mixDiff() { return 1.0; }
  public double maxApproxError() { return 0.001; }
}


