/*
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

import berkeley.cs.dmc.numericalMethods.Integrator;
import berkeley.cs.dmc.numericalMethods.Derivative;

public class TiltCarBurnIn implements Derivative {

  private static final double BURN_TIME = 20.0;
  private static final double OMEGA_DOT = TiltCarSystem.OMEGA_RUN / BURN_TIME;

  private double t_start = 0.0;
  private double theta_start = 0.0;

  public final int getDimension() { return 2; }

  public final void derivs( double x, int n, double y[], double dydx[] ) {

    double t = x - t_start;
    double omega = t * OMEGA_DOT;
    double theta = 3.0 * ( theta_start + 0.5 * OMEGA_DOT * t * t );
    double alpha = TiltCarSystem.ALPHA0
                 - TiltCarSystem.ALPHA1 * Math.cos(theta);
    double beta = 3.0 * TiltCarSystem.ALPHA1 * Math.sin(theta);
    
    double sin_x = Math.sin(y[0]);
    double cos_x = Math.cos(y[0]);

    double r1 = TiltCarSystem.R1;
    double r2 = TiltCarSystem.R2;

    dydx[0] = y[1];
    dydx[1] = ( 9.81 * r2 * ( alpha * sin_x - beta * cos_x )
              - r1 * r2 * omega * omega * sin_x
	      - r1 * r2 * cos_x * OMEGA_DOT
	      - TiltCarSystem.RHO_M * y[1] )
              / ( r2 * r2 );
  }

  public double burnIn(double[] vars) {

    double theta_diff = Math.random() * 100.0;

    theta_start = -0.5 * OMEGA_DOT * BURN_TIME * BURN_TIME - theta_diff;
    t_start = -theta_diff / TiltCarSystem.OMEGA_RUN - BURN_TIME;

    double alpha = TiltCarSystem.ALPHA0
                 - TiltCarSystem.ALPHA1 * Math.cos(3.0 * theta_start);
    double beta = 3.0 * TiltCarSystem.ALPHA1 * Math.sin(3.0 * theta_start);
    double temp = Math.sqrt(alpha * alpha + beta * beta);
    vars[0] = Math.atan2(-beta/temp, -alpha/temp);
    vars[1] = 0.0;

    Integrator integ = new Integrator(this);

    integ.ODEIntegrate(vars, t_start, t_start + BURN_TIME, 1.0e-4, 0.5, 1.0e-6);

    return t_start + BURN_TIME;
  }

}


