/*
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
import vrml.*;
import vrml.field.*;
import vrml.node.*;
import berkeley.cs.dmc.system.DynamicsVariable;

public class Pendulum2 extends Script {
  double time = 0.0;
  double last_time = 0.0;
  boolean initialized = false;
  double[] old_params = new double[2];
  double[] seen_params = new double[2];
  double[] new_params = new double[2];
  private Pendulum2Variable phi = null;
  private MFFloat phi_buffer = null;
  private SFFloat phi_lastTime = null;
  private SFFloat phi_lastSeen = null;
  private float[] phi_elements = new float[1];
  private boolean phi_visible = true;
  private Pendulum2Theta theta = null;
  private MFFloat theta_buffer = null;
  private SFFloat theta_lastTime = null;
  private SFFloat theta_lastSeen = null;
  private float[] theta_elements = new float[1];
  private boolean theta_visible = true;
  private SFRotation Beam1Rot_event = null;
  private SFRotation ChairRot_event = null;
  private SFRotation Beam2Rot_event = null;
  private SFRotation ChairUnRoll_event = null;
  public void initialize() {
    if ( initialized ) return;
    phi = new Pendulum2Variable();
    phi_buffer = (MFFloat)getField("phi_buffer");
    phi_lastTime = (SFFloat)getField("phi_lastTime");
    phi_lastSeen = (SFFloat)getField("phi_lastSeen");
    theta = new Pendulum2Theta();
    theta_buffer = (MFFloat)getField("theta_buffer");
    theta_lastTime = (SFFloat)getField("theta_lastTime");
    theta_lastSeen = (SFFloat)getField("theta_lastSeen");
    Beam1Rot_event = (SFRotation)getEventOut("Beam1Rot");
    ChairRot_event = (SFRotation)getEventOut("ChairRot");
    Beam2Rot_event = (SFRotation)getEventOut("Beam2Rot");
    ChairUnRoll_event = (SFRotation)getEventOut("ChairUnRoll");
    initialized = true;
  }
  public void processEvent( Event e ) {
    String name = e.getName();
    if ( "currentTime".equals( name ) ) {
      time=((ConstSFFloat)e.getValue()).getValue();
    } else if ( "Pendulum2Bound".equals( name ) ) {
      theta_visible = (boolean)((ConstSFBool)e.getValue()).getValue();
      phi_visible = (boolean)((ConstSFBool)e.getValue()).getValue();
      ((SFBool)getEventOut("visible")).setValue((boolean)((ConstSFBool)e.getValue()).getValue());
    }
  }
  private void timeStep(MFFloat buffer, SFFloat lastTime,
                        SFFloat lastSeen, double bufStep, double phase,
                        DynamicsVariable var, double time, float[] fields,
                        int num_p, int num_b) {
    time += phase;

    double last_time = (double)lastTime.getValue();
    double last_seen = (double)lastSeen.getValue();

    if ( num_b == 1 )
    {
      for ( int i = 0 ; i < num_p ; i++ )
        old_params[i] = (double)buffer.get1Value(i);
      var.newState(last_time, old_params,
                   last_time, old_params,
                   time, new_params);
      for ( int i = 0 ; i < num_p ; i++ )
        buffer.set1Value(i, (float)new_params[i]);
      lastTime.setValue((float)time);
      var.evaluateFields(time, new_params, fields);
      return;
    }

    int    low_index = (int)Math.floor(time / bufStep);
    double low_time = low_index * bufStep;
    int    high_index = ( low_index + 1 ) % num_b;
    int    last_index = (int)Math.floor((last_time + 0.001) / bufStep) % num_b;
    int    seen_index = (int)Math.floor((last_seen + 0.001) / bufStep) % num_b;

    float work_done = 0.0f;

    int base = 0;

    low_index = low_index % num_b;

    base = last_index * num_p;
    for ( int i = 0 ; i < num_p ; i++ )
      old_params[i] = (double)buffer.get1Value(base + i);
    base = seen_index * num_p;
    for ( int i = 0 ; i < num_p ; i++ )
      seen_params[i] = (double)buffer.get1Value(base + i);

    if ( low_time > last_time + 0.00001 )
    {
      work_done = var.newState(last_time, old_params,
                               last_seen, seen_params,
                               low_time, new_params);
      for ( int i = 0 ; i < num_p ; i++ )
        old_params[i] = new_params[i];
      base = low_index * num_p;
      for ( int i = 0 ; i < num_p ; i++ )
        buffer.set1Value(base + i, (float)new_params[i]);
      last_time = low_time;
      last_index = low_index;
    }
    if ( low_time + bufStep > last_time + 0.00001 )
    {
      work_done += var.newState(low_time, old_params,
                                last_seen, seen_params,
                                low_time + bufStep, new_params);
      for ( int i = 0 ; i < num_p ; i++ )
        old_params[i] = new_params[i];
      base = high_index * num_p;
      for ( int i = 0 ; i < num_p ; i++ )
        buffer.set1Value(base + i, (float)new_params[i]);
      last_time = low_time + bufStep;
      last_index = high_index;
    }


    // Guaranteed to have low and high time buffer positions filled.
    int slots_used = last_index - low_index + 1;
    if ( slots_used < 1 )
      slots_used += num_b;
    base = 0;
    while (  work_done < 1.0f && slots_used < num_b )
    {
      work_done += var.newState(last_time, old_params,
                                last_seen, seen_params,
                                last_time + bufStep, new_params);
      last_index = ( last_index + 1 ) % num_b;
      base = last_index * num_p;
      for ( int i = 0 ; i < num_p ; i++ )
        old_params[i] = new_params[i];
      for ( int i = 0 ; i < num_p ; i++ )
        buffer.set1Value(base + i, (float)new_params[i]);
      last_time += bufStep;
      last_time = Math.floor(last_time / bufStep + 0.1) * bufStep;
      slots_used++;
    }

    lastTime.setValue((float)last_time);
    lastSeen.setValue((float)(low_time + bufStep));

    double fraction = ( time - low_time ) / bufStep;
    int base1 = high_index * num_p;
    base = low_index * num_p;
    for ( int i = 0 ; i < num_p ; i++ )
      new_params[i] = fraction * buffer.get1Value(base1 + i)
                    + ( 1.0 - fraction ) * buffer.get1Value(base + i);
    var.evaluateFields(time, new_params, fields);
  }
  public void eventsProcessed() {
    if ( ! initialized ) return;
    if ( time <= last_time )
        return;
    last_time = time;
    if ( phi_visible )
      timeStep(phi_buffer, phi_lastTime, phi_lastSeen, 0.1f, 0.0f, phi, last_time, phi_elements, 2, 200);
    if ( theta_visible )
      timeStep(theta_buffer, theta_lastTime, theta_lastSeen, 0.01f, 0.0f, theta, last_time, theta_elements, 1, 1);
    if ( theta_visible )
      Beam1Rot_event.setValue(0.0f, 0.0f, 1.0f, theta_elements[0]);
    if ( phi_visible )
      ChairRot_event.setValue(0.0f, 0.0f, 1.0f, phi_elements[0]);
    if ( theta_visible )
      Beam2Rot_event.setValue(0.0f, 0.0f, 1.0f, theta_elements[0]);
    if ( theta_visible )
      ChairUnRoll_event.setValue(0.0f, 0.0f, -1.0f, theta_elements[0]);
  }
}

