/*
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
import berkeley.cs.dmc.system.DynamicsSystem;
import berkeley.cs.dmc.numericalMethods.Derivative;
import berkeley.cs.dmc.numericalMethods.Integrator;


public class SimplePendulumSystem implements DynamicsSystem, Derivative {

    public static final double LENGTH = 5.0;

    private static final double INIT_ANGLE = 0.75 * Math.PI;
    private static final double ENERGY =
			       9.81 * LENGTH * ( 1.0 - Math.cos(INIT_ANGLE) );

    private double[] params = new double[2];

    Integrator integ = null;

    public SimplePendulumSystem() {
	integ = new Integrator(this);
    }

    public int getDimension() { return 2; }

    public void derivs(double x, int n, double[] y, double[] dydx) {
	dydx[0] = y[1];
	dydx[1] = - 9.81 * Math.sin(y[0]) / LENGTH;
    }

    public int numVariables() { return 2; }

    public int numFields() { return 1; }

    public int subsetLength() { return 2; }

    public int evaluationTechnique(int ind) {
	if ( ind == 0 )
	    return ITERATIVE;
	else
	    return FIRST_DERIVATIVE;
    }

    public double period() { return 0.0; }

    public boolean isCyclic(int ind) { return false; }

    public double[] getBound(int ind) { return new double[] { -1.0, 1.0 }; }

    public boolean initialCondition(double[] p) {
	p[0] = INIT_ANGLE;
	p[1] = 0.0;

	return false;
    }

    public float evaluate(double t_last, double[] p_last,
			  double t_now, double[] p_now) {
	params[0] = p_last[0];
	params[1] = p_last[1];

	double min_step = 1.0e-4;
	if ( min_step > t_now - t_last )
	    min_step = t_now - t_last;

	integ.ODEIntegrate(params, t_last, t_now, 1.0e-3, 0.5, min_step);

	boolean pos = params[1] >= 0.0;
	double temp = 2.0 * ( ENERGY + 9.81*LENGTH*( Math.cos(params[0]) - 1 ));
	if ( temp < 0.0 )
	    params[1] = 0.0;
	else
	    params[1] = Math.sqrt(temp) / LENGTH;
	if ( ! pos )
	    params[1] = -params[1];

	p_now[0] = params[0];
	p_now[1] = params[1];

	return 0.05f;
    }

    public void evaluateFields(double t, double[] params, float[] fields) {
	fields[0] = (float)params[0];
    }


    public boolean checkState(double t, double[] p, boolean discard) {
	return true;
    }

    public float bufferStep() { return 0.05f; }

    public int bufferSize() { return 100; }

    public double[] resolutions() { return new double[] { 0.1, 0.1 }; }

    public double mixTestMax() { return 0.5; }

    public double mixDiff() { return 0.5; }

    public double maxApproxError() { return 0.0001; }

}


