/*
 * BSpline.java - uniform cubic b-splines
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package berkeley.cs.dmc.bspline;

/** This class provides functions for uniform cubic b-splines.
 *
 * @author Stephen Chenney
 * @verison 1.0
 */
public class BSpline {
    
    private int		dimension = 0;
    private int		num_controls = 0;
    private float[][]	control_pts = null;
    private float[]     g_basis = null;
    private float[]     vel = null;

    /**
     * The no argument constructor generates a b-spline with
     * no control points.
     */
    public BSpline() { }

    /**
     * This constructor takes the dimensionof the output vector, the
     * number of control points and the control points themselves.
     *
     * @param dim The dimension of the output vector.
     * @param num The number of control points.
     * @param The control points. A num by dim array.
     */
    public BSpline(int dim, int num, float[][] cont)
    {
	dimension = dim;
	num_controls = num;
	control_pts = cont;
	g_basis = new float[4];
	vel = new float[dimension];
    }

    /**
     * Change the control points.
     *
     * @param num The new number of controls.
     * @param cont The new control points.
     */
    public void newControls(int num, float[][] cont)
    {
	num_controls = num;
	control_pts = cont;
    }


    private void expandBasis(float u, float[] basis)
    {
	float   u2, u3;

	u2 = u * u;
	u3 = u * u2;

	basis[0] = ( -u3 + 3.0f * u2 - 3.0f * u + 1.0f ) / 6.0f;
	basis[1] = ( 3.0f * u3 - 6.0f * u2 + 4.0f ) / 6.0f;
	basis[2] = ( -3.0f * u3 + 3.0f * u2 + 3.0f * u + 1.0f ) / 6.0f;
	basis[3] = ( u3 ) / 6.0f;
    }

    private void expandPdotBasis(float u, float u_dot, float[] basis)
    {
	float   u2, u3;

	u2 = 2.0f * u * u_dot;
	u3 = 3.0f * u * u * u_dot;

	basis[0] = ( -u3 + 3.0f * u2 - 3.0f * u_dot ) / 6.0f;
	basis[1] = ( 3.0f * u3 - 6.0f * u2 ) / 6.0f;
	basis[2] = ( -3.0f * u3 + 3.0f * u2 + 3.0f * u_dot ) / 6.0f;
	basis[3] = ( u3 ) / 6.0f;
    }

    private void expandDpDuBasis(float u, float[] basis)
    {
	float   u2, u3;

	u2 = 2.0f * u;
	u3 = 3.0f * u * u;

	basis[0] = ( -u3 + 3.0f * u2 - 3.0f ) / 6.0f;
	basis[1] = ( 3.0f * u3 - 6.0f * u2 ) / 6.0f;
	basis[2] = ( -3.0f * u3 + 3.0f * u2 + 3.0f ) / 6.0f;
	basis[3] = ( u3 ) / 6.0f;
    }


    private void expandDpdotDuBasis(float u, float u_dot, float[] basis)
    {
	float   u2, u3;

	u2 = 2.0f * u_dot;
	u3 = 6.0f * u * u_dot;

	basis[0] = ( -u3 + 3.0f * u2 ) / 6.0f;
	basis[1] = ( 3.0f * u3 - 6.0f * u2 ) / 6.0f;
	basis[2] = ( -3.0f * u3 + 3.0f * u2 ) / 6.0f;
	basis[3] = ( u3 ) / 6.0f;
    }


    /**
     * Evaluate the spline at value u. u must be greater than or equal to 0
     * and less than the nummber of controls - 3.
     *
     * @param u The parameter value to evaluate at.
     * @param pt Return space for the output vector.
     *
     * @exception ArithmeticException Thrown if the parameter value is out
     *                                of range.
     */
    public void evaluate(float u, float[] pt)
	throws ArithmeticException
    {
	float   u_local = u - (float)Math.floor((double)u);
	int     first_control = (int)Math.floor((double)u);

	expandBasis(u_local, g_basis);

	if ( first_control >= num_controls - 3 || first_control < 0 )
	    throw new ArithmeticException("control " + first_control
                                        + ", num_controls " + num_controls
					+ ", u = " + u);

	for ( int i = 0 ; i < dimension ; i++ )
	    pt[i] = g_basis[0] * control_pts[first_control][i]
	          + g_basis[1] * control_pts[first_control+1][i]
	          + g_basis[2] * control_pts[first_control+2][i]
	          + g_basis[3] * control_pts[first_control+3][i];
    }

    /**
     * Evaluate the magnitude of the derivative of the output vector
     * with respect to some variable, t, given the parameter and derivative
     * of the parameter with respect to t. The parameter value must be
     * greater than or equal to 0
     * and less than the nummber of controls - 3.
     *
     * @param u The parameter value to evaluate at.
     * @param u_dot The derivative of the parameter value wrt t.
     * 
     * @return The magnitude of the derivative of the output vector wrt t.
     *
     * @exception ArithmeticException Thrown if the parameter value is out
     *                                of range.
     */
    public float evaluateSpeed(float u, float u_dot)
	throws ArithmeticException
    {
	float   res = 0.0f;
	float   u_local = u - (float)Math.floor((double)u);
	int     first_control = (int)Math.floor((double)u);

	expandPdotBasis(u_local, u_dot, g_basis);

	if ( first_control >= num_controls - 3 )
	    throw new ArithmeticException("control " + first_control
                                        + ", num_controls " + num_controls
					+ ", u = " + u);

	for ( int i = 0 ; i < dimension ; i++ )
	{
	    vel[i] = g_basis[0] * control_pts[first_control][i]
	           + g_basis[1] * control_pts[first_control+1][i]
	           + g_basis[2] * control_pts[first_control+2][i]
	           + g_basis[3] * control_pts[first_control+3][i];
	    res += vel[i] * vel[i];
	}
	
	return res;
    }


    /**
     * Evaluate the derivative of the output vector with respect to the
     * parameter at the parameter value u. u must be greater than or equal to 0
     * and less than the nummber of controls - 3.
     *
     * @param u The parameter value to evaluate at.
     * @param pt Return space for the derivative vector.
     *
     * @exception ArithmeticException Thrown if the parameter value is out
     *                                of range.
     */
    public void evaluateDpDu(float u, float[] pt)
	throws ArithmeticException
    {
	float   u_local = u - (float)Math.floor((double)u);
	int     first_control = (int)Math.floor((double)u);

	expandDpDuBasis(u_local, g_basis);

	if ( first_control >= num_controls - 3 )
	    throw new ArithmeticException("control " + first_control
                                        + ", num_controls " + num_controls
					+ ", u = " + u);

	for ( int i = 0 ; i < dimension ; i++ )
	    pt[i] = g_basis[0] * control_pts[first_control][i]
	          + g_basis[1] * control_pts[first_control+1][i]
	          + g_basis[2] * control_pts[first_control+2][i]
	          + g_basis[3] * control_pts[first_control+3][i];
    }


    /**
     * Evaluate the derivative of Dpdot with respect to the parameter.
     * Dpdot is the derivative of the output vector with respect to
     * some other variable t. The derivative is evaluated for the parameter
     * value u and parameter derivative u_dot with
     * respect to t.
     *
     * @param u The parameter value to evaluate at.
     * @param u_dot The derivative of the parameter value wrt t.
     * @param pt Return space for the derivative vector.
     *
     * @exception ArithmeticException Thrown if the parameter value is out
     *                                of range.
     */
    public void evaluateDpdotDu(float u, float u_dot, float[] pt)
	throws ArithmeticException
    {
	float   u_local = u - (float)Math.floor((double)u);
	int     first_control = (int)Math.floor((double)u);

	expandDpdotDuBasis(u_local, u_dot, g_basis);

	if ( first_control >= num_controls - 3 )
	    throw new ArithmeticException("control " + first_control);

	for ( int i = 0 ; i < dimension ; i++ )
	    pt[i] = g_basis[0] * control_pts[first_control][i]
	          + g_basis[1] * control_pts[first_control+1][i]
	          + g_basis[2] * control_pts[first_control+2][i]
	          + g_basis[3] * control_pts[first_control+3][i];
    }


    /**
     * Evaluate a single output field of the spline at value u.
     * u must be greater than or equal to 0
     * and less than the nummber of controls - 3.
     *
     * @param i The index of the output value to evaluate.
     * @param u The parameter value to evaluate at.
     *
     * @return The ith output value for parameter value u.
     *
     * @exception ArithmeticException Thrown if the parameter value is out
     *                                of range.
     */
    public float evaluateOne(int i, float u)
    {
	float   u_local = u - (float)Math.floor((double)u);
	int     first_control = (int)Math.floor((double)u);

	expandBasis(u_local, g_basis);

	if ( first_control >= num_controls - 3 )
	    throw new ArithmeticException("control " + first_control);

	return g_basis[0] * control_pts[first_control][i]
	     + g_basis[1] * control_pts[first_control+1][i]
	     + g_basis[2] * control_pts[first_control+2][i]
	     + g_basis[3] * control_pts[first_control+3][i];
    }
}

