/*
 * Bounds.java - bounding code for the DMC compiler
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.compiler;

import berkeley.cs.dmc.system.DynamicsSystem;

class Bounds {
    private static final int BOUND_SAMPLES = 10000;
    private static final int BOUND_CHANGE_SAMPLES = 10000;
    private static final double BOUND_MIN_STEP = 0.1;
    private static final double BOUND_MAX_STEP = 0.5;
    private static final double BOUND_EPSILON = 1.0e-10;

    private static final Object lock = new Object();

    static DynamicsSystem system;
    static double in[];
    static double out[];
    static double min[], max[];
    static double burnin[];
    static int numVars;
    static int subLength;

    static void boundSystem( DynamicsWrapper wrapper ) {
	synchronized( lock ) {
	    system = wrapper.system;
	    numVars = system.numVariables();
	    subLength = system.subsetLength();
	    min = wrapper.min_bounds = new double[numVars];
	    max = wrapper.max_bounds = new double[numVars];
	    in  = new double[numVars];
	    out = new double[numVars];

	    burnin = new double[numVars];
	    system.initialCondition(burnin);

	    for( int i=0 ; i<subLength ; i++ ) {
		if( system.isCyclic(i) ) {
		    min[i] = system.getBound(i)[0];
		    max[i] = system.getBound(i)[1];
		} else {
		    boundVariable(i);
		}
		System.out.println( "RANGE["+i+"] = "+min[i]+" to "+max[i]);
	    }

	    in  = null;
	    out = null;
	    min = null;
	    max = null;
	}
    }

    static double[][] boundDelta( DynamicsWrapper wrapper, double time ) {
	DynamicsSystem system = wrapper.system;
	int numVars = system.numVariables();
	double bounds[][] = new double[subLength][2];
	double change;
	double in[] = new double[numVars];
	double out[] = new double[numVars];
	double tmp[];
	double mins[] = wrapper.min_bounds;
	double ranges[] = new double[subLength];
	boolean cyclic[] = new boolean[subLength];
	int j, i;
        double t = 0.0;

	system.initialCondition(in);

	for( j=0 ; j<subLength ; j++ ) {
	    bounds[j][0] = 1.0e+100;
	    bounds[j][1] = -1.0e+100;
	    cyclic[j] = system.isCyclic(j);
	    ranges[j] = wrapper.max_bounds[j] - wrapper.min_bounds[j];
	}

	for( j=0 ; j<BOUND_CHANGE_SAMPLES ; j++ ) {
	    system.evaluate( t, in, t + time, out );
	    for( i=0 ; i<numVars ; i+=subLength ) {
		for ( int k = 0 ; k < subLength ; k++ )
		{
		    change = out[i+k] - in[i+k];
		    if( change < bounds[k][0] ) bounds[k][0] = change;
		    if( change > bounds[k][1] ) bounds[k][1] = change;
		    if( cyclic[k] )
			out[i+k] = Math.IEEEremainder( out[i+k], ranges[k] );
		}
	    }
	    tmp = in; in = out; out = tmp;
            t += time;
	}

	System.out.println( "DELTA bounds for dt="+time );
	for( i=0 ; i<subLength ; i++ )
	    System.out.println( "  ["+i+"] : " + bounds[i][0] + " to "
			      + bounds[i][1] );

	return bounds;
    }

    static void boundVariable( int i ) {
	int numSteps = BOUND_SAMPLES-1;
	double tlast = 0.0, tnow;
	double tmin, tmax, tmid;
	double minside[] = new double[numVars];
	double maxside[] = new double[numVars];
	double midside[] = new double[numVars];
	int    numSys = (int)(numVars / subLength);
	double dx[] = new double[numSys];
	double lastdx[] = new double[numSys];

	for( int j=0 ; j<numVars ; j++ )
	    in[j] = burnin[j];

	min[i] = 1.0e+100;
	max[i] = -1.0e+100;

	tnow = tlast + BOUND_MIN_STEP;
	system.evaluate( tlast, in, tnow, out );
	for ( int k = 0 ; k < numVars ; k += subLength )
	    lastdx[k%subLength] = out[k+i] - in[k+i];
	tlast = tnow;

	for( int j=0 ; j<numVars ; j++ )
	    in[j] = out[j];

	while( numSteps-- > 0 ) {
	    tnow = tlast + BOUND_MIN_STEP +
		Math.random() * (BOUND_MAX_STEP - BOUND_MIN_STEP);

	    system.evaluate( tlast, in, tnow, out );

	    for ( int j = 0, k = 0 ; j < numVars ; j += subLength, k++ )
	    {
		dx[k] = out[i+j] - in[i+j];
		
		if( ( lastdx[k] > 0 && dx[k] < 0)
		    || ( in[j+i] == min[j+i] && out[j+i] >= min[j+i]) ) {
		    // find a local min
		    double lmin = Math.min( in[j+i], out[j+i] );

		    tmin = tlast;
		    tmax = tnow;
		    do {
			tmid = (tmin + tmax) / 2.0;
			system.evaluate( tlast, in, (tmin+tmid)/2.0, minside );
			system.evaluate( tlast, in, (tmax+tmid)/2.0, maxside );
			system.evaluate( tlast, in, tmid, midside );
			if( minside[j+i] < lmin ) {
			    lmin = minside[j+i];
			    tmax = tmid;
			} else if( maxside[j+i] < lmin ) {
			    lmin = maxside[j+i];
			    tmin = tmid;
			} else if( midside[j+i] < lmin ) {
			    lmin = midside[j+i];
			    tmin = (tmin+tmid)/2.0;
			    tmax = (tmax+tmid)/2.0;
			} else {
			    // System.out.println( "BREAK!" );
			    break;
			}
		    } while( (tmax - tmin) > BOUND_EPSILON );
		    // System.out.println( "new ranges (dt="+(float)(tmax-tmin)+"):" );

		    if( lmin < min[i] ) {
			min[i] = lmin;
			numSteps = BOUND_SAMPLES; // start again
			// System.out.println( "new min = " + lmin );
		    }
		} else if( (lastdx[k] < 0 && dx[k] > 0)
			|| (in[j+i] == max[j+i] && out[j+i] <= max[j+i]) ) {
		    // find a local max
		    double lmax = Math.max( in[j+i], out[j+i] );

		    tmin = tlast;
		    tmax = tnow;
		    do {
			tmid = (tmin + tmax) / 2.0;
			system.evaluate( tlast, in, (tmin+tmid)/2.0, minside );
			system.evaluate( tlast, in, (tmax+tmid)/2.0, maxside );
			system.evaluate( tlast, in, tmid, midside );
			if( minside[j+i] > lmax ) {
			    lmax = minside[j+i];
			    tmax = tmid;
			} else if( maxside[j+i] > lmax ) {
			    lmax = maxside[j+i];
			    tmin = tmid;
			} else if( midside[j+i] > lmax ) {
			    lmax = midside[j+i];
			    tmin = (tmin+tmid)/2.0;
			    tmax = (tmax+tmid)/2.0;
			} else {
			    // System.out.println( "BREAK!" );
			    break;
			}
		    } while( (tmax - tmin) > BOUND_EPSILON );
		    // System.out.println( "new ranges (dt="+(float)(tmax-tmin)+"):" );

		    if( lmax > max[i] ) {
			max[i] = lmax;
			numSteps = BOUND_SAMPLES; // start again
			// System.out.println( "new max = " + lmax );
		    }
		} else {
		    if( out[j+i] > max[i] ) {
			max[i] = out[j+i];
			numSteps = BOUND_SAMPLES;
			// System.out.println( "MAX!" );
		    }
		    if( out[j+i] < min[i] ) {
			min[i] = out[j+i];
			numSteps = BOUND_SAMPLES;
			// System.out.println( "MIN!" );
		    }
		}
	    }

	    for( int j=0 ; j<numVars ; j++ )
		in[j] = out[j];

	    lastdx = dx;
	    tlast = tnow;
	}
    }
}
