/*
 * CodeGeneration.java - Code generation part of the DMC compiler
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.compiler;

import java.io.PrintWriter;
import java.io.FileWriter;
import berkeley.cs.dmc.probtree.ProbabilityTree;

class CodeGeneration {
    private static final String[] templates = { "_super_name_",
						"_class_name_",
						"_num_variables_",
						"_period_",
						"_sample_thresh_",
						"_approx_thresh_",
						"_sub_size_",
						"_buffer_step_" };

    private static final String[] first_template = {
"// This file is automatically generated",
"import berkeley.cs.dmc.system.DynamicsVariable;",
"import java.util.Random;",
"public class _class_name_ extends _super_name_ implements DynamicsVariable {",
"    private static double period = _period_;",
"    private static final double buffer = _buffer_step_;",
"",
"    public _class_name_() {",
"        super();",
"    }",
""
    };

    private static final String[] approx_template = {
"    public void doApprox(double last_time, double[] last_p,",
"                         double time, double[] p) {",
"        // Integrate to a period boundary.",
"        double sub_s_vals[] = new double[_sub_size_];",
"        double sub_e_vals[] = new double[_sub_size_];",
"        double start_vals[] = new double[_num_variables_];",
"        double start_time = Math.ceil(last_time / period) * period;",
"        super.evaluate(last_time, last_p, start_time, start_vals);",
"",
"        // Set next time to the end of the approximation period.",
"        double end_vals[] = new double[_num_variables_];",
"        double end_time = Math.floor(time / period) * period;",
"        int num_periods = (int)((end_time - start_time) / period + 0.1);",
"        for ( int i = 0 ; i < _num_variables_ ; i += _sub_size_ ) {",
"            for ( int j = 0 ; j < _sub_size_ ; j++ )",
"                sub_s_vals[j] = start_vals[j+i];",
"            approxValues(sub_s_vals, sub_e_vals, num_periods);",
"            for ( int j = 0 ; j < _sub_size_ ; j++ )",
"                end_vals[j + i] = sub_s_vals[j];",
"        }",
"        super.checkState(end_time, end_vals, false);",
"",
"        super.evaluate(end_time, end_vals, time, p);",
"    }",
""
    };

    private static final String[] second_template = {
"    private void doSample(double time, double[] p) {",
"        boolean good_sample = false;",
"        int num_sampled = 0;",
"        double sub_vals[] = new double[_sub_size_];",
"        double sample_vals[] = new double[_num_variables_];",
"        double sample_time;",
"        // Just sample values, then figure out what the time is.",
"        sample_time = Math.floor(time / period) * period;",
"        while ( ! good_sample ) {",
"            for ( int i = 0 ; i < _num_variables_ ; i +=  _sub_size_ ) {",
"                sampleValues(sub_vals);",
"                for ( int j = 0 ; j < _sub_size_ ; j++ )",
"                    sample_vals[j + i] = sub_vals[j];",
"            }",
"            num_sampled++;",
"            if ( num_sampled > 100 ) {",
"              super.checkState(sample_time, sample_vals, false);",
"              good_sample = true;",
"            }",
"            else",
"              good_sample = super.checkState(sample_time, sample_vals, true);",
"        };",
"        // Bring it to required time.",
"        super.evaluate(sample_time, sample_vals, time, p);",
"    }",
"",
"    public float newState(double last_time, double[] last_p,",
"                          double seen_time, double[] seen_p,",
"                          double new_time, double[] new_p) {",
"	float res = 1.0f;",
"	if ( new_time - last_time < buffer + 0.00001 )",
"	  res = super.evaluate(last_time, last_p, new_time, new_p);",
"	else",
"	{",
"	  if ( new_time - seen_time >= _sample_thresh_ )",
"	      doSample(new_time, new_p);"
    };

    private static final String[] approx2_template = {
"        else if ( new_time - seen_time >= _approx_thresh_ )",
"            doApprox(last_time, last_p, new_time, new_p);"
    };

    private static final String[] final_template = {
"        else",
"            res = super.evaluate(last_time, last_p, new_time, new_p);",
"       }",
"       return res;",
"    }",
"}"
    };

    DynamicsWrapper system;

    private static String substitute(String basis, String old_sub,
				     String new_sub)
    {
	int old_index = basis.indexOf(old_sub);

	if ( old_index == -1 )
	    return basis;

	return basis.substring(0, old_index) + new_sub
	     + basis.substring(old_index + old_sub.length(), basis.length());
    }

    private static String substituteMulti(String basis, String[] olds,
                                          String[] news)
    {
	if ( olds.length != news.length )
	    return null;

	String result = basis;
	for ( int i = 0 ; i < olds.length ; i++ )
	    result = substitute(result, olds[i], news[i]);
	return result;
    }

    static void generate(DynamicsWrapper system)
    {
	String[] new_names = new String[templates.length];
	String lineSep = System.getProperty("line.separator");

	new_names[0] = system.system_name;
	new_names[1] = substitute(new_names[0], "System", "Variable");
	new_names[2] = new Integer(system.system.numVariables()).toString();
	new_names[3] = Double.toString(system.system.period());
	new_names[4] = Double.toString(system.sample_threshold);
	new_names[5] = Double.toString(system.approx_threshold);
	new_names[6] = new Integer(system.system.subsetLength()).toString();
	new_names[7] = Double.toString(system.system.bufferStep()) + "f";

	try {
	    PrintWriter out =
		new PrintWriter(new FileWriter(new_names[1] + ".java"));

	    for ( int i = 0 ; i < first_template.length ; i++ )
		out.println(substituteMulti(first_template[i], templates,
					    new_names));
	    int max_buffer = (int)Math.ceil(system.approx_threshold
	                                    / system.system.bufferStep());
	    if ( max_buffer < system.system.bufferSize() )
		out.println("  public int bufferSize() { return "
		          + max_buffer + "; }");

	    system.equilibrium_distrib.output(out);

	    if ( system.approx_network != null )
	    {
		system.approx_network.output(out);
		for ( int i = 0 ; i < approx_template.length ; i++ )
		    out.println(substituteMulti(approx_template[i], templates,
						new_names));
	    }

	    for ( int i = 0 ; i < second_template.length ; i++ )
		out.println(substituteMulti(second_template[i], templates,
					    new_names));

	    if ( system.approx_network != null )
	    {
		for ( int i = 0 ; i < approx2_template.length ; i++ )
		    out.println(substituteMulti(approx2_template[i], templates,
						new_names));
	    }

	    for ( int i = 0 ; i < final_template.length ; i++ )
		out.println(substituteMulti(final_template[i], templates,
					    new_names));

	    out.close();
	} catch (Exception e) {
	    System.err.println("Error generating code");
	    System.err.println(e);
	}
    }
}


