/*
 * Main.java - main entry point for the DMC compiler
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.compiler;

import java.io.FileOutputStream;
import java.io.PrintStream;
import java.io.IOException;
import java.util.Date;
import berkeley.cs.dmc.system.DynamicsSystem;


/**
 * Transform a dynamical system described by a DynamicsSystem implementation
 * into a new system which may be effectively culled, described by a
 * DynamicsVariable implementation.
 *
 * @author Stephen Chenney and Jeffrey Ichnowski
 * @version 1.0
 */
public class Main {
    private static Class dynamicsClass = null;
    private static DynamicsSystem dynamicsSystem = null;
    private static DynamicsWrapper dynamicsWrapper = null;

    private static void doWork()
    {
	System.out.println("Analysis started " + (new Date()));

	System.out.println("Class " + dynamicsClass.getName() + " loaded." );
	System.out.println(" - numVariables = "+dynamicsSystem.numVariables() );
	System.out.println(" - subsetLength = "+dynamicsSystem.subsetLength() );
	System.out.println(" - period = " + dynamicsSystem.period() );
	for ( int i = 0 ; i < dynamicsSystem.subsetLength() ; i++ )
	    System.out.println( "  - evaluationTechnique["+i+"] = " + 
			    (new String [] {"CLOSED_FORM", "ITERATIVE",
			            "FIRST_DERIVATIVE", "SECOND_DERIVATIVE"})
			    [dynamicsSystem.evaluationTechnique(i)]
			    + " - Cyclic " + dynamicsSystem.isCyclic(i));

	Bounds.boundSystem( dynamicsWrapper );

	SampleThreshold thresholder = new SampleThreshold(dynamicsWrapper);
	thresholder.findThreshold();

	System.out.println("Sampling threshold is: "
			 + dynamicsWrapper.sample_threshold);

	int approx_steps = (int)Math.floor(dynamicsWrapper.sample_threshold
					 / dynamicsWrapper.period + 0.1) - 1;
	if ( approx_steps > 0 )
	{
	    dynamicsWrapper.approx_network =
		new NeuralNetSystem(dynamicsWrapper);
	    dynamicsWrapper.approx_network.learn(dynamicsWrapper.period,
		     (int)Math.floor(dynamicsWrapper.sample_threshold
				   / dynamicsWrapper.period + 0.1) - 1);
	}
	else
	    dynamicsWrapper.approx_network = null;

	/* Set the approx threshold. */
	dynamicsWrapper.approx_threshold = dynamicsWrapper.period;

	/* Need to do code generation here. */
	CodeGeneration.generate(dynamicsWrapper);

	System.out.println("Analysis complete " + (new Date()));

    }

    /**
     * Application that takes a description of a dynamical system and
     * transforms it into a new description that may be effectively culled.
     * It should be involed as:
     *
     * <PRE>
     * java dynamics.compiler.Main [ -l logfile ] <DynamicsSystem>
     * </PRE>
     *
     * The -l option will cause all the standard output to be written to a log
     * file of the given name.
     *
     * The required argument must be the name of a class that implements
     * the dmc.system.DynamicsSystem interface. It must appear somewhere
     * in the class search path. The name must include the string "System"
     * somewhere within it.
     *
     * The program will do a lot of munging of the input file (maybe several
     * hours worth), and produce another file, derived from the input class
     * by replacing System with Variable (so, for instance, RollerSystem becomes
     * RollerVariable). The new file contains a class definition that implements
     * the DynamicsVariable interface.
     *
     * See the associated documentation for full details about using this
     * application.
     */
    public static void main( String args[] ) {
	if( args.length < 1 ) {
	    System.out.println( "Usage: java dynamics.compiler.Main [ -l logfile ] <DynamicsSystem>" );
	    System.exit(-1);
	}

	String  logname = null;
	if ( args.length == 3 )
	{
	    logname = args[1];
	    args[0] = args[2];
	}
	
	try { 
	    dynamicsClass = Class.forName( args[0] );
	    dynamicsSystem = (DynamicsSystem)dynamicsClass.newInstance();
	} catch( ClassNotFoundException e ) {
	    System.err.println( "Class '"+args[0]+"' not found." );
	    System.exit(1);
	    return;
	} catch( InstantiationException e ) {
	    System.err.println( "Could not instantiate class. " + e.getMessage() );
	    System.exit(2);
	    return;
	} catch( IllegalAccessException e ) {
	    System.err.println( "Illegal access. " + e.getMessage() );
	    System.exit(3);
	    return;
	} catch( ClassCastException e ) {
	    System.err.println( "Class does not implement berkeley.cs.dmc.system.DynamicsSystem" );
	    System.exit(4);
	    return;
	}

	dynamicsWrapper = new DynamicsWrapper(args[0], dynamicsSystem);

	PrintStream log = null;

	if ( logname != null )
	{
	    try {
		log = new PrintStream(new FileOutputStream(logname), false);
		System.setOut(new PrintStream(
				   new StreamSplitter(System.out,log)));
	    } catch (Exception e) {
		System.err.println("Error creating log file");
		System.err.println(e);
		log = null;
	    }
	}

	doWork();

	if( log != null )
	    log.close();

	System.exit(0);
    }
}

