/*
 * NeuralNetSystem.java - Neural network training code for the DMC compiler
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.compiler;

import java.awt.Button;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.PrintWriter;
import java.io.IOException;
import java.util.Random;
import berkeley.cs.dmc.system.DynamicsSystem;
import berkeley.cs.dmc.neuralnet.*;
import java.awt.*;

class NeuralNetSystem {
    boolean userQuit;
    BailOutButton bailout;

    class BailOutButton extends Frame implements ActionListener {
	Button button = new Button("Stop Learning");
	public BailOutButton() {
	    super( "Neural Nets" );
	    add( button );
	    button.addActionListener(this);
	    userQuit = false;
	    pack();
	    show();
	}
	public void actionPerformed( ActionEvent evt ) {
	    button.setLabel( "Stopping..." );
	    button.setEnabled( false );
	    userQuit = true;
	}
    }

    private static final int MAX_HIDDEN_NODES = 33;

    private static final int HIDDEN_NODES = 8;
    private static final int HIDDEN_LAYERS = 2;
    private static final ActFunc ACT_FUNC = ActFunc.logistic;
    private static final OutFunc OUT_FUNC = OutFunc.identity;

    private static final int PATTERNS_PER_LEARN = 25000;

    private static final int    ERROR_BUFFER = 16;
    private static final double MIN_IMPROVEMENT = 0.20;

    private static final int MAX_STEPS = 100000000;

    private static final boolean PLOT_NET = true;

    private DynamicsWrapper wrapper;
    private DynamicsSystem system;
    private int numVars;
    private int subsetSize;
    private int numSteps;
    private double timeSteps[];
    private FeedForwardNeuralNet nets[];
    private double error[];

    private double inmins[];
    private double inscales[];
    private double outmins[][];
    private double outranges[][];
    private double bounds[][][];

    NeuralNetSystem( DynamicsWrapper wrapper ) {
	this.wrapper = wrapper;
	system = wrapper.system;
	numVars = system.numVariables();
	subsetSize = system.subsetLength();
    }

    synchronized void learn( double period, int maxperiods ) {
	// Thread.currentThread().setPriority(Thread.MIN_PRIORITY);
	Random rand = new Random();
	double maxtime = maxperiods*period;
	double timestep;
	int i, j, k;

	System.out.println( "Learning, "+period+", " + maxtime );


	//
	// Calculate number of nets in heirarchy
	//
	for( numSteps=0, timestep=period ; timestep<=maxtime ; timestep *= 2 )
	    numSteps++;

	if( numSteps == 0 ) throw new InternalError( "num steps = 0" );

	//
	// Allocate nets, initialize
	//
	timeSteps = new double[numSteps];
	nets = new FeedForwardNeuralNet[numSteps];
	error = new double[numSteps];

	System.out.println( "Time Steps:" );
	System.out.print( "  "+period );

	timeSteps[0] = period;
	for( i=1 ; i<numSteps ; i++ ) {
	    timeSteps[i] = timeSteps[i-1] * 2;
	    System.out.print( ", " + timeSteps[i] );
	}
	System.out.println();

	for( i=0 ; i<numSteps ; i++ ) {
	    nets[i] = new FeedForwardNeuralNet( subsetSize, 
					 HIDDEN_LAYERS, HIDDEN_NODES,
					 subsetSize );
	    // nets[i].randomize( 0.0, 1.0 );
	}

	Learn backprop = Learn.backprop;
	int bufferPtr = 0;
	int bufferSize = numSteps + PATTERNS_PER_LEARN;
	double in[], out[];
	double inbuffer[][] = new double[bufferSize][numVars];
	// double dxbuffer[][] = new double[bufferSize][numVars];
	double scaledIn[][] = new double[bufferSize][subsetSize];
	double scaledOut[][] = new double[bufferSize][subsetSize];
	Pattern patterns[] = new Pattern[PATTERNS_PER_LEARN];
	inmins = wrapper.min_bounds;
	inscales = new double[subsetSize];
	outmins = new double[numSteps][subsetSize];
	outranges = new double[numSteps][subsetSize];
	double outscales[][] = new double[numSteps][subsetSize];
	boolean cyclic[] = new boolean[numVars];
	double ranges[] = new double[subsetSize];
	int times = 0;
	bounds = new double[numSteps][][];
	double errors[][] = new double[numSteps][ERROR_BUFFER];
	int errorIdx[] = new int[numSteps];
	int last_err = 0;
	boolean goodenough = false;

	for( i=0 ; i<subsetSize; i++ ) {
	    ranges[i] = wrapper.max_bounds[i] - inmins[i];
	    inscales[i] = 1.0 / ranges[i];
	}
	for( i=0 ; i<numVars ; i++ )
	    cyclic[i] = system.isCyclic(i);

	for( i=0 ; i<numSteps ; i++ ) {
	    bounds[i] = Bounds.boundDelta( wrapper, timeSteps[i] );
	    for( j=0 ; j<subsetSize ; j++ ) {
		double rng[] = bounds[i][j];
		outmins[i][j] = rng[0];
		outranges[i][j] = rng[1] - rng[0];
		outscales[i][j] = 1.0 / (outranges[i][j]);
	    }
	}

	// --- test output ---
	// System.out.println( "outputting" );
	// java.io.PrintWriter pw = new java.io.PrintWriter(
	//			  new java.io.OutputStreamWriter( System.out ) );
	// nets[0].outputProc( pw, "net0", bounds[0] );
	// pw.flush();

	//
	// initialize Patterns
	//
	for( i=0 ; i<PATTERNS_PER_LEARN ; i++ ) {
	    patterns[i] = new Pattern(numVars,numVars);
	    patterns[i].input = scaledIn[i];
	    patterns[i].output = scaledOut[i];
	}

	system.initialCondition(inbuffer[bufferSize-1]);

	//
	// Burn in
	//
	// System.out.println( "Burning in." );

	// for( i=0 ; i<BURNIN_PERIODS ; i++ )
	    // system.evaluate( 0, inbuffer[i&1], period, inbuffer[(~i)&1] );
	
	// for( j=0 ; j<numVars ; j++ )
	    // inbuffer[bufferSize-1][j] = inbuffer[i&1][j];

	//
	// Start training
	//

	PlotNet plots[];
	if( PLOT_NET ) {
	    plots = new PlotNet[numSteps];

	    for( i=0 ; i<numSteps ; i++ ) {
		Frame frame = new Frame( "dt="+((float)timeSteps[i]) );
		plots[i] = new PlotNet( nets[i], 150, 150 );
		System.out.println( "showing frame " + i );
		frame.add( plots[i] );
		frame.pack();
		frame.show();
		
		plots[i].init( wrapper, timeSteps[i], outmins[i], outscales[i] );
	    }
	} else
	    plots = null;

	try {
	bailout = new BailOutButton();

	while( !userQuit && !goodenough && times < MAX_STEPS ) {
	    //
	    // Fill in the buffer
	    //
	    // System.out.println( "Generating Patterns." );
	    system.evaluate( 0, inbuffer[bufferSize-1], period, inbuffer[0] );
	    for( j=0 ; j<numVars ; j++ ) {
		if( cyclic[j] ) // avoid unbounded results
		    inbuffer[0][j] = Math.IEEEremainder( inbuffer[0][j], ranges[j] );
		// smin[j] = smax[j] = buffer[0][j];
	    }

	    for( i=1 ; i<bufferSize ; i++ )
		system.evaluate( 0, inbuffer[i-1], period, inbuffer[i] );
	    
	    // System.out.println( "Randomizing Order..." );
	    for( j=patterns.length-1 ; j>0 ; j-- ) {
		Pattern swaptmp;
		k = (rand.nextInt()&0x7fffffff)%(j+1);
		swaptmp = patterns[j];
		patterns[j] = patterns[k];
		patterns[k] = swaptmp;
	    }

	    // double smin[] = new double[numVars];
	    //double smax[] = new double[numVars];

	    //for( j=0 ; j<numVars ; j++ ){
	    //smin[j] = 1.0e+100;
	    //smax[j] = -1.0e+100;
	    //}
		
	    
	    for( j=0 ; j<subsetSize ; j++ ) {
		double scale = inscales[j];
		double min = inmins[j];
		if( cyclic[j] ) {
		    double range = ranges[j];
		    for( i=0 ; i<bufferSize ; i++ ) {
			scaledIn[i][j] = (Math.IEEEremainder( inbuffer[i][j], range ) - min) * scale;

			/*
			double t = inbuffer[i][j];
			double r = t - range * Math.floor( t / range );
			scaledIn[i][j] = r * scale;
			*/
			/*
			double t = Math.IEEEremainder( inbuffer[i][j] - inmins[j], ranges[j] );
			if( t < 0.0 )
			    t += ranges[j];
			scaledIn[i][j] = t * inscales[j];
			*/
		    }
		} else {
		    for( i=0 ; i<bufferSize ; i++ )
			scaledIn[i][j] = (inbuffer[i][j] - min) * scale;
		}
	    }

	    // for( j=0 ; j<numVars ; j++ )
	    // System.out.println( "  range["+j+"]: " + smin[j] + " to " + smax[j] );

	    for( i=0 ; i<numSteps ; i++ ) {
		//double smins[] = new double[numVars];
		//double smaxs[] = new double[numVars];
		//for( j=0 ; j<numVars ; j++ ) {
		//    smins[j] = 1.0e+100;
		//    smaxs[j] = -1.0e+100;
		//}

		// Scale scaledOut.
		for( j=0 ; j<PATTERNS_PER_LEARN ; j++ )
		    for( k=0 ; k<subsetSize ; k++ ) {
			scaledOut[j][k] = ((inbuffer[j+i+1][k] - inbuffer[j][k]) - 
					   outmins[i][k]) * outscales[i][k];
			//if( t > smaxs[k] ) smaxs[k] = t;
			//if( t < smins[k] ) smins[k] = t;
		    }
		
		//for( j=0 ; j<numVars ; j++ )
		//    System.out.println( "output range["+j+"] : "+smins[j]+" to "+smaxs[j] );

		// if( PLOT_NET )
		//     plots[i].plotdist( patterns );

		double err = 0;
		for( j=0 ; j<patterns.length ; j++ )
		    err += nets[i].backprop( patterns[j].input,
					     patterns[j].output,
					     0.75, 0.5, 0.00 );
		
		error[i] = err / patterns.length;

		System.out.println( (times+1) + "-"
		                  + (times+PATTERNS_PER_LEARN) + " "
				  + "Error (dt = "+((float)timeSteps[i])+"): "
				  + error[i] );

		if( PLOT_NET )
		    plots[i].plot();

		errors[i][errorIdx[i]%ERROR_BUFFER] = error[i];
		last_err = errorIdx[i]%ERROR_BUFFER;
	    }

	    goodenough = true;

	    for( i=0 ; i<numSteps ; i++ ) {
		if( error[i] > system.maxApproxError() ) {
		    goodenough = false;

		    if( nets[i].numHidden() < MAX_HIDDEN_NODES && 
			errorIdx[i] >= ERROR_BUFFER &&
			error[i] > system.maxApproxError() ) {

//			double emin = 1.0e+100, emax = -1.0e+100;
//			for( j=0 ; j<ERROR_BUFFER ; j++ ) {
//			    if( errors[i][j] > emax ) emax = errors[i][j];
//			    if( errors[i][j] < emin ) emin = errors[i][j];
//			}
//			double improvement = (emax - emin) / emax;

			double e_first_avg = 0.0 , e_last_avg = 0.0;
			for ( j = 0 ; j < ERROR_BUFFER / 2 ; j++ )
			{
			    e_first_avg +=
			        errors[i][(last_err+j+1)%ERROR_BUFFER];
			    e_last_avg +=
			        errors[i][(last_err+j+1+ERROR_BUFFER/2)
				          %ERROR_BUFFER];
			}
			e_first_avg /= ( ERROR_BUFFER * 0.5 );
			e_last_avg /= ( ERROR_BUFFER * 0.5 );
			double improvement = (e_first_avg - e_last_avg)
			                   / e_first_avg;
			System.out.println("A1 " + e_first_avg + "; A2 "
			                 + e_last_avg + "; Improv "
					 + improvement);
			    
			if( improvement < MIN_IMPROVEMENT ) {
			    System.out.println( "Growing net " + i
			                      + ", improvement was "
					      + improvement);
			    nets[i].growNet(4);
			    errorIdx[i] = -1; // reset buffer
			}
		    }
		}

		errorIdx[i]++;
	    }

	    times+=PATTERNS_PER_LEARN;
	}
	} finally {
	    bailout.setVisible(false);
	    bailout.dispose();
	}
	System.out.println( "Done training." );
    }

    void output( PrintWriter out ) throws IOException {
	final String indent = "    ";
	final String indent2 = indent + indent;
	String loopindent;
	String loopoffset;
	int i, j;

	out.println( indent + "public static void approxValues(double in[], double out[], int periods ) {" );
	out.println( indent2 + "double sin[] = new double["+subsetSize+"];" );
	out.println( indent2 + "double sout[] = new double["+subsetSize+"];" );

	out.println( indent2+"System.arraycopy(in,0,out,0,"+numVars+");" );

	if( subsetSize != numVars ) {
	    out.println( indent2 + "for( int i=0 ; i<"+numVars+" ; i+="+subsetSize+" ) {" );
	    loopindent = indent2 + indent;
	    loopoffset = "i+";
	} else {
	    loopindent = indent2;
	    loopoffset = "";
	}

	final String loop1 = loopindent;
	final String loop2 = loopindent + indent;
	final String loop3 = loopindent + indent2;

	for( i=0 ; i<numVars ; i++ )
	    out.println( loop1 + "sin["+i+"] = (in["+loopoffset+i+"] - "+inmins[i]+") * "+inscales[i] + ";" );
	
	for( i=numSteps-1 ; i>=0 ; i-- ) {
	    out.println( loop1+"if( (periods & "+ (1 << i) + ") != 0 ) {" );
	    out.println( loop2+"evalNet"+i+"( sin, sout );" );
	    for( j=0 ; j<numVars ; j++ ) {
		if( system.isCyclic(j) ) {
		    double rng = wrapper.max_bounds[j] - wrapper.min_bounds[j];
		    out.println( loop2+"while( sout["+j+"] > "+wrapper.max_bounds[j]+" )" );
		    out.println( loop3+"sout["+j+"] -= "+rng+";" );
		    out.println( loop2+"while( sout["+j+"] < "+wrapper.min_bounds[j]+" )" );
		    out.println( loop3+"sout["+j+"] += "+rng+";" );
		}
		out.println( loop2+"sin["+j+"] += (sout["+j+"] - "+inmins[j]+") * "+inscales[j]+ ";" );
		out.println( loop2+"out["+loopoffset+j+"] += sout["+j+"];" );
	    }
	    out.println( loop1+"}" );
	}

	if( subsetSize != numVars ) {
	    out.println( indent2 + "}" );
	}

	out.println( indent+"}" );

	out.println();
	FeedForwardNeuralNet.outputActFunc( out );

	for( i=0 ; i<numSteps ; i++ ) {
	    out.println();
	    nets[i].outputCode( out, "evalNet"+i , bounds[i] );
	}
    }
}
