/*
 * PlotNet.java - Neural network plotting code for the DMC compiler
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.compiler;

import java.awt.Canvas;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.image.IndexColorModel;
import java.awt.image.MemoryImageSource;
import berkeley.cs.dmc.system.DynamicsSystem;
import berkeley.cs.dmc.neuralnet.FeedForwardNeuralNet;
import berkeley.cs.dmc.neuralnet.Pattern;

import berkeley.cs.dmc.probtree.ProbabilityTree;

class PlotNet extends Canvas { // implements Runnable {
    private int width, height;
    private byte syspixels0[];
    private byte syspixels1[];
    private byte netpixels0[];
    private byte netpixels1[];
    private MemoryImageSource syssource0;
    private MemoryImageSource syssource1;
    private MemoryImageSource netsource0;
    private MemoryImageSource netsource1;
    private Image sysimage0;
    private Image sysimage1;
    private Image netimage0;
    private Image netimage1;
    private FeedForwardNeuralNet net;
    private ProbabilityTree distrib;
    private double inmin0 = 0.0;
    private double inmin1 = 0.0;
    private double inscale0 = 1.0;
    private double inscale1 = 1.0;
    /*
    byte dstpixels0[];
    // byte dstpixels1[];
    MemoryImageSource dstsource0;
    // MemoryImageSource dstsource1;
    Image dstimage0;
    // Image dstimage1;
    */
    private static IndexColorModel grayScale;
    static {
	byte gray[] = new byte[256];
	for( int i=0 ; i<256 ; i++ )
	    gray[i] = (byte)i;
	grayScale = new IndexColorModel( 8, 256, gray, gray, gray );
    }

    PlotNet( FeedForwardNeuralNet net, int width, int height ) {
	int i, size = width * height;
	this.width = width;
	this.height = height;
	netpixels0 = new byte[width * height];
	netpixels1 = new byte[width * height];
	syspixels0 = new byte[width * height];
	syspixels1 = new byte[width * height];
	// dstpixels0 = new byte[(width * 2 + 1) * (height * 2 + 1)];
	// dstpixels1 = new byte[width * height];

	syssource0 = new MemoryImageSource(width, height, grayScale, syspixels0, 0, width);
	syssource1 = new MemoryImageSource(width, height, grayScale, syspixels1, 0, width);
	syssource0.setAnimated(true);
	syssource1.setAnimated(true);
	sysimage0 = createImage(syssource0);
	sysimage1 = createImage(syssource1);

	netsource0 = new MemoryImageSource( width, height, grayScale, netpixels0, 0, width );
	netsource1 = new MemoryImageSource( width, height, grayScale, netpixels1, 0, width );
	netsource0.setAnimated(true);
	netsource1.setAnimated(true);
	netimage0 = createImage(netsource0);
	netimage1 = createImage(netsource1);

	/*
	dstsource0 = new MemoryImageSource( width*2+1, height*2+1, grayScale, dstpixels0, 0, width*2+1 );
	// dstsource1 = new MemoryImageSource( width, height, grayScale, dstpixels1, 0, width );
	dstsource0.setAnimated(true);
	// dstsource1.setAnimated(true);
	dstimage0 = createImage(dstsource0);
	// dstimage1 = createImage(dstsource1);
	*/
	this.net = net;
    }

    void init( DynamicsWrapper wrapper, double time, double mins[], double scales[] ) {
	DynamicsSystem system = wrapper.system;
	int x, y, offset, val;
	double in[] = new double[2];
	double in1[] = new double[2];
	double out[] = new double[2];
	double outmin0 = mins[0];
	double outmin1 = mins[1];
	double outscale0 = 255.0 * scales[0];
	double outscale1 = 255.0 * scales[1];
	boolean cyclic0 = system.isCyclic(0);
	boolean cyclic1 = system.isCyclic(1);
	double dval;

	inmin0 = wrapper.min_bounds[0];
	inmin1 = wrapper.min_bounds[1];
	inscale0 = (wrapper.max_bounds[0] - inmin0) / width;
	inscale1 = (wrapper.max_bounds[1] - inmin1) / height;

	distrib = wrapper.equilibrium_distrib;

	for( y=offset=0 ; y<height ; y++ ) {
	    in[1] = ( height - y - 1 ) * inscale1 + inmin1;
	    for( x=0 ; x<width ; x++,offset++ ) {
		in[0] = x * inscale0 + inmin0;
		in1[0] = in[0];
		in1[1] = in[1];
		if ( distrib.density(in1) <= 1.0e-4 )
		{
		    syspixels0[offset] = (byte)0;
		    syspixels1[offset] = (byte)0;
		    continue;
		}
		system.evaluate( 0.0, in, time, out );

		dval = out[0] - in[0];
		val = (int)((dval - outmin0) * outscale0);
		if( val > 255 ) val = 255; else if( val < 0 ) val = 0;
		syspixels0[offset] = (byte)val;

		dval = out[1] - in[1];
		val = (int)((dval - outmin1) * outscale1);
		if( val > 255 ) val = 255; else if( val < 0 ) val = 0;
		syspixels1[offset] = (byte)val;
	    }
	    syssource0.newPixels(0,y,width,1);
	    syssource1.newPixels(0,y,width,1);
	}
	
	syssource0.newPixels(0,0,width,height);
	syssource1.newPixels(0,0,width,height);
    }

    public void update( Graphics g ) {
	paint( g );
    }
    public void paint( Graphics g ) {
	g.drawImage( sysimage0, 0, 0, this );
	g.drawImage( sysimage1, width+1, 0, this );
	g.drawImage( netimage0, 0, height+1, this );
	g.drawImage( netimage1, width+1, height+1, this );
	// g.drawImage( dstimage0, 0, 2*(height+1), this );
	// g.drawImage( dstimage1, width+1, 2*(height+1), this );
    }

    void plot() {
	FeedForwardNeuralNet net = this.net;
	int x, y;
	int offset;
	double in[] = new double[2];
	double in1[] = new double[2];
	double out[] = new double[2];
	int val;
	
	for( y=offset=0 ; y<height ; y++ ) {
	    in[1] = ( height - y - 1 ) / (double)height;
	    for( x=0 ; x<width ; x++,offset++ ) {
		in[0] = x / (double)width;
		in1[1] = ( height - y - 1 ) * inscale1 + inmin1;
		in1[0] = x * inscale0 + inmin0;
		if ( distrib.density(in1) <= 1.0e-4 )
		{
		    netpixels0[offset] = (byte)0;
		    netpixels1[offset] = (byte)0;
		    continue;
		}
		net.evaluate( in, out );
		val = (int)(out[0] * 255.0) & 0xff;
		netpixels0[offset] = (byte)val;
		val = (int)(out[1] * 255.0) & 0xff;
		netpixels1[offset] = (byte)val;
	    }
	    // netsource0.newPixels(0,y,width,1);
	    // netsource1.newPixels(0,y,width,1);
	}
	
	netsource0.newPixels(0,0,width,height);
	netsource1.newPixels(0,0,width,height);
    }
    /*
    public void plotdist( Pattern pat[] ) {
	int wd = width * 2 + 1;
	int ht = height * 2 + 1;

	double in[];
	int i, x, y, off;
	byte val;
	// int outofrange = 0;
	for( i=0 ; i<pat.length ; i++ ) {
	    in = pat[i].input;
	    x = (int)(in[0] * wd);
	    y = (int)(in[1] * ht);
	    if( x < 0 || x >= wd || y < 0 || y >= ht ) {
		// outofrange++;
		continue;
	    }

	    off = y*wd + x;
	    val = dstpixels0[off];
	    if( val != (byte)0xff )
		dstpixels0[off] = (byte)(val+1);
	}
	dstsource0.newPixels(0,0,wd,ht);
	// System.out.println( "Distribution plot: "+outofrange+" (of "+pat.length+") values out of range" );
    }
    */
    public Dimension getPreferredSize() { return new Dimension( 2*width+1, 2*height+1 ); }
    public Dimension getMinimumSize() { return new Dimension( 2*width+1, 2*height+1 ); }
}
