/*
 * SampleThreshold.java - Sampling and threshold code for DMC compiler
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.compiler;

import berkeley.cs.dmc.system.DynamicsSystem;
import java.util.Random;
import berkeley.cs.dmc.probtree.ProbabilityTree;
import berkeley.cs.dmc.probtree.ProbabilityTreePlot;

/**
 * A class wrapper for the functions that determine a dynamic system's
 * threshold for sampling. This is defined to be the time that must
 * ellapse while out of view before the object's state may be sampled
 * from the equilibrium distribution.
 */
class SampleThreshold {

    private static final int	PATH_NUMBER = 20000;

    private static final double ACCUMULATE_ERROR = 0.002;
    private static final double EQUIL_CONVERGENCE_DISTANCE = 0.002;

    private static final double CELL_MAX_MIX_TIME = 60.0 * 60.0; // One hour

    private DynamicsWrapper system = null;
    private int 	    num_vars = 0;
    private int 	    dimension = 0;
    private double[][]	    paths = null;
    private Random	    random_gen = new Random();

    private boolean rand_init = false;

    private double[]	    resolutions = null;
    private int		    max_res = -1;

    private ProbabilityTree     temp_distrib1;
    private ProbabilityTree     temp_distrib2;
    private ProbabilityTree     equilibrium_distrib;
    private ProbabilityTreePlot equil_plot;
    private ProbabilityTreePlot path_plot;

    /**
     * @param sys The system to be analysed.
     */
    SampleThreshold(DynamicsWrapper sys)
    {
	system = sys;
	num_vars = system.system.numVariables();
	dimension = system.system.subsetLength();
    }

    private void Init_Paths()
    {
	/* Allocate a set of paths to trace. */
	paths = new double[PATH_NUMBER][];
	for ( int i = 0 ; i < paths.length ; i++ )
	    paths[i] = new double[num_vars];
    }

    private void Step_Paths(double t_last, double t_now, int num)
    {
	for ( int i = 0 ; i < num ; i++ )
	{
	    system.system.evaluate(t_last, paths[i], t_now, paths[i]);
	    for ( int j = 0 ; j < num_vars ; j++ )
	    {
		if ( system.system.isCyclic(j%dimension) )
		{
		    while ( paths[i][j] <= system.min_bounds[j%dimension] )
			paths[i][j] +=
			    ( system.max_bounds[j%dimension]
			    - system.min_bounds[j%dimension] );
		    while ( paths[i][j] > system.max_bounds[j%dimension] )
			paths[i][j] -=
			    ( system.max_bounds[j%dimension]
			    - system.min_bounds[j%dimension] );
		}
	    }
	}
    }

    private void Add_Samples(ProbabilityTree distrib, int num)
    {
	double[]   val = new double[dimension];
	for ( int i = 0 ; i < num ; i++ )
	{
	    for ( int j = 0 ; j < num_vars ; j += dimension )
	    {
		for ( int k = 0 ; k < dimension ; k++ )
		    val[k] = paths[i][j + k];
		distrib.addSample(val);
	    }
	}
    }


    private void Build_Equilibrium(double sample_time)
    {
	double t = 0.0;

	/* Initialize the first distribution. */
	System.out.println("Generating initial conditions");
        rand_init = system.system.initialCondition(paths[0]);
	if ( rand_init )
	{
	    for ( int i = 1 ; i < paths.length ; i++ )
		system.system.initialCondition(paths[i]);
	}
	else
	{
	    for ( int i = 1 ; i < paths.length ; i++, t += sample_time )
	    {
		system.system.evaluate(t, paths[i-1],
		                       t + sample_time, paths[i]);
		for ( int j = 0 ; j < num_vars ; j++ )
		{
		    if ( system.system.isCyclic(j%dimension) )
		    {
			while ( paths[i][j] <= system.min_bounds[j%dimension] )
			    paths[i][j] +=
				( system.max_bounds[j%dimension]
				- system.min_bounds[j%dimension] );
			while ( paths[i][j] > system.max_bounds[j%dimension] )
			    paths[i][j] -=
				( system.max_bounds[j%dimension]
				- system.min_bounds[j%dimension] );
		    }
		}
	    }
	}

	Add_Samples(equilibrium_distrib, paths.length);
	temp_distrib1 = (ProbabilityTree)equilibrium_distrib.clone();

	java.awt.Frame frame = new java.awt.Frame("Equilibrium");
	equil_plot = new ProbabilityTreePlot(equilibrium_distrib, 512, 512);
	frame.add(equil_plot);
	frame.pack();
	frame.show();

	while ( true )
	{
	    if ( rand_init )
		Step_Paths(t, t + sample_time, paths.length);
	    else
	    {
		system.system.evaluate(t, paths[paths.length-1],
				       t + sample_time, paths[0]);
		for ( int i = 1 ; i < paths.length ; i++, t += sample_time )
		{
		    system.system.evaluate(t, paths[i-1],
		                           t + sample_time, paths[i]);
		    for ( int j = 0 ; j < num_vars ; j++ )
		    {
			if ( system.system.isCyclic(j%dimension) )
			{
			    while ( paths[i][j]<=system.min_bounds[j%dimension])
				paths[i][j] +=
				    ( system.max_bounds[j%dimension]
				    - system.min_bounds[j%dimension] );
			    while ( paths[i][j]>system.max_bounds[j%dimension] )
				paths[i][j] -=
				    ( system.max_bounds[j%dimension]
				    - system.min_bounds[j%dimension] );
			}
		    }
		}
	    }
	    Add_Samples(equilibrium_distrib, paths.length);
	    temp_distrib2 = temp_distrib1;
	    try {
		temp_distrib1 = (ProbabilityTree)equilibrium_distrib.clone();
	    } catch (Exception e) {
	    }
	    temp_distrib1.accumulate(ACCUMULATE_ERROR);
	    double distance = temp_distrib1.difference(temp_distrib2);

	    equil_plot.refreshData();

	    System.out.println("\tDist " + distance);

	    if ( distance < EQUIL_CONVERGENCE_DISTANCE )
		break;

	    if ( rand_init )
		t += sample_time;
	}

	try {
	    equilibrium_distrib = (ProbabilityTree)temp_distrib1.clone();
	} catch (Exception e) {
	}
    }

    private double Cell_Threshold(double sample_step, double[] center, int major)
    {
	double	max_thresh;
	double  distance;
	double	this_val;
	double	dont_care_prob;

	/* Initialize the paths for this cell. */
	for ( int i = 0 ; i < paths.length ; i++ )
	{
	    do {
		for ( int j = 0 ; j < dimension ; j++ )
		    paths[i][j] = ( random_gen.nextDouble() - 0.5 )
                                  * resolutions[j] + center[j%dimension];
	    } while ( ! system.system.checkState(0.0, paths[i], true) );
	}

	max_thresh = sample_step;
        Step_Paths(0, max_thresh, paths.length);

	int minor = 1;
	while ( true )
	{
	    temp_distrib1.clearData();
	    Add_Samples(temp_distrib1, paths.length);
	    temp_distrib1.accumulate(ACCUMULATE_ERROR);
//temp_distrib1.plotPPM(major,  minor);
//minor *= 2;

	    distance = temp_distrib1.difference(equilibrium_distrib);

	    System.out.println("\tt = " + max_thresh + " : Dist " + distance);

	    if ( distance < system.system.mixDiff()
              || max_thresh > CELL_MAX_MIX_TIME )
		break;

	    Step_Paths(max_thresh, max_thresh * 2.0, paths.length);
	    max_thresh *= 2.0;
	}
	path_plot.refreshData();
	System.out.print("\tCell ");
	for ( int i = 0 ; i < center.length ; i++ )
	    System.out.print(" " + center[i]);
	System.out.println(" threshold is " + max_thresh);

	return max_thresh;
    }

    private double Find_Threshold(double sample_time, double[] cell_steps)
    {
	int	    num_cells;
	double[][]  cell_info = null;
	int	    num_matter;
	double	    sum;
	double	    this_val;

	java.awt.Frame frame = new java.awt.Frame("Paths");
	path_plot = new ProbabilityTreePlot(temp_distrib1, 512, 512);
	frame.add(path_plot);
	frame.pack();
	frame.show();

	// Get all the probabilities.
	num_cells = equilibrium_distrib.numNonZero();
	cell_info = new double[num_cells][];
	for ( int i = 0 ; i < num_cells ; i++ )
	    cell_info[i] = new double[dimension*2+2];
	equilibrium_distrib.findNonZero(cell_info);
	// This should be changed to some more efficient sort thing.
	for ( int i = 0 ; i < cell_info.length - 1 ; i++ )
	{
	    for ( int j = i + 1 ; j < cell_info.length ; j++ )
	    {
		if ( cell_info[j][0] > cell_info[i][0] )
		{
		    double[]  temp_d = cell_info[i];
		    cell_info[i] = cell_info[j];
		    cell_info[j] = temp_d;
		}
	    }
	}

	// Find the first N that together sum to CELL_TEST_MAX
	// These are the cells we'll check.
	for ( num_matter = 0, sum = 0.0 ; sum < system.system.mixTestMax() ;
	      num_matter++ )
	    sum += cell_info[num_matter][1];

	System.out.println("Will be testing " + num_matter + " cells");

	double[]   center = new double[dimension];
	double[]   thresholds = new double[num_matter];
	for ( int i = 0 ; i < num_matter ; i++ )
	{
	    double  size = cell_info[i][2 + dimension + max_res]
			 / ( system.max_bounds[max_res]
			   - system.min_bounds[max_res] );
	    int     num_subs = (int)Math.ceil(size / resolutions[max_res]);
	    num_subs = (int)Math.pow(num_subs, dimension);

	    System.out.println("Cell " + i + " " + num_subs + " sub cells");

	    thresholds[i] = 0.0;
	    for ( int j = 0 ; j < num_subs ; j++ )
	    {
		for ( int k = 0 ; k < dimension ; k++ )
		    center[k] = random_gen.nextDouble()
			      * cell_info[i][2 + dimension + k]
			      + cell_info[i][2 + k];
		this_val = Cell_Threshold(sample_time, center, i);
		if ( this_val > thresholds[i] )
		    thresholds[i] = this_val;
	    }
	}

	// Sort the thresholds.
	// This should be changed to some more efficient sort thing.
	for ( int i = 0 ; i < thresholds.length - 1 ; i++ )
	{
	    for ( int j = i + 1 ; j < thresholds.length ; j++ )
	    {
		if ( thresholds[j] > thresholds[i] )
		{
		    double  temp_d = thresholds[i];
		    thresholds[i] = thresholds[j];
		    thresholds[j] = temp_d;
		}
	    }
	}
	num_matter = (int)Math.floor(num_matter * 0.2);

	return thresholds[num_matter];
    }


    /**
     * Find the sampling threshold.
     *
     * @return The threshold.
     */
    void findThreshold()
    {
	double[]    cell_steps;
	int	    tree_max_depth = 0;

	/* First need to build up some representation for the long term
	** distribution.
	*/

	Init_Paths();

	System.out.println("SAMPLING THRESHOLD");

	/* Check for periodicity. */
	/* If periodic, we want to sample on the period time, otherwise
	** we want to sample at some pre-defined time interval. This time
	** really depends on how fast the system mixes, and probably should
	** be a parameter of some form.
	*/
	if ( system.system.period() != 0.0 )
	    system.period = system.system.period();
	else
	    system.period = system.system.bufferStep();

	/* Set up the cell size. */
	cell_steps = new double[dimension];
	for ( int i = 0 ; i < dimension ; i++ )
	    cell_steps[i] = system.max_bounds[i] - system.min_bounds[i];

	resolutions = system.system.resolutions();
	for ( int i = 0 ; i < dimension ; i++ )
	{
	    int	this_num = (int)Math.ceil(cell_steps[i] / resolutions[i]) - 1;
	    int	k;
	    for ( k = 0 ; this_num > 0 ; this_num /= 2, k++ )
	    if ( k > tree_max_depth )
	    {
		tree_max_depth = k;
		max_res = i;
	    }
	}

	System.out.println("Maximum tree depth = " + tree_max_depth);

	System.out.println("\tBuilding equilibrium distribution");
	equilibrium_distrib = new ProbabilityTree(dimension, tree_max_depth,
						cell_steps, system.min_bounds);

	Build_Equilibrium(system.period);

	system.sample_threshold = Find_Threshold(system.period, cell_steps);

	system.equilibrium_distrib = equilibrium_distrib;
    }
}


