/*
 * Distributor.java - Distrubutes timer events to visible nodes
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.distributor;


import java.io.PrintWriter;
import java.io.FileWriter;

/**
 * Generates code that will distribute timer events to visible nodes.
 *
 * Takes as input a list of node names to manage, as arguments.
 *
 * Produces a chunk of VRML for a script node and routes, and a chunk of
 * java to do the distribution and event management.
 *
 * @auther Stephen Chenney
 * @version 1.0
 */
public class Main {

    private static void writeVRML(PrintWriter out, String[] names)
    {
	out.println("DEF DistributorScript Script {");
	out.println("    url \"Distributor.class\"");
	out.println("    mustEvaluate TRUE");
	out.println("    eventIn SFTime cycleTime");
	out.println("    eventIn SFFloat fraction_changed");
	for ( int i = 0 ; i < names.length ; i++ )
	{
	    out.println("    eventIn SFBool " + names[i] + "Visible");
	    out.println("    eventOut SFFloat " + names[i] + "CurrentTime");
	}
	out.println("}");
	for ( int i = 0 ; i < names.length ; i++ )
	{
	    out.println("ROUTE DistributorScript." + names[i]
		       + "CurrentTime TO " + names[i] + ".currentTime");
	    out.println("ROUTE " + names[i] + ".visible TO DistributorScript."
		       + names[i] + "Visible");
	}
    }


    private static void writeJava(PrintWriter out, String[] names)
    {
	out.println("import vrml.*;");
	out.println("import vrml.field.*;");
	out.println("import vrml.node.*;");
	out.println();

	out.println("public class Distributor extends Script {");

	out.println("  double start_time = 0.0;");
	out.println("  double time = 0.0;");
	out.println("  double last_time = 0.0;");
	out.println("  double fraction = 0.0;");

	for ( int i = 0 ; i < names.length ; i++ )
	    out.println("  boolean " + names[i] + "_visible = true;");

	// Dump the event processing routine - processes timer events.
	out.println("  public void processEvent( Event e ) {");
	out.println("    String name = e.getName();");
	out.println("    if ( \"cycleTime\".equals( name ) ) {");
	out.println("      time = ((ConstSFTime)e.getValue()).getValue();");
	out.println("      fraction = 0.0;");
	out.println("      if ( start_time == 0.0 )");
	out.println("        start_time = time;");
	for ( int i = 0 ; i < names.length ; i++ )
	{
	    out.println("    } else if ( \""
			+ names[i] + "Visible\".equals(name) ) {");
	    out.println("      " + names[i] + "_visible = "
		      + "(boolean)((ConstSFBool)e.getValue()).getValue();");
	}
	out.println("    } else if ( \"fraction_changed\".equals( name ) )");
	out.println("      fraction = " +
			"(double)((ConstSFFloat)e.getValue()).getValue();");
	out.println("  }");

        out.println("  public void eventsProcessed() {");
	out.println("    if ( time + fraction - start_time <= last_time )");
	out.println("        return;");
	out.println("    last_time = time + fraction - start_time;");
	for ( int i = 0 ; i < names.length ; i++ )
	{
	    out.println("    if ( " + names[i] + "_visible )");
	    out.println("      ((SFFloat)getEventOut(\""
		      + names[i]
		      + "CurrentTime\")).setValue((float)last_time);");
	}
        out.println("  }");

	out.println("}");
    }


    /**
     * Generate VRML and Java to manage events for dynamics culling.
     * Input is a list of DEFed node names to be managed. Each of
     * these is assumed to be of the type output by the tovrml.Main()
     * application or the dmod.Main() modeller. In other words, the
     * nodes given have been defined as:
     *
     * <PRE>
     * PROTO name [ eventIn SFFloat currentTime eventOut SFBool visible ] { }
     * </PRE>
     *
     * Two files are generated, Distributor.wrl and Distributor.java.
     * The former should be included in the VRML file, either directly
     * or by inlining. The Script node defined should have events from
     * a TimeSensor node routed to it. In other words, it defines:
     *
     * <PRE>
     * eventIn SFTime cycleTime
     * eventIn SFFloat fraction_changed
     * </PRE>
     *
     * Routes are also defined between the script node and each of the
     * nodes is manages.
     *
     * The Java file should be compiled and placed where a browser can
     * locate it.
     *
     * @param args An array of the names of nodes to manage.
     */
    public static void main(String[] args)
    {
	PrintWriter vrml_out = null;
	PrintWriter java_out = null;

	if ( args.length == 0 )
	    System.exit(0);

	try {
	    vrml_out = new PrintWriter(new FileWriter("Distributor.wrl"));
	    java_out = new PrintWriter(new FileWriter("Distributor.java"));
	} catch (Exception e) {
	    System.err.println(e);
	    System.exit(1);
	}

	writeVRML(vrml_out, args);

	writeJava(java_out, args);

	try {
	    vrml_out.close();
	    java_out.close();
	} catch (Exception e) {
	    System.err.println(e);
	    System.exit(1);
	}

	System.exit(0);
    }
}

