package berkeley.cs.dmc.dmod;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Vector;

import berkeley.cs.dmc.ji3d.*;
import berkeley.cs.dmc.vrml.parser.VRMLParser;
import berkeley.cs.dmc.vrml.parser.Scene;
import berkeley.cs.dmc.vrml.parser.ParseException;
import berkeley.cs.dmc.vrml.RenderState;
import berkeley.cs.dmc.model.*;

public class ModelCanvas extends Canvas
{
    Main main;
    Image doubleBuffer = null;
    Dimension size = new Dimension( -1, -1 );

    Vec3 from;
    Vec3 at;
    Vec3 up;
    Vec3 viewdir = new Vec3();
    float viewdist;
    View view;

    Material whiteMaterial = new Material( new Vec3( 0.1f, 0.1f, 0.1f ),
                                           new Vec3( 0.7f, 0.7f, 0.7f ),
                                           new Vec3( 0.2f, 0.2f, 0.2f ),
                                           20.0f );

    PointLight light = new PointLight( new Vec3( 3.0f, 3.0f, 3.0f ),
				       new Vec3( 10.0f, 3.0f, 5.0f ) );

    Light lights[] = new Light[] { light };
    FaceArray faceList = new FaceArray();

    public ModelCanvas( Main main ) {
	this( main,
	      new Vec3( 16.0f, 3.0f, 0.0f ),
	      new Vec3( 0.0f, 0.0f, 2.0f ), 
	      new Vec3( 0.0f, 0.0f, 1.0f ) );
    }

    public ModelCanvas( Main main, Vec3 from, Vec3 at, Vec3 up ) {
	this.main = main;
	this.from = from;
	this.at = at;
	this.up = up;
	view = new View( from, at, up, 100, 100 );
	enableEvents( AWTEvent.MOUSE_EVENT_MASK | AWTEvent.MOUSE_MOTION_EVENT_MASK );
	setBackground( Color.gray );
	resetBounds();
    }

    public void updateDisplay() {
	repaint();
    }

    public void updateFrame() {
	Graphics g = getGraphics();
	if(g!=null) update(g);
	g.dispose();
    }

    protected synchronized void updateBufferSize() {
	Dimension sz = getSize();
	if( sz.width != size.width || sz.height != size.height ) {
	    size = sz;
	    doubleBuffer = createImage( size.width, size.height );
	    setView();
	}
    }

    void setView() {
	Vec3 u, v;
	view.set( from, at, up, size.width, size.height );
	u = view.u;
	v = view.v;
	light.center.x = from.x + u.x * 3 + v.x * 3;
	light.center.y = from.y + u.y * 3 + v.y * 3;
	light.center.z = from.z + u.z * 3 + v.z * 3;
    }
    
    Scene load( String filename ) {
	InputStream in = null;
	VRMLParser parser;
	
	System.out.println( "loading "+filename+"..." );
	File file = new File( main.directory+filename );
	
	try {
	    if( file.exists() )
		in = new FileInputStream(file);
	    else
		in = new URL(filename).openStream();

	    parser = new VRMLParser(in);
	    parser.file();
	    return parser.getScene().getFirstProto();
	} catch( berkeley.cs.dmc.vrml.parser.ParseException e ) {
	    System.out.println( "ParseException: " );
	    e.printStackTrace();
	    return null;
	} catch( MalformedURLException e ) {
	    System.out.println( "Malformed URL: "+e.getMessage() );
	    return null;
	} catch( IOException e ) {
	    System.out.println( "IO Error: "+e.getMessage() );
	    return null;
	} finally {
	    if( in != null ) {
		try { in.close(); } catch( IOException e ) {}
	    }
	}
    }


    float toFloat( Object obj ) {
	if( obj instanceof Number )
	    return ((Number)obj).floatValue();
	else if( obj instanceof VariableField )
	    return ((VariableField)obj).getValue();
	else
	    return 0.0f;
    }

    void drawTree( Graphics g, RenderState state, TreeNode node ) {
	if( node instanceof VRMLModelNode ) {
	    VRMLModelNode n = (VRMLModelNode)node;
	    Scene scene = n.scene;
	    if( scene == null ) {
		scene = load( n.getURL().toString() );
		if( scene == null )
		    scene = new Scene();

		scene.build();
		n.scene = scene;
	    }
	    scene.render( state );
	} else if( node instanceof TranslateNode ) {
	    TranslateNode n = (TranslateNode)node;
	    Mat4 mat = new Mat4( state.model );
	    RenderState newstate = new RenderState();
	    newstate.copyFrom( state );
	    newstate.model = mat;
	    mat.appendTranslate( toFloat(n.getX()),
				 toFloat(n.getY()),
				 toFloat(n.getZ()) );
	    state = newstate;
	} else if( node instanceof RotateNode ) {
	    RotateNode n = (RotateNode)node;
	    Mat4 mat = new Mat4( state.model );
	    Vec3 axis = new Vec3( toFloat(n.getX()), toFloat(n.getY()), toFloat(n.getZ()) );
	    RenderState newstate = new RenderState();
	    newstate.copyFrom( state );
	    newstate.model = mat;
	    mat.prependRotate( axis, toFloat(n.getAngle()) );
	    state = newstate;
	}

	int i, size = node.getChildCount();
	for( i=0 ; i<size ; i++ )
	    drawTree( g, state, node.getChild(i) );
    }

    float bounds[][] = new float[3][2];

    public synchronized void resetBounds() {
	bounds[0][0] = bounds[1][0] = bounds[2][0] = Float.MAX_VALUE;
	bounds[0][1] = bounds[1][1] = bounds[2][1] = -Float.MAX_VALUE;
    }

    public synchronized void drawObjects( Graphics g ) {
	if( g == null )
	    return;

	faceList.clear();
	RenderState state = new RenderState();
	Mat4 model = new Mat4();
	model.identity();
	state.model = model;
	state.faceList = faceList;
	state.material = whiteMaterial;
	state.view = view;

	TreeNode root = main.getTree();
	if( root != null )
	    drawTree( g, state, root );

	if( faceList.minx < bounds[0][0] ) bounds[0][0] = faceList.minx;
	if( faceList.maxx > bounds[0][1] ) bounds[0][1] = faceList.maxx;
	if( faceList.miny < bounds[1][0] ) bounds[1][0] = faceList.miny;
	if( faceList.maxy > bounds[1][1] ) bounds[1][1] = faceList.maxy;
	if( faceList.minz < bounds[2][0] ) bounds[2][0] = faceList.minz;
	if( faceList.maxz > bounds[2][1] ) bounds[2][1] = faceList.maxz;

	faceList.depthsort();

	if( main.viewBounds.getState() &&
	    bounds[0][0] < bounds[0][1] &&
	    bounds[1][0] < bounds[1][1] &&
	    bounds[2][0] < bounds[2][1] ) {

	    Vec4 t = new Vec4();
	    Vec4 p = new Vec4();
	    Mat4 m = new Mat4();
	    int xpts[] = new int[8];
	    int ypts[] = new int[8];
	    float z, closest_z = Float.MAX_VALUE;
	    int closest_index = 0;
	    int i, j;

	    m.mul( view.lookat, view.projectionMatrix );

	    for( i=0 ; i<8 ; i++ ) {
		p.x = bounds[0][i&1];
		p.y = bounds[1][(i&2) >> 1];
		p.z = bounds[2][(i&4) >> 2];

		t.xform( m, p );
		
		xpts[i] = (int)(t.x/t.w) + view.half_width;
		ypts[i] = (int)(t.y/t.w) + view.half_height;
		z = t.z / t.w;

		if( z < closest_z ) {
		    closest_z = z;
		    closest_index = i;
		}
	    }

	    g.setColor( Color.yellow );
	    for( i=0 ; i<8 ; i++ ) {
		if( i != closest_index ) {
		    for( j=1 ; j<=4 ; j<<=1 ) {
			int k = i^j;
			if( i<k && k != closest_index )
			    g.drawLine( xpts[i], ypts[i], xpts[k], ypts[k] );
		    }
		}
	    }

	    if( main.wireFrame.getState() )
		faceList.paintWireFrame( g, from, lights );
	    else
		faceList.paint( g, from, lights );

	    g.setColor( Color.yellow );
	    for( i=1 ; i<=4 ; i<<=1 ) {
		int k = closest_index^i;
		g.drawLine( xpts[closest_index], ypts[closest_index],
			    xpts[k], ypts[k] );
	    }
	} else {

	    if( main.wireFrame.getState() )
		faceList.paintWireFrame( g, from, lights );
	    else
		faceList.paint( g, from, lights );
	}
    }

    public synchronized void update( Graphics g ) {
	paint( g );
    }

    public synchronized void paint( Graphics g ) {
	updateBufferSize();
	Graphics ig = doubleBuffer.getGraphics();
	ig.setColor( getBackground() );
	ig.fillRect( 0,0, size.width, size.height );
	drawObjects( ig );
	ig.dispose();
      
	g.drawImage( doubleBuffer, 0,0 , this );
    }
  
    /*
     * Event Handling
     */

    int startx, starty;
    int currentx, currenty;
    boolean viewChanging = false;
    double init_phi;
    double init_theta;
    double scale_to_top, scale_to_bottom, horiz_scale;
    boolean button1;

    void updateView()
    {
	double  delta_theta, delta_phi;
	double  sinx;
	int x = currentx;
	int y = currenty;
	
	if( !viewChanging )
	    return;

	if( button1 ) {
	    /* Cut it to the top and bottom of the screen. */
	    if( y >= view.height ) y = view.height - 1;
	    if( y <= 0 )           y = 1;
	    
	    delta_theta = -(startx - x) * horiz_scale;
	    delta_phi = -(starty - y) *
		((y < starty) ? scale_to_top : scale_to_bottom);
	    sinx = Math.sin(init_phi + delta_phi);
	    viewdir.x = (float)(sinx * Math.cos(init_theta + delta_theta));
	    viewdir.y = (float)(sinx * Math.sin(init_theta + delta_theta));
	    viewdir.z = (float)Math.cos(init_phi + delta_phi);
	
	    viewdir.scale( viewdist );
	    from.sub( at, viewdir );
	    // view.set( from, at, up, view.width, view.height );
	    setView();
	} else { // Button 2
	    float factor = (float)Math.pow(2.0, 4.0 * (starty - y) / view.height);
	    float dist = viewdist * factor;
	    from.scale( viewdir, dist ); // use 'from' as a temp var
	    from.sub( at, from );
	    // view.set( from, at, up, view.width, view.height );
	    setView();
	}
    }
    
    public void processEvent( AWTEvent e )
    {
	if( e instanceof MouseEvent ) {
	    MouseEvent event = (MouseEvent)e;
	    switch( event.getID() ) {
	    case MouseEvent.MOUSE_PRESSED:
		double  cosx, sinx;
		viewdir.sub( at, from );
		viewdist = viewdir.length();
		viewdir.scale( 1.0f / viewdist );

		button1 = (event.getModifiers() & MouseEvent.BUTTON3_MASK) == 0;
		startx = event.getX();
		starty = event.getY();
		
		if( viewdir.z >  1.0f ) viewdir.z =  1.0f;
		if( viewdir.z < -1.0f ) viewdir.z = -1.0f;
		init_phi = Math.acos( viewdir.z);
		sinx = Math.sin(init_phi);
		if( sinx == 0.0 ) {
		    cosx = 1.0;
		    sinx = 0.0;
		} else {
		    cosx = viewdir.x / sinx;
		    sinx = viewdir.y / sinx;
		}
		if( cosx >  1.0 ) cosx = 1.0;
		if( cosx < -1.0 ) cosx = -1.0;
		init_theta = (sinx > 0) ? Math.acos(cosx) : -Math.acos(cosx);
		
		if( starty == 0 ) starty++;
		if( starty == (int)view.height ) starty--;
		scale_to_top = (Math.PI - init_phi) / (double)starty;
		scale_to_bottom = init_phi / ((double)view.height - starty);
		horiz_scale = 2.0 * Math.PI / (double)view.width;
		
		viewChanging = true;
		
		// fall through
		
	    case MouseEvent.MOUSE_DRAGGED:
		currentx = event.getX();
		currenty = event.getY();
		updateView();
		repaint();
		break;
	    case MouseEvent.MOUSE_RELEASED:
		viewChanging = false;
		break;
	    }

	}
	super.processEvent( e );
    }
    /*
    public boolean handleEvent( Event event )
    {
        switch( event.id ) {
        case Event.MOUSE_DOWN:
            double  cosx, sinx;

            viewdir.sub( from, at );
            viewdist = viewdir.length();
            viewdir.scale( 1.0f / viewdist );

            startx = event.x;
            starty = event.y;

            if( viewdir.z >  1.0f ) viewdir.z =  1.0f;
            if( viewdir.z < -1.0f ) viewdir.z = -1.0f;
            init_phi = Math.acos( viewdir.z);
            sinx = Math.sin(init_phi);
            if( sinx == 0.0 ) {
                cosx = 1.0;
                sinx = 0.0;
            } else {
                cosx = viewdir.x / sinx;
                sinx = viewdir.y / sinx;
            }
            if( cosx >  1.0 ) cosx = 1.0;
            if( cosx < -1.0 ) cosx = -1.0;
            init_theta = (sinx > 0) ? Math.acos(cosx) : -Math.acos(cosx);

            if( starty == 0 ) starty++;
            if( starty == (int)view.height ) starty--;
            scale_to_top = (Math.PI - init_phi) / (double)starty;
            scale_to_bottom = init_phi / ((double)view.height - starty);
            horiz_scale = 2.0 * Math.PI / (double)view.width;

            viewChanging = true;
            // fall through

        case Event.MOUSE_DRAG:
            currentx = event.x;
            currenty = event.y;
	    updateView();
	    repaint();
            break;
        case Event.MOUSE_UP:
            viewChanging = false;
            break;
        }

        return true; // return super.handleEvent(event);
    }
    */
}

