package berkeley.cs.dmc.dmod;

import java.awt.*;
import java.util.StringTokenizer;

public class MultiLineLabel extends Canvas {
    protected String lines[];
    protected int align;
    protected int valign;
    protected int vspace;
    protected boolean transparent;
    protected transient Dimension size = null;

    public static final int LEFT = 0;
    public static final int RIGHT = 1;
    public static final int CENTER = 2;

    public static final int TOP = 3;
    public static final int BOTTOM = 4;
    public static final int FILL = 5;

    public MultiLineLabel( String text, int halign, int valign ) {
	setText( text );
	setAlignment( halign );
	setVAlign( valign );
	setVSpace( -4 );
	setTransparent( false );
    }

    public MultiLineLabel( String text, int align ) {
	this( text, align, FILL );
    }

    public MultiLineLabel( String text ) {
	this( text, CENTER, FILL );
    }

    public /*sync*/ void setAlignment( int align ) {
	if( align != LEFT && align != RIGHT && align != CENTER )
	    throw new IllegalArgumentException( "alignment must be LEFT, RIGHT, or CENTER" );
	this.align = align;
    }

    public /*sync*/ void setVAlign( int align ) {
	if( align != TOP && align != BOTTOM && align != FILL && align != CENTER )
	    throw new IllegalArgumentException( "alignment must be TOP, BOTTOM, FILL, or CENTER" );
	this.valign = align;
    }

    public /*sync*/ void setVSpace( int spc ) {
	vspace = spc;
    }

    public /*sync*/ void setTransparent( boolean trans ) {
	transparent = trans;
    }

    public /*sync*/ void setText( String text ) {
	synchronized( getTreeLock() ) {
	    StringTokenizer st = new StringTokenizer( text, "\r\n" );
	    int numlines = st.countTokens();
	    int i;

	    lines = new String[ numlines ];
	    for( i=0 ; st.hasMoreTokens() ; i++ )
		lines[i] = st.nextToken();

	    size = null;
	    invalidate();
	    if( isVisible() ) {
		validate();
		repaint();
	    }
	}
    }

    public /*sync*/ String getText() {
	StringBuffer str = new StringBuffer();
	for( int i=0 ; i<lines.length; i++ ) {
	    if( i>0 ) str.append( '\n' );
	    str.append( lines[i] );
	}
	return new String(str);
    }

    private Dimension calcSize() {
	if( size != null )
	    return size;

	Font font = getFont();
	FontMetrics fm = getFontMetrics( font );
	int width = 0;
	int height = lines.length * (fm.getHeight() + vspace) - vspace;

	for( int i=0 ; i<lines.length ; i++ ) {
	    int linewidth = fm.stringWidth( lines[i] );
	    if( width < linewidth )
		width = linewidth;
	}

	size = new Dimension( width, height );
	return size;
    }

    public /*sync*/ void update( Graphics g ) {
	Dimension sz = getSize();
	if( !transparent ) {
	    g.setColor( getBackground() );
	    g.fillRect( 0,0, sz.width, sz.height );
	}
	paint( g, sz );
    }

    public /*sync*/ void paint( Graphics g ) {
	paint( g, getSize() );
    
    }

    private final int stringX( String str, Dimension sz, FontMetrics fm ) {
	switch( align ) {
	case LEFT: return 0;
	case RIGHT: return sz.width - fm.stringWidth(str);
	case CENTER: return (sz.width - fm.stringWidth(str)) / 2;
	default: throw new InternalError();
	}
    }

    void paint( Graphics g, Dimension sz ) {

	if( !transparent ) {
	    g.setColor( getBackground() );
	    g.fillRect( 0,0, sz.width, sz.height );
	}

	g.setFont( getFont() );
	g.setColor( getForeground() );
	FontMetrics fm = g.getFontMetrics();
	int y = fm.getAscent();
	int fontheight = fm.getHeight();
	int fh = fontheight + vspace;
	int i, width;

	switch( valign ) {
	case FILL: break;
	case TOP: break;
	case BOTTOM: y += sz.height - fh*lines.length + vspace; break;
	case CENTER: y += (sz.height - fh*lines.length + vspace)/2; break;
	default:
	    throw new InternalError();
	}

	if( valign != FILL ) {
	    for( i=0 ; i<lines.length ; i++, y+=fh )
		g.drawString( lines[i], stringX( lines[i], sz, fm ), y );
	} else {
	    if( lines.length == 1 ) // FILL with one line act as CENTER
		g.drawString( lines[0], stringX( lines[0], sz, fm ), y+(sz.height - fontheight)/2 );
	    else {
		int dy = Math.max( ((sz.height - fontheight) << 16) / (lines.length-1), fh << 16 );
		y <<= 16;
		for( i=0 ; i<lines.length ; i++, y+=dy )
		    g.drawString( lines[i], stringX( lines[i], sz, fm ), y >> 16 );
	    }
	}
    }

    private final int PstringX( String str, Dimension sz, FontMetrics fm ) {
	switch( align ) {
	case LEFT: return 0;
	case RIGHT: return sz.width - fm.stringWidth(str) - 3;
	case CENTER: return (sz.width - fm.stringWidth(str)) / 2;
	default: throw new InternalError();
	}
    }
    
    /*
      public void print(Graphics g) {
      if(!(g instanceof PrintGraphics)) {
      throw new IllegalArgumentException ("Graphics context not PrintGraphics");
      } else {
      // move the text down (so it's vertically centered)
      g.translate(0,3);
      paint(g);
      }
      }
      */
    public /*sync*/ void setFont( Font font ) {
	size = null;
	super.setFont( font );
    }

    /**
     * @deprecated As of JDK version 1.1,
     * replaced by getPreferredSize().
     */
    public Dimension minimumSize() { return calcSize(); }
    /**
     * @deprecated As of JDK version 1.1,
     * replaced by getPreferredSize().
     */
    public Dimension preferredSize() { return calcSize(); }
    public Dimension getPreferredSize() { return calcSize(); }
    public Dimension getMinimumSize() { return calcSize(); }
    // public Dimension getMaximumSize() { return calcSize(); }
}
