package berkeley.cs.dmc.dmod;

import java.awt.AWTEvent;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.LayoutManager;
import java.awt.Panel;
import java.awt.event.MouseEvent;

public class VResizePanel extends Panel {
    public static final Object TOP = new Object();
    public static final Object BOTTOM = new Object();

    public static final int MIN_HEIGHT = 32;

    public static final int GROW_TOP = 0;
    public static final int GROW_BOTTOM = 1;
    public static final int PROPORTIONAL = 2; 

    Component top;
    Component bottom;
    Dimension oldsize = new Dimension(-1,-1);
    int resizePolicy;
    int gapy;
    int gapheight;

    public VResizePanel( int resize, int gap ) {
	super(null);
	if( resize != GROW_TOP && resize != GROW_BOTTOM && resize != PROPORTIONAL )
	    throw new IllegalArgumentException( "Invalid resize policy" );
	if( gap < 1 )
	    throw new IllegalArgumentException( "Gap height must be at least 1" );
	resizePolicy = resize;
	gapheight = gap;
	enableEvents( AWTEvent.MOUSE_EVENT_MASK |
		      AWTEvent.MOUSE_MOTION_EVENT_MASK );
    }

    public void setLayout( LayoutManager mgr ) {
	if( mgr != null )
	    throw new IllegalArgumentException( "VResizePanel handles layout" );

	super.setLayout(null);
    }

    protected void addImpl( Component c, Object constraints, int index ) {
	if( constraints == TOP ) {
	    if( top != null )
		remove( top );
	    top = c;
	} else if( constraints == BOTTOM ) {
	    if( bottom != null )
		remove( bottom );
	    bottom = c;
	} else
	    throw new IllegalArgumentException( "component constraints must be TOP or BOTTOM" );

	super.addImpl( c, constraints, index );
    }

    public void remove( int index ) {
	Component c = getComponent(index);
	if( c == top )
	    top = null;
	else if( c == bottom )
	    bottom = null;
	super.remove(index);
    }

    public void removeAll() {
	top = bottom = null;
	super.removeAll();
    }

    public void doLayout() {
	Dimension size = getSize();

	// System.out.println( "doLayout" );

	if( top == null ) {
	    gapy = -1;
	    if( bottom == null )
		return;
	    // System.out.println( "doLayout: bottom" );
	    bottom.setBounds( 0,0 , size.width, size.height );
	} else if( bottom == null ) {
	    gapy = -1;
	    // System.out.println( "doLayout: top" );
	    top.setBounds( 0,0 , size.width, size.height );
	} else {
	    Dimension tsize = top.getSize();
	    Dimension bsize = bottom.getSize();

	    // System.out.println( "doLayout: both" );
	    if( size.height != oldsize.height ) {
		switch( resizePolicy ) {
		case GROW_TOP: {
		    int newheight = Math.max( MIN_HEIGHT, size.height - gapheight - bsize.height );
		    top.setBounds( 0,0 , size.width, newheight );
		    bottom.setBounds( 0, newheight+gapheight, size.width, bsize.height );
		    gapy = newheight;
		    break;
		}
		case GROW_BOTTOM: {
		    int newheight = Math.max( MIN_HEIGHT, size.height - gapheight - tsize.height );
		    top.setBounds( 0,0 , size.width, tsize.height );
		    bottom.setBounds( 0, tsize.height + gapheight, size.width, newheight );
		    gapy = tsize.height;
		    break;
		}
		case PROPORTIONAL: {
		    int lheight=Math.max( MIN_HEIGHT, tsize.height * size.height / oldsize.height);
		    top.setBounds( 0,0, size.width, lheight );
		    bottom.setBounds( 0, lheight+gapheight, size.width,
				      (size.height - gapheight - lheight) );
		    gapy = lheight;
		    break;
		}
		}
	    } else if( size.width != oldsize.width ) {
		top.setSize( size.width, tsize.height );
		bottom.setSize( size.width, bsize.height );
		gapy = tsize.height;
	    }
	    oldsize = size;
	}
    }

    boolean dragging;
    boolean inResizeGap;
    int startY;
    
    private final void inResizeGap( MouseEvent evt ) {
	if( top != null && bottom != null ) {
	    int y = evt.getY();
	    if( y >= gapy && y <= (gapy + gapheight) ) {
		if( !inResizeGap ) {
		    setCursor( Cursor.getPredefinedCursor( Cursor.N_RESIZE_CURSOR ) );
		    inResizeGap = true;
		}
		return;
	    }
	}

	if( inResizeGap ) {
	    setCursor( Cursor.getPredefinedCursor( Cursor.DEFAULT_CURSOR ) );
	    inResizeGap = false;
	}
    }

    private final void resizeDrag( MouseEvent evt ) {
	if( dragging ) {
	    int y = evt.getY();
	    int lheight = Math.max( y, MIN_HEIGHT );
	    int rheight = oldsize.height - lheight - gapheight;
	    if( rheight < MIN_HEIGHT )
		return;
	    top.setSize( oldsize.width, lheight );
	    bottom.setBounds( 0, lheight + gapheight, oldsize.width, rheight );
	    gapy = lheight;
	    validate();
	}
    }

    protected void processMouseEvent( MouseEvent evt ) {
	// System.out.println( evt );
	switch( evt.getID() ) {
	case MouseEvent.MOUSE_PRESSED:
	    inResizeGap(evt);
	    if( inResizeGap ) {
		startY = evt.getY();
		dragging = true;
	    }
	    break;
	case MouseEvent.MOUSE_RELEASED:
	    resizeDrag(evt);
	    dragging = false;
	    break;
	case MouseEvent.MOUSE_ENTERED:
	    if( !dragging )
		inResizeGap(evt);
	    break;
	case MouseEvent.MOUSE_EXITED:
	    if( !dragging && inResizeGap ) {
		setCursor( Cursor.getPredefinedCursor( Cursor.DEFAULT_CURSOR ) );
		inResizeGap = false;
	    }
	    break;
	}
    }

    protected void processMouseMotionEvent( MouseEvent evt ) {
	// System.out.println( evt );
	switch( evt.getID() ) {
	case MouseEvent.MOUSE_MOVED:
	    inResizeGap(evt);
	    break;
		
	case MouseEvent.MOUSE_DRAGGED:
	    resizeDrag(evt);
	    break;
	}
    }

    public Dimension getPreferredSize() {
	if( top == null && bottom == null )
	    return new Dimension(0,0);
	else if( top == null )
	    return top.getSize();
	else if( bottom == null )
	    return bottom.getSize();
	else {
	    Dimension tsize = top.getSize();
	    Dimension bsize = bottom.getSize();
	    return new Dimension( Math.max( tsize.width, bsize.width ),
				  tsize.height + gapheight + bsize.height );
	}
    }

    public Dimension getMinimumSize() {
	Dimension min;

	if( top == null && bottom == null )
	    return new Dimension(0,0);
	else if( top == null )
	    min = top.getMinimumSize();
	else if( bottom == null )
	    min = bottom.getMinimumSize();
	else {
	    Dimension tsize = top.getSize();
	    Dimension bsize = bottom.getSize();
	    min = new Dimension( Math.max( tsize.width, bsize.width ),
				 tsize.height + gapheight + bsize.height );
	}

	Dimension pref = getPreferredSize();
	return new Dimension( Math.min( pref.width, min.width ),
			      Math.min( pref.height, min.height ) );
    }

    public static void main( String args[] ) {
	java.awt.Frame frame = new java.awt.Frame( "VResizePanel" );
	VResizePanel test = new VResizePanel( GROW_BOTTOM, 3 );
	java.awt.Button top = new java.awt.Button( "Top" );
	java.awt.Button bottom = new java.awt.Button( "Bottom" );
	test.add( top, TOP );
	test.add( bottom, BOTTOM );
	top.setSize( 100, 100 ); // only width needs to be set.
	frame.add( test );
	frame.addNotify();
	frame.setSize( 300, 200 );
	frame.validate();
	frame.show();
    }
}
