/*
 * Face.java - Face of a 3D object
 * Copyright (C) 1998 Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.ji3d;

import java.awt.*;

final public class Face extends Polygon
{
    public static final int MAX_POLY_SIZE = 32;
    
    public Material material;
    public Vec3  normal = new Vec3();
    public Vec4  center = new Vec4();

    public int   minx, maxx;
    public int   miny, maxy;
    public float minz, maxz;

    Face next, prev; // used in FaceList;

    public Vec4  points[];
    public float zpoints[];

    // public SurfaceDetail surfaceDetail = null;

    public Face() {
	xpoints = new int[MAX_POLY_SIZE];
	ypoints = new int[MAX_POLY_SIZE];
	zpoints = new float[MAX_POLY_SIZE];
	points  = new Vec4[MAX_POLY_SIZE];
	npoints = 0;

	for( int i=0 ; i<MAX_POLY_SIZE ; i++ )
	    points[i] = new Vec4();
    }

    public void set( int idx[], Vec4 pts[], ScrPt scrpts[], Material material ) {
	this.material = material;

	npoints = (idx.length > MAX_POLY_SIZE) ? MAX_POLY_SIZE : idx.length;

	maxx = maxy = Integer.MIN_VALUE;
	minx = miny = Integer.MAX_VALUE;
	minz = Float.MAX_VALUE;
	maxz = -Float.MAX_VALUE;

	for( int i=0 ; i<npoints ; i++ ) {
	    int j = idx[i];
	    Vec4 pi = points[i], pj = pts[j];
	    pi.x = pj.x;
	    pi.y = pj.y;
	    pi.z = pj.z;
	    pi.w = pj.w;

	    ScrPt sp = scrpts[j];
	    int x = sp.x, y = sp.y;
	    float z = sp.z;
	    xpoints[i] = x;
	    ypoints[i] = y;
	    zpoints[i] = z;

	    if( x < minx ) minx = x;
	    if( x > maxx ) maxx = x;
	    if( y < miny ) miny = y;
	    if( y > maxy ) maxy = y;
	    if( z < minz ) minz = z;
	    if( z > maxz ) maxz = z;
	}
    }

    /*
    public void setBounds()
    {
        int i;

        minx = maxx = scrpts[0].x;
        miny = maxy = scrpts[0].y;
        minz = maxz = scrpts[0].z;

        for( i=1 ; i<scrpts.length ; i++ ) {
            ScrPt pt = scrpts[i];

            if(      pt.x < minx ) minx = pt.x;
            else if( pt.x > maxx ) maxx = pt.x;
            if(      pt.y < miny ) miny = pt.y;
            else if( pt.y > maxy ) maxy = pt.y;
            if(      pt.z < minz ) minz = pt.z;
            else if( pt.z > maxz ) maxz = pt.z;
        }
    }
    */

    /*
    public Face( int pt[], Obj object )
    {
        int i;

        this.indecies = pt;
        this.object = object;

        scrpts = new ScrPt[pt.length];
        points = new Vec4[pt.length];
        for( i=0 ; i<pt.length ; i++ ) {
            scrpts[i] = object.projected[pt[i]];
            points[i] = object.xformPoints[pt[i]];
        }
    }
    */
    /*
    public void addDetail( Vec2 sd[], Vec3 base, Vec3 normal, Vec3 up, Material mat )
    {
        surfaceDetail = new SurfaceDetail( sd, base, normal, up, mat );
    }
    */
    /*
    public void projectSurface( Mat4 m, int half_width, int half_height )
    {
        if( surfaceDetail != null )
            surfaceDetail.project( m, half_width, half_height );
    }
    */

    public boolean backfacing()
    {
	return (xpoints[1] - xpoints[0]) * (ypoints[2] - ypoints[1]) >
	       (xpoints[2] - xpoints[1]) * (ypoints[1] - ypoints[0]);
	/*
        return (scrpts[1].x - scrpts[0].x) * (scrpts[2].y - scrpts[1].y) >
               (scrpts[2].x - scrpts[1].x) * (scrpts[1].y - scrpts[0].y);
	*/
    }

    public static final float eps = 0.005f;

    public final boolean behind( Face f ) {
        float a = f.normal.x;
        float b = f.normal.y;
        float c = f.normal.z;
        float d = a * f.points[0].x + b * f.points[0].y + c * f.points[0].z;
	d *= (1.0f - eps);

        // System.out.println( f.normal + " " + f.points[0] + " " + d );

        for( int i=0 ; i<npoints ; i++ ) {
            if( a*points[i].x + b*points[i].y + c*points[i].z < d )
                return false;
        }

        return true;
    }

    public final boolean before( Face f ) {
        float a = f.normal.x;
        float b = f.normal.y;
        float c = f.normal.z;
        float d = a * f.points[0].x + b * f.points[0].y + c * f.points[0].z;
	d *= (1.0f + eps);

        // System.out.println( f.normal + " " + f.points[0] + " " + d );

        for( int i=0 ; i<npoints ; i++ ) {
            if( a*points[i].x + b*points[i].y + c*points[i].z > d )
                return false;
        }

        return true;
    }

    public static final boolean overlap( Face A, Face B ) {
        return true;
    }

    public static final float near = -1.0f;

    /*
    private void clip()
    {
        int xp[] = poly.xpoints;
        int yp[] = poly.ypoints;
        ScrPt jp, sp;
        int j, npts = 0;
        float jdot, sdot, t;
        int dx, dy;
        float dz;

        sp = scrpts[scrpts.length - 1];
        sdot = sp.z * near;

        for( j=0 ; j<scrpts.length ; j++ ) {
            jp = scrpts[j];
            jdot = jp.z * near;

            if( jdot >= 0.0 ) {
                if( sdot >= 0.0 ) {
                    xp[npts] = jp.x;
                    yp[npts] = jp.y;
                    npts++;
                } else {
                    t = -jdot / ((jp.z - sp.z) * near);
                    xp[npts] = jp.x + (int)(t * (jp.x - sp.x));
                    yp[npts] = jp.y + (int)(t * (jp.y - sp.y));
                    npts++;
                    xp[npts] = jp.x;
                    yp[npts] = jp.y;
                    npts++;
                }
            } else if( sdot >= 0.0 ) {
                t = -jdot / ((jp.z - sp.z) * near);
                xp[npts] = jp.x + (int)(t * (jp.x - sp.x));
                yp[npts] = jp.y + (int)(t * (jp.y - sp.y));
                npts++;
            }

            sp = jp;
            sdot = jdot;
        }

        poly.npoints = npts;
    }
    */

    public void paintWireFrame( Graphics g, Vec3 from, Light lights[] ) {
	g.setColor( Light.illuminate( lights, from, center, material, normal ) );
	g.drawPolygon( this );
    }

    public void paint( Graphics g, Vec3 from, Light lights[] )
    {
        SurfaceDetail surf;
        Light l;
        int i;

	/*
        if( object.needsClipping )
            clip();
	    */

	// if( swapped ) {
	//  g.setColor( Color.white );
	//  swapped = false;
	// } else	
	g.setColor( Light.illuminate( lights, from, center, material, normal ) );

	if( Options.WIREFRAME )
	  g.drawPolygon( this );
	else
	  g.fillPolygon( this );

	/*
        for( surf = surfaceDetail; surf != null ; surf = surf.next ) {
            g.setColor( Light.illuminate( lights, from, center, surf.material, normal ) );

	    if( Options.WIREFRAME )
	      g.drawPolygon( surf );
	    else
	      g.fillPolygon( surf );
        }
	*/
    }
}
