/*
 * FaceArray.java - Simplified painter's algorithm
 * Copyright (C) 1998 Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.ji3d;

import java.awt.Graphics;
import java.util.Vector;

public final class FaceArray {

    public int  numfaces;
    public Face array[];

    public float minx, maxx;
    public float miny, maxy;
    public float minz, maxz;    

    public FaceArray( int initsize ) {
	clear();
        array = new Face[initsize];
	for( int i=0 ; i<initsize ; i++ )
	    array[i] = new Face();
    }

    public FaceArray() { this( 16 ); }

    public final void ensureCapacity( int c ) {
        if( c >= array.length ) {
            Face newarray[] = new Face[c + array.length];
            System.arraycopy( array, 0, newarray, 0, array.length );
	    for( int i=array.length ; i<newarray.length ; i++ )
		newarray[i] = new Face();
            array = newarray;
        }
    }

    public final void clear() {
	numfaces = 0;
	minx = miny = minz = Float.MAX_VALUE;
	maxx = maxy = maxz = -Float.MAX_VALUE;
    }

    public final int length() { return numfaces; }

    public final Face nextFace() {
	ensureCapacity( numfaces + 1 );
	return array[numfaces];
    }
    
    public final Face nextFaceNoCheck() {
	return array[numfaces];
    }
    
    public final void allocFace() { numfaces++; }

    public final Face elementAt( int i ) {
        return array[i];
    }

    public final boolean inOrder( int i, int j ) {
        return array[i].maxz >= array[j].maxz;

	/*
        Face P = array[i];
        Face Q = array[j];

	if( P.maxz >= Q.maxz ) {
	  if( P.minz >= Q.maxz )
	    return true;
	  
	  if( (P.minx > Q.maxx) || (P.maxx < Q.minx) ||
	      (P.miny > Q.maxy) || (P.maxy < Q.miny) )
	    return true; // xy bounds dont overlap

	  if( P.behind( Q ) )
	    return true; // in proper order

	  if( Q.before( P ) )
	    return true; // in proper order

	  return false;

	} else {
	  if( Q.minz >= P.maxz )
	    return false;
	  
	  if( (Q.minx > P.maxx) || (Q.maxx < P.minx) ||
	      (Q.miny > P.maxy) || (Q.maxy < P.miny) )
	    return false; // xy bounds dont overlap

	  if( Q.behind( P ) )
	    return false; // in proper order

	  if( P.before( Q ) )
	    return false; // in proper order

	  return true;
	}
	*/
    }

    public final void swap( int i, int j ) {
        Face tmp = array[j];
        array[j] = array[i];
        array[i] = tmp;
    }

    public final void depthsort() {
        int inc, lim, si, i, j;

        for( inc=1 ; inc < numfaces ; inc = 3*inc + 1 );

        for( ; inc > 0 ; inc = (inc - 1) / 3 ) {
	    lim = numfaces - inc;

	    for( si=0 ; si<inc ; si++ ) {
		for( i=si ; i<lim ; i+=inc ) {
		    for( j=i; !inOrder( j, j+inc ) ; j -= inc ) {
			swap( j, j+inc );
			if( j < inc ) break;
		    }
		}
	    }
        }
    }

    final void moveUp( int i, int j ) {
        Face tmp = array[j];
        System.arraycopy( array, i, array, i+1, j-i );
        array[i] = tmp;
    }

    Face head;

    final void sortOverlaps( Face P, Face tail, String indent ) {
	Face start, Q, nextQ;

	for( ; P != tail ; P=P.next ) {
	    start = P;
	    // System.out.println( indent + P.index );

	    for( Q=P.next ; Q != tail ; Q=nextQ ) {
		if( P.minz >= Q.maxz ) break; // no more overlapping polygons possible

		nextQ = Q.next;

		if( (P.minx > Q.maxx) || (P.maxx < Q.minx) ||
		    (P.miny > Q.maxy) || (P.maxy < Q.miny) )
		    continue; // xy bounds dont overlap

		if( P.behind( Q ) )
		    continue; // in proper order

		if( Q.before( P ) )
		    continue; // in proper order

		/*
		if( !Face.overlap( P, Q ) )
		    continue; // projections dont overlap

		if( !Q.behind( P ) ) // needs to be split
		    continue;

		if( !P.before( Q ) ) // ditto
		    continue;
		    */

		// unlink Q
		Q.prev.next = nextQ;
		if( nextQ != null )
		    nextQ.prev = Q.prev;

		// relink Q
		Q.next = P;
		Q.prev = P.prev;
		P.prev = Q;
		if( Q.prev == null )
		    head = Q;
		else
		    Q.prev.next = Q;

		if( start == P )
		    start = Q;
		/*
		Q.prev = start.prev;
		Q.next = start;
		start.prev = Q;
		if( Q.prev == null )
		    head = Q;
		else
		    Q.prev.next = Q;
		
		start = Q;
		*/
		// Q = nextQ;
	    }

	    if( start != P )
		sortOverlaps( start, P, indent + "  " );
	}
    }

    public synchronized final void resolveOverlaps() {
        int i,j;
	int out=0;

        if( numfaces <= 1 ) {
	    head = null;
	    return;
	}

        Face P, Q;
	// int swaps = 0;

	for( i=1 ; i<numfaces-1 ; i++ ) {
	    array[i].prev = array[i-1];
	    array[i].next = array[i+1];
	    // array[i].index = i;
	}
	head = array[0];
	head.prev = null;
	head.next = array[1];
	// head.index = 0;
	array[numfaces-1].prev = array[numfaces-2];
	array[numfaces-1].next = null;
	// array[numfaces-1].index = numfaces-1;

	// System.out.println( "sorting " + numfaces );

	sortOverlaps( head, null, "" );
	
	/*
	for( i=0 ; i<numfaces ; i++ ) {
	    P = array[i];

            for( Q=P.next ; Q != null ; Q=Q.next ) {
                if( P.minz >= Q.maxz ) break; // no more overlapping polygons possible

                if( (P.minx > Q.maxx) || (P.maxx < Q.minx) ||
                    (P.miny > Q.maxy) || (P.maxy < Q.miny) )
                    continue; // xy bounds dont overlap

                if( P.behind( Q ) )
                    continue; // in proper order

                if( Q.before( P ) )
                    continue; // in proper order

                if( !Face.overlap( P, Q ) )
		    continue; // projections dont overlap

		if( !Q.behind( P ) ) // needs to be split
		    continue;

		if( !P.behind( Q ) ) // ditto
		    continue;

		Face nextQ = Q.prev; // for loop sets Q = Q.next

		Q.prev.next = Q.next;

		if( Q.next != null )
		    Q.next.prev = Q.prev;

		Q.next = P;
		Q.prev = P.prev;
		if( Q.prev == null )
		    head = Q;
		else
		    Q.prev.next = Q;
		P.prev = Q;

                P = Q;
		Q = nextQ;
            }
	}
	*/

	/*
        for( i=0 ; i<numfaces ; i++ ) {
            P = array[i];
            for( j=i+1 ; j<numfaces ; j++ ) {
                Q = array[j];

                // given P.maxz > Q.maxz

                if( P.minz >= Q.maxz ) break; // no more overlapping polygons possible

                if( (P.minx > Q.maxx) || (P.maxx < Q.minx) ||
                    (P.miny > Q.maxy) || (P.maxy < Q.miny) )
                    continue; // xy bounds dont overlap

                if( P.behind( Q ) )
                    continue; // in proper order

                if( Q.before( P ) )
                    continue; // in proper order

                if( !Face.overlap( P, Q ) )
		    continue; // projections dont overlap

                // if( !Q.behind( P ) )
		//    continue; // split;

                // if( !P.behind( Q ) )
		//    continue; // split;

                // System.out.println( "swap" );

                moveUp( i,j );
		// swaps++;
		// Q.swapped = true;
		// P.swapped = true;
                P = Q;
            }
        }
	*/

	// System.out.println( "Swaps: " + swaps );
    }

    public final void paintWireFrame( Graphics g, Vec3 from, Light lights[] ) {
	for( int i=0 ; i<numfaces ; i++ )
	    array[i].paintWireFrame( g, from, lights );
    }

    public final void paint( Graphics g, Vec3 from, Light lights[] ) {
        int i;

	if( !Options.WIREFRAME ) {
	    resolveOverlaps();

	    for( Face face = head ; face != null ; face = face.next )
		face.paint( g, from, lights );
	} else {
	    for( i=0 ; i<numfaces ; i++ )
		array[i].paint( g, from, lights );
	}
    }

}
