/*
 * Light.java - base class for light models.
 * Copyright (C) 1998 Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.ji3d;

import java.awt.Color;

public abstract class Light extends Base implements Options
{
    public Vec3 color;

    // public Light next = null;

    public Light( Vec3 c ) {
        color = c;
    }
    protected abstract float lightDir( Vec3 ldir, Vec4 v );
    protected abstract float attenuate( float dist );

    public static final Color illuminate( Light lights[], Vec3 eye, Vec4 vertex, Material m, Vec3 normal )
    {
        Vec3 reflect = new Vec3();
        Vec3 eyedir  = new Vec3();
        Vec3 lightdir = new Vec3();
        float ldist, latten = 1.0f;
        float ndl;
        float eyedist;
        float cx, cy, cz;

        eyedir.x = eye.x - vertex.x;    // eyeDir.sub( eye, this.vertex );
        eyedir.y = eye.y - vertex.y;
        eyedir.z = eye.z - vertex.z;

        eyedist = eyedir.length();
        eyedir.scale( 1.0f/eyedist );

        reflect.scale( normal, -2.0f * Vec3.dot( normal, eyedir ) );
        reflect.sub( eyedir );
        reflect.normalize();

        float dcompx = 0.0f, // diffuse component
              dcompy = 0.0f,
              dcompz = 0.0f,
              scompx = 0.0f, // specular component
              scompy = 0.0f,
              scompz = 0.0f;

        for( int i=0 ; i<lights.length ; i++ ) {
            Light light = lights[i];

            ldist = light.lightDir( lightdir, vertex );
            ndl   = Vec3.dot( normal, lightdir );

            if( attenuateOn ) {
                latten = light.attenuate( ldist );
                if( latten >= 1.0f )
                    latten = 1.0f;
            }
            if( diffuseOn ) {
                if( ndl > 0 ) {
                    if( attenuateOn ) ndl *= latten;
                    dcompx += light.color.x * ndl;
                    dcompy += light.color.y * ndl;
                    dcompz += light.color.z * ndl;
                }
            }

            if( specularOn ) {
                if( m.shine != 0.0f ) {
                    float rdl = Vec3.dot( reflect, lightdir );

                    if( rdl > 0 ) {
                        float shine_comp = (float)Math.pow( (double)rdl, (double)m.shine );
                        if( attenuateOn ) shine_comp *= latten;
                        scompx += light.color.x * shine_comp;
                        scompy += light.color.y * shine_comp;
                        scompz += light.color.z * shine_comp;
                    }
                }
            }
        }

        if( ambientOn ) {
            cx = m.ambient.x;
            cy = m.ambient.y;
            cz = m.ambient.z;
        } else
            cx = cy = cz = 0.0f;

        if( diffuseOn ) {
            cx += m.diffuse.x * dcompx;
            cy += m.diffuse.y * dcompy;
            cz += m.diffuse.z * dcompz;
        }
        if( specularOn ) {
            cx += m.specular.x * scompx;
            cy += m.specular.y * scompy;
            cz += m.specular.z * scompz;
        }

        if( depthOn ) {
            float atten = 1.0f / (depthque.x +
                                  depthque.y * eyedist +
                                  depthque.z * eyedist * eyedist);

            if( atten < 1.0f ) {
                float invattn = 1.0f - atten;

                cx = cx * atten + background.x * invattn;
                cy = cy * atten + background.y * invattn;
                cz = cz * atten + background.z * invattn;
            }
        }

        int r = (int)(cx * (float)(255 << 16));
        int g = (int)(cy * (float)(255 << 8));
        int b = (int)(cz * (float)255);

        r = r > (255<<16) ? (255<<16) : (r & 0xff0000);
        g = g > (255<<8) ? (255<<8) : (g & 0xff00);
        b = b > 255 ? 255 : b;

        return new Color( r | g | b );
    }
}
