/*
 * Mat4.java - 4x4 matrix code
 * Copyright (C) 1998 Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.ji3d;

import java.io.PrintStream;

public final class Mat4 extends Base implements Cloneable
{
    public float xx, yx, zx, wx,
                 xy, yy, zy, wy,
                 xz, yz, zz, wz,
                 xw, yw, zw, ww;

    public Mat4() {}

    public Mat4( Mat4 m ) {
        xx = m.xx; yx = m.yx; zx = m.zx; wx = m.wx;
        xy = m.xy; yy = m.yy; zy = m.zy; wy = m.wy;
        xz = m.xz; yz = m.yz; zz = m.zz; wz = m.wz;
        xw = m.xw; yw = m.yw; zw = m.zw; ww = m.ww;
    }

    public Mat4( float a, float b, float c, float d,
                 float e, float f, float g, float h,
                 float i, float j, float k, float m,
                 float n, float o, float p, float q ) {
        xx = a; yx = b; zx = c; wx = d;
        xy = e; yy = f; zy = g; wy = h;
        xz = i; yz = j; zz = k; wz = m;
        xw = n; yw = o; zw = p; ww = q;
    }

    public final void set( float a, float b, float c, float d,
                     float e, float f, float g, float h,
                     float i, float j, float k, float m,
                     float n, float o, float p, float q ) {
        xx = a; yx = b; zx = c; wx = d;
        xy = e; yy = f; zy = g; wy = h;
        xz = i; yz = j; zz = k; wz = m;
        xw = n; yw = o; zw = p; ww = q;
    }


    // public final Object clone() { return new Mat4( this ); }

    public final boolean equals( Object anObject ) {
        if((anObject != null) && (anObject instanceof Mat4)) {
            Mat4 m = (Mat4)anObject;

            return (xx == m.xx && yx == m.yx && zx == m.zx && wx == m.wx &&
                    xy == m.xy && yy == m.yy && zy == m.zy && wy == m.wy &&
                    xz == m.xz && yz == m.yz && zz == m.zz && wz == m.wz &&
                    xw == m.xw && yw == m.yw && zw == m.zw && ww == m.ww);
        }
        else
            return false;
    }

    public final Mat4 identity() {
        xx = 1.0f; yx = 0.0f; zx = 0.0f; wx = 0.0f;
        xy = 0.0f; yy = 1.0f; zy = 0.0f; wy = 0.0f;
        xz = 0.0f; yz = 0.0f; zz = 1.0f; wz = 0.0f;
        xw = 0.0f; yw = 0.0f; zw = 0.0f; ww = 1.0f;
        return this;
    }

    public final Mat4 transpose( Mat4 m ) {
        xx = m.xx; yx = m.xy; zx = m.xz; wx = m.xw;
        xy = m.yx; yy = m.yy; zy = m.yz; wy = m.yw;
        xz = m.zx; yz = m.zy; zz = m.zz; wz = m.zw;
        xw = m.wx; yw = m.wy; zw = m.wz; ww = m.ww;
        return this;
    }

    public final Mat4 inverse( Mat4 m )
        throws ArithmeticException
    {
        float xxyy = m.xx*m.yy, zzww = m.zz*m.ww, wzzw = m.wz*m.zw, xxyz = m.xx*m.yz,
              zyww = m.zy*m.ww, wyzw = m.wy*m.zw, xxyw = m.xx*m.yw, zywz = m.zy*m.wz,
              wyzz = m.wy*m.zz, xyyx = m.xy*m.yx, xyyz = m.xy*m.yz, zxww = m.zx*m.ww,
              wxzw = m.wx*m.zw, xyyw = m.xy*m.yw, zxwz = m.zx*m.wz, wxzz = m.wx*m.zz,
              xzyx = m.xz*m.yx, xzyy = m.xz*m.yy, xzyw = m.xz*m.yw, zxwy = m.zx*m.wy,
              wxzy = m.wx*m.zy, xwyx = m.xw*m.yx, xwyy = m.xw*m.yy, xwyz = m.xw*m.yz;

        float d = (((xxyy*zzww + xxyz*wyzw) + (xxyw*zywz + xyyx*wzzw))
                 + ((xyyw*wxzz + xzyx*zyww) + (xzyy*wxzw + xzyw*zxwy))
                 + (xwyx*wyzz + xwyy*zxwz + xwyz*wxzy))
                - (((xxyy*wzzw + xxyz*zyww) + (xxyw*wyzz + xyyx*zzww))
                 + ((xyyz*wxzw + xyyw*zxwz) + (xyyz*zxww + xzyx*wyzw))
                 + ((xzyy*zxww + xzyw*wxzy) + (xwyx*zywz + xwyy*wxzz + xwyz*zxwy)));

        if( d == 0.0 ) throw new ArithmeticException( "Matrix not invertible" );

        xx = ((m.yy*zzww + m.yz*wyzw + m.yw*zywz) - (m.yw*wyzz + m.yy*wzzw + m.yz*zyww)) / d;
        xy = ((m.xy*wzzw + m.xz*zyww + m.xw*wyzz) - (m.xy*zzww + m.xz*wyzw + m.xw*zywz)) / d;
        xz = ((xyyz*m.ww + xzyw*m.wy + xwyy*m.wz) - (xyyw*m.wz + xzyy*m.ww + xwyz*m.wy)) / d;
        xw = ((xyyw*m.zz + xzyy*m.zw + xwyz*m.zy) - (xyyz*m.zw + xzyw*m.zy + xwyy*m.zz)) / d;
        yx = ((m.yx*wzzw + m.yz*zxww + m.yw*wxzz) - (m.yx*zzww + m.yz*wxzw + m.yw*zxwz)) / d;
        yy = ((m.xx*zzww + m.xz*wxzw + m.xw*zxwz) - (m.xx*wzzw + m.xz*zxww + m.xw*wxzz)) / d;
        yz = ((xxyw*m.wz + xzyx*m.ww + xwyz*m.wx) - (xxyz*m.ww + xzyw*m.wx + xwyx*m.wz)) / d;
        yw = ((xxyz*m.zw + xzyw*m.zx + xwyx*m.zz) - (xxyw*m.zz + xzyx*m.zw + xwyz*m.zx)) / d;
        zx = ((m.yx*zyww + m.yy*wxzw + m.yw*zxwy) - (m.yx*wyzw + m.yy*zxww + m.yw*wxzy)) / d;
        zy = ((m.xx*wyzw + m.xy*zxww + m.xw*wxzy) - (m.xx*zyww + m.xy*wxzw + m.xw*zxwy)) / d;
        zz = ((xxyy*m.ww + xyyw*m.wx + xwyx*m.wy) - (xxyw*m.wy + xyyx*m.ww + xwyy*m.wx)) / d;
        zw = ((xxyw*m.zy + xyyx*m.zw + xwyy*m.zx) - (xxyy*m.zw + xyyw*m.zx + xwyx*m.zy)) / d;
        wx = ((m.yx*wyzz + m.yy*zxwz + m.yz*wxzy) - (m.yx*zywz + m.yy*wxzz + m.yz*zxwy)) / d;
        wy = ((m.xx*zywz + m.xy*wxzz + m.xz*zxwy) - (m.xx*wyzz + m.xy*zxwz + m.xz*wxzy)) / d;
        wz = ((xxyz*m.wy + xyyx*m.wz + xzyy*m.wx) - (xxyy*m.wz + xyyz*m.wx + xzyx*m.wy)) / d;
        ww = ((xxyy*m.zz + xyyz*m.zx + xzyx*m.zy) - (xxyz*m.zy + xyyx*m.zz + xzyy*m.zx)) / d;

        return this;
    }

    public final Mat4 mul( Mat4 a, Mat4 b ) {
        xx = a.xx * b.xx + a.xy * b.yx + a.xz * b.zx + a.xw * b.wx;
        xy = a.xx * b.xy + a.xy * b.yy + a.xz * b.zy + a.xw * b.wy;
        xz = a.xx * b.xz + a.xy * b.yz + a.xz * b.zz + a.xw * b.wz;
        xw = a.xx * b.xw + a.xy * b.yw + a.xz * b.zw + a.xw * b.ww;

        yx = a.yx * b.xx + a.yy * b.yx + a.yz * b.zx + a.yw * b.wx;
        yy = a.yx * b.xy + a.yy * b.yy + a.yz * b.zy + a.yw * b.wy;
        yz = a.yx * b.xz + a.yy * b.yz + a.yz * b.zz + a.yw * b.wz;
        yw = a.yx * b.xw + a.yy * b.yw + a.yz * b.zw + a.yw * b.ww;

        zx = a.zx * b.xx + a.zy * b.yx + a.zz * b.zx + a.zw * b.wx;
        zy = a.zx * b.xy + a.zy * b.yy + a.zz * b.zy + a.zw * b.wy;
        zz = a.zx * b.xz + a.zy * b.yz + a.zz * b.zz + a.zw * b.wz;
        zw = a.zx * b.xw + a.zy * b.yw + a.zz * b.zw + a.zw * b.ww;

        wx = a.wx * b.xx + a.wy * b.yx + a.wz * b.zx + a.ww * b.wx;
        wy = a.wx * b.xy + a.wy * b.yy + a.wz * b.zy + a.ww * b.wy;
        wz = a.wx * b.xz + a.wy * b.yz + a.wz * b.zz + a.ww * b.wz;
        ww = a.wx * b.xw + a.wy * b.yw + a.wz * b.zw + a.ww * b.ww;

        return this;
    }

    public final Mat4 append( Mat4 b ) { // return mul( new Mat4(this), b );
        float x = xx, y = xy, z = xz, w = xw;
        xx = x*b.xx + y*b.yx + z*b.zx + w*b.wx;
        xy = x*b.xy + y*b.yy + z*b.zy + w*b.wy;
        xz = x*b.xz + y*b.yz + z*b.zz + w*b.wz;
        xw = x*b.xw + y*b.yw + z*b.zw + w*b.ww;
        x = yx; y = yy; z = yz; w = yw;
        yx = x*b.xx + y*b.yx + z*b.zx + w*b.wx;
        yy = x*b.xy + y*b.yy + z*b.zy + w*b.wy;
        yz = x*b.xz + y*b.yz + z*b.zz + w*b.wz;
        yw = x*b.xw + y*b.yw + z*b.zw + w*b.ww;
        x = zx; y = zy; z = zz; w = zw;
        zx = x*b.xx + y*b.yx + z*b.zx + w*b.wx;
        zy = x*b.xy + y*b.yy + z*b.zy + w*b.wy;
        zz = x*b.xz + y*b.yz + z*b.zz + w*b.wz;
        zw = x*b.xw + y*b.yw + z*b.zw + w*b.ww;
        x = wx; y = wy; z = wz; w = ww;
        wx = x*b.xx + y*b.yx + z*b.zx + w*b.wx;
        wy = x*b.xy + y*b.yy + z*b.zy + w*b.wy;
        wz = x*b.xz + y*b.yz + z*b.zz + w*b.wz;
        ww = x*b.xw + y*b.yw + z*b.zw + w*b.ww;
        return this;
    }

    public final Mat4 prepend( Mat4 a ) { // return mul( a, new Mat4( this ) );
        float x,y,z,w;
        x=xx; y=yx; z=zx; w=wx;
        xx = a.xx * x + a.xy * y + a.xz * z + a.xw * w;
        yx = a.yx * x + a.yy * y + a.yz * z + a.yw * w;
        zx = a.zx * x + a.zy * y + a.zz * z + a.zw * w;
        wx = a.wx * x + a.wy * y + a.wz * z + a.ww * w;
        x=xy; y=yy; z=zy; w=wy;
        xy = a.xx * x + a.xy * y + a.xz * z + a.xw * w;
        yy = a.yx * x + a.yy * y + a.yz * z + a.yw * w;
        zy = a.zx * x + a.zy * y + a.zz * z + a.zw * w;
        wy = a.wx * x + a.wy * y + a.wz * z + a.ww * w;
        x=xz; y=yz; z=zz; w=wz;
        xz = a.xx * x + a.xy * y + a.xz * z + a.xw * w;
        yz = a.yx * x + a.yy * y + a.yz * z + a.yw * w;
        zz = a.zx * x + a.zy * y + a.zz * z + a.zw * w;
        wz = a.wx * x + a.wy * y + a.wz * z + a.ww * w;
        x=xw; y=yw; z=zw; w=ww;
        xw = a.xx * x + a.xy * y + a.xz * z + a.xw * w;
        yw = a.yx * x + a.yy * y + a.yz * z + a.yw * w;
        zw = a.zx * x + a.zy * y + a.zz * z + a.zw * w;
        ww = a.wx * x + a.wy * y + a.wz * z + a.ww * w;
        return this;
    }


    public final Mat4 mulTxM( Mat4 a, Mat4 b ) {
        xx = a.xx * b.xx + a.yx * b.yx + a.zx * b.zx + a.wx * b.wx;
        xy = a.xx * b.xy + a.yx * b.yy + a.zx * b.zy + a.wx * b.wy;
        xz = a.xx * b.xz + a.yx * b.yz + a.zx * b.zz + a.wx * b.wz;
        xw = a.xx * b.xw + a.yx * b.yw + a.zx * b.zw + a.wx * b.ww;

        yx = a.xy * b.xx + a.yy * b.yx + a.zy * b.zx + a.wy * b.wx;
        yy = a.xy * b.xy + a.yy * b.yy + a.zy * b.zy + a.wy * b.wy;
        yz = a.xy * b.xz + a.yy * b.yz + a.zy * b.zz + a.wy * b.wz;
        yw = a.xy * b.xw + a.yy * b.yw + a.zy * b.zw + a.wy * b.ww;

        zx = a.xz * b.xx + a.yz * b.yx + a.zz * b.zx + a.wz * b.wx;
        zy = a.xz * b.xy + a.yz * b.yy + a.zz * b.zy + a.wz * b.wy;
        zz = a.xz * b.xz + a.yz * b.yz + a.zz * b.zz + a.wz * b.wz;
        zw = a.xz * b.xw + a.yz * b.yw + a.zz * b.zw + a.wz * b.ww;

        wx = a.xw * b.xx + a.yw * b.yx + a.zw * b.zx + a.ww * b.wx;
        wy = a.xw * b.xy + a.yw * b.yy + a.zw * b.zy + a.ww * b.wy;
        wz = a.xw * b.xz + a.yw * b.yz + a.zw * b.zz + a.ww * b.wz;
        ww = a.xw * b.xw + a.yw * b.yw + a.zw * b.zw + a.ww * b.ww;

        return this;
    }

    public final void rotationX( float c, float s ) {
        xx = 1.0f; yx = 0.0f; zx = 0.0f; wx = 0.0f;
        xy = 0.0f; yy =  c;   zy = -s;   wy = 0.0f;
        xz = 0.0f; yz =  s;   zz =  c;   wz = 0.0f;
        xw = 0.0f; yw = 0.0f; zw = 0.0f; ww = 1.0f;
    }
    public final void rotationXrad( double rad ) {
        rotationX( (float)Math.cos(rad), (float)Math.cos(rad) );
    }
    public final void rotationXdeg( float angle ) {
        rotationXrad( angle * (Math.PI / 180.0f) );
    }

    public final void appendRotateX( float c, float s ) {
        float tmp;
        tmp = yx; yx = c*tmp + s*zx; zx = c*zx - s*tmp;
        tmp = yy; yy = c*tmp + s*zy; zy = c*zy - s*tmp;
        tmp = yz; yz = c*tmp + s*zz; zz = c*zz - s*tmp;
        tmp = yw; yw = c*tmp + s*zw; zw = c*zw - s*tmp;
    }
    public final void appendRotateXrad( double rad ) {
        appendRotateX( (float)Math.cos(rad), (float)Math.sin(rad) );
    }
    public final void appendRotateXdeg( float angle )  {
        appendRotateXrad( angle * (Math.PI / 180.0) );
    }

    public final void prependRotateX( float c, float s ) {
        float tmp;
        tmp = xy; xy = c*tmp - s*xz; xz = s*tmp + c*xz;
        tmp = yy; yy = c*tmp - s*yz; yz = s*tmp + c*yz;
        tmp = zy; zy = c*tmp - s*zz; zz = s*tmp + c*zz;
        tmp = wy; wy = c*tmp - s*wz; wz = s*tmp + c*wz;
    }
    public final void prependRotateXrad( double rad ) {
        prependRotateX( (float)Math.cos(rad), (float)Math.sin(rad) );
    }
    public final void prependRotateXdeg( float angle ) {
        prependRotateXrad( angle * (Math.PI / 180.0) );
    }

    public final void rotationY( float c, float s ) {
        xx =  c;   yx = 0.0f; zx =  s;   wx = 0.0f;
        xy = 0.0f; yy = 1.0f; zy = 0.0f; wy = 0.0f;
        xz = -s;   yz = 0.0f; zz =  c;   wz = 0.0f;
        xw = 0.0f; yw = 0.0f; zw = 0.0f; ww = 1.0f;
    }
    public final void rotationYrad( double rad ) {
        rotationY( (float)Math.cos(rad), (float)Math.sin(rad) );
    }
    public final void rotationYdeg( float angle ) {
        rotationYrad( angle * (Math.PI / 180.0) );
    }

    public final void appendRotateY( float c, float s ) {
        float tmp;
        tmp=xx; xx = c*tmp - s*zx; zx = s*tmp + c*zx;
        tmp=xy; xy = c*tmp - s*zy; zy = s*tmp + c*zy;
        tmp=xz; xz = c*tmp - s*zz; zz = s*tmp + c*zz;
        tmp=xw; xw = c*tmp - s*zw; zw = s*tmp + c*zw;
    }
    public final void appendRotateYrad( double rad ) {
        appendRotateY( (float)Math.cos(rad), (float)Math.sin(rad) );
    }
    public final void appendRotateYdeg( float angle ) {
        appendRotateYrad( angle * (Math.PI / 180.0) );
    }

    public final void prependRotateY( float c, float s ) {
        float tmp;
        tmp=xx; xx = s*xz + c*tmp; xz = c*xz - s*tmp;
        tmp=yx; yx = s*yz + c*tmp; yz = c*yz - s*tmp;
        tmp=zx; zx = s*zz + c*tmp; zz = c*zz - s*tmp;
        tmp=wx; wx = s*wz + c*tmp; wz = c*wz - s*tmp;
    }
    public final void prependRotateYrad( double rad ) {
        prependRotateY( (float)Math.cos(rad), (float)Math.sin(rad) );
    }
    public final void prependRotateYdeg( float angle ) {
        prependRotateYrad( angle * (Math.PI / 180.0) );
    }

    public final void rotationZ( float c, float s ) {
        xx =  c;   yx = -s;   zx = 0.0f; wx = 0.0f;
        xy =  s;   yy =  c;   zy = 0.0f; wy = 0.0f;
        xz = 0.0f; yz = 0.0f; zz = 1.0f; wz = 0.0f;
        xw = 0.0f; yw = 0.0f; zw = 0.0f; ww = 1.0f;
    }
    public final void rotationZrad( double rad ) {
        rotationZ( (float)Math.cos(rad), (float)Math.sin(rad) );
    }
    public final void rotationZdeg( float angle ) {
        rotationZrad( angle * (Math.PI / 180.0) );
    }

    public final void appendRotateZ( float c, float s ) {
        float tmp;
        tmp=xx; xx = s*yx + c*tmp; yx = c*yx - s*tmp;
        tmp=xy; xy = s*yy + c*tmp; yy = c*yy - s*tmp;
        tmp=xz; xz = s*yz + c*tmp; yz = c*yz - s*tmp;
        tmp=xw; xw = s*yw + c*tmp; yw = c*yw - s*tmp;
    }
    public final void appendRotateZrad( double rad ) {
        appendRotateZ( (float)Math.cos(rad), (float)Math.sin(rad) );
    }
    public final void appendRotateZdeg( float angle ) {
        appendRotateZrad( angle * (Math.PI / 180.0) );
    }

    public final void prependRotateZ( float c, float s ) {
        float tmp;
        tmp=xx; xx = c*tmp - s*xy; xy = s*tmp + c*xy;
        tmp=yx; yx = c*tmp - s*yy; yy = s*tmp + c*yy;
        tmp=zx; zx = c*tmp - s*zy; zy = s*tmp + c*zy;
        tmp=wx; wx = c*tmp - s*wy; wy = s*tmp + c*wy;
    }
    public final void prependRotateZrad( double rad ) {
        prependRotateZ( (float)Math.cos(rad), (float)Math.sin(rad) );
    }
    public final void prependRotateZdeg( float angle ) {
        prependRotateZrad( angle * (Math.PI / 180.0) );
    }

    public final void rotation( Vec3 axis, double rad ) {
        float len = axis.length();

        if( len <= 1e-6f ) {
            identity();
            return;
        }

        float s = (float)Math.sin(rad);
        float c = (float)Math.cos(rad);

        float norm = 1.0f / len,
              x = axis.x * norm,
              y = axis.y * norm,
              z = axis.z * norm;

        float ic = 1 - c,
              x2 = x * x, y2 = y * y, z2 = z * z,
              axyc = x * y * ic,
              azxc = z * x * ic,
              ayzc = y * z * ic,
              xs = x * s, ys = y * s, zs = z * s;

        xx = x2+c*(1-x2); yx = axyc - zs;   zx = azxc + ys;   wx = 0.0f;
        xy = axyc + zs;   yy = y2+c*(1-y2); zy = ayzc - xs;   wy = 0.0f;
        xz = azxc - ys;   yz = ayzc + xs;   zz = z2+c*(1-z2); wz = 0.0f;
        xw = 0.0f;        yw = 0.0f;        zw = 0.0f;        ww = 1.0f;
    }

    public final void appendRotate( Vec3 axis, double rad ) {
        Mat4 rot = new Mat4();
        rot.rotation( axis, rad );
        append( rot );
    }

    public final void prependRotate( Vec3 axis, double rad ) {
        Mat4 rot = new Mat4();
        rot.rotation( axis, rad );
        prepend( rot );
    }

    public final float decompose( Vec3 axis, Vec3 trans ) {
      double cos = (xx+yy+zz) / 2;
      double angle = Math.acos( cos );
      float invsin2 = (float)(1.0 / (2 * Math.sin( angle )));

      axis.x = (yz - zy) * invsin2;
      axis.y = (zx - xz) * invsin2;
      axis.z = (xy - yz) * invsin2;

      trans.x = wx;
      trans.y = wy;
      trans.z = wz;

      return (float)angle;
    }

    public final void scaling( float x, float y, float z ) {
        xx =  x;   yx = 0.0f; zx = 0.0f; wx = 0.0f;
        xy = 0.0f; yy =  y;   zy = 0.0f; wy = 0.0f;
        xz = 0.0f; yz = 0.0f; zz =  z;   wz = 0.0f;
        xw = 0.0f; yw = 0.0f; zw = 0.0f; ww = 1.0f;
    }

    public final void appendScale( float x, float y, float z ) {
        xx *= x; xy *= x; xz *= x; xw *= x;
        yx *= y; yy *= y; yz *= y; yw *= y;
        zx *= z; zy *= z; zz *= z; zw *= z;
    }
    public final void prependScale( float x, float y, float z ) {
        xx *= x; yx *= x; zx *= x; wx *= x;
        xy *= y; yy *= y; zy *= y; wy *= y;
        xz *= z; yz *= z; zz *= z; wz *= z;
    }

    public final float determinant3() {
        return xx*yy*zz + yx*zy*xz + zx*xy*yz
             - zx*yy*xz - yx*xy*zz - xx*zy*yz;
    }

    public final float determinant() {
        float xxyy = xx*yy, zzww = zz*ww, wzzw = wz*zw, xxyz = xx*yz,
              zyww = zy*ww, wyzw = wy*zw, xxyw = xx*yw, zywz = zy*wz,
              wyzz = wy*zz, xyyx = xy*yx, xyyz = xy*yz, zxww = zx*ww,
              wxzw = wx*zw, xyyw = xy*yw, zxwz = zx*wz, wxzz = wx*zz,
              xzyx = xz*yx, xzyy = xz*yy, xzyw = xz*yw, zxwy = zx*wy,
              wxzy = wx*zy, xwyx = xw*yx, xwyy = xw*yy, xwyz = xw*yz;

        return (((xxyy*zzww + xxyz*wyzw) + (xxyw*zywz + xyyx*wzzw))
              + ((xyyw*wxzz + xzyx*zyww) + (xzyy*wxzw + xzyw*zxwy))
               + (xwyx*wyzz + xwyy*zxwz + xwyz*wxzy))
             - (((xxyy*wzzw + xxyz*zyww) + (xxyw*wyzz + xyyx*zzww))
              + ((xyyz*wxzw + xyyw*zxwz) + (xyyz*zxww + xzyx*wyzw))
              + ((xzyy*zxww + xzyw*wxzy) + (xwyx*zywz + xwyy*wxzz + xwyz*zxwy)));
    }

    public final void translation( float x, float y, float z ) {
        xx = 1.0f; yx = 0.0f; zx = 0.0f; wx = x;
        xy = 0.0f; yy = 1.0f; zy = 0.0f; wy = y;
        xz = 0.0f; yz = 0.0f; zz = 1.0f; wz = z;
        xw = 0.0f; yw = 0.0f; zw = 0.0f; ww = 1.0f;
    }
    public final void appendTranslate( float x, float y, float z ) {
        wx += x * xx + y * yx + z * zx;
        wy += x * xy + y * yy + z * zy;
        wz += x * xz + y * yz + z * zz;
        ww += x * xw + y * yw + z * zw;
    }
    public final void prependTranslate( float x, float y, float z ) {
        xx += x*xw; yx += x*yw; zx += x*zw; wx += x*ww;
        xy += y*xw; yy += y*yw; zy += y*zw; wy += y*ww;
        xz += z*xw; yz += z*yw; zz += z*zw; wz += z*ww;
    }

    public final Mat4 lookat( Vec3 view, Vec3 cop, float twistangle ) {
        Vec3 n = new Vec3();
        Vec3 u = new Vec3();
        Vec3 v = new Vec3();
        Mat4 m;

        appendRotateZdeg( -twistangle-90.0f );
        n.sub( view, cop ); n.normalize();
        u.cross( new Vec3(0.0f,1.0f,0.0f), n ); u.normalize();
        v.cross( n, u );

        m = new Mat4( u.x,  u.y,  u.z, -view.x,
                      v.x,  v.y,  v.z, -view.y,
                      n.x,  n.y,  n.z, -view.z,
                      0.0f, 0.0f, 0.0f, 1.0f );

        prepend( m );

        return this;
    }

    public Mat4 lookat( Vec3 from, Vec3 at, Vec3 up ) {
        Vec3 n = new Vec3();
        Vec3 u = new Vec3();
        Vec3 v = new Vec3();
        Mat4 m;

        n.sub( at, from );  n.normalize();
        u.cross( n, up );   u.normalize();
        v.cross( u, n );    v.normalize();

        m = new Mat4( u.x,  u.y,  u.z, -from.x, // Warning: GC
                      v.x,  v.y,  v.z, -from.y,
                      n.x,  n.y,  n.z, -from.z,
                      0.0f, 0.0f, 0.0f, 1.0f );

        prepend( m );

        return this;
    }

    public final Mat4 perspective( float fovy, float aspect, float near, float far ) {
        float yscale;
        yscale = (float)(1.0 / Math.tan( fovy * (Math.PI / 180.0) / 2.0 ));
        identity();
        xx = yscale / (aspect * far);
        yy = yscale / far;
        zz = -1.0f / (far - near);
        zw = -1.0f / far;
        wz = near / (far - near);
        ww = 0.0f;

        return this;
    }

    public final void print( PrintStream out, String name ) {
        out.println( "Matrix \"" + name + "\":" );
        out.println( "  " + xx + ' ' + yx + ' ' + zx + ' ' + wx );
        out.println( "  " + xy + ' ' + yy + ' ' + zy + ' ' + wy );
        out.println( "  " + xz + ' ' + yz + ' ' + zz + ' ' + wz );
        out.println( "  " + xw + ' ' + yw + ' ' + zw + ' ' + ww );
    }
/*
    public static void main( String args[] ) {
        PrintStream out = System.out;

        out.println( "Mat4 regression tests" );

        Mat4 a = new Mat4( 1,2,3,4, 0,5,6,7 , 4,0,8,9, 0,0,0,10 );
        Mat4 b = new Mat4();
        Mat4 c = new Mat4();

        try {
            a.print( out, "A" );
            b.inverse( a );
            b.print( out, "A^-1" );
            c.mul( a, b );
            c.print( out, "A * A^-1" );
        } catch( ArithmeticException e ) {
            out.println( e.getMessage() );
        }

        System.exit(0);
    }
*/
}
