/*
 * Obj.java - 3D object model in ji3d package
 * Copyright (C) 1998 Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.ji3d;

import java.awt.*;

public final class Obj extends Base
{
    public Vec3 points[];
    public Vec3 normals[];
    public Vec4 centers[];

    // bounding sphere
    public Vec3  center = new Vec3();
    public float radius;
    public Vec3  xformCenter = new Vec3();;
    public float xformRadius;

    public Vec4  xformPoints[];
    public Vec3  xformNormals[];
    public Vec4  xformCenters[];
    public ScrPt projected[];
    public boolean needsClipping;

    // public Face  faces[];
    public int faceIdx[][];

    public Mat4 modelMatrix;
    public Material material;

    // public boolean inFacelist = false; // used to ensure one object is not inserted twice.

    final void calculateNormals( int facePts[][] )
    {
        int i;
        int face[] = null;
        Vec3 a = new Vec3();
        Vec3 b = new Vec3();
	Vec3 norm;

        for( i=0 ; i<facePts.length ; i++ ) {
	    face = facePts[i];
	    norm = normals[i];

	    try {
		a.sub( points[face[1]], points[face[0]] );
		b.sub( points[face[2]], points[face[1]] );

		norm.cross( a, b );
		norm.normalize();
	    } catch( Exception e ) {
		norm.set(0,0,1);
	    }
        }
    }

    final void calculateCenters( int facePts[][] )
    {
        int i, j;
        int face[];
        Vec4 center;

        for( i=0 ; i<facePts.length ; i++ ) {
            face = facePts[i];
            center = centers[i];

            center.zero();

            for( j=0 ; j<face.length ; j++ )
                center.add( points[face[j]] );

            center.scale( 1.0f / face.length );
        }
    }

    // An semi-inefficient bounding sphere
    public final void calculateBounds()
    {
        float dist;
        int i;
        float minx, maxx;
        float miny, maxy;
        float minz, maxz;

        minx = maxx = points[0].x;
        miny = maxy = points[0].y;
        minz = maxz = points[0].z;

        for( i=1 ; i<points.length ; i++ ) {
            if(      points[i].x < minx ) minx = points[i].x;
            else if( points[i].x > maxx ) maxx = points[i].x;
            if(      points[i].y < miny ) miny = points[i].y;
            else if( points[i].y > maxy ) maxy = points[i].y;
            if(      points[i].z < miny ) minz = points[i].z;
            else if( points[i].z > maxy ) maxz = points[i].z;
        }

        center.x = (maxx + minx) / 2.0f;
        center.y = (maxy + miny) / 2.0f;
        center.z = (maxz + minz) / 2.0f;

        float dx = maxx - minx;
        float dy = maxy - miny;
        float dz = maxz - minz;

        xformRadius = radius = (float)Math.sqrt( dx*dx + dy*dy + dz*dz ) / 2.0f;

        /*
        for( i=0 ; i<points.length ; i++ )
            center.add( points[i] );

        center.scale( 1.0f / points.length );

        radius = 0;
        for( i=0 ; i<points.length ; i++ ) {
            dist = Vec3.dist( points[i], center );
            if( dist > radius )
                radius = dist;
        }

        xformRadius = radius;
        */
    }

    private final static Vec4[] newVec4Array( int size )
    {
        Vec4 a[] = new Vec4[size];
        for( int i=0; i<size ; i++ ) a[i] = new Vec4();
        return a;
    }

    private final static Vec3[] newVec3Array( int size )
    {
        Vec3 a[] = new Vec3[size];
        for( int i=0; i<size ; i++ ) a[i] = new Vec3();
        return a;
    }

    public Obj( Vec3 points[], int faceIdx[][], Mat4 modelMatrix, Material material )
    {
        int i;

        this.points = points;
        this.modelMatrix = modelMatrix;
        this.material = material;

        centers = newVec4Array( faceIdx.length );
        normals = newVec3Array( faceIdx.length );
        xformNormals = newVec3Array( faceIdx.length );
        xformPoints  = newVec4Array( points.length );

        projected = new ScrPt[points.length];
        for( i=0 ; i<projected.length ; i++ )
            projected[i] = new ScrPt();

	this.faceIdx = faceIdx;
	/*
        this.faces = new Face[faceIdx.length];
        for( i=0 ; i<faceIdx.length ; i++ )
            this.faces[i] = new Face( faceIdx[i], this );
        */

        calculateNormals( faceIdx );
        calculateCenters( faceIdx );
        calculateBounds();
    }
    /*
    public final void addDetail( int i, Vec2 surface[], Material mat )
    {
        int face[] = faces[i].indecies;
        Vec3 up = new Vec3().sub( points[face[face.length-1]], points[face[0]] );
        faces[i].addDetail( surface, new Vec3(centers[i]), normals[i], up, mat );
    }
    */
    /*
    public final void project( Mat4 projectionMatrix, int half_width, int half_height )
    {
        int i,j;
        Vec4 t = new Vec4();

        for( i=0 ; i<points.length ; i++ ) {
            t.xform( projectionMatrix, xformPoints[i] );
            projected[i].x = (int)((t.x/t.w) * half_width) + half_width;
            projected[i].y = (int)((t.y/t.w) * half_height) + half_height;
            projected[i].z  = t.z/t.w;
        }
    }
    */
    public final void applyTransform()
    {
        int i;

        for( i=0 ; i<points.length ; i++ )
            xformPoints[i].xform( modelMatrix, points[i] );

        xformCenter.xform( modelMatrix, center );
        // xformRadius = radius * (float)Math.pow(modelMatrix.determinant3(), 1.0/3.0);
    }

    public final void drawBounds( Graphics g, View view )
    {
        Vec4 t = new Vec4();
        Vec3 rp = new Vec3( view.u );
        rp.scale( xformRadius );
        rp.add( xformCenter );

        t.xform( view.projectionMatrix, xformCenter );
        int x = (int)(t.x/t.w) + view.half_width;
        int y = (int)(t.y/t.w) + view.half_height;
        t.xform( view.projectionMatrix, rp );
        int x2 = (int)(t.x/t.w) + view.half_width;
        int rad = x - x2;

        g.drawOval( x - rad, y - rad, 2*rad, 2*rad );
    }

    public final void appendProjectedFacesToFaceList( View view, FaceArray list )
    {
        int i, j;
        Vec4 t = new Vec4();
	Mat4 vm = new Mat4();

        vm.mul( modelMatrix, view.lookat );
        xformCenter.xform( vm, center );

	//
	// Transform all the points
	//
	float minx = list.minx;
	float maxx = list.maxx;
	float miny = list.miny;
	float maxy = list.maxy;
	float minz = list.minz;
	float maxz = list.maxz;

        for( i=0 ; i<points.length ; i++ ) {
	    t.xform( modelMatrix, points[i] );
	    xformPoints[i].xform( view.lookat, t );

	    // xformPoints[i].xform( vm, points[i] );

	    if( t.x < minx ) minx = t.x;
	    if( t.x > maxx ) maxx = t.x;
	    if( t.y < miny ) miny = t.y;
	    if( t.y > maxy ) maxy = t.y;
	    if( t.z < minz ) minz = t.z;
	    if( t.z > maxz ) maxz = t.z;	    

            t.xform( view.projectionMatrix, xformPoints[i] );
            projected[i].x = (int)(t.x/t.w) + view.half_width;
            projected[i].y = (int)(t.y/t.w) + view.half_height;
            projected[i].z  = t.z/t.w;
        }

	list.minx = minx; list.maxx = maxx;
	list.miny = miny; list.maxy = maxy;
	list.minz = minz; list.maxz = maxz;
        list.ensureCapacity( list.length() + faceIdx.length );

	for( i=0 ; i<faceIdx.length ; i++ ) {
	    Face face = list.nextFace();
	    int idx[] = faceIdx[i];

	    face.normal.xformNormal( vm, normals[i] );
	    face.center.xform( vm, centers[i] );

	    if( !Options.BACKFACE_CULL ||
		(face.normal.x * face.center.x +
		 face.normal.y * face.center.y +
		 face.normal.z * face.center.z) >= 0.0 ) {

		list.allocFace();
		face.set( idx, xformPoints, projected, material );
		// face.setBounds();
	    }
	}		

	/*
        for( i=0 ; i<faces.length ; i++ ) {
            Face face = list.nextFace();

            face.normal.xformNormal( vm, normals[i] );
            face.center.xform( vm, centers[i] );

            // if( face.normal.z >= 0 )
            // if( !view.backfacing( face.center, face.normal ) )

	    if( !Options.BACKFACE_CULL ||
		(face.normal.x * face.center.x +
		 face.normal.y * face.center.y +
		 face.normal.z * face.center.z) >= 0.0 ) {
                face.setBounds();
                face.projectSurface( m, view.half_width, view.half_height );
                list.appendNoCheck( face );
            }
        }
	*/
    }
}
