/*
 * Vec.java - 3x1 Vector
 * Copyright (C) 1998 Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.ji3d;

import java.lang.Math;

final public class Vec implements Cloneable
{
    public float x, y, z;

    public Vec( float x, float y, float z ) { set( x, y, z ); }
    public Vec( Vec v )                     { set( v.x, v.y, v.z ); }
    public Vec() 		                    { zero(); }

    public Object clone()
    {
        return new Vec( this );
    }

    public Vec zero()
    {
        return set( 0.0f, 0.0f, 0.0f );
    }

    public Vec set( float x, float y, float z )
    {
        this.x = x;
        this.y = y;
        this.z = z;
        return this;
    }

    public Vec add( Vec v ) { x += v.x; y += v.y; z += v.z; return this; }
    public static Vec add( Vec a, Vec b ) { return new Vec(a).add(b); }

    public Vec subtract( Vec v ) { x -= v.x; y -= v.y; z -= v.z; return this; }
    public static Vec subtract( Vec a, Vec b ) { return new Vec(a).subtract(b); }

    public Vec multiply( Vec v ) { x *= v.x; y *= v.y; z *= v.z; return this; }
    public static Vec multiply( Vec a, Vec b ) { return new Vec(a).multiply(b); }

    public Vec scale( float s ) { x *= s; y *= s; z *= s; return this; }
    public static Vec scale( Vec v, float s ) { return new Vec(v).scale(s); }

    public static Vec cross( Vec a, Vec b ) {
        return new Vec( a.y * b.z - a.z * b.y,
                        a.z * b.x - a.x * b.z,
                	    a.x * b.y - a.y * b.x );
    }
    public float dot( Vec v ) { return (x * v.x) + (y * v.y) + (z * v.z); }
    public static float dot( Vec a, Vec b ) { return a.dot( b ); }

    public float length()
    {
        try {
            return (float)Math.sqrt( this.dot(this) );
        }
        catch( ArithmeticException e ) {
            return 0;
        }
    }

    public void normalize() throws ArithmeticException
    {
        float len = length();

        if( len <= 0.0 ) throw new ArithmeticException();

        scale( 1.0f / len );
    }

    public static Vec normalize( Vec v ) throws ArithmeticException
    {
        Vec n = new Vec(v);
        n.normalize();
        return n;
    }

    public String toString() { return "( " + x + ", " + y + ", " + z + " )"; }

    public boolean equals( Object anObject ) {
        if((anObject != null) && (anObject instanceof Vec)) {
            Vec v = (Vec)anObject;
            return (v.x == x) && (v.y == y) && (v.z == z);
        }
        else
            return false;
    }

/*
    public static boolean Test() {
        Vector3D a = new Vector3D( (float)Math.random(), (float)Math.random(), (float)Math.random() );
        Vector3D b = new Vector3D( (float)Math.random(), (float)Math.random(), (float)Math.random() );
        Vector3D c = new Vector3D( (float)Math.random(), (float)Math.random(), (float)Math.random() );

        if( a.length() * 3.0f != scale( a, 3.0f ).length() )
            return false;
        if( ! add( a, a ).equals( scale( a, 2.0f ) ) )
            return false;
        if( subtract( a, a ).length() != 0.0f )
            return false;
        if( ! multiply( a, new Vector3D(2,2,2) ).equals( scale( a, 2 ) ) )
            return false;

        return true;
    }

    public static void main( String argv[] ) {
        if( Vector3D.Test() )
            System.out.print( "Vector3D test passed.\n" );
        else
            System.out.print( "Vector3D test failed.\n" );

        System.exit(0);
    }
*/
}
