/*
 * Vec2.java - 2x1 vector
 * Copyright (C) 1998 Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.ji3d;

import java.lang.Math;

final public class Vec2 extends Base implements Cloneable
{
    public float x, y;

    public Vec2( float x, float y ) { set( x, y ); }
    public Vec2( Vec2 v )           { set( v.x , v.y ); }
    public Vec2()                   {}

    public final Object clone() {
        return new Vec2( this );
    }

    public final Vec2 zero() {
        return set( 0.0f, 0.0f );
    }

    public final Vec2 set( float x, float y ) {
        this.x = x;
        this.y = y;
        return this;
    }

    public final Vec2 add( Vec2 a, Vec2 b ) { x = a.x + b.x; y = a.y + b.y; return this; }
    public final Vec2 add( Vec2 v )         { x += v.x; y += v.y; return this; }

    public final Vec2 sub( Vec2 a, Vec2 b ) { x = a.x - b.x; y = a.y - b.y; return this; }
    public final Vec2 sub( Vec2 v )         { x -= v.x; y -= v.y; return this; }

    public final Vec2 mul( Vec2 a, Vec2 b ) { x = a.x * b.x; y = a.y * b.y; return this; }
    public final Vec2 mul( Vec2 v )         { x *= v.x; y *= v.y; return this; }

    public final Vec2 scale( float s )  { x *= s; y *= s; return this; }
    public final Vec2 scale( Vec2 v, float s ) { x = v.x*s; y = v.y*s; return this; }

    public final float dot( Vec2 v ) { return (x * v.x) + (y * v.y); }
    public final static float dot( Vec2 a, Vec2 b ) { return (a.x*b.x) + (a.y*b.y); }

    public final float length() {
        try {
            return (float)Math.sqrt( this.dot( this ) );
        } catch( ArithmeticException e) {
            return 0.0f;
        }
    }

    public final void normalize() throws ArithmeticException
    {
        float len = length();
        if( len <= 0.0f ) throw new ArithmeticException( "zero length vector" );
        scale( 1.0f / len );
    }

    public final void normalize( Vec2 v ) throws ArithmeticException
    {
        float len = v.length();
        if( len <= 0.0f ) throw new ArithmeticException( "zero length vector" );
        scale( v, 1.0f / len );
    }

    public final String toString() { return "( " + x + ", " + y + " )"; }

    public final boolean equals( Object anObject ) {
        if( (anObject != null) && (anObject instanceof Vec2) ) {
            Vec2 v = (Vec2)anObject;
            return (v.x == x) && (v.y == y);
        } else
            return false;
    }
}
