/*
 * Vec3 - 3x1 vector
 * Copyright (C) 1998 Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.ji3d;

import java.lang.Math;

final public class Vec3 extends Base implements Cloneable
{
    public float x, y, z;

    public Vec3( float x, float y, float z ) { set( x, y, z ); }
    public Vec3( Vec3 v )                    { set( v.x, v.y, v.z ); }
    public Vec3( Vec4 v )                    { set( v.x/v.w, v.y/v.w, v.z/v.w ); }
    public Vec3() 		                     {}

    public final Object clone() {
        return new Vec3( this );
    }

    public final Vec3 zero() {
        return set( 0.0f, 0.0f, 0.0f );
    }

    public final Vec3 set( Vec4 v ) {
        return set( v.x/v.w, v.y/v.w, v.z/v.w );
    }

    public final Vec3 set( float x, float y, float z ) {
        this.x = x;
        this.y = y;
        this.z = z;
        return this;
    }

    public final Vec3 add( Vec3 a, Vec3 b ) { x = a.x+b.x; y = a.y+b.y; z = a.z+b.z; return this; }
    public final Vec3 add( Vec3 v ) { x += v.x; y += v.y; z += v.z; return this; }

    public final Vec3 sub( Vec3 v ) { x -= v.x; y -= v.y; z -= v.z; return this; }
    public final Vec3 sub( Vec3 a, Vec3 b ) { x = a.x-b.x; y = a.y-b.y; z = a.z-b.z; return this; }
    public final Vec3 sub( Vec4 a, Vec4 b ) { x = a.x-b.x; y = a.y-b.y; z = a.z-b.z; return this; }

    public final Vec3 mul( Vec3 v ) { x *= v.x; y *= v.y; z *= v.z; return this; }
    public final Vec3 mul( Vec3 a, Vec3 b ) { x = a.x*b.x; y = a.y*b.y; z = a.z*b.z; return this; }

    public final Vec3 scale( float s ) { x *= s; y *= s; z *= s; return this; }
    public final Vec3 scale( Vec3 v, float s ) { x = v.x*s; y = v.y*s; z = v.z*s; return this; }

    public final Vec3 cross( Vec3 a, Vec3 b ) {
        x = a.y * b.z - a.z * b.y;
        y = a.z * b.x - a.x * b.z;
        z = a.x * b.y - a.y * b.x;
        return this;
    }

    public final float dot( Vec3 v ) { return (x * v.x) + (y * v.y) + (z * v.z); }
    public final static float dot( Vec3 a, Vec3 b ) { return a.dot( b ); }

    public final float length() {
        try {
            return (float)Math.sqrt( this.dot(this) );
        } catch( ArithmeticException e ) {
            return 0.0f;
        }
    }

    public final static float dist( Vec3 a, Vec3 b ) {
        float dx = a.x - b.x;
        float dy = a.y - b.y;
        float dz = a.z - b.z;

        try {
            return (float)Math.sqrt( dx*dx + dy*dy + dz*dz );
        } catch( ArithmeticException e ) {
            return 0;
        }
    }

    public final void normalize() throws ArithmeticException
    {
        float len = length();

        if( len <= 0.0 ) throw new ArithmeticException();

        scale( 1.0f / len );
    }

    public final void normalize( Vec3 v ) throws ArithmeticException
    {
        float len = v.length();
        if( len <= 0.0 ) throw new ArithmeticException();
        this.scale( v, 1.0f / len );
    }

    public final Vec3 xform( Mat4 m, Vec2 v ) {
        x = ( v.x * m.xx + v.y * m.yx + m.wx );
        y = ( v.x * m.xy + v.y * m.yy + m.wy );
        z = ( v.x * m.xz + v.y * m.yz + m.wz );
        return this;
    }
    public final Vec3 xform( Mat4 m, Vec3 v ) {
        x = ( v.x * m.xx + v.y * m.yx + v.z * m.zx + m.wx );
        y = ( v.x * m.xy + v.y * m.yy + v.z * m.zy + m.wy );
        z = ( v.x * m.xz + v.y * m.yz + v.z * m.zz + m.wz );
        return this;
    }

    public final Vec3 xform( Mat4 m, Vec4 v ) {
        float w;
        x = ( v.x * m.xx + v.y * m.yx + v.z * m.zx + v.w * m.wx );
        y = ( v.x * m.xy + v.y * m.yy + v.z * m.zy + v.w * m.wy );
        z = ( v.x * m.xz + v.y * m.yz + v.z * m.zz + v.w * m.wz );
        w = ( v.x * m.xw + v.y * m.yw + v.z * m.zw + v.w * m.ww );
        scale( 1.0f / w );
        return this;
    }

    public final Vec3 xformNormal( Mat4 m, Vec3 v ) {
        x = ( v.x * m.xx + v.y * m.yx + v.z * m.zx );
        y = ( v.x * m.xy + v.y * m.yy + v.z * m.zy );
        z = ( v.x * m.xz + v.y * m.yz + v.z * m.zz );
        return this;
    }

    public final Vec3 xformNormalT( Mat4 m, Vec3 v ) {
        x = ( v.x * m.xx + v.y * m.xy + v.z * m.xz );
        y = ( v.x * m.yx + v.y * m.yy + v.z * m.yz );
        z = ( v.x * m.zx + v.y * m.zy + v.z * m.zz );
        return this;
    }

    public final String toString() { return "( " + x + ", " + y + ", " + z + " )"; }

    public final boolean equals( Object anObject ) {
        if((anObject != null) && (anObject instanceof Vec3)) {
            Vec3 v = (Vec3)anObject;
            return (v.x == x) && (v.y == y) && (v.z == z);
        }
        else
            return false;
    }
}
