/*
 * Vec4.java - 4x1 Homogeneous vector
 * Copyright (C) 1998 Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.ji3d;

import java.lang.Math;

final public class Vec4 extends Base implements Cloneable
{
    public float x, y, z, w;

    public Vec4( float x, float y, float z, float w ) { set( x, y, z, w ); }
    public Vec4( float x, float y, float z )          { set( x, y, z ); }
    public Vec4( Vec4 v )                             { set( v.x, v.y, v.z, v.w ); }
    public Vec4( Vec3 v )                             { set( v.x, v.y, v.z ); }
    public Vec4() 		                              { zero(); }

    public final Object clone() { return new Vec4( this ); }

    public final Vec4 zero() { return set( 0.0f, 0.0f, 0.0f ); }

    public final Vec4 set( float x, float y, float z ) {
        return set( x, y, z, 1.0f );
    }

    public final Vec4 set( float x, float y, float z, float w ) {
        this.x = x;
        this.y = y;
        this.z = z;
        this.w = w;
        return this;
    }

    public final Vec4 add( Vec4 a, Vec3 b ) { x = a.x+b.x; y = a.y+b.y; z = a.z+b.z; return this; }
    public final Vec4 add( Vec3 v ) { x += v.x; y += v.y; z += v.z; return this; }
    public final Vec4 add( Vec4 v ) { x += v.x; y += v.y; z += v.z; return this; }

    public final Vec4 sub( Vec3 v ) { x -= v.x; y -= v.y; z -= v.z; return this; }
    public final Vec4 sub( Vec4 a, Vec3 b ) { x = a.x-b.x; y = a.y-b.y; z = a.z-b.z; return this; }

    public final Vec4 scale( float s ) { x *= s; y *= s; z *= s; return this; }
    public final Vec4 scale( Vec4 v, float s ) { x = v.x*s; y = v.y*s; z = v.z*s; return this; }

    public final float dot( Vec4 v ) { return x*v.x + y*v.y + z*v.z + w*v.w; }
    public final static float dot( Vec4 a, Vec4 b ) { return a.dot( b ); }

    public final Vec4 xform( Mat4 m, Vec4 v ) {
        x = ( v.x * m.xx + v.y * m.yx + v.z * m.zx + v.w * m.wx );
        y = ( v.x * m.xy + v.y * m.yy + v.z * m.zy + v.w * m.wy );
        z = ( v.x * m.xz + v.y * m.yz + v.z * m.zz + v.w * m.wz );
        w = ( v.x * m.xw + v.y * m.yw + v.z * m.zw + v.w * m.ww );
        return this;
    }

    public final Vec4 xformT( Mat4 m, Vec4 v ) {
        x = ( v.x * m.xx + v.y * m.xy + v.z * m.xz + v.w * m.xw );
        y = ( v.x * m.yx + v.y * m.yy + v.z * m.yz + v.w * m.yw );
        z = ( v.x * m.zx + v.y * m.zy + v.z * m.zz + v.w * m.zw );
        w = ( v.x * m.wx + v.y * m.wy + v.z * m.wz + v.w * m.ww );
        return this;
    }

    public final Vec4 xform( Mat4 m, Vec3 v ) {
        x = ( v.x * m.xx + v.y * m.yx + v.z * m.zx + m.wx );
        y = ( v.x * m.xy + v.y * m.yy + v.z * m.zy + m.wy );
        z = ( v.x * m.xz + v.y * m.yz + v.z * m.zz + m.wz );
        w = ( v.x * m.xw + v.y * m.yw + v.z * m.zw + m.ww );
        return this;
    }

    public final Vec4 xformT( Mat4 m, Vec3 v ) {
        x = ( v.x * m.xx + v.y * m.xy + v.z * m.xz + m.xw );
        y = ( v.x * m.yx + v.y * m.yy + v.z * m.yz + m.yw );
        z = ( v.x * m.zx + v.y * m.zy + v.z * m.zz + m.zw );
        w = ( v.x * m.wx + v.y * m.wy + v.z * m.wz + m.ww );
        return this;
    }

    public final String toString() { return "( " + x + ", " + y + ", " + z + ", " + w + " )"; }

    public final boolean equals( Object anObject ) {
        if((anObject != null) && (anObject instanceof Vec4)) {
            Vec4 v = (Vec4)anObject;
            float dw = v.w / w;
            return (v.x == x * dw) && (v.y == y * dw) && (v.z == z * dw);
        }
        else
            return false;
    }
}
