/*
 * View.java - View coordinates
 * Copyright (C) 1998 Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.ji3d;

import java.lang.Math;

public final class View
{
    // compile time parameters
    public static final float viewAngle = 45.0f;

    // precalculations
    public static final float  cosY = (float)Math.cos( viewAngle * Math.PI / 360.0 );
    public static final float  sinY = (float)Math.sin( viewAngle * Math.PI / 360.0 );
    public static final double tanAngle = Math.tan( viewAngle * Math.PI / 360.0);

    // view parameters
    public Vec3 from;
    public Vec3 at;
    public Vec3 up;
    public int width;
    public int height;

    // projection matrix
    public Mat4 projectionMatrix = new Mat4();
    public Mat4 lookat = new Mat4();
    public int half_width;
    public int half_height;

    // clipping planes
    public Vec3 clips[] = new Vec3[5];
    public Vec3 frompt = new Vec3();

    // viewer coordinate system
    public Vec3 n = new Vec3();
    public Vec3 u = new Vec3();
    public Vec3 v = new Vec3();

    public View( Vec3 from, Vec3 at, Vec3 up, int width, int height )
    {
        this();
        set( from, at, up, width, height );
    }

    public View()
    {
        int i;
        for( i=0 ; i<clips.length ; i++ )
            clips[i] = new Vec3();
    }

    public void set( Vec3 from, Vec3 at, Vec3 up, int width, int height )
    {
        this.from = from;
        this.at = at;
        this.up = up;
        this.width = width;
        this.height = height;
        this.half_width = width / 2;
        this.half_height = height / 2;

        float aspect = (float)width / height;

        /*
         * viewer coordinate system (copied from Mat4.lookat())
         */
        n.sub( at, from ); n.normalize();
        u.cross( n, up );  u.normalize();
        v.cross( u, n );   v.normalize();

        lookat.set( u.x,  u.y,  u.z,  0.0f,
                    v.x,  v.y,  v.z,  0.0f,
                    n.x,  n.y,  n.z,  0.0f,
                    0.0f, 0.0f, 0.0f, 1.0f );
        lookat.appendTranslate( -from.x, -from.y, -from.z );

        /*
         * perspective transform
         */
        projectionMatrix.perspective( viewAngle, aspect, 1, 10 );
        projectionMatrix.appendScale( half_width, half_height, 1.0f );
        // projectionMatrix.prepend( lookat );        // Warning: GC

        /*
         * clip planes
         */
        frompt.x = from.x * u.x + from.y * v.x + from.z * n.x;
        frompt.y = from.x * u.y + from.y * v.y + from.z * n.y;
        frompt.z = from.x * u.z + from.y * v.z + from.z * n.z;

        float snx = sinY*n.x, sny = sinY*n.y, snz = sinY*n.z;
        float cvx = cosY*v.x, cvy = cosY*v.y, cvz = cosY*v.z;

        clips[0].x = snx + cvx;
        clips[0].y = sny + cvy;
        clips[0].z = snz + cvz;

        clips[1].x = snx - cvx;
        clips[1].y = sny - cvy;
        clips[1].z = snz - cvz;

        double hfov = Math.atan( aspect * tanAngle );
        float s = (float)Math.sin( hfov );
        float c = (float)Math.cos( hfov );
        snx = s*n.x; sny = s*n.y; snz = s*n.z;
        cvx = c*u.x; cvy = c*u.y; cvz = c*u.z;

        clips[2].x = snx + cvx;
        clips[2].y = sny + cvy;
        clips[2].z = snz + cvz;

        clips[3].x = snx - cvx;
        clips[3].y = sny - cvy;
        clips[3].z = snz - cvz;

        clips[4].x = n.x;
        clips[4].y = n.y;
        clips[4].z = n.z;
    }

    public static final int
                    NOT_VISIBLE = 0,
                    VISIBLE     = 1,
                    NEEDS_CLIPPING = 2;

    public int visibility( Obj object )
    {
        int i;
        Vec3 c = new Vec3().sub( object.xformCenter, from );

        for( i=0 ; i<4 ; i++ ) {
            if( (c.x * clips[i].x) + (c.y * clips[i].y) + (c.z * clips[i].z)
                < -object.xformRadius )
                return NOT_VISIBLE;
        }

        float dot = (c.x * clips[4].x) + (c.y * clips[4].y) + (c.z * clips[4].z);

        if( dot < -object.xformRadius )
            return NOT_VISIBLE;
        else if( dot < object.xformRadius )
            return NEEDS_CLIPPING;
        else
            return VISIBLE;
    }

    public boolean backfacing( Vec4 center, Vec3 normal )
    {
        float dx, dy, dz;

        dx = center.x - from.x;
        dy = center.y - from.y;
        dz = center.z - from.z;

        return (normal.x * dx + normal.y * dy + normal.z * dz) < 0.0f;
    }
}
