/*
 * Main.java - Roller coaster generation
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.makeRoller;

import java.io.PrintWriter;
import java.io.FileWriter;
import java.io.FileReader;
import java.io.StreamTokenizer;
import java.io.IOException;
import berkeley.cs.dmc.bspline.BSpline;

/**
 * An application that takes a track description and generates a VRML and Java
 * roller coaster.
 *
 * @author Stephen Chenney
 * @version 1.0
 */
public class Main {
    static String eoln = System.getProperty("line.separator");

    static String track_header =
        "#VRML V2.0 utf8" + eoln
      + "PROTO Track [] {" + eoln
      + " Group {" + eoln
      + "  children [";

    static String sub_track_head =
        "    Shape {" + eoln
      + "      geometry IndexedFaceSet {" + eoln
      + "        solid FALSE" + eoln
      + "        coord Coordinate {" + eoln
      + "          point [";

    static String sub_track_tail =
	"        ]" + eoln
      + "      }" + eoln
      + "      appearance Appearance {" + eoln
      + "        material Material {" + eoln
      + "          diffuseColor 0.0 1.0 0.0" + eoln
      + "          ambientIntensity 1.0" + eoln
      + "        }" + eoln
      + "      }" + eoln
      + "    }";

    static String track_tail =
        "  ]" + eoln
      + " }" + eoln
      + "}" + eoln;

    static String sys_header =
	"import berkeley.cs.dmc.system.DynamicsSystem;" + eoln
      + "import berkeley.cs.dmc.numericalMethods.Integrator;" + eoln
      + "import berkeley.cs.dmc.numericalMethods.Derivative;" + eoln
      + "import berkeley.cs.dmc.bspline.BSpline;" + eoln
      + "public class RollerSystem implements DynamicsSystem {" + eoln
      + "  BSpline     pspline;" + eoln
      + "  BSpline     nspline;" + eoln
      + "  private Derivative deriv = null;" + eoln
      + "  private Integrator integ = null;" + eoln
      + "  private double[] params = new double[2];" + eoln
      + "  private double[] a = new double[3];" + eoln
      + "  private double[] v = new double[3];" + eoln
      + "  private float[] n = new float[3];" + eoln
      + "  private float[] t = new float[3];" + eoln
      + "  private float[] pt = new float[3];" + eoln
      + "  private static final double INIT_VEL = 2.0;" + eoln
      + "  private double energy = 0.0;" + eoln;

    static String sys_mid =
        "  public int numVariables() { return 2; }" + eoln
      + "  public int numFields() { return 11; }" + eoln
      + "  public int subsetLength() { return 2; }" + eoln
      + "  public int evaluationTechnique(int i) {" + eoln
      + "    if ( i == 0 )" + eoln
      + "      return ITERATIVE;" + eoln
      + "    else" + eoln
      + "      return FIRST_DERIVATIVE;" + eoln
      + "  }" + eoln
      + "  public double period() { return 0.0; }" + eoln
      + "  public boolean isCyclic(int ind) {" + eoln
      + "    if ( ind == 0 )" + eoln
      + "      return true;" + eoln
      + "    else" + eoln
      + "      return false;" + eoln
      + "  }" + eoln
      + "  public double[] getBound(int ind) {" + eoln
      + "    double[] res = new double[2];" + eoln
      + "    if ( ind == 0 ) {" + eoln
      + "      res[0] = 0.0; res[1] = PERIOD;" + eoln
      + "    } else {" + eoln
      + "      res[0] = 0.0; res[1] = 1.0;" + eoln
      + "    }" + eoln
      + "    return res;" + eoln
      + "  }" + eoln
      + "  public boolean initialCondition(double[] p) {";

    static String sys_tail =
        "    float[] dp = new float[3];" + eoln
      + "    pspline.evaluateDpDu((float)p[0], dp);" + eoln
      + "    p[1] = Math.sqrt(INIT_VEL*INIT_VEL"
      + "/(dp[0]*dp[0]+dp[1]*dp[1]+dp[2]*dp[2]));" + eoln
      + "    return false;" + eoln
      + "  }" + eoln
      + "  public float evaluate(double t_last, double[] p_last," + eoln
      + "                        double t_now, double[] p_now) {" + eoln
      + "    params[0] = p_last[0];" + eoln
      + "    params[1] = p_last[1];" + eoln
      + "    double  min_step = 1.0e-4;" + eoln
      + "    if ( t_now - t_last < min_step )" + eoln
      + "      min_step = t_now - t_last;" + eoln
      + "    integ.ODEIntegrate(params, t_last, t_now, 1.0e-3, 0.01,min_step);"
      + eoln
      + "    float   u = (float)Math.IEEEremainder(params[0], PERIOD);" + eoln
      + "    if ( u < 0.0f )" + eoln
      + "       u += PERIOD;" + eoln
      + "    if ( u >= PERIOD )" + eoln
      + "       u -= PERIOD;" + eoln
      + "    float[] x = new float[3];" + eoln
      + "    pspline.evaluate(u, x);" + eoln
      + "    float[] dp = new float[3];" + eoln
      + "    pspline.evaluateDpDu(u, dp);" + eoln
      + "    float t_dot = dp[0] * dp[0] + dp[1] * dp[1] + dp[2] * dp[2];" +eoln
      + "    double v_req_sq = 2.0 * ( energy - x[1] * 9.81 );" + eoln
      + "    params[1] = Math.sqrt(v_req_sq / t_dot);" + eoln
      + "    p_now[0] = params[0];" + eoln
      + "    p_now[1] = params[1];" + eoln
      + "    return 1.0f;" + eoln
      + "  }" + eoln
      + "  public void evaluateFields(double time," + eoln
      + "                     double[] p_in, float[] f_out) {" + eoln
      + "    float u = (float)Math.IEEEremainder(p_in[0], PERIOD);" + eoln
      + "    if ( u < 0.0f )" + eoln
      + "      u += PERIOD;" + eoln
      + "    if ( u >= PERIOD )" + eoln
      + "      u -= PERIOD;" + eoln
      + "    pspline.evaluate(u, pt);" + eoln
      + "    f_out[0] = (float)pt[0];" + eoln
      + "    f_out[1] = (float)pt[1];" + eoln
      + "    f_out[2] = (float)pt[2];" + eoln
      + "    pspline.evaluateDpDu(u, t);" + eoln
      + "    double l = Math.sqrt(t[0] * t[0] + t[1] * t[1] + t[2] * t[2]);"
      + eoln
      + "    a[0] = 0.0;" + eoln
      + "    a[1] = -t[2] / l;" + eoln
      + "    a[2] = t[1] / l;" + eoln
      + "    double sin_a = Math.sqrt(a[1] * a[1] + a[2] * a[2]);" + eoln
      + "    if ( sin_a < 1.0e-5 && sin_a > -1.0e-5 ) {" + eoln
      + "      f_out[7] = 1.0f;" + eoln
      + "      f_out[8] = 0.0f;" + eoln
      + "      f_out[9] = 0.0f;" + eoln
      + "      f_out[10] = 0.0f;" + eoln
      + "    } else {" + eoln
      + "      f_out[7] = 0.0f;" + eoln
      + "      f_out[8] = (float)(a[1] / sin_a);" + eoln
      + "      f_out[9] = (float)(a[2] / sin_a);" + eoln
      + "      if ( t[0] >= 0.0 )" + eoln
      + "        f_out[10] = (float)Math.asin(sin_a);" + eoln
      + "      else" + eoln
      + "        f_out[10] = (float)(Math.PI - Math.asin(sin_a));" + eoln
      + "    }" + eoln
      + "    nspline.evaluate(u, n);" + eoln
      + "    v[0] = n[1] * t[2] - n[2] * t[1];" + eoln
      + "    v[1] = n[2] * t[0] - n[0] * t[2];" + eoln
      + "    v[2] = n[0] * t[1] - n[1] * t[0];" + eoln
      + "    n[0] = (float)(t[1] * v[2] - t[2] * v[1]);" + eoln
      + "    n[1] = (float)(t[2] * v[0] - t[0] * v[2]);" + eoln
      + "    n[2] = (float)(t[0] * v[1] - t[1] * v[0]);" + eoln
      + "    l = Math.sqrt(n[0] * n[0] + n[1] * n[1] + n[2] * n[2]);" + eoln
      + "    n[0] /= l;" + eoln
      + "    n[1] /= l;" + eoln
      + "    n[2] /= l;" + eoln
      + "    double cos_a = ( 1.0 - Math.cos(f_out[10]) );" + eoln
      + "    v[0] = a[1];" + eoln
      + "    v[1] = f_out[8] * f_out[9] * cos_a;" + eoln
      + "    v[2] = 1.0 - f_out[8] * f_out[8] * cos_a;" + eoln
      + "    a[0] = v[1] * n[2] - n[1] * v[2];" + eoln
      + "    a[1] = v[2] * n[0] - n[2] * v[0];" + eoln
      + "    a[2] = v[0] * n[1] - n[0] * v[1];" + eoln
      + "    sin_a = Math.sqrt(a[0] * a[0] + a[1] * a[1] + a[2] * a[2]);" + eoln
      + "    if ( sin_a < 1.0e-5 && sin_a > -1.0e-5 ) {" + eoln
      + "      f_out[3] = 1.0f;" + eoln
      + "      f_out[4] = 0.0f;" + eoln
      + "      f_out[5] = 0.0f;" + eoln
      + "      f_out[6] = 0.0f;" + eoln
      + "    } else {" + eoln
      + "      f_out[3] = (float)(a[0] / sin_a);" + eoln
      + "      f_out[4] = (float)(a[1] / sin_a);" + eoln
      + "      f_out[5] = (float)(a[2] / sin_a);" + eoln
      + "      if ( v[0] * n[0] + v[1] * n[1] + v[2] * n[2] >= 0.0 )" + eoln
      + "        f_out[6] = (float)Math.asin(sin_a);" + eoln
      + "      else" + eoln
      + "        f_out[6] = (float)(Math.PI - Math.asin(sin_a));" + eoln
      + "    }" + eoln
      + "  }" + eoln
      + "  public boolean checkState(double t, double[] p, boolean discard) {"
      + eoln
      + "    return true;" + eoln
      + "  }" + eoln
      + "  public float bufferStep() { return 0.1f; }" + eoln
      + "  public int bufferSize() { return 100; }" + eoln
      + "  public double[] resolutions() {" + eoln
      + "    double[] res = new double[2];" + eoln
      + "    res[0] = 0.2;" + eoln
      + "    res[1] = 0.2;" + eoln
      + "    return res;" + eoln
      + "  }" + eoln
      + "  public double mixTestMax() { return 0.5; }" + eoln
      + "  public double mixDiff() { return 1.0; }" + eoln
      + "  public double maxApproxError() { return 0.001; }" + eoln
      + "}" + eoln + eoln;


    static void readControls(StreamTokenizer in, float[][] tc, float[][] nc,
			     int num)
	throws IOException
    {
	for ( int i = 0 ; i < num ; i++ )
	{
	    tc[i] = new float[3];
	    for ( int j = 0 ; j < 3 ; j++ )
	    {
		in.nextToken();
		tc[i][j] = (float)in.nval;
	    }
	    nc[i] = new float[3];
	    for ( int j = 0 ; j < 3 ; j++ )
	    {
		in.nextToken();
		nc[i][j] = (float)in.nval;
	    }
	}
	for ( int i = num ; i < num + 3 ; i++ )
	{
	    tc[i] = tc[i - num];
	    nc[i] = nc[i - num];
	}
    }

    static void readLifters(StreamTokenizer in, float[][] l, int num)
	throws IOException
    {
	for ( int i = 0 ; i < num ; i++ )
	{
	    l[i] = new float[3];
	    for ( int j = 0 ; j < 3 ; j++ )
	    {
		in.nextToken();
		l[i][j] = (float)in.nval;
	    }
	}
    }


    private static float writeTrack(PrintWriter trackfile, BSpline pspline,
				    BSpline nspline, int num_c)
    {
	int         count = 0;
	float       high_pt = 0.0f;
	float       high_u = 0.0f;

	trackfile.println(track_header);

	float[] pt = new float[3];
	float[] tang = new float[3];
	float[] norm = new float[3];
	float[] cross = new float[3];

	for ( float v = 0.0f ; v < num_c - 0.1f ; v += 1.0f )
	{
	    trackfile.println(sub_track_head);
	    count = 0;
	    for ( float u = 0.0f ; u <= 1.0001f ; u += 0.1f )
	    {
	      try {
	        if ( v + u >= num_c )
		    u = 0.999f;
		pspline.evaluate(v + u, pt);
		pspline.evaluateDpDu(v + u, tang);
		nspline.evaluate(v + u, norm);
		cross[0] = tang[1] * norm[2] - tang[2] * norm[1];
		cross[1] = tang[2] * norm[0] - tang[0] * norm[2];
		cross[2] = tang[0] * norm[1] - tang[1] * norm[0];
		float length = (float)Math.sqrt(cross[0] * cross[0]
		                              + cross[1] * cross[1]
		                              + cross[2] * cross[2]);
		cross[0] /= length;
		cross[1] /= length;
		cross[2] /= length;
		trackfile.println("            " + ( pt[0] + 0.4 * cross[0] )
		                           + " " + ( pt[1] + 0.4 * cross[1] )
					   + " " + ( pt[2] + 0.4 * cross[2] ));
		trackfile.println("            " + ( pt[0] - 0.4 * cross[0] )
		                           + " " + ( pt[1] - 0.4 * cross[1] )
					   + " " + ( pt[2] - 0.4 * cross[2] ));
		if ( pt[1] > high_pt )
		{
		    high_pt = pt[1];
		    high_u = v + u;
		}
		count++;
	      } catch (ArithmeticException e) {
		System.err.println(e);
	      }
	    }
	    if ( norm[1] > 0.0 )
	    {
		trackfile.println("            " + ( pt[0] + 0.4 * cross[0] )
				       + " 0.0 " + ( pt[2] + 0.4 * cross[2] ));
		trackfile.println("            " + ( pt[0] - 0.4 * cross[0] )
				       + " 0.0 " + ( pt[2] - 0.4 * cross[2] ));
	    }
	    trackfile.println("          ]");
	    trackfile.println("        }");
	    trackfile.println("        coordIndex [");

	    for ( int i = 0 ; i < count - 1 ; i++ )
		trackfile.println("          " + ( 2 * i )
		                         + " " + ( 2 * i + 2 )
					 + " " + ( 2 * i + 1 )
				      + " -1 " + ( 2 * i + 1 )
				         + " " + ( 2 * i + 2 )
					 + " " + ( 2 * i + 3 ) + " -1");
	    if ( norm[1] > 0.0 )
		trackfile.println("          " + ( 2 * count )
					 + " " + ( 2 * count - 2 )
					 + " " + ( 2 * count + 1 )
				      + " -1 " + ( 2 * count - 2 )
					 + " " + ( 2 * count - 1 )
					 + " " + ( 2 * count + 1 ) + " -1");

	    trackfile.println(sub_track_tail);
	}

	trackfile.println(track_tail);

	return high_u;
    }


    /**
     * Generate a roller coaster given the control points and lifter
     * information.
     *
     * If there are to be n control points for the track, t_control_pts
     * is an (n+3) by 3 array of the track location controls, and
     * n_control_pts is an (n+3) by 3 array of the normal control points.
     * The last 3 entries of each array need not be set. They are copied
     * from the first 3 to loop the track.
     * Lifters is a k by 3 array of lifter data. See main() for
     * information on the files generated.
     *
     * @param t_control_pts The track position controls.
     * @param n_control_pts The track normal controls.
     * @param lifters The lifter data.
     */
    public static void writeRoller(float[][] t_control_pts,
				   float[][] n_control_pts,
				   float[][] lifters)
    {
	PrintWriter trackfile =  null;
	PrintWriter sysfile =  null;

	int num_c = t_control_pts.length - 3;
	int num_l = lifters == null ? 0 : lifters.length;

	for ( int i = 0 ; i < 3 ; i++ )
	{
	    t_control_pts[num_c+i] = t_control_pts[i];
	    n_control_pts[num_c+i] = n_control_pts[i];
	}

	BSpline     pspline = new BSpline(3, num_c + 3, t_control_pts);
	BSpline     nspline = new BSpline(3, num_c + 3, n_control_pts);
	int         count = 0;

	float       high_u = 0.0f;

	try {
	    trackfile = new PrintWriter(new FileWriter("Track.wrl"));
	    sysfile = new PrintWriter(new FileWriter("RollerSystem.java"));
	} catch (Exception e) {
	    System.err.println(e);
	    System.exit(1);
	}

	high_u = writeTrack(trackfile, pspline, nspline, num_c); 

	sysfile.println(sys_header);

	sysfile.println("  static final double PERIOD = " + num_c + ".0;");
	sysfile.println("  private float[][]   t_control_pts = {");
	for ( int i = 0 ; i < num_c + 2 ; i++ )
	    sysfile.println("    { " + t_control_pts[i][0] + "f, "
	                             + t_control_pts[i][1] + "f, "
				     + t_control_pts[i][2] + "f},");
	sysfile.println("    { "
	              + t_control_pts[num_c+2][0] + "f, "
		      + t_control_pts[num_c+2][1] + "f, "
		      + t_control_pts[num_c+2][2] + "f }");
	sysfile.println("  };");
	sysfile.println("  private float[][]   n_control_pts = {");
	for ( int i = 0 ; i < num_c + 2 ; i++ )
	    sysfile.println("    { " + n_control_pts[i][0] + "f, "
	                             + n_control_pts[i][1] + "f, "
				     + n_control_pts[i][2] + "f},");
	sysfile.println("    { "
	              + n_control_pts[num_c + 2][0] + "f, "
		      + n_control_pts[num_c + 2][1] + "f, "
		      + n_control_pts[num_c + 2][2] + "f }");
	sysfile.println("  };");
	sysfile.println("  public float[][] lifters = "
		      + ( num_l == 0 ? "null;" : "{"));
	if ( num_l > 0 )
	{
	    for ( int i = 0 ; i < num_l - 1 ; i++ )
		sysfile.println("    { " + lifters[i][0] + "f, "
					 + lifters[i][1] + "f, "
					 + lifters[i][2] + "f},");
	    sysfile.println("    { "
			  + lifters[num_l - 1][0] + "f, "
			  + lifters[num_l - 1][1] + "f, "
			  + lifters[num_l - 1][2] + "f }");
	    sysfile.println("  };");
	}
	sysfile.println("  public RollerSystem() {");
	sysfile.println("    pspline = new BSpline(3, " + ( num_c + 3 )
		      + ", t_control_pts);");
        sysfile.println("    float u = 9.0f;");
	sysfile.println("    float[] x = new float[3];");
	sysfile.println("    pspline.evaluate(u, x);");
	sysfile.println("    energy = x[1] * 9.81 + 0.5 * INIT_VEL*INIT_VEL;");

	sysfile.println("    nspline = new BSpline(3, " + ( num_c + 3 )
		      + ", n_control_pts);");
	sysfile.println("    deriv = new RollerDerivative(this);");
	sysfile.println("    integ = new Integrator(deriv);");
	sysfile.println("  }");
	sysfile.println(sys_mid);
        sysfile.println("    p[0] = " + high_u + ";");
	sysfile.println(sys_tail);

	try {
	    trackfile.close();
	    sysfile.close();
	} catch (Exception e) {
	}
    }

    /**
     * Given a track description, generate a VRML and Java roller coaster.
     * The track description consists of a number, the number of control
     * points, followed by that many 6-vectors, where the first 3 number
     * give a position control point, and the next 3 numbers give a normal
     * control point. Then comes another number giving the number of lifters
     * defined, and that many 3-vectors, where each vector gives the start
     * and stop parameter values for the lifter, and the minimum speed
     * for the car while on that lifter.
     *
     * For example, the simplest track (circular with upright normals) and
     * a lifter to keep the car speed above 1.0 between parameter values 1
     * and 2, is:
     *
     * <PRE>
     * 4
     * 10 1 10 0 1 0
     * -10 1 10 0 1 0
     * -10 1 -10 0 1 0
     * 10 1 -10 0 1 0
     * 1
     * 1 2 1
     * </PRE>
     * 
     * Two files are output. Track.wrl is a VRML file containing the track
     * geometry. It references a prototype Car.wrl, which should be made
     * available in the same directory.
     *
     * RollerSystem.java is a java file conforming to the
     * dmc.system.DynamicsSystem interface. It should be processed by
     * dynamid.periodic.Main() to generate a file suitable for use with
     * dmc.dmod.Main() or dmc.tovrml.Main().
     *
     * Read the complete documentation for a full description of how to
     * make roller coasters with this and other utilities.
     *
     * @param args A single element array containing the track file name.
     */
    public static void main(String[] args)
    {
	StreamTokenizer input = null;
	float[][]       t_control_pts = null;
	float[][]       n_control_pts = null;
	float[][]       lifters = null;
	int             num_c = 0;
	int             num_l = 0;

	try {
	    input = new StreamTokenizer(new FileReader(args[0]));
	    input.nextToken();
	    num_c = (int)input.nval;
	    t_control_pts = new float[num_c + 3][];
	    n_control_pts = new float[num_c + 3][];
	    readControls(input, t_control_pts, n_control_pts, num_c);
	    input.nextToken();
	    num_l = (int)input.nval;
	    if ( num_l > 0 )
	    {
		lifters = new float[num_l][];
		readLifters(input, lifters, num_l);
	    }
	} catch (Exception e) {
	    System.err.println(e);
	    System.exit(1);
	}

	writeRoller(t_control_pts, n_control_pts, lifters);
    }
}

