/*
 * TreeNode.java - Root class of all nodes in the model tree
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.model;

import java.util.Vector;
import java.io.PrintStream;

public class TreeNode {

    String name = "";
    TreeNode parent = null;
    Vector children = new Vector();

    // The following are used by TreeCanvas
    public int x, y;     // position of label
    public int strwidth; // width of name
    public int bx, by;   // upper-left corner of bounding box of tree rooted at this node
    public int width;    // width of tree rooted at this node
    public int height;   // height of tree at this node
    public boolean needslayout = true;
    public boolean collapsed = false;
    // End of TreeCanvas stuff

    static int hashCodeCounter = 0;

    int hashCode;

    public TreeNode() {
	hashCode = hashCodeCounter++;
    }

    public int hashCode() { return hashCode; }    

    public final String getName() { return name; }
    public final void setName( String name ) { this.name = name; }
    public final TreeNode getParent() { return parent; }
    public final int getChildCount() { return children.size(); }
    public final TreeNode getChild( int i ) { return (TreeNode)children.elementAt(i); }

    public final boolean isLeaf() { return children.size() == 0; }

    public String[] getFieldList() {
	return new String[0];
    }
    public void setField( String field, Object value ) {
	throw new IllegalArgumentException( "unknown field: '" + field + "'" );
    }
    public Object getField( String field ) {
	return null;
    }

    public final void removeParent() {
	if( parent != null )
	    parent.removeChild( this );
    }

    public final void addChild( TreeNode child ) {
	if( child.parent != null )
	    child.parent.removeChild( child );

	children.addElement( child );
	child.parent = this;
    }

    public final void removeChild( TreeNode child ) {
	if( child.parent == this ) {
	    children.removeElement( child );
	    child.parent = null;
	}
    }

    public final void removeChild( int i ) {
	TreeNode child = (TreeNode)children.elementAt(i);
	children.removeElementAt(i);
	child.parent = null;
    }
    
    public final boolean isParentOf( TreeNode child ) {
	if( child == this )
	    return true;
	else if( child == null )
	    return false;
	else
	    return isParentOf( child.parent );
    }
    
    public final boolean isChildOf( TreeNode parent ) {
	if( parent == null )
	    return false;
	else
	    return parent.isParentOf( this );
    }

    public static String escapeString( String str ) {
	int i, len = str.length();
	StringBuffer buf = new StringBuffer( len*2 );
	buf.append( '"' );
	for( i=0 ; i<len ; i++ ) {
	    char ch = str.charAt(i);
	    switch( ch ) {
	    case '"':
	    case '\\':
		buf.append( '\\' );
		// fallthrough
	    default:
		buf.append( ch );
	    }
	}
	return new String(buf.append('"'));
    }

    public static String formatValue( Object obj ) {
	if( obj instanceof String )
	    return escapeString( (String)obj );
	else
	    return obj.toString();
    }
    
    public String getType() { return "group"; }
    public void printFields( PrintStream out, String indent ) {
	String fields[] = getFieldList();
	for( int i=0 ; i<fields.length ; i++ )
	    out.println( indent + fields[i] + " " + formatValue(getField(fields[i])) );
    }
    public final void print( PrintStream out, String indent, String tabstop ) {
	String newindent = indent + tabstop;
	out.println( indent + getType() + " " + getName() + " {" );
	printFields( out, newindent );
	for( int i=0 ; i<getChildCount() ; i++ )
	    getChild(i).print(out,newindent,tabstop);
	out.println( indent + "}" );
    }
    public final void print( PrintStream out ) { print( out, "", "  " ); }
    public final void print() { print( System.out ); }
}
