/*
 * ActFunc.java - Activation functions
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.neuralnet;

import java.io.PrintWriter;

public abstract class ActFunc {
    // Activation functions...
    public static final ActFunc logistic = new Logistic();
    public static final ActFunc elliott = new Elliott();

    // Activation function class prototype...
    public abstract String getName();
    public abstract String outputCall( Neuron n, String x );
    public abstract void   outputFunc( PrintWriter out );
    public abstract double eval( Neuron n );
    public abstract double deriv1( Neuron n );
    public abstract double deriv2( Neuron n );
    public int hashCode() { return getName().hashCode(); }

    // Useful helper functions...
    public static final double sumInputs( Neuron n ) {
	int i, num = n.numInputs();
	double sum = 0.0;

	for( i=0 ; i<num ; i++ ) {
	    Link l = n.getInput(i);
	    sum += l.in.out * l.weight;
	}

	return sum;
    }

    public static final double safe_exp( double v ) {
	return ((v>709.7827128916) ? Double.MAX_VALUE :
		((v<-745.133219093) ? Double.MIN_VALUE :
		 Math.exp(v)));
	/* for floats:
	return ((v>88.72f) ? Float.MAX_VALUE :
		((v<-102.0f) ? Float.MIN_VALUE :
		 Math.exp(v)));
	*/
    }

    //
    // Activation function classes follow....
    //

    private static final class Logistic extends ActFunc { // in trans_f.c
	public String getName() { return "logistic"; }
	public String outputCall( Neuron n, String x ) {
	    return "logistic("+x+"+"+n.bias+")";
	}
	public void outputFunc( PrintWriter out ) {
	    out.println("    public static final double logistic( double x ) {");
	    out.println("        return 1.0 / (1.0 + Math.exp(-x));");
	    out.println("    }");
	}
	public double eval( Neuron n ) {
	    return 1.0 / (1.0 + ActFunc.safe_exp( -ActFunc.sumInputs(n) - n.bias ));
	}
	public double deriv1( Neuron n ) {
	    return n.act * (1.0 - n.act);
	}
	public double deriv2( Neuron n ) {
	    return n.act * (1.0 - n.act) * (2.0 * n.act - 1.0);
	}
    }

    private static final class Elliott extends ActFunc {
	public String getName() { return "elliott"; }
	public String outputCall( Neuron n, String x ) {
	    return "elliott("+x+")";
	}
	public void outputFunc( PrintWriter out ) {
	    out.println("    public final double elliott( double x ) {");
	    out.println("        return (x <= 0.0) ? (x / (1.0-x)) : (x / (1.0+x));");
	    out.println("    }");
	}
	public double eval( Neuron n ) {
	    double sum = ActFunc.sumInputs(n);
	    return (sum <= 0.0) ?
		(sum / (1.0 - sum)):
		(sum / (1.0 + sum));
	}
	public double deriv1( Neuron n ) {
	    double act = n.act;
	    act = (act <= 0.0) ? (1.0 + act) : (1.0 - act);
	    return act*act;
	}
	public double deriv2( Neuron n ) {
	    double act = n.act;
	    act = (act <= 0.0) ? (1.0 + act) : (1.0 - act);
	    return (act <= 0.0) ? (2 * act*act) : (-2 * act*act);
	}
    }
} 


