/*
 * FeedForwardNeuralNet.java - feed forward neural network
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.neuralnet;

import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.awt.*;
import java.awt.image.*;

public class FeedForwardNeuralNet {

    int numInputs;
    int numOutputs;
    int numHiddenLayers;
    int numHiddenPerLayer;

    static class Node {
	double act;
	double bias;
	double learn; // for backprop
	double momentum;
	double weights[] = null;
	Node() { 
	    this.act = Math.random();
	    this.bias = Math.random();
	}
    }
    
    public static final double BIAS_ACT = 0.5;

    Node inputs[];
    Node hidden[][];
    Node outputs[];

    public int numInputs() { return inputs.length; }
    public int numOutputs() { return outputs.length; }
    public int numLayers() { return hidden.length; }
    public int numHidden() { return hidden[0].length; }

    public FeedForwardNeuralNet( int in, int layers, int hid, int out ) {
	if( in < 1 || layers < 1 || hid < 1 || out < 1 )
	    throw new IllegalArgumentException("all parameters must be >= 1");

	inputs = new Node[numInputs = in];
	hidden = new Node[numHiddenLayers = layers][numHiddenPerLayer = hid];
	outputs = new Node[numOutputs = out];

	int i, j, k;
	double wgts[];

	for( i=0 ; i<in ; i++ )
	    inputs[i] = new Node();

	for( j=0 ; j<hid ; j++ ) {
	    hidden[0][j] = new Node();
	    wgts = hidden[0][j].weights = new double[in];
	    for( k=0 ; k<in ; k++ )
		wgts[k] = Math.random()*2 - 1;
	}

	for( i=1 ; i<layers ; i++ ) {
	    for( j=0 ; j<hid ; j++ ) {
		hidden[i][j] = new Node();
		wgts = hidden[i][j].weights = new double[hid];
		for( k=0 ; k<hid ; k++ )
		    wgts[k] = Math.random()*2 - 1;
	    }
	}

	for( j=0 ; j<out ; j++ ) {
	    outputs[j] = new Node();
	    wgts = outputs[j].weights = new double[hid];
	    for( k=0 ; k<hid ; k++ )
		wgts[k] = Math.random()*2 - 1;
	}
    }

    public synchronized void growNet( int amt ) {
	int newsize = numHiddenPerLayer + amt;
	Node hid[];
	int i, j, k;
	double wgts[];

	hid = new Node[newsize];
	System.arraycopy( hidden[0], 0, hid, 0, numHiddenPerLayer );
	for( i=numHiddenPerLayer ; i<newsize ; i++ ) {
	    hid[i] = new Node();
	    wgts = hid[i].weights = new double[numInputs];
	    for( j=0 ; j<numInputs ; j++ )
		wgts[j] = Math.random()*2 - 1;
	}
	hidden[0] = hid;

	for( i=1 ; i<numHiddenLayers ; i++ ) {
	    hid = new Node[newsize];
	    System.arraycopy( hidden[i], 0, hid, 0, numHiddenPerLayer );
	    for( j=0 ; j<numHiddenPerLayer ; j++ ) {
		wgts = new double[newsize];
		System.arraycopy( hid[j].weights, 0, wgts, 0, numHiddenPerLayer );
		// wgts[newsize] = hid[j].weights[numHiddenPerLayer]; // bias

		for( k=numHiddenPerLayer ; k<newsize ; k++ )
		    wgts[k] = Math.random()*2 - 1;

		hid[j].weights = wgts;
	    }

	    for( ; j<newsize ; j++ ) {
		hid[j] = new Node();
		wgts = new double[newsize];
		for( k=0 ; k<newsize ; k++ )
		    wgts[k] = Math.random()*2 - 1;
		hid[j].weights = wgts;
	    }

	    hidden[i] = hid;
	}

	for( i=0 ; i<numOutputs ; i++ ) {
	    wgts = new double[newsize];
	    System.arraycopy( outputs[i].weights, 0, wgts, 0, numHiddenPerLayer );
	    // wgts[newsize] = outputs[i].weights[numHiddenPerLayer]; // bias

	    for( j=numHiddenPerLayer ; j<newsize ; j++ )
		wgts[j] = Math.random()*2 - 1;

	    outputs[i].weights = wgts;
	}

	numHiddenPerLayer = newsize;
    }

    public static final double activation( Node n, Node inputs[] ) {
	double wgts[] = n.weights;
	double sum = n.bias; // wgts[wgts.length-1] * BIAS_ACT;
	int i;

	for( i=0 ; i<inputs.length ; i++ )
	    sum += wgts[i] * inputs[i].act;

	return 1.0 / (1.0 + Math.exp( -sum )); // - n.bias ));
    }

    public static final double deriv( double act ) {
	return act * (1.0 - act);
    }

    public synchronized void evaluate( double in[], double out[] ) {
	feedforward( in );
	for( int i=0 ; i<numOutputs ; i++ )
	    out[i] = outputs[i].act;
    }

    public synchronized void feedforward( double in[] ) {
	int i, j, k;

	for( i=0 ; i<numInputs ; i++ )
	    inputs[i].act = in[i];

	for( j=0 ; j<numHiddenPerLayer ; j++ )
	    hidden[0][j].act = activation( hidden[0][j], inputs );

	for( i=1 ; i<numHiddenLayers ; i++ )
	    for( j=0 ; j<numHiddenPerLayer ; j++ )
		hidden[i][j].act = activation( hidden[i][j], hidden[i-1] );

	for( i=0 ; i<numOutputs ; i++ )
	    outputs[i].act = activation( outputs[i], hidden[numHiddenLayers-1] );
    }
    
    protected final void clearlearnlayer( Node nodes[], int num ) {
	for( int i=0 ; i<num ; i++ )
	    nodes[i].learn = 0.0;
    }

    protected final void clearlearn() {
	clearlearnlayer( inputs, numInputs );
	for( int i=0 ; i<numHiddenLayers ; i++ )
	    clearlearnlayer( hidden[i], numHiddenPerLayer );
	clearlearnlayer( outputs, numOutputs );
    }

    public final double backprop( double in[], double target[],
				  double learning, double deltaMax ) {
	return backprop( in, target, learning, 0.0, deltaMax );
    }

    public synchronized double backprop( double in[], double target[],
					 double learning, double momentumWeight, double deltaMax ) {
	feedforward( in );
	clearlearn();
	int i, j, k;
	double sumerror = 0.0;
	double learnerror;
	double error;
	double dev, dev2;
	double eta = learning;
	Node n;
	Node layer[];
	Node inputLayer[];
	double wgts[];

	// outputs
	for( i=numOutputs-1 ; i>=0 ; i-- ) {
	    n = outputs[i];

	    dev = target[i] - n.act;
	    dev2 = dev * dev;
	    if( dev2 <= deltaMax ) // note: deltaMax is really deltaMax^2
		continue;

	    sumerror += dev2;
	    error = dev * deriv( n.act );

	    learnerror = eta * (momentumWeight * n.momentum + error);
	    n.bias += learnerror;
	    n.momentum = learnerror;

	    inputLayer = hidden[numHiddenLayers-1];
	    wgts = n.weights;

	    for( j=0 ; j<numHiddenPerLayer ; j++ ) {
		inputLayer[j].learn += wgts[j] * error;
		wgts[j] += learnerror * inputLayer[j].act;
	    }
	    // wgts[j] += learnerror * BIAS_ACT; // bias node
	}

	for( i=numHiddenLayers-1 ; i>0 ; i-- ) {
	    layer = hidden[i];
	    for( j=numHiddenPerLayer-1 ; j >= 0 ; j-- ) {
		n = layer[j];
		error = deriv(n.act) * n.learn;

		learnerror = eta * (n.momentum * momentumWeight + error);
		n.bias += learnerror;
		n.momentum = learnerror;

		inputLayer = hidden[i-1];
		wgts = n.weights;

		for( k=0 ; k<numHiddenPerLayer ; k++ ) {
		    inputLayer[k].learn += wgts[k] * error;
		    wgts[k] += learnerror * inputLayer[k].act;
		}
		// wgts[k] += learnerror * BIAS_ACT; // bias node
	    }
	}

	layer = hidden[0];
	for( j=numHiddenPerLayer-1 ; j >= 0 ; j-- ) {
	    n = layer[j];
	    error = deriv(n.act) * n.learn;

	    learnerror = eta * (n.momentum * momentumWeight + error);
	    n.bias += learnerror;
	    n.momentum = learnerror;

	    inputLayer = inputs;
	    wgts = n.weights;
	    
	    for( k=0 ; k<numInputs ; k++ )
		wgts[k] += learnerror * inputLayer[k].act;

	    // wgts[k] += learnerror * BIAS_ACT; // bias node
	}

	return sumerror;
    }

    public static void outputActFunc( PrintWriter out ) {
	out.println( "    public static final double logistic( double x ) {" );
	out.println( "        return 1.0 / (1.0 + Math.exp(-x));" );
	out.println( "    }" );
    }

    public synchronized void outputCode( PrintWriter out, String name, double outranges[][] ) {
	final String indent = "    ";
	final String indent2 = indent + indent;
	String str;
	double wgts[];
	Node n;
	int i, j, k;

	out.println( indent+"public static final void "+name+"(double in[], double out[]) {" );
	
	for( i=0 ; i<numInputs ; i++ )
	    out.println( indent2 + "double i"+i+" = in["+i+"];" );

	out.println();

	for( i=0 ; i<numHiddenPerLayer ; i++ ) {
	    n = hidden[0][i];
	    wgts = n.weights;

	    out.print( indent2 + "double h0_"+i+" = logistic("+n.bias );
		       // (n.bias + wgts[numInputs] * BIAS_ACT) );

	    
	    for( j=0 ; j<numInputs ; j++ )
		out.print( " + " + wgts[j] + "*i"+j );

	    out.println( ");" );
	}

	out.println();

	for( i=1 ; i<numHiddenLayers ; i++ ) {
	    str = "*h"+(i-1)+"_";

	    for( j=0 ; j<numHiddenPerLayer ; j++ ) {
		n = hidden[i][j];
		wgts = n.weights;

		out.print( indent2 + "double h"+i+"_"+j+" = logistic("+n.bias );
			   // (n.bias + wgts[numHiddenPerLayer] * BIAS_ACT) );

		for( k=0 ; k<numHiddenPerLayer ; k++ )
		    out.print( " + " + wgts[k] + str + k );

		out.println( ");" );
	    }
	}

	out.println();

	str = "*h"+(numHiddenLayers-1)+"_";

	for( i=0 ; i<numOutputs ; i++ ) {
	    n = outputs[i];
	    wgts = n.weights;

	    out.print( indent2 + "out["+i+"] = logistic(" + n.bias );
		       // (n.bias + wgts[numHiddenPerLayer] * BIAS_ACT) );

	    for( j=0 ; j<numHiddenPerLayer ; j++ )
		out.print( " + " + wgts[j] + str + j );

	    out.println( ") * "+(outranges[i][1]-outranges[i][0]) +
			  " + "+outranges[i][0] + ";" );
	}

	out.println( indent + "}" );
    }

    static class Plot extends Canvas {
	byte pixels[];
	MemoryImageSource source;
	Image image;
	int width, height;

	static ColorModel greyScale;
	static {
	    byte lvl[] = new byte[256];
	    for(int i=0 ; i<256 ; i++) lvl[i] = (byte)i;
	    greyScale = new IndexColorModel(8,256,lvl,lvl,lvl);
	}

	Plot( int width, int height, boolean animated ) {
	    this.width = width;
	    this.height = height;
	    pixels = new byte[width * height];
	    source = new MemoryImageSource( width, height, greyScale, pixels, 0, width );
	    source.setAnimated(animated);
	    image = createImage(source);
	}

	public void update( Graphics g ) { paint(g); }
	public void paint( Graphics g ) {
	    g.drawImage(image,0,0,this);
	}

	public Dimension getPreferredSize() { return new Dimension(width,height); }
	public Dimension getMinimumSize() { return new Dimension(width,height); }
    }

    //
    // Testing stuff follows...
    //
    static class NetPlot extends Plot {
	FeedForwardNeuralNet net;
	NetPlot( FeedForwardNeuralNet net, int width, int height ) {
	    super(width,height,true);
	    this.net = net;
	}
	public void plot() {
	    int x, y, offset;
	    double in[] = new double[2];
	    double out[] = new double[1];

	    for( y=offset=0 ; y<height ; y++ ) {
		in[1] = (double)y / height;
		for( x=0 ; x<width ; x++,offset++ ) {
		    in[0] = (double)x / width;
		    net.evaluate( in, out );
		    
		    pixels[offset] = (byte)(out[0] * 255);
		}

		// source.newPixels(0,y,width,1);
	    }

	    source.newPixels(0,0,width,height);
	}
    }

    static class FuncPlot extends Plot {
	FuncPlot( int width, int height ) {
	    super( width, height, false );

	    int x, y, offset;
	    double in[] = new double[2];
	    for( y=offset=0 ; y<height ; y++ ) {
		in[1] = (double)y / height;
		for( x=0 ; x<width ; x++,offset++ ) {
		    in[0] = (double)x / width;
		    pixels[offset] = (byte)(testFunc(in)*255);
		}
	    }

	    pixels = null;
	    source = null;
	}
    }

    static final double testFunc( double in[] ) {
	// return (Math.sin( x*2 + y*3 ) + 1) * 0.5;
	double x = in[0] * 2 * Math.PI - Math.PI;
	double y = in[1] * 4 - 2;

	double v = Math.sin(x - 5.23*Math.cos(y*y)) +
	    Math.cos( Math.sin( 1.43*y ) * 2.34*Math.cos(x+1.2) );

	return (v + 2) * 0.25;
    }

    public static final int TEST_RUNS = 0x7fffffff;
    public static final int TEST_UPDATE = 25000;
    public static final int RES = 150;
    public static void main( String argv[] ) {
	FeedForwardNeuralNet net = new FeedForwardNeuralNet( 2, 2, 6, 1 );
	NetPlot plot = new NetPlot( net, RES,RES );
	
	Frame frame = new Frame( "FF-net test" );
	frame.setLayout( new GridLayout( 1, 2, 1, 1 ) ); 
	frame.add( new FuncPlot(RES,RES) );
	frame.add( plot );
	frame.pack();
	frame.show();

	double in[] = new double[2];
	double target[] = new double[1];
	double error;

	net.growNet( 6 );

	double rate = 1.0;

	for( int i=0 ; i<TEST_RUNS ; i++ ) {
	    error = 0;
	    System.out.print( "["+i+"] rate = " + rate + "  " );
	    System.out.flush();
	    for( int j=0 ; j<TEST_UPDATE ; j++ ) {
		in[0] = Math.random();
		in[1] = Math.random();
		target[0] = testFunc(in);
		error += net.backprop( in, target, 0.75, 0.5, 0.0 );
	    }
	    error /= TEST_UPDATE;
	    rate = Math.min( rate, error * 20.0 );

	    System.out.println( "error = " + error );
	    plot.plot();
	}

	PrintWriter out = null;
	try {
	    out = new PrintWriter( new FileWriter( "FFNetTest.java" ) );
	    net.outputActFunc( out );
	    net.outputCode( out, "eval", new double[][] { { 0.0, 1.0 } } );
	} catch( IOException e ) {
	    System.out.println( "IO Error: "+ e.getMessage() );
	} finally {
	    if( out != null )
		out.close();
	}
    }
}
