/*
 * NeuralNet.java - A neural network
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.neuralnet;

import java.io.PrintWriter;
import java.util.Hashtable;

public class NeuralNet {
    Neuron input[];
    Neuron hidden[];
    Neuron output[];

    int inputCount;
    int hiddenCount;
    int outputCount;

    boolean sorted = false;

    public String toString() {
	return "input=" + inputCount + ", hidden=" + hiddenCount + ", output=" + outputCount;
    }

    public void print( java.io.PrintStream out ) {
	int i;
	out.println( "INPUT " + inputCount );
	for( i=0 ; i<inputCount ; i++ )
	    out.println( "  "+input[i] );
	out.println( "HIDDEN " + hiddenCount );
	for( i=0 ; i<hiddenCount ; i++ )
	    out.println( "  "+hidden[i] );
	out.println( "OUTPUT " + outputCount );
	for( i=0 ; i<outputCount ; i++ )
	    out.println( "  "+output[i] );
    }

    public NeuralNet() {
	input = new Neuron[1];
	hidden = new Neuron[1];
	output = new Neuron[1];
    }

    private int dfs( Neuron topo[], Neuron n, int i ) {
	if( n.type == Neuron.HIDDEN ) {
	    if( n.visited )
		return i;

	    n.visited = true;
	}

	int j, num = n.numOutputs();
	for( j=0 ; j<num ; j++ ) {
	    Neuron t = n.getOutput(j).out;
	    i = dfs( topo, t, i );
	}
	if( n.type == Neuron.HIDDEN )
	    topo[--i] = n;

	return i;
    }

    protected Neuron[] addToArray( Neuron array[], int idx, Neuron n ) {
	if( idx >= array.length ) {
	    Neuron newarray[] = new Neuron[array.length+idx+1];
	    System.arraycopy( array, 0, newarray, 0, idx );
	    array = newarray;
	}
	array[idx] = n;
	return array;
    }

    protected void removeFromArray( Neuron array[], int size, Neuron n ) {
	for( int i=0 ; i<size ; i++ ) {
	    if( array[i] == n ) {
		System.arraycopy( array, i, array, i+1, size-(i+1) );
		return;
	    }
	}
	throw new InternalError( "Neuron not in net" );
    }

    protected void addInput( Neuron n ) {
	input = addToArray( input, inputCount++, n );
    }
    protected void addHidden( Neuron n ) {
	hidden = addToArray( hidden, hiddenCount++, n );
    }
    protected void addOutput( Neuron n ) {
	output = addToArray( output, outputCount++, n );
    }
    protected void removeInput( Neuron n ) {
	removeFromArray( input, inputCount, n );
	inputCount--;
    }
    protected void removeHidden( Neuron n ) {
	removeFromArray( hidden, hiddenCount, n );
	hiddenCount--;
    }
    protected void removeOutput( Neuron n ) {
	removeFromArray( output, outputCount, n );
	outputCount--;
    }

    public void addNeuron( Neuron n ) {
	if( n.net == this )
	    return;

	sorted = false;

	// if( n.net != this )
	//    n.net.removeNeuron( n );

	n.net = this;

	switch( n.type ) {
	case Neuron.INPUT: addInput( n ); break;
	case Neuron.HIDDEN: addHidden( n ); break;
	case Neuron.OUTPUT: addOutput( n ); break;
	default:
	    throw new InternalError( "Neuron type: " + n.type );
	}
	
	int i, num = n.numInputs();
	for( i=0 ; i<num ; i++ )
	    addNeuron(n.getInput(i).in);
	num = n.numOutputs();
	for( i=0 ; i<num ; i++ )
	    addNeuron(n.getOutput(i).out);
    }

    public void removeNeuron( Neuron n ) {
	if( n.net != this )
	    return;

	n.net = null;

	switch( n.type ) {
	case Neuron.INPUT: removeInput(n); break;
	case Neuron.HIDDEN: removeHidden(n); break;
	case Neuron.OUTPUT: removeOutput(n); break;
	default:
	    throw new InternalError( "Neuron type: " + n.type );
	}

	int i, num = n.numInputs();
	for( i=0 ; i<num ; i++ )
	    removeNeuron( n.getInput(i).in );
	num = n.numOutputs();
	for( i=0 ; i<num ; i++ )
	    removeNeuron( n.getOutput(i).out );
    }

    public void topoSort() {
	if( sorted )
	    return;

	synchronized( this ) {
	    int i, idx = hiddenCount;
	    // Neuron topo[] = new Neuron[inputCount + hiddenCount + outputCount];
	    Neuron topo[] = new Neuron[hiddenCount];

	    // for( i=0 ; i<inputCount ; i++ )
	    //     input[i].visited = false;
	    for( i=0 ; i<hiddenCount ; i++ )
		hidden[i].visited = false;
	    // for( i=0 ; i<outputCount ; i++ )
	    //     output[i].visited = false;

	    // for( i=0 ; i<inputCount ; i++ )
	    //     if( !input[i].visited )
	    //         idx = dfs( topo, input[i], idx );

	    for( i=0 ; i<inputCount ; i++ )
		idx = dfs( topo, input[i], idx );
	    for( i=0 ; i<hiddenCount ; i++ )
		idx = dfs( topo, hidden[i], idx );

	    // for( i=0 ; i<outputCount ; i++ )
	    //     if( !output[i].visited )
	    //          idx = dfs( topo, output[i], idx );

	    // int ii = 0;
	    int hi = 0;
	    // int oi = 0;

	    for( i=0 ; i<topo.length ; i++ ) {
		Neuron n = topo[i];
		// switch( n.type ) {
		// case Neuron.INPUT: input[ii++] = n; break;
		// case Neuron.HIDDEN:
		hidden[hi++] = n; // break;
		// case Neuron.OUTPUT: output[oi++] = n; break;
		// }
	    }

	    sorted = true;
	}
    }

    public synchronized void test( double in[], double out[] ) {
	setInput(in);
	propagateForward();
	getOutput(out);
    }

    public synchronized void setInput( double in[] ) {
	if( inputCount > in.length )
	    throw new IllegalArgumentException( "net input size differs from pattern size" );

	for( int i=0 ; i<inputCount ; i++ ) {
	    input[i].eval( in[i] );
	    input[i].learn1 = 0.0;
	}
    }

    public synchronized void getOutput( double out[] ) {
	if( outputCount > out.length )
	    throw new IllegalArgumentException( "net output size differs" );

	for( int i=0 ; i<outputCount ; i++ )
	    out[i] = output[i].out;
    }

    private final void outputActivation( PrintWriter out, Neuron n, Hashtable varMap ) {
	Link link;
	StringBuffer buf = new StringBuffer();
	int num = n.numInputs();

	link = n.getInput(0);
	buf.append( (String)varMap.get( link.in ) ).append("*").append(link.weight);

	for( int j=1 ; j<num ; j++ ) {
	    link = n.getInput(j);
	    buf.append("+").append((String)varMap.get( link.in )).append("*").append(link.weight);
	}

	out.print( n.actFunc.outputCall( n, buf.toString() ) );
    }

    public void outputProc( PrintWriter out, String procName, double outranges[][] ) {
	final String indent = "    ";
	final String indent2 = indent + indent;
	out.println( indent + "public static final void " + procName + "(double in[], double out[]) {" );

	Hashtable varMap = new Hashtable();
	Hashtable actFuncs = new Hashtable();
	Neuron n;
	int i, j;

	topoSort();

	for( i=0 ; i<inputCount ; i++ )
	    varMap.put( input[i], "i"+i );
	for( i=0 ; i<hiddenCount ; i++ ) {
	    n = hidden[i];
	    if( n.numInputs() == 0 ) // bias node
		varMap.put( n, Double.toString(n.out) );
	    else
		varMap.put( n, "h"+i );
	}
	for( i=0 ; i<outputCount ; i++ )
	    varMap.put( output[i], "out["+i+"]" );

	for( i=0 ; i<inputCount ; i++ ) {
	    out.println( indent2 + "double i"+i+" = in["+i+"];" );
	    // out.println( indent2+"double i"+i+" = "+
	    //		 input[i].actFunc.outputCall(input[i],"in["+i+"]")+
	    //		 ";" );
	}

	for( i=0 ; i<hiddenCount ; i++ ) {
	    if( hidden[i].numInputs() > 0 ) {
		out.print( indent2 + "double h"+i+" = " );
		outputActivation( out, hidden[i], varMap );
		out.println( ";" );
	    }
	}
	for( i=0 ; i<outputCount ; i++ ) {
	    out.print( indent2 + "out["+i+"] = " );
	    outputActivation( out, output[i], varMap );
	    out.println( " * "+(outranges[i][1]-outranges[i][0]) +
			 " + "+outranges[i][0] + ";" );
	}
	out.println( indent + "}" );
    }

    public synchronized void propagateForward() {
	int i, num = hiddenCount;
	for( i=0 ; i<num ; i++ ) {
	    hidden[i].eval();
	    hidden[i].learn1 = 0.0;
	}

	num = outputCount;
	for( i=0 ; i<num ; i++ ) {
	    output[i].eval();
	    output[i].learn1 = 0.0;
	}
    }

    public synchronized double propagateBackward2( Pattern pat, 
						   double learning, 
						   double deltaMax ) {
	int i, j, numj, num = outputCount;
	double sumerror = 0.0;
	double learnerror;
	double error;
	double dev;
	double eta = learning;

	// outputs
	for( i=num-1 ; i>=0 ; i-- ) {
	    Neuron n = output[i];

	    dev = pat.output[i] - n.out;
	    if( Math.abs(dev) <= deltaMax )
		continue;

	    sumerror += dev * dev;
	    error = dev * n.actFunc.deriv1(n);
	    
	    // if( IS_SPECIAL_UNIT(n) )
	    //     learn_error = 0.0;
	    // else
	    learnerror = eta * error;
	    
	    n.bias += learnerror;
	 
	    numj = n.numInputs();
	    for( j=0 ; j<numj ; j++ ) {
		Link l = n.getInput(j);
		l.in.learn1 += l.weight * error;
		l.weight += learnerror * l.in.out;
	    }
	}
	
	// hidden
	num = hiddenCount;
	for( i=num-1 ; i>=0 ; i-- ) {
	    Neuron n = hidden[i];
	    error = n.actFunc.deriv1(n) * n.learn1;
	    // if( IS_SPECIAL_UNIT(n) )
	    //     learnerror = 0;
	    // else
	    learnerror = eta * error;
	    n.bias += learnerror;
	    
	    numj = n.numInputs();
	    for( j=0 ; j<numj ; j++ ) {
		Link l = n.getInput(j);
		Neuron src = l.in;

		if( src.type == Neuron.HIDDEN )
		    src.learn1 += l.weight * error;

		l.weight += learnerror * src.out;
	    }
	}
	return sumerror;
    }

    private final void randomizeNode( Neuron n, double min, double range ) {
	n.act =  Math.random() * range + min;
	n.out =  Math.random() * range + min;
	n.bias = Math.random() * range + min;
	int i, num = n.numInputs();
	for( i=0 ; i<num ; i++ )
	    n.getInput(i).weight = Math.random() * range + min;
	num = n.numOutputs();
	for( i=0 ; i<num ; i++ )
	    n.getOutput(i).weight = Math.random() * range + min;
    }

    public void randomize( double min, double max ) {
	double range = max - min;
	int i, j;

	for( i=0 ; i<inputCount ; i++ )
	    randomizeNode( input[i], min, range );

	for( i=0 ; i<hiddenCount ; i++ )
	    randomizeNode( hidden[i], min, range );

	for( i=0 ; i<outputCount ; i++ )
	    randomizeNode( output[i], min, range );
    }

    public int numHidden() { return hiddenCount; }

    /**
     * Adds neurons to the hidden layer(s)
     */
    int numAdded = 0;
    public void growNet( int hid, ActFunc actf, OutFunc outf ) {
	Neuron prot = input[0].getOutput(0).out;
	int i, j, numin, numout;

	for( ; prot.type == Neuron.HIDDEN ; prot = prot.getOutput(0).out ) {
	    numin = prot.numInputs();
	    numout = prot.numOutputs();
	    for( i=0 ; i<hid ; i++ ) {
		Neuron n = new Neuron( "add"+(numAdded++), Neuron.HIDDEN, outf, actf );
		for( j=0 ; j<numin ; j++ )
		    new Link( prot.getInput(j).in, n, 0.0 );
		for( j=0 ; j<numout ; j++ )
		    new Link( n, prot.getOutput(j).out, 0.0 );

		addNeuron( n );
		randomizeNode( n, 0.0, 1.0 );
	    }
	}
    }

    public static NeuralNet makeNet( int in, int numhid, 
				     int hid, int out, 
				     ActFunc actf, OutFunc outf )
    {
	Neuron inlayer[] = new Neuron[in];
	Neuron hidlayer[][] = new Neuron[numhid][hid];
	Neuron outlayer[] = new Neuron[out];
	Neuron bias[] = new Neuron[numhid];
	int i, j, k;

	for( i=0 ; i<in ; i++ )
	    inlayer[i] = new Neuron( "input"+(i+1), Neuron.INPUT, outf, actf );
	for( i=0 ; i<numhid ; i++ ) {
	    bias[i] = new Neuron( "bias"+(i+1), Neuron.HIDDEN, outf, actf );
	    for( j=0 ; j<hid ; j++ )
		hidlayer[i][j] = new Neuron( "hidden"+(i+1)+":"+(j+1), Neuron.HIDDEN, outf, actf );
	}
	for( i=0 ; i<out ; i++ )
	    outlayer[i] = new Neuron( "output"+(i+1), Neuron.OUTPUT, outf, actf );

	for( i=0 ; i<in ; i++ )
	    for( j=0 ; j<hid ; j++ )
		new Link( inlayer[i], hidlayer[0][j], 1.0 );
	
	for( i=1 ; i<numhid ; i++ )
	    for( j=0 ; j<hid ; j++ )
		for( k=0 ; k<hid ; k++ )
		    new Link( hidlayer[i-1][j], hidlayer[i][k], 1.0 );

	for( i=0 ; i<numhid ; i++ )
	    for( j=0 ; j<hid ; j++ )
		new Link( bias[i], hidlayer[i][j], 1.0 );

	k = numhid-1;
	for( i=0 ; i<hid ; i++ )
	    for( j=0 ; j<out ; j++ )
		new Link( hidlayer[k][i], outlayer[j], 1.0 );
		 
	NeuralNet net = new NeuralNet();

	for( i=0 ; i<in ; i++ )
	    net.addNeuron( inlayer[i] );

	return net;
    }

    public static NeuralNet makeNet2( int in, int numhid, int hid, int out, 
				      ActFunc actf, OutFunc outf )
    {
	Neuron inlayer[] = new Neuron[in];
	Neuron hidlayer[][][] = new Neuron[out][numhid][hid];
	Neuron outlayer[] = new Neuron[out];
	Neuron bias[][] = new Neuron[out][numhid];
	int i, j, k, l;

	for( i=0 ; i<in ; i++ )
	    inlayer[i] = new Neuron( "input"+(i+1), Neuron.INPUT, outf, actf );
	for( i=0 ; i<out ; i++ ) {
	    for( j=0 ; j<numhid ; j++ ) {
		bias[i][j] = new Neuron( "bias"+(i+1), Neuron.HIDDEN, outf, actf );
		for( k=0 ; k<hid ; k++ )
		    hidlayer[i][j][k] = new Neuron( "hidden"+(i+1)+":"+(j+1)+":"+(k+1), 
						    Neuron.HIDDEN, outf, actf );
	    }
	}
	for( i=0 ; i<out ; i++ )
	    outlayer[i] = new Neuron( "output"+(i+1), Neuron.OUTPUT, outf, actf );

	for( i=0 ; i<out ; i++ )
	    for( j=0 ; j<in ; j++ )
		for( k=0 ; k<hid ; k++ )
		    new Link( inlayer[j], hidlayer[i][0][k], 1.0 );
	
	for( i=0 ; i<out ; i++ )
	    for( j=1 ; j<numhid ; j++ )
		for( k=0 ; k<hid ; k++ )
		    for( l=0 ; l<hid ; l++ )
			new Link( hidlayer[i][j-1][k], hidlayer[i][j][l], 1.0 );
	
	for( i=0 ; i<out ; i++ )
	    for( j=0 ; j<numhid ; j++ )
		for( k=0 ; k<hid ; k++ )
		    new Link( bias[i][j], hidlayer[i][j][k], 1.0 );
	
	k = numhid-1;
	for( i=0 ; i<hid ; i++ )
	    for( j=0 ; j<out ; j++ )
		new Link( hidlayer[j][k][i], outlayer[j], 1.0 );
	
	NeuralNet net = new NeuralNet();
	
	for( i=0 ; i<in ; i++ )
	    net.addNeuron( inlayer[i] );
	
	return net;
    }
}



