/*
 * Integrator.java - numerical integration of differential equations
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.numericalMethods;

/**
 * The Integrator class defines methods for numerically integrating
 * differential equations. The actual routine used is a 5th order
 * Runge-Kutta method given in "Numerical Recipes", Cambridge University Press,
 * 1992.
 *
 * @author Stephen Chenney and Jeffry Ichnowski
 * @version 1.0
 */
public final class Integrator {

    private static final double TINY = 1.0e-20;

    private static final double  SAFETY = 0.9;
    private static final double  GROW   = 1.5;
    private static final double  PGROW  = -0.25;
    private static final double  SHRNK  = 0.5;
    private static final double  PSHRNK = ( -1.0 / 3.0 );
    private static final double  ERRCON = 0.1296;
    private static final int     MAXTRY = 40;

    private static final double  R_A2 = 0.2;
    private static final double  R_A3 = 0.3;
    private static final double  R_A4 = 0.6;
    private static final double  R_A5 = 1.0;
    private static final double  R_A6 = 0.875;
    private static final double  R_B21 = 0.2;
    private static final double  R_B31 = ( 3.0 / 40.0 );
    private static final double  R_B32 = ( 9.0 / 40.0 );
    private static final double  R_B41 = 0.3;
    private static final double  R_B42 = -0.9;
    private static final double  R_B43 = 1.2;
    private static final double  R_B51 = ( -11.0 / 54.0 );
    private static final double  R_B52 = 2.5;
    private static final double  R_B53 = ( -70.0 / 27.0 );
    private static final double  R_B54 = ( 35.0 / 27.0 );
    private static final double  R_B61 = ( 1631.0 / 55296.0 );
    private static final double  R_B62 = ( 175.0 / 512.0 );
    private static final double  R_B63 = ( 575.0 / 13824.0 );
    private static final double  R_B64 = ( 44275.0 / 110592.0 );
    private static final double  R_B65 = ( 253.0 / 4096.0 );
    private static final double  R_C1 = ( 37.0 / 378.0 );
    private static final double  R_C3 = ( 250.0 / 621.0 );
    private static final double  R_C4 = ( 125.0 / 594.0 );
    private static final double  R_C6 = ( 512.0 / 1771.0 );
    private static final double  R_DC5 = ( -277.0 / 14336.0 );
    private static final double  R_DC1 = ( R_C1 - 2825.0 / 27648.0 );
    private static final double  R_DC3 = ( R_C3 - 18575.0 / 48384.0 );
    private static final double  R_DC4 = ( R_C4 - 13525.0 / 55296.0 );
    private static final double  R_DC6 = ( R_C6 - 0.25 );

    private double ytemp1[];
    private double ytemp2[];
    private double ak2[], ak3[], ak4[], ak5[], ak6[];
    private double yscal[], y[], dydx[], err[];

    Derivative derivs;

    /**
     * An instance of the class that evaluates the function's derivatives is
     * given to the constructor. The derivatives routine is the complete
     * description of the function.
     *
     * @param derivs An instance of the class to evaluate the function's
     *               derivatives. This class must implement the Derivative
     *               interface.
     *
     * @see Derivative
     */
    public Integrator( Derivative derivs ) {
	int nvar = derivs.getDimension();
	this.derivs = derivs;

	yscal	= new double[nvar];
	y	= new double[nvar];
	dydx	= new double[nvar];
	err	= new double[nvar];
	ytemp1	= new double[nvar];
	ytemp2	= new double[nvar];
	ak2	= new double[nvar];
	ak3	= new double[nvar];
	ak4	= new double[nvar];
	ak5	= new double[nvar];
	ak6	= new double[nvar];
    }

    private final void Runge_Kutta_Step(double y[], double dydx[], int n,
					double h, double yout[], double yerr[])
    {
	int i;

	for ( i = 0 ; i < n ; i++ )
	    ytemp2[i] = y[i] + R_B21 * h * dydx[i];

	derivs.derivs(x + R_A2 * h, n, ytemp2, ak2);

	for ( i = 0 ; i < n ; i++ )
	    ytemp2[i] = y[i] + h * ( R_B31 * dydx[i] + R_B32 * ak2[i] );

	derivs.derivs(x + R_A3 * h, n, ytemp2, ak3);

	for ( i = 0 ; i < n ; i++ )
	    ytemp2[i] = y[i] + h
			* ( R_B41*dydx[i] + R_B42*ak2[i] + R_B43*ak3[i] );

	derivs.derivs(x + R_A4 * h, n, ytemp2, ak4);

	for ( i = 0 ; i < n ; i++ )
	    ytemp2[i] = y[i] + h *
		( R_B51*dydx[i] + R_B52*ak2[i] + R_B53*ak3[i] + R_B54*ak4[i] );

	derivs.derivs(x + R_A5 * h, n, ytemp2, ak5);

	for ( i = 0 ; i < n ; i++ )
	    ytemp2[i] = y[i] + h
		* ( R_B61*dydx[i] + R_B62*ak2[i] + R_B63*ak3[i]
		    + R_B64*ak4[i] + R_B65*ak5[i] );

	derivs.derivs(x + R_A6 * h, n, ytemp2, ak6);

	for ( i = 0 ; i < n ; i++ )
	    yout[i] = y[i] + h * ( R_C1*dydx[i] + R_C3*ak3[i] + R_C4*ak4[i] +
				   R_C6*ak6[i] );
	for ( i = 0 ; i < n ; i++ )
	    yerr[i] = h * ( R_DC1 * dydx[i] + R_DC3 * ak3[i] + R_DC4 * ak4[i] +
			    R_DC5 * ak5[i] + R_DC6 * ak6[i] );
    }

    private double x;
    private double hdid;
    private double hnext;

    private final void Runge_Kutta(double y[], double dydx[], int n,
				   double htry, double eps, double yscal[])
	throws ArithmeticException
    {
	int     i;
	double  errmax, h, htemp, xnew;

	h = htry;
	while( true ) {
	    Runge_Kutta_Step(y, dydx, n, h, ytemp1, err);
	    errmax = 0.0;
	    for ( i = 0 ; i < n ; i++ )
		errmax = Math.max(errmax, Math.abs(err[i] / yscal[i]));

	    errmax /= eps;
	    if ( errmax < 1.0 )
		break;

	    htemp = SAFETY * h * Math.pow(errmax, PSHRNK);
	    h = (h > 0.0 ? Math.max(htemp, 0.1 * h) : Math.min(htemp, 0.1 * h));
	    xnew = x + h;
	    if ( xnew == x ) {
		System.err.println( "Stepsize underflow in Runge_Kutta" );
		throw new ArithmeticException(
		    "Stepsize underflow in Runge_Kutta" );
	    }
	}
	if ( errmax > ERRCON )
	    hnext = SAFETY * h * Math.pow(errmax, PGROW);
	else
	    hnext = 5.0 * h;
	x += ( hdid = h );
        System.arraycopy(ytemp1, 0, y, 0, n);
    }


    private static final int MAXSTP = 10000;


    /**
     * The integration routine. Integrates from x1 to x2, with error
     * tolerance eps, initial step size h1 and minimum step size hmin.
     *
     * @param ystart The initial values for all the variables.
     * @param x1 The initial ordinate value (generally time).
     * @param x2 The final ordinate value (generally time).
     * @param eps The error tolerance.
     * @param h1 The initial step size.
     * @param hmin The minimum step size. An exception wil be thrown if
     *             the step size drops below this value.
     *
     * @exception ArithmeticException Thrown if the step size becomes
     *                                too small or the number of steps is
     *                                too large.
     */
    public synchronized void ODEIntegrate( double ystart[], double x1,
					   double x2, double eps,
					   double h1, double hmin )
	throws ArithmeticException
    {
	double  h;

	x = x1;
	h = Math.abs(h1);
	if ( x2 - x1 < 0.0 )
	    h = -h;

	// Remove this when the jit compiler stops misbehaving
	System.out.print("");
	
        System.arraycopy(ystart, 0, y, 0, ystart.length);

	for ( int nstp = 0 ; nstp < MAXSTP ; nstp++ ) {

	    derivs.derivs(x, ystart.length, y, dydx);

	    for ( int i = 0 ; i < ystart.length ; i++ )
		yscal[i] = Math.abs(y[i]) + Math.abs(dydx[i] * h) + TINY;

	    if ( ( x + h - x2 ) * ( x + h - x1 ) > 0.0 )
		h = x2 - x;

	    Runge_Kutta(y, dydx, ystart.length, h, eps, yscal );

	    if ( ( x - x2 ) * ( x2 - x1 ) >= 0.0 ) {
		System.arraycopy(y, 0, ystart, 0, ystart.length);
		return;
	    }

	    if ( Math.abs(hnext) <= hmin )
		throw new ArithmeticException(
			"Stepsize too small in ODE_Integrate: hmin = "
			+ hmin + " and hnext = " + hnext );

	    h = hnext;
	}

	throw new ArithmeticException( "Too many steps in ODE_Integrate" );
    }
}

