/*
 * Minimize.java - One dimensional minimization routines
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package berkeley.cs.dmc.numericalMethods;

/**
 * One dimensional minimization routines.
 */
public class Minimize {
    private static final double GOLD = 1.618034;
    private static final double TINY = 1.0e-20;
    private static final double GLIMIT = 100.0;

    private static double sign(double a, double b) {
	return ( b >= 0.0 ? Math.abs(a) : -Math.abs(a) );
    }

    private static double sqr(double a) {
	return ( a == 0.0 ? 0.0 : a * a );
    }

    private static void shift(double[] a, double[] b, double[] c, double d) {
	a[0] = b[0];
	b[0] = c[0];
	c[0] = d;
    }


    /**
     * Given a function func, and given distinct initial points ax and bx, this
     * routine searches in the downhill direction (defined by the function as
     * evaluated at the initial points) and returns new points ax, bx, cx that
     * bracket a minimum if the function. Also returned are the function values
     * at the three points, fa, fb, fc.
     *
     * @param ax One element array. On input, initial value. On output, one
     *           of the bracketing point.
     * @param bx One element array. On input, initial value. On output, the
     *           lowest of the bracketing points.
     * @param cx One element array. On output, one of bracketing points.
     * @param fa On output, elmt 0 contains the function value at ax[0].
     * @param fb On output, elmt 0 contains the function value at bx[0].
     * @param fc On output, elmt 0 contains the function value at cx[0].
     * @param func An object to evaluate the function.
     * @param fdata User data passed to func when it is evaluated.
     */
    public static void minimumBracket1D(double[] ax, double[] bx, double[] cx,
				        double[] fa, double[] fb, double[] fc,
				        MinimizeFunction func, Object[] fdata)
    {
	double      ulim, r, q;
	double[]    u = new double[1];
	double[]    fu = new double[1];
	double[]    dum = new double[1];

	fa[0] = func.evaluate(ax[0], fdata);
	fb[0] = func.evaluate(bx[0], fdata);
	if ( fb[0] > fa[0] )
	{
	    shift(dum, ax, bx, dum[0]);
	    shift(dum, fa, fb, dum[0]);
	}
	cx[0]= bx[0] + GOLD * ( bx[0] - ax[0] );
	fc[0] = func.evaluate(cx[0], fdata);
	while ( fb[0] > fc[0] )
	{
	    r = ( bx[0] - ax[0] ) * ( fb[0] - fc[0] );
	    q = ( bx[0] - cx[0] ) * ( fb[0] - fa[0] );
	    u[0] = ( bx[0] ) - ( ( bx[0] - cx[0] ) * q - ( bx[0] - ax[0] ) * r )
		   / ( 2.0 * sign(Math.max(Math.abs(q-r),TINY), q - r) );
	    ulim = bx[0] + GLIMIT * ( cx[0] - bx[0] );
	    if ( ( bx[0] - u[0] ) * ( u[0] - cx[0] ) > 0.0 )
	    {
		fu[0] = func.evaluate(u[0], fdata);
		if ( fu[0] < fc[0] )
		{
		    ax[0] = bx[0];
		    bx[0] = u[0];
		    fa[0] = fb[0];
		    fb[0] = fu[0];
		    return;
		}
		else if ( fu[0] > fb[0] )
		{
		    cx[0] = u[0];
		    fc[0] = fu[0];
		    return;
		}
		u[0] = cx[0] + GOLD * ( cx[0] - bx[0] );
		fu[0] = func.evaluate(u[0], fdata);
	    }
	    else if ( ( cx[0] - u[0] ) * ( u[0] - ulim ) > 0.0 )
	    {
		fu[0] = func.evaluate(u[0], fdata);
		if ( fu[0] < fc[0] )
		{
		    shift(bx, cx, u, cx[0] + GOLD * ( cx[0] - bx[0] ));
		    shift(fb, fc, fu, func.evaluate(u[0], fdata));
		}
	    }
	    else if ( ( u[0] - ulim ) * ( ulim - cx[0] ) >= 0.0 )
	    {
		u[0] = ulim;
		fu[0] = func.evaluate(u[0], fdata);
	    }
	    else
	    {
		u[0] = cx[0] + GOLD * ( cx[0] - bx[0] );
		fu[0] = func.evaluate(u[0], fdata);
	    }
	    shift(ax, bx, cx, u[0]);
	    shift(fa, fb, fc, fu[0]);
	}
    }

    private static final int ITMAX = 100;
    private static final double ZEPS = 1.0e-10;
    private static final double CGOLD = 0.3819660;

    /**
     * Method that uses Brent's algorithm to find the minimum of a function
     * between a pair of bracketing values.
     *
     * @param ax One enpt of the bracketing region.
     * @param bx An intermediate point of the bracketing region.
     * @param cx The other enpt of the bracketing region.
     * @param func An object to evaluate the function being minimized.
     * @param fdata User data passed to func when it is evaluated.
     * @param tol The tolerance for the minimum.
     * @param xmin The x value at the minimum is returned in xmin[0].
     *
     * @return The minimum value.
     */
    public static double brentMinimize1D(double ax, double bx, double cx,
					 MinimizeFunction func, Object[] fdata,
					 double tol, double[] xmin) {
	int	iter;
	double  a, b, etemp, fu;
	double	p, q, r, tol1, tol2, u, xm;
	double[]    x = new double[1];
	double[]    fx = new double[1];
	double[]    v = new double[1];
	double[]    fv = new double[1];
	double[]    w = new double[1];
	double[]    fw = new double[1];
	double  d = 0.0;
	double  e = 0.0;

	a = ( ax < cx ? ax : cx );
	b = ( ax > cx ? ax : cx );
	x[0] = w[0] = v[0] = bx;
	fw[0] = fv[0] = fx[0] = func.evaluate(x[0], fdata);
	for ( iter = 0 ; iter < ITMAX ; iter++ )
	{
	    xm = 0.5 * ( a + b );
	    tol2 = 2.0 * ( tol1 = tol * Math.abs(x[0]) + ZEPS );
	    if ( Math.abs(x[0]-xm) <= ( tol2 - 0.5 * ( b - a ) ) )
	    {
		xmin[0] = x[0];
		return fx[0];
	    }
	    if ( Math.abs(e) > tol1 )
	    {
		r = ( x[0] - w[0] ) * ( fx[0] - fv[0] );
		q = ( x[0] - v[0] ) * ( fx[0] - fw[0] );
		p = ( x[0] - v[0] ) * q - ( x[0] - w[0] ) * r;
		q = 2.0 * ( q - r );
		if ( q > 0.0 )
		    p = -p;
		q = Math.abs(q);
		etemp = e;
		e = d;
		if ( Math.abs(p) >= Math.abs(0.5 * q * etemp)
		  || p <= q * ( a - x[0] )
		  || p >= q * ( b - x[0] ) )
		{
		    d = CGOLD * ( e = ( x[0] >= xm ? a - x[0] : b - x[0] ) );
		}
		else
		{
		    d = p / q;
		    u = x[0] + d;
		    if ( u - a < tol2 || b - u < tol2 )
			d = sign(tol1, xm - x[0]);
		}
	    }
	    else
	    {
		d = CGOLD * ( e = ( x[0] >= xm ? a - x[0] : b - x[0] ) );
	    }
	    u = ( Math.abs(d) >= tol1 ? x[0] + d : x[0] + sign(tol1, d) );
	    fu = func.evaluate(u, fdata);
	    if ( fu <= fx[0] )
	    {
		if ( u >= x[0] )
		    a = x[0];
		else
		    b = x[0];
		shift(v, w, x, u);
		shift(fv, fw, fx, fu);
	    }
	    else
	    {
		if ( u < x[0] )
		    a = u;
		else
		    b = u;
		if ( fu <= fw[0] || w[0] == x[0] )
		{
		    v[0] = w[0];
		    w[0] = u;
		    fv[0] = fw[0];
		    fw[0] = fu;
		}
		else if ( fu <= fv[0] || v[0] == x[0] || v[0] == w[0] )
		{
		    v[0] = u;
		    fv[0] = fu;
		}
	    }
	}
	System.err.println("Too many iterations in Brent_Minimize_1D");
	System.exit(1);

	return 1.0;	// Never reached
    }
}


