/*
 * BoundFinder.java - bounds finding code
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.periodic;

import berkeley.cs.dmc.system.DynamicsSystem;
import berkeley.cs.dmc.numericalMethods.Minimize;
import berkeley.cs.dmc.numericalMethods.MinimizeFunction;

class BoundFinder {
    DynamicsSystem   system = null;
    double           period = 0.0;
    int              dimension = 0;
    int              num_vars = 0;
    MinimizeFunction min_func = null;
    MinimizeFunction max_func = null;

    BoundFinder(DynamicsSystem sys, double p) {
	system = sys;
	period = p;
	num_vars = system.numVariables();
	dimension = system.subsetLength();
	min_func = new BoundMinFunc(system, num_vars, dimension);
	max_func = new BoundMaxFunc(system, num_vars, dimension);
    }


    double[] boundParameter(int index) {
	double[] res = new double[2];
	double   t_2 = 0.0;
	double   t_1 = 0.0;
	double   t_0 = 0.0;
	double   t_step = system.bufferStep();
	double[] p_2 = new double[num_vars];
	double[] p_1 = new double[num_vars];
	double[] p_0 = new double[num_vars];
	Object[] min_data = new Object[3];
	double[] min_t = new double[1];

	// Init for minimization routines.
	min_data[0] = new Integer(index);
	min_data[2] = p_2;

	system.initialCondition(p_2);
	t_2 = 0.0;
	t_1 = t_2 + t_step;
	system.evaluate(t_2, p_2, t_1, p_1);

	res[0] = p_2[index];
	res[1] = p_2[index];
	for ( int i = index + dimension ; i < num_vars ; i += dimension )
	{
	    if ( res[0] > p_2[i] )
		res[0] = p_2[i];
	    if ( res[1] < p_2[i] )
		res[1] = p_2[i];
	}

	// Search for local maxima and minima
	while ( t_1 < period )
	{
	    t_0 = t_1 + t_step;
	    system.evaluate(t_1, p_1, t_0, p_0);

	    // Check for local minima
	    boolean is_minima = false;
	    for ( int i = index ; i < num_vars ; i += dimension )
		if ( p_1[i] < p_2[i] && p_1[i] < p_0[i] )
		{
		    is_minima = true;
		    break;
		}
	    if ( is_minima )
	    {
		min_data[1] = new Double(t_2);
		double min = Minimize.brentMinimize1D(t_2, t_1, t_0,
		                                      min_func, min_data,
						      1.0e-6, min_t);
		if ( min < res[0] )
		    res[0] = min;
	    }

	    // Now look for local maxima
	    boolean is_maxima = false;
	    for ( int i = index ; i < num_vars ; i += dimension )
		if ( p_1[i] > p_2[i] && p_1[i] > p_0[i] )
		{
		    is_maxima = true;
		    break;
		}
	    if ( is_maxima )
	    {
		min_data[1] = new Double(t_2);
		double max = Minimize.brentMinimize1D(t_2, t_1, t_0,
		                                      max_func, min_data,
						      1.0e-6, min_t);
		if ( -max > res[1] )
		    res[1] = -max;
	    }

	    for ( int i = 0 ; i < num_vars ; i++ )
	    {
		p_2[i] = p_1[i];
		p_1[i] = p_0[i];
	    }
	    t_2 = t_1;
	    t_1 = t_0;
	}

	// Need to check the endpt for a strictly increasing function.
	system.evaluate(t_2, p_2, period, p_1);
	for ( int i = index ; i < num_vars ; i += dimension )
	{
	    if ( res[0] > p_1[i] )
		res[0] = p_1[i];
	    if ( res[1] < p_1[i] )
		res[1] = p_1[i];
	}

	double range = res[1] - res[0];
        res[1] += range * 0.05;
        res[0] -= range * 0.05;

	return res;
    }


    double[][] findBound() {
	double[][] min_max = new double[dimension][];
	for ( int i = 0 ; i < dimension ; i++ )
		min_max[i] = boundParameter(i);

	return min_max;
    }

}

