/*
 * CodeGeneration.java
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.periodic;

import berkeley.cs.dmc.system.DynamicsSystem;
import java.io.PrintWriter;
import java.io.FileWriter;

class CodeGeneration {
    private static final String[] templates = { "_super_name_",
						"_class_name_",
						"_num_variables_",
						"_period_",
						"_approx_thresh_",
						"_sub_size_",
						"_buffer_step_" };

    private static final String[] first_template = {
"// This file is automatically generated",
"import berkeley.cs.dmc.system.DynamicsVariable;",
"public class _class_name_ extends _super_name_ implements DynamicsVariable {",
"    private static double period = _period_;",
"    private static final double buffer = _buffer_step_;",
"",
"    public _class_name_() {",
"        super();",
"    }",
""
    };

    private static final String[] final_template = {
"",
"    public void doApprox(double time, double[] p) {",
"        double net_out[] = new double[_sub_size_];",
"",
"        double end_vals[] = new double[_num_variables_];",
"        double eval_time = Math.IEEEremainder(time, period);",
"        for ( int i = 0 ; i < _num_variables_ ; i += _sub_size_ ) {",
"            approxValues(net_out, eval_time);",
"            for ( int j = 0 ; j < _sub_size_ ; j++ )",
"                p[j + i] = net_out[j];",
"        }",
"        super.checkState(time, p, false);",
"    }",
"",
"    public float newState(double last_time, double[] last_p,",
"                          double seen_time, double[] seen_p,",
"                          double new_time, double[] new_p) {",
"	float res = 1.0f;",
"	if ( new_time - last_time < buffer + 0.00001 )",
"	  res = super.evaluate(last_time, last_p, new_time, new_p);",
"	else",
"	{",
"         if ( new_time - seen_time >= _approx_thresh_ )",
"            doApprox(new_time, new_p);",
"        else",
"            res = super.evaluate(last_time, last_p, new_time, new_p);",
"       }",
"       return res;",
"    }",
"}"
    };

    private static String substitute(String basis, String old_sub,
				     String new_sub)
    {
	int old_index = basis.indexOf(old_sub);

	if ( old_index == -1 )
	    return basis;

	return basis.substring(0, old_index) + new_sub
	     + basis.substring(old_index + old_sub.length(), basis.length());
    }

    private static String substituteMulti(String basis, String[] olds,
                                          String[] news)
    {
	if ( olds.length != news.length )
	    return null;

	String result = basis;
	for ( int i = 0 ; i < olds.length ; i++ )
	    result = substitute(result, olds[i], news[i]);
	return result;
    }

    static void generate(String name, DynamicsSystem system, double period,
                         double[][] bounds, NeuralNetSystem net)
    {
	String[] new_names = new String[templates.length];
	String lineSep = System.getProperty("line.separator");

	new_names[0] = name;
	new_names[1] = substitute(new_names[0], "System", "Variable");
	new_names[2] = new Integer(system.numVariables()).toString();
	new_names[3] = Double.toString(period);
	new_names[4] = Double.toString(period * 0.1);
	new_names[5] = new Integer(system.subsetLength()).toString();
	new_names[6] = Double.toString(system.bufferStep()) + "f";

	try {
	    PrintWriter out =
		new PrintWriter(new FileWriter(new_names[1] + ".java"));

	    for ( int i = 0 ; i < first_template.length ; i++ )
		out.println(substituteMulti(first_template[i], templates,
					    new_names));

	    int max_buffer = (int)Math.ceil(period / system.bufferStep());
	    if ( max_buffer < system.bufferSize() )
	    {
		out.println("    public int bufferSize() { return "
		          + max_buffer + "; }");
		out.println();
	    }

            net.output(out);

	    for ( int i = 0 ; i < final_template.length ; i++ )
		out.println(substituteMulti(final_template[i], templates,
					    new_names));

	    out.close();
	} catch (Exception e) {
	    System.err.println("Error generating code");
	    System.err.println(e);
	}
    }
}


