/*
 * Main.java
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.periodic;

import berkeley.cs.dmc.system.DynamicsSystem;


/**
 * Application that searches for periodic behaviour in a dynamical
 * system, and if it finds it, generates a more efficient version
 * of the system, suitable for culling.
 *
 * @author Stephen Chenney and Jeffry Ichnowski
 * @version 1.0
 */
public class Main {

    /**
     * The main method for the application. It should be invoked as:
     *
     * <PRE>
     * java dynamics.period.Main <DynamicsSystem>
     * </PRE>
     *
     * <DynamicsSystem> is a class name that must implement the
     * berkeley.cs.dmc.system.DynamicsSystem
     * interface, and should contain the string System somewhere in the
     * name. The class file must be in the class search path.
     *
     * If the system is found to be periodic in output over time,
     * a new class is created, with a name derived from the original
     * class by replacing the string System with the string Variable.
     * The new class will implement the
     * berkeley.cs.dmc.system.DynamicsVariable interface.
     *
     * If no period under one hour is found, the system exits
     * without generating anything.
     */
    public static void main(String args[]) {
	if ( args.length < 1 ) {
	    System.out.println(
		"Usage: java dynamics.period.Main <DynamicsSystem>");
	    System.exit(-1);
	}

	Class	       dynamics_class = null;
	DynamicsSystem dynamics_system = null;

	try { 
	    dynamics_class = Class.forName(args[0]);
	    dynamics_system = (DynamicsSystem)dynamics_class.newInstance();
	} catch( ClassNotFoundException e ) {
	    System.err.println( "Class '"+args[0]+"' not found." );
	    System.exit(1);
	    return;
	} catch( InstantiationException e ) {
	    System.err.println(
		"Could not instantiate class. " + e.getMessage() );
	    System.exit(2);
	    return;
	} catch( IllegalAccessException e ) {
	    System.err.println( "Illegal access. " + e.getMessage() );
	    System.exit(3);
	    return;
	} catch( ClassCastException e ) {
	    System.err.println(
		"Class does not implement dynamics.system.DynamicsSystem" );
	    System.exit(4);
	    return;
	}

	PeriodFinder period_finder = new PeriodFinder(dynamics_system);

	double period = period_finder.findPeriod();

	if ( period == -1.0 )
	{
	    System.out.println("System not periodic");
	    System.exit(0);
	}

	System.out.println("Periodic: Period = " + period);

	BoundFinder bound_finder = new BoundFinder(dynamics_system, period);
	double[][]  bounds = bound_finder.findBound();

	for ( int i = 0 ; i < bounds.length ; i++ )
	    System.out.println("Bound[" + i + "]: " + bounds[i][0]
	                     + " to " + bounds[i][1]);

	// Now need to learn neural net approximators. One for each parameter.
	// It will take 0.0 <= t < PERIOD and return each param.
	NeuralNetSystem net =
		new NeuralNetSystem(dynamics_system, bounds, period);
	net.learn();

	// Then, with approximators in hand, we can write the DynamicsVariable
	// class.
	CodeGeneration.generate(args[0], dynamics_system, period, bounds, net);

	System.exit(0);

    }
}


