/*
 * NeuralNetSystem.java
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.periodic;

import berkeley.cs.dmc.system.DynamicsSystem;
import berkeley.cs.dmc.neuralnet.*;
import java.io.*;
import java.util.Random;
import java.awt.*;

public class NeuralNetSystem extends Canvas {
    public static final int    MAX_CYCLES = 10000000;
    public static final int    HIDDEN_LAYERS = 2;
    public static final int    HIDDEN_NODES = 10;
    public static final int    BUFFER_SIZE = 16;
    public static final int    BUFFER_CYCLES_PER_UPDATE = 4096;

    DynamicsSystem system;
    double bounds[][];
    double period;
    FeedForwardNeuralNet net;

    Object systemLock = new Object();

    public void paint( Graphics g ) {
        Dimension size = getSize();
	int vars = system.subsetLength();
	int yorg[] = new int[vars];
	int ht = size.height / vars;
	int width = size.width;
	int i, j;
	double in[] = new double[vars];
	double out[] = new double[vars];
	double netin[] = new double[1];
	int sy[] = new int[vars];
	int ny[] = new int[vars];
	int sy2[] = new int[vars];
	int ny2[] = new int[vars];

	synchronized( systemLock ) {

	for( i=0 ; i<vars ; i++ )
	    yorg[i] = i * size.height / vars;

	system.initialCondition( in );

	for( i=0 ; i<vars ; i++ )
	    sy[i] = yorg[i] + (int)(ht - ((in[i] - bounds[i][0]) * ht / (bounds[i][1] - bounds[i][0])));

	netin[0] = 0.0;

	net.evaluate( netin, out );

	for( i=0 ; i<vars ; i++ )
	    ny[i] = yorg[i] + (int)(ht - out[i] * ht);

	double t_eval = 0.0;
        double t_step = period / (double)(width-1);
	for( int t=1 ; t<width ; t++ ) {
	    system.evaluate( t_eval, in, t_eval + t_step, out );
            t_eval += t_step;
	    t_eval = Math.floor(t_eval / t_step + 0.1) * t_step;
	    for( i=0 ; i<vars ; i++ )
		in[i] = out[i];
	    g.setColor( Color.red );
	    for( i=0 ; i<vars ; i++ ) {
	        sy2[i] = yorg[i] + (int)(ht - ((out[i] - bounds[i][0]) * ht / (bounds[i][1] - bounds[i][0])));
		g.drawLine( t-1, sy[i], t, sy2[i] );
	    }

	    netin[0] = (double)t / (double)(width - 1);
	    net.evaluate( netin, out );
	    g.setColor( Color.blue );
	    for( i=0 ; i<vars ; i++ ) {
	        ny2[i] = yorg[i] + (int)(ht - out[i] * ht);
		g.drawLine( t-1, ny[i], t, ny2[i] );
	    }

	    int tmp[];
	    tmp = ny;
	    ny = ny2;
	    ny2 = tmp;
	    tmp = sy;
	    sy = sy2;
	    sy2 = tmp;
	}
	}
    }
  
    public Dimension getPreferredSize() { return new Dimension(100,100*system.subsetLength()); }
    public Dimension getMinimumSize() { return new Dimension(50,50*system.subsetLength()); }

    public NeuralNetSystem( DynamicsSystem system, double bounds[][], double period ) {
	this.system = system;
	this.bounds = bounds;
	this.period = period;

	setBackground( Color.black );
	setForeground( Color.white );
    }

    public synchronized void learn() {
	Random random = new java.util.Random();
	int subsetLength = system.subsetLength();
	Learn backprop = Learn.backprop;
	// double netin[] = new double[1];
	// double netout[] = new double[subsetLength];
	double sysin[] = new double[subsetLength];
	double sysout[] = new double[subsetLength];
	double timestep;
	int i, j;
	double error = 0.0;
	int count = 0;
	Frame frame;

	double netin[][] = new double[BUFFER_SIZE][1];
	double netout[][] = new double[BUFFER_SIZE][subsetLength];
	int netindex = 0;

	net = new FeedForwardNeuralNet( 1, HIDDEN_LAYERS, HIDDEN_NODES,
					subsetLength );
	
	frame = new Frame( "Neural Networks" );
	frame.setLayout( new BorderLayout() );
	frame.add( this, "Center" );
	frame.pack();
	frame.show();

	synchronized( systemLock ) {
	system.initialCondition( sysin );
	for( i=0 ; i<BUFFER_SIZE ; i++ ) {
	    netin[i][0] = random.nextDouble();
	    timestep = netin[i][0] * period;
	    system.evaluate( 0, sysin, timestep, sysout );
	    for( j=0 ; j<subsetLength ; j++ )
	    {
		netout[i][j] = (sysout[j] - bounds[j][0])
		             / (bounds[j][1] - bounds[j][0]);
	    }
	}
	}
	while( true ) {
	    netin[netindex][0] = random.nextDouble();
	    timestep = netin[netindex][0] * period;

	    synchronized( systemLock ) {
		system.evaluate( 0, sysin, timestep, sysout );
		
		for( i=0 ; i<subsetLength ; i++ )
		    netout[netindex][i] = (sysout[i] - bounds[i][0]) /
			(bounds[i][1] - bounds[i][0]);

		for( i=0 ; i<BUFFER_SIZE ; i++ )
		    error += net.backprop( netin[i], netout[i], 0.75, 0.5, 0.00 );
	    }

	    netindex = (netindex + 1) % BUFFER_SIZE;
	    count++;

	    if( count % BUFFER_CYCLES_PER_UPDATE == 0 ) {
		error /= BUFFER_SIZE * BUFFER_CYCLES_PER_UPDATE;
		System.out.println( "  Cycle "+(count*BUFFER_SIZE)+", error = "+error );

		repaint();
		
		if( error < system.maxApproxError() ||
		    (count*BUFFER_SIZE) > MAX_CYCLES )
		    break;

		error = 0;
	    }
	}
	System.out.println( "  Neural network trained." );

	frame.setVisible(false);
	frame.dispose();
    }

    public void output( PrintWriter out ) throws IOException {
	out.println( "    public static void approxValues(double out[], double timestep) {" );
	out.println( "        double in[] = new double[1];" );
	out.println( "        in[0] = timestep * "+(1.0 / period)+";" );
	out.println( "        evalNet( in, out );" );
	out.println( "    }" );
	out.println();
	FeedForwardNeuralNet.outputActFunc( out );
	out.println();
	net.outputCode( out, "evalNet", bounds );
    }
}
