/*
 * PeriodFinder.java
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.periodic;

import berkeley.cs.dmc.numericalMethods.Minimize;
import berkeley.cs.dmc.numericalMethods.MinimizeFunction;
import berkeley.cs.dmc.system.DynamicsSystem;

class PeriodFinder implements MinimizeFunction {

    private static final int WINDOW_SIZE = 512;
    private static final double MAX_PERIOD = 60.0 * 60.0; // 1 hr
    private static final double BURN_TIME = 2000.0;

    private DynamicsSystem dynamics_system = null;

    private double[][]	static_window = new double[WINDOW_SIZE][];
    private double[][]  sliding_window = new double[WINDOW_SIZE+2][];
    private int         window_zero = 0;
    private double[]	min_params = null;
    private double	period = 0.0;
    private double	t_step = 0.0;
    private int		num_vars = 0;
    private int	    	dimension = 0;
    private double[]	min_bounds = null;
    private double[]	max_bounds = null;

    PeriodFinder(DynamicsSystem sys) {
	dynamics_system = sys;
    }

    public double evaluate(double x, Object[] data) {
	double	diff = 0.0;
	double	t = period - t_step;
	double	offset = x - t;

	for ( int i = 0 ; i < WINDOW_SIZE ; i++ )
	{
	    dynamics_system.evaluate(t,
			     sliding_window[(window_zero+i)%(WINDOW_SIZE+2)],
			     t + offset, min_params);
	    for ( int j = 0 ; j < num_vars ; j++ )
	    {
		if ( dynamics_system.isCyclic(j%dimension) )
		{
		    while ( min_params[j] <= min_bounds[j%dimension] )
			min_params[j] += ( max_bounds[j%dimension]
					 - min_bounds[j%dimension] );
		    while ( min_params[j] > max_bounds[j%dimension] )
			min_params[j] -= ( max_bounds[j%dimension]
					 - min_bounds[j%dimension] );
		}
	    }
	    for ( int j = 0 ; j < num_vars ; j++ )
		diff += ( static_window[i][j] - min_params[j] )
		      * ( static_window[i][j] - min_params[j] );
	    t += t_step;
	    t = Math.floor(t / t_step + 0.1) * t_step;
	}

	return diff;
    }

    double findPeriod() {
	Object[]    min_data = new Object[1];
	double[]    params = null;
	double      magnitude = 0.0;
	double	    t, t_new;

	num_vars = dynamics_system.numVariables();
	dimension = dynamics_system.subsetLength();
	min_params = new double[num_vars];

	min_bounds = new double[dimension];
	max_bounds = new double[dimension];

	min_data[0] = dynamics_system;

	for ( int j = 0 ; j < dimension ; j++ )
	    if ( dynamics_system.isCyclic(j) )
	    {
		double[]    b = dynamics_system.getBound(j);
		min_bounds[j] = b[0];
		max_bounds[j] = b[1];
	    }


	for ( int i = 0 ; i < WINDOW_SIZE ; i++ )
	{
	    static_window[i] = new double[num_vars];
	    sliding_window[i] = new double[num_vars];
	}
	sliding_window[WINDOW_SIZE] = new double[num_vars];
	sliding_window[WINDOW_SIZE+1] = new double[num_vars];

	dynamics_system.initialCondition(static_window[0]);
	for ( int j = 0 ; j < num_vars ; j++ )
	{
	    sliding_window[0][j] = static_window[0][j];
	    magnitude += static_window[0][j] * static_window[0][j];
	}

	// Fill the static window.
	t = 0.0;
	t_step = (double)dynamics_system.bufferStep();
	for ( int i = 1 ; i < WINDOW_SIZE ; i++ )
	{
	    t_new = t + t_step;
	    t_new = Math.floor(t_new / t_step + 0.1) * t_step;
	    dynamics_system.evaluate(t, static_window[i-1],
				     t_new, static_window[i]);
	    for ( int j = 0 ; j < num_vars ; j++ )
	    {
		if ( dynamics_system.isCyclic(j%dimension) )
		{
		    while ( static_window[i][j] <= min_bounds[j%dimension] )
			static_window[i][j] += ( max_bounds[j%dimension]
					       - min_bounds[j%dimension] );
		    while ( static_window[i][j] > max_bounds[j%dimension] )
			static_window[i][j] -= ( max_bounds[j%dimension]
					       - min_bounds[j%dimension] );
		}
		sliding_window[i][j] = static_window[i][j];
		magnitude += static_window[i][j] * static_window[i][j];
	    }
	    t = t_new;
	}
	t_new = t + t_step;
	t_new = Math.floor(t_new / t_step + 0.1) * t_step;
	dynamics_system.evaluate(t, sliding_window[WINDOW_SIZE-1],
				 t_new, sliding_window[WINDOW_SIZE]);
	for ( int j = 0 ; j < num_vars ; j++ )
	{
	    if ( dynamics_system.isCyclic(j%dimension) )
	    {
		while ( sliding_window[WINDOW_SIZE][j]
		     <= min_bounds[j%dimension] )
		    sliding_window[WINDOW_SIZE][j] += ( max_bounds[j%dimension]
					            - min_bounds[j%dimension] );
		while ( sliding_window[WINDOW_SIZE][j]
		      > max_bounds[j%dimension] )
		    sliding_window[WINDOW_SIZE][j] -= ( max_bounds[j%dimension]
						    - min_bounds[j%dimension] );
	    }
	}

	double	diffm1 = 0.0;
	double	diff = 0.0;
	double	diffp1;
	for ( int i = 0 ; i < WINDOW_SIZE ; i++ )
	{
	    for ( int j = 0 ; j < num_vars ; j++ )
	    {
		diff += ( static_window[i][j] - sliding_window[i+1][j] )
		      * ( static_window[i][j] - sliding_window[i+1][j] );
	    }
	}
	period = t_step;
	window_zero = 0;

	double	t_window = t_new;
	Main	eval_func = new Main();

	// Look for local mimima.
	while ( period < MAX_PERIOD )
	{
	    int	i1 = ( window_zero + WINDOW_SIZE ) % ( WINDOW_SIZE + 2 );
	    int	i2 = ( window_zero + WINDOW_SIZE + 1 ) % ( WINDOW_SIZE + 2 );
	    dynamics_system.evaluate(t_window + period - t_step,
				     sliding_window[i1],
				     t_window + period,
				     sliding_window[i2]);
	    for ( int j = 0 ; j < num_vars ; j++ )
	    {
		if ( dynamics_system.isCyclic(j%dimension) )
		{
		    while ( sliding_window[i2][j] <= min_bounds[j%dimension] )
			sliding_window[i2][j] += ( max_bounds[j%dimension]
						 - min_bounds[j%dimension] );
		    while ( sliding_window[i2][j] > max_bounds[j%dimension] )
			sliding_window[i2][j] -= ( max_bounds[j%dimension]
						 - min_bounds[j%dimension] );
		}
	    }

	    diffp1 = 0.0;
	    for ( int i = 0 ; i < WINDOW_SIZE ; i++ )
	    {
		int	k = ( window_zero + 2 + i ) % ( WINDOW_SIZE + 2 );
		for ( int j = 0 ; j < num_vars ; j++ )
		    diffp1 += ( static_window[i][j] - sliding_window[k][j] )
			    * ( static_window[i][j] - sliding_window[k][j] );
	    }

	    if ( diffp1 > diff && diffm1 > diff ) // Local minima
	    {
		double[]    t_min = new double[1];
		double	min =
		    Minimize.brentMinimize1D(period - t_step,
					     period,
					     period + t_step,
					     this, min_data,
					     1.0e-6, t_min);
                System.out.println("Local min " + (min/magnitude) + " at "
		                 + t_min[0]);

		if ( min / magnitude < 1.0e-4 )
		    return t_min[0];
	    }

	    window_zero++; 
	    if ( window_zero == WINDOW_SIZE + 2 )
		window_zero = 0;
	    period += t_step;
	    period = Math.floor(period / t_step + 0.1) * t_step;

	    diffm1 = diff;
	    diff = diffp1;
	}

	return -1.0;
    }

}

