/*
 * ProbabilityTree.java
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.probtree;

import java.io.PrintWriter;
import java.util.Random;

/**
 * Probability trees are region subdivision trees (quadtrees, octrees)
 * that represent probability densities. The actual distribution stored is
 * a multi-variate discrete distribution, obtained by binning samples
 * according to the subdivision.
 *
 * @author Stephen Chenney
 * @version 1.0
 */
public class ProbabilityTree implements Cloneable {
    ProbabilityTreeNode	tree = null;
    int                 dimension = 0;
    int			max_depth;
    double		num_samples = 0;
    double[]	    	steps = null;
    double[]		mins = null;
    boolean		accumulated = false;
    Random              random = new Random();

    /**
     * Create a new tree with specified dimension and maximum depth.
     * Here, dimension means the size of the random vector that must
     * be stored. The maximum depth defines the resolution of the
     * subdivision. It is the number of times space will be subdivided in
     * each dimension. The steps and mins define the region of interest.
     *
     * @param dim The dimension of the tree.
     * @param max_depth The maximum depth.
     * @param steps The length of the region of interest in each dimension.
     * @param mins The minimum value in each dimension.
     */
    public ProbabilityTree(int dim, int max_depth, double[] steps,
			   double[] mins) {
	tree = new ProbabilityTreeNode(dim, max_depth);
	dimension = dim;
	this.max_depth = max_depth;
	try {
	    this.steps = (double[])steps.clone();
	    this.mins = (double[])mins.clone();
	} catch ( Exception e) {
	}
    }

    /**
     * The clone routine.
     */
    public Object clone() {
	ProbabilityTree copy = new ProbabilityTree(dimension, max_depth,
						   steps, mins);
	copy.num_samples = num_samples;
	copy.accumulated = accumulated;
	copy.tree.copyData(tree);

	return copy;
    }


    /**
     * Add the information from one sample to the tree.
     * 
     * @param value The value of the sample. This value is destroyed.
     */
    public void addSample(double[] value) {
	for ( int i = 0 ; i < value.length ; i++ )
	    value[i] = ( value[i] - mins[i] ) / steps[i];
	tree.addSample(value);
	num_samples += 1.0;
    }

    /**
     * Get the value of the density function at a particular point.
     * 
     * @param value The point. This value is destroyed.
     *
     * @return The density at that point.
     */
    public double density(double[] value) {
	for ( int i = 0 ; i < value.length ; i++ )
	    value[i] = ( value[i] - mins[i] ) / steps[i];
	return tree.density(value, 1.0);
    }


    /**
     * Accumulate the data in the tree. This normalizes the distribution,
     * builds cumulative distribution functions, and contracts the tree.
     * If the maximum difference between the probability of the child
     * nodes at any level is less than the threshold, then those children
     * will be combined. Lower thresholds result in less compression.
     * 
     * @param threshold The minimum difference between child nodes.
     */
    public void accumulate(double threshold) {
	tree.accumulate(threshold, num_samples);
	accumulated = true;
    }


    /**
     * Returns the difference between 2 trees, measured using the L1 norm.
     *
     * @param other The other tree.
     *
     * @return The difference.
     */
    public double difference(ProbabilityTree other) {
	return tree.differenceTree(other.tree);
    }


    /**
     * Write code to sample from the distribution.
     *
     * @param out The output writer.
     */
    public void output(PrintWriter out) {
	String	    lineSep = System.getProperty("line.separator");
	boolean[]   first = new boolean[1];
	double[]    last = new double[1];

	if ( ! accumulated )
	    accumulate(0.0);

	out.println("    private static final float[] cumul_distrib = {");
	first[0] = true;
	last[0] = 0.0;
	tree.outputCumulative(out, first, last, lineSep);
	out.println(" };");
	out.println();

	out.println("    private static final float[][][] prob_cells = {");
	first[0] = true;
	tree.outputCells(out, first, lineSep, mins, steps);
	out.println(" };");
	out.println();

	out.println("    private Random rand = new Random();");

	out.println("    private void sampleValues(double[] vals) {");
	out.println("        double prob = rand.nextDouble();");
	out.println("        int cell = cumul_distrib.length / 2;");
	out.println("        int min = 0, max = cumul_distrib.length - 1;");
	out.println("        while ( min < max ) {");
	out.println("          if ( prob < cumul_distrib[cell] )");
	out.println("            max = cell;");
	out.println("          else");
	out.println("            min = cell + 1;");
	out.println("          cell = ( max + min ) / 2;");
	out.println("        }");
	for ( int i = 0 ; i < dimension ; i++ )
	{
	    out.println("        vals[" + i + "] = "
		      + "prob_cells[cell][" + i + "][0] + rand.nextDouble() "
		      + "* prob_cells[cell][" + i + "][1];");
	}
	out.println("    }");
    }

    /**
     * Samples a value. It is assumed that the distribution is uniform within
     * each cell of the subdivision, so any real value within the cell
     * maybe returned..
     *
     * @params vals Will be filled with the sample.
     */
    public void getSample(double[] vals) {
	if ( ! accumulated )
	    accumulate(0.0);
	tree.getSample(vals, random, mins, steps);
    }


    /**
     * Returns the maximum density occuring in this distribution.
     */
    public double maxDensity() {
        double  volume = 1.0;
	for ( int i = 0 ; i < steps.length ; i++ )
	    volume *= steps[i];
        return tree.maxDensity(volume);
    }


    /**
     * Returns the number of cells with non-zero probability.
     *
     * @return The number of cells.
     */
    public int numNonZero() {
	return tree.numNonZero();
    }

    /**
     * Returns info about all the non-zero cells. In entries it puts
     * a weighted probability, the actual probability, and the dimensions
     * of the cell. Entries should have dimensions at least numNonZero()
     * by 2 * dimension + 2.
     *
     * @param entries Space for the requested data.
     */
    public void findNonZero(double[][] entries) {
	int[]   next = new int[] { 0 };
	tree.findNonZero(entries, next, mins, steps, 1.0);
    }


    /**
     * Clears the tree and resets all the children to have zero density.
     */
    public void clearData() {
	tree.clearData(dimension, max_depth);
	num_samples = 0;
	accumulated = false;
    }


    /**
     * Plots the distribution in the given pixel array.
     *
     * @param pixels The pixel array to fill.
     * @param width The width of the image.
     * @param height The height of the image.
     * @param max The maximum probability in the tree, from maxDensity().
     */
    public void plotPixels(int[] pixels, int width, int height, double max) {
        double  volume = 1.0;
	for ( int i = 0 ; i < steps.length ; i++ )
	    volume *= steps[i];
        tree.plotPixels(pixels, 0, 0, width, height, width, volume, max);
    }


    /**
     * Prints a representation of the tree on system output.
     */
    public void print()
    {
	tree.print("");
    }

}

