/*
 * ProbabilityTreeNode.java
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.probtree;

import java.io.PrintWriter;
import java.util.Random;

class ProbabilityTreeNode {
    double		    prob = 0.0;
    ProbabilityTreeNode[]   children = null;
    double[]		    cumulative = null;

    ProbabilityTreeNode(int dim, int height) {
	int length = 1;
	for ( int i = 0 ; i < dim ; i++ )
	    length *= 2;
	if ( height > 0 )
	{
	    children = new ProbabilityTreeNode[length];
	    for ( int i = 0 ; i < children.length ; i++ )
		children[i] = new ProbabilityTreeNode(dim, height - 1);
	}
    }

    void copyData(ProbabilityTreeNode other) {
	prob = other.prob;
	if ( other.cumulative != null )
	{
	    try {
		cumulative = (double[])other.cumulative.clone();
	    } catch (Exception e) {
	    }
	}
	if ( other.children == null )
	{
	    children = null;
	    return;
	}
	for ( int i = 0 ; i < children.length ; i++ )
	    children[i].copyData(other.children[i]);
    }


    private int stepDown(double[] val) {
	int res = 0;
	for ( int i = 0 ; i < val.length ; i++ )
	{
	    res <<= 1;
	    if ( val[i] >= 0.5 )
	    {
		val[i] = val[i] * 2.0 - 1.0;
		res++;
	    }
	    else
		val[i] *= 2.0;
	}
	return res;
    }

    void addSample(double[] posn) {
	if ( children == null )
	    prob += 1.0;
	else
	{
	    int	ind = stepDown(posn);
	    children[ind].addSample(posn);
	}
    }


    double density(double[] posn, double weight) {
	if ( children == null )
	    return prob / weight;
	else
	{
	    int	ind = stepDown(posn);
	    return children[ind].density(posn, weight / children.length);
	}
    }


    boolean accumulate(double threshold, double num_samples) {
	boolean	can_merge = true;

	if ( children == null )
	{
	    prob /= num_samples;
	    return true;
	}

	cumulative = new double[children.length];
	for ( int i = 0 ; i < children.length ; i++ )
	{
	    can_merge &= children[i].accumulate(threshold, num_samples);
	    if ( i == 0 )
		cumulative[i] = children[i].prob;
	    else
		cumulative[i] = children[i].prob + cumulative[i-1];

	}
	prob = cumulative[children.length-1];
	if ( prob != 0.0 )
	{
	    for ( int i = 0 ; i < children.length ; i++ )
		cumulative[i] /= cumulative[children.length-1];
	}
	else
	{
	    // Zero prob is mergable. All children must be 0.
	    cumulative = null;
	    children = null;
	    return true;
	}

	if ( can_merge )
	{
	    double  diff;
	    double  max_diff = 0.0;
	    for ( int i = 0 ; i < children.length ; i++ )
		for ( int j = i + 1 ; j < children.length ; j++ )
		{
		    diff = Math.abs(children[i].prob - children[j].prob);
		    if ( diff > max_diff )
			max_diff = diff;
		}

	    if ( max_diff < threshold )
	    {
		cumulative = null;
		children = null;
		return true;
	    }
	}

	return false;
    }

    private double differenceConst(double other) {
	if ( children == null )
	    return Math.abs(prob - other);

	double	sum = 0.0;
	for ( int i = 0 ; i < children.length ; i++ )
	    sum += children[i].differenceConst(other / children.length);
	return sum;
    }

    double differenceTree(ProbabilityTreeNode other) {
	if ( children == null )
	{
	    if ( other.children == null )
		return Math.abs(prob - other.prob);
	    else
		return other.differenceConst(prob);
	}

	if ( other.children == null )
	    return differenceConst(other.prob);

	double	sum = 0.0;
	for ( int i = 0 ; i < children.length ; i++ )
	    sum += children[i].differenceTree(other.children[i]);
	return sum;
    }


    void getSample(double[] vals, Random rand, double[] mins, double[] steps) {
	if ( children == null )
	{
	    for ( int i = 0 ; i < vals.length ; i++ )
		vals[i] = rand.nextDouble() * steps[i] + mins[i];
	    return;
	}

	double  rval = rand.nextDouble();
	int     index = 0;
	for ( index = 0 ; index < children.length && cumulative[index] < rval ;
	      index++ );

	double[]    new_mins = new double[mins.length];
	double[]    new_steps = new double[steps.length];

	for ( int i = 0 ; i < steps.length ; i++ )
	    new_steps[i] = steps[i] * 0.5;

	int	ind = index;
	for ( int i = mins.length - 1 ; i >= 0 ; i-- )
	{
	    if ( ( ind & 0x1 ) == 1 )
		new_mins[i] = mins[i] + new_steps[i];
	    else
		new_mins[i] = mins[i];
	    ind >>= 1;
	}
	children[index].getSample(vals, rand, new_mins, new_steps);
    }


    int numNonZero() {
	if ( children == null )
	{
	    if ( prob == 0.0 )
		return 0;
	    else
		return 1;
	}
	int sum = 0;
	for ( int i = 0 ; i < children.length ; i++ )
	    sum += children[i].numNonZero();
	return sum;
    }


    void findNonZero(double[][] entries, int[] next,
		     double[] min, double[] step, double weight) {
	if ( children == null )
	{
	    if ( prob != 0.0 )
	    {
		entries[next[0]][0] = prob / weight;
		entries[next[0]][1] = prob;
		for ( int i = 0 ; i < min.length ; i++ )
		    entries[next[0]][i+2] = min[i];
		for ( int i = 0 ; i < min.length ; i++ )
		    entries[next[0]][i+min.length+2] = step[i];
		next[0]++;
	    }
	    return;
	}

	double[]    new_mins = new double[min.length];
	double[]    new_steps = new double[step.length];

	for ( int i = 0 ; i < step.length ; i++ )
	    new_steps[i] = step[i] * 0.5;

	for ( int i = 0 ; i < children.length ; i++ )
	{
	    int	ind = i;
	    for ( int j = min.length - 1 ; j >= 0 ; j-- )
	    {
		if ( ( ind & 0x1 ) == 1 )
		    new_mins[j] = min[j] + new_steps[j];
		else
		    new_mins[j] = min[j];
		ind >>= 1;
	    }
	    children[i].findNonZero(entries, next, new_mins, new_steps,
				    weight / children.length);
	}
    }


    void outputCumulative(PrintWriter out, boolean[] first, double[] last,
			  String line_sep) {
	if ( children == null )
	{
	    if ( first[0] )
		first[0] = false;
	    else
		out.print("," + line_sep);
	    last[0] += prob;
	    out.print("        " + last[0] + "f");
	    return;
	}

	for ( int i = 0 ; i < children.length ; i++ )
	    children[i].outputCumulative(out, first, last, line_sep);
    }


    void outputCells(PrintWriter out, boolean[] first, String line_sep,
		     double[] mins, double[] steps) {
	if ( children == null )
	{
	    if ( first[0] )
		first[0] = false;
	    else
		out.print("," + line_sep);
	    out.print("        { { " + mins[0] + "f, " + steps[0] + "f }");
	    for ( int i = 1 ; i < mins.length ; i++ )
		out.print("," + line_sep
			+ "          { " + mins[i] + "f, " + steps[i] + "f }");
	    out.print(" }");
	    return;
	}

	double[]    new_mins = new double[mins.length];
	double[]    new_steps = new double[steps.length];

	for ( int i = 0 ; i < steps.length ; i++ )
	    new_steps[i] = steps[i] * 0.5;

	for ( int i = 0 ; i < children.length ; i++ )
	{
	    int	ind = i;
	    for ( int j = mins.length - 1 ; j >= 0 ; j-- )
	    {
		if ( ( ind & 0x1 ) == 1 )
		    new_mins[j] = mins[j] + new_steps[j];
		else
		    new_mins[j] = mins[j];
		ind >>= 1;
	    }
	    children[i].outputCells(out, first, line_sep, new_mins, new_steps);
	}
    }

    double maxDensity(double volume) {
	if ( children == null )
	    return prob / volume;
	
	double  max = 0.0;
	double  cell;

	for ( int i = 0 ; i < children.length ; i++ )
	{
	    cell = children[i].maxDensity(volume / children.length);
	    if ( cell > max )
		max = cell;
	}
	return max;
    }


    void clearData(int dim, int height) {
	if ( height > 0 )
	{
	    if ( children == null )
	    {
		int length = 1;
		for ( int i = 0 ; i < dim ; i++ )
		    length *= 2;
		children = new ProbabilityTreeNode[length];
		for ( int i = 0 ; i < children.length ; i++ )
		    children[i] = new ProbabilityTreeNode(dim, 0);
	    }
	    for ( int i = 0 ; i < children.length ; i++ )
		children[i].clearData(dim, height - 1);
	}
	prob = 0;
	cumulative = null;
    }


    void plotPixels(int[] pixels, int x, int y, int width, int height,
		    int full_width, double volume, double max) {
	if ( children == null )
	{
	    int intensity = (int)Math.floor((prob * 255) / (volume * max));
	    for ( int i = y ; i < y + height ; i++ )
	    {
		int offset = i * full_width;
		for ( int j = x ; j < x + width ; j++ )
		    pixels[offset + j] =
		      0xff000000 | (intensity<<16) | (intensity<<8) | intensity;
	    }
	    return;
	}

	int nw = width / 2;
	if ( children.length == 2 )
	{
	    children[0].plotPixels(pixels, x, y, nw, height, full_width,
				   volume * 0.5, max);
	    children[1].plotPixels(pixels, x+nw, y, nw, height, full_width,
				   volume * 0.5, max);
	    return;
	}

	int nh = height / 2;
	for ( int i = 0 ; i < 4 ; i++ )
	{
	    int min_x, min_y;
	    if ( ( i & 0x1 ) == 0 )
		min_y = y + nh;
	    else
	        min_y = y;
	    if ( ( i & 0x2 ) == 0 )
		min_x = x;
	    else
	        min_x = x + nw;
	    children[i].plotPixels(pixels, min_x, min_y, nw, nh, full_width,
				   volume / children.length, max);
	}
    }


    void print(String indent)
    {
	System.out.println(indent + prob);
	if ( children != null )
	    for ( int i = 0 ; i < children.length ; i++ )
		children[i].print(indent + " ");
    }
}

