/*
 * ProbabilityTreePlot.java
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.probtree;

import java.awt.Canvas;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.image.MemoryImageSource;


/**
 * Used to produce plots of probability trees, but it will only work for
 * two dimensional distributions, or it will plot the first two variables.
 * For one dimensional variables, it will plot stripes.
 *
 * @author Stephen Chenney
 * @version 1.0
 */
public class ProbabilityTreePlot extends Canvas implements Runnable {
    int width, height;
    int pixels[];
    MemoryImageSource source;
    Image image;
    Thread thread;
    ProbabilityTree distribution;

    /**
     * The constructor sets up a pixel array and window, and sets a
     * thread running to manage redraws and updates.
     *
     * @param distrib The distribution to be plotted.
     * @param width The desired plot width, in pixels.
     * @param height The desired plot height, in pixels.
     */
    public ProbabilityTreePlot(ProbabilityTree distrib, int width, int height )
    {
	int i, size = width * height;
	this.width = width;
	this.height = height;

	pixels = new int[width * height];

	for( i=0 ; i<size ; i++ )
	    pixels[i] = 0xff000000;

	source = new MemoryImageSource( width, height, pixels, 0, width );
	source.setAnimated(true);
	image = createImage(source);
	distribution = distrib;

	thread = new Thread(this);
	thread.start();
    }

    public void update( Graphics g ) {
	paint( g );
    }

    public void paint( Graphics g ) {
	Dimension size = getSize();
	g.drawImage(image, (size.width - width) / 2,
		    (size.height - height) / 2, this );
    }


    /**
     * Should be called whenever the distribution to plot is changed
     * and a plot update is required.
     */
    public void refreshData()
    {
	double  max_density = distribution.maxDensity();
	distribution.plotPixels(pixels, width, height, max_density);
	source.newPixels(0,0,width,height);
    }

    public void run()
    {
	if( thread != Thread.currentThread() )
	    throw new InternalError( "Unknown thread tried to run: "
				   + Thread.currentThread().getName() );

	refreshData();
    }

    public Dimension getPreferredSize()
    {
	return new Dimension( width, height );
    }

    public Dimension getMinimumSize()
    {
	return new Dimension( width, height );
    }
}
