/*
 * Box.java
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.rapid2d;

import java.util.Vector;
import java.io.PrintStream;

public class Box {
    Mat2D rotate    = new Mat2D();
    Vec2D translate = new Vec2D();

    Box P;
    Box N;
    Box U;

    Vector segs = new Vector();    // segments
    Vec2D  dim  = new Vec2D();  // half side length -- "radius"
    float b_size;                // max( dim.x, dim.y )
    float rad;                    // Radius = sqrt(dim.x^2 + dim.y^2)

    private static final double EPSILON = 1.0e-4;

    // box to model space: x_m = R*x_b + T
    // model to box space: x_b = R.T()*(x_m - T)

    public String toString() {
        return "box( R=" + rotate + ", T=" + translate + ", dim=" + dim + ", segments = " + ((segs == null) ? 0 : segs.size());
    }

    static int debug_level = 0;
    private static void debug( String s ) {
        for( int i=0 ; i<debug_level ; i++ )
            System.out.print( "  " );
        System.out.println( s );
    }

    private static void debugpush() { debug_level++; }
    private static void debugpop()  { debug_level--; }

    private boolean isLeaf() { return P == null && N == null; }
    private float  size()   { return b_size; }

    /**
     * Feed model into a Box a segment at a time.
     *
     * @param p end point of segment
     * @param q end point of segment
     * @param id segment id number (returned in collision test)
     */
    public void addSeg( Vec2D p, Vec2D q, int id ) {
        segs.addElement( new Segment(p,q,id) );
    }

    /**
     * Feed model into a Box a segment at a time
     *
     * @param seg the segment to add
     */
    public void addSeg( Segment seg ) {
        segs.addElement( seg );
    }

    public Box() {
        U = P = N = null;
        rotate.identity();
        translate.set( 0, 0 );
        dim.set( 1.0f, 1.0f );
        b_size = 1.0f;
        rad = 1.732051f;
    }

    public void buildHierarchy() {
        reorient();
        refit();
        putSegsInSelfCS();
        buildRecurse();
    }

    private void buildRecurse() {
        if( isLeaf() ) splitOnce();
        if( isLeaf() ) return;

        if( P != null ) P.buildRecurse();
        if( N != null ) N.buildRecurse();
    }

    private float meanPointOnAxis( int a ) {
        int size = segs.size();
        float p = 0;

        for( int i=0 ; i<size ; i++ ) {
            Segment seg = (Segment)segs.elementAt(i);
            p += seg.p1.comp(a) + seg.p2.comp(a);
        }

        return p / (size * 2.0f);
    }

    private void splitOnce() {
        int size = segs.size();
        int i;

        if( size < 3 ) return; // no possible split

        // want axis of largest dimension
        float md;
        int a;

        if( dim.x > dim.y ) { md = dim.x; a = 0; }
        else                { md = dim.y; a = 1; }

        if( md == 0.0 ) return;        // no split possible

        // create subboxes
        P = new Box();
        N = new Box();

        P.U = N.U = this;

        // partition elements accordingly
        float mp = meanPointOnAxis(a);
        float eps = dim.comp(a)/100.0f;

        for( i=0 ; i<size ; i++ ) {
            Segment seg = (Segment)segs.elementAt(i); // already on self CS

            float p = (seg.p1.comp(a) + seg.p2.comp(a)) * 0.5f;

            if     ( p > (mp+eps) )                  P.addSeg( seg );
            else if( p < (mp-eps) )                  N.addSeg( seg );
            else if( P.segs.size() < N.segs.size() ) P.addSeg( seg );
            else                                     N.addSeg( seg );
       }

       // abort the split if either P or N is still empty
       if( P.segs.size() == 0 || N.segs.size() == 0 ) {
           System.err.println( "@ERR: splitOnce - empty children" );
           System.err.println( "      axis = " + a );
           System.err.println( "      n = " + segs.size() + ", P.n = " + P.segs.size() + ", N.n = " + N.segs.size() );
           P = N = null;
           return;
       }

       // assign positions and extents to each box
       P.dim = new Vec2D( dim );
       P.reorient();
       P.refit();
       P.putSegsInSelfCS();

       N.dim = new Vec2D( dim );
       N.reorient();
       N.refit();
       N.putSegsInSelfCS();

       // free the data in this box
       segs = null;
   }


    /**
     * Translates segments from parent CS to self CS.  This routine
     * assumes that the segments are on the parent CS.
     */
    private void putSegsInSelfCS() {
        int size = segs.size();
        Vec2D t1 = new Vec2D();
        Vec2D t2 = new Vec2D();

        for( int i=0 ; i<size ; i++ ) {
            Segment seg = (Segment)segs.elementAt(i);

            t1.sub( seg.p1, translate );
            t2.sub( seg.p2, translate );
            seg.p1.mulT( rotate, t1 );
            seg.p2.mulT( rotate, t2 );
        }
    }

    /**
     * Keeps the orientation the same, but determines a new origin
     * and dimensions to fit the enclosed tris exactly.
     */
    private void refit() {
        int i;
        Segment seg;
        Vec2D p1 = new Vec2D();
        Vec2D p2 = new Vec2D();
        int size = segs.size();

        float maxx, minx;
        float maxy, miny;

        if( size == 0 ) return;

        seg = (Segment)segs.elementAt(0);
        p1.mulT( rotate, seg.p1 );
        minx = maxx = p1.x;
        miny = maxy = p1.y;

        for(i=0; i<size; i++ ) {
            seg = (Segment)segs.elementAt(i);
            p1.mulT( rotate, seg.p1 );
            p2.mulT( rotate, seg.p2 );

            if( maxx < p1.x ) maxx = p1.x;
            if( maxx < p2.x ) maxx = p2.x;
            if( minx > p1.x ) minx = p1.x;
            if( minx > p2.x ) minx = p2.x;
            if( maxy < p1.y ) maxy = p1.y;
            if( maxy < p2.y ) maxy = p2.y;
            if( miny > p1.y ) miny = p1.y;
            if( miny > p2.y ) miny = p2.y;
        }

        // place the origin in the proper position.
        p1.set( 0.5f*(maxx + minx), 0.5f*(maxy + miny) );

        translate.mul( rotate, p1 );

        // correct the dimensions
        dim.set( 0.5f*(maxx - minx), 0.5f*(maxy - miny) );

        b_size = Math.max( dim.x, dim.y );
        rad = (float)Math.sqrt( dim.x*dim.x + dim.y*dim.y );
    }


    private Mat2D getCov() {
        // find the max and min along each axis
        int i;
        int size = segs.size();
        Vec2D p   = new Vec2D( 0, 0 ); // center of mass
        Mat2D Cov = new Mat2D();

        Vec2D s1 = new Vec2D( 0, 0 );
        Mat2D s2 = new Mat2D( 0, 0, 0, 0 );

        for( i=0; i<size ; i++ ) {
            Segment seg = (Segment)segs.elementAt(i);
            Vec2D p1 = seg.p1;
            Vec2D p2 = seg.p2;

            s1.x += p1.x + p2.x;
            s1.y += p1.y + p2.y;

            s2.xx += (p1.x*p1.x + p2.x*p2.x);
            s2.yy += (p1.y*p1.y + p2.y*p2.y);
            s2.xy += (p1.x*p1.y + p2.x*p2.y);
        }

        float n = 2 * size;

        // Calculate covariances
        Cov.xx = s2.xx - s1.x*s1.x / n;
        Cov.yy = s2.yy - s1.y*s1.y / n;
        Cov.xy = s2.xy - s1.x*s1.y / n;
        Cov.yx = Cov.xy;

        return Cov;
    }

    void reorient() {
        if( segs.size() <= 0 )
            rotate.identity();
        else {
            Mat2D Cov = getCov();

            // now get eigenvectors
            Vec2D Evals = new Vec2D();
            Cov.eigen( rotate, Evals );

            if( Evals.y < Evals.x )
            {
                float u, xy, yy;
                // u = Evals.y;
                xy = rotate.xy;
                yy = rotate.yy;
                // Evals.y = Evals.x;
                rotate.xy = rotate.xx;
                rotate.yy = rotate.yx;
                // Evals.x = u;
                rotate.xx = xy;
                rotate.yx = yy;
            }
        }
    }

    static boolean intersect( Vec2D p1, Vec2D p2, Vec2D q1, Vec2D q2 ) {
        float x1, y1, x2, y2;
        float c1, c2;

        // Quick reject...
        if( p1.x < p2.x ) {
            if( q1.x < p1.x - EPSILON && q2.x < p1.x - EPSILON ) return false;
            if( q1.x - EPSILON > p2.x && q2.x - EPSILON > p2.x ) return false;
        } else {
            if( q1.x < p2.x - EPSILON && q2.x < p2.x - EPSILON ) return false;
            if( q1.x - EPSILON > p1.x && q2.x - EPSILON > p1.x ) return false;
        }

        if( p1.y < p2.y ) {
            if( q1.y < p1.y - EPSILON && q2.y < p1.y - EPSILON ) return false;
            if( q1.y - EPSILON > p2.y && q2.y - EPSILON > p2.y ) return false;
        } else {
            if( q1.y < p2.y - EPSILON && q2.y < p2.y - EPSILON ) return false;
            if( q1.y - EPSILON > p1.y && q2.y - EPSILON > p1.y ) return false;
        }

        /*
         * Straddle test:
         *  Check to see if the end points on a line are on opposite sides
         * of the reference line.  Opposite signs on the cross products
         * indicates this, thus check to see if their product is negative.
         * Also, if one end point lies on reference line, it straddles that
         * line.  Thus if either cross product is 0, it straddles.
         *
         * Two line segments intersect if and only if they straddle each
         * other, provided that passed trivial rejection.
         */

        // q1q2 straddles p1p2?
        x1 = p2.x - p1.x; y1 = p2.y - p1.y;
        x2 = q1.x - p1.x; y2 = q1.y - p1.y;
        c1 = x1*y2 - x2*y1;
        x2 = q2.x - p1.x; y2 = q2.y - p1.y;
        c2 = x1*y2 - x2*y1;
        if( (c2 * c1) > EPSILON ) return false;

        // p1p2 straddles q1q2?
        x1 = q2.x - q1.x; y1 = q2.y - q1.y;
        x2 = p1.x - q1.x; y2 = p1.y - q1.y;
        c1 = x1*y2 - x2*y1;
        x2 = p2.x - q1.x; y2 = p2.y - q1.y;
        c2 = x1*y2 - x2*y1;
        if( (c2 * c1) > EPSILON ) return false;

        return true;
    }

    static int segContact( Vector c, Box b1, Box b2, Mat2D R, Vec2D V ) {

        //  fprintf(stderr, "tri_contact: %d and %d tris\n",
        //      b1->tri_list.num_tris, b2->tri_list.num_tris);

        int i,j;
        int numContacts = 0;
        int size1 = b1.segs.size();
        int size2 = b2.segs.size();

        Vec2D j1 = new Vec2D();
        Vec2D j2 = new Vec2D();
        Vec2D tmp = new Vec2D();

        for( j=0; j<size2 ; j++ ) {
            Segment segj = (Segment)b2.segs.elementAt(j);

            tmp.mul( R, segj.p1 );
            j1.add( tmp, V );

            tmp.mul( R, segj.p2 );
            j2.add( tmp, V );

            for( i=0; i<size1 ; i++ ) {
                Segment segi = (Segment)b1.segs.elementAt(i);

                if( intersect( segi.p1, segi.p2, j1, j2 ) ) {
                    c.addElement( new Contact( segi.id, segj.id ) );
                    numContacts++;
                }
            }
        }

        /*
        for( i=0; i<size1 ; i++ ) {
            Segment segi = (Segment)b1.segs.elementAt(i);

            Vec2D i1 = segi.p1;
            Vec2D i2 = segi.p2;

            for( j=0; j<size2 ; j++ ) {
                Segment segj = (Segment)b2.segs.elementAt(j);

                Vec2D j1 = Vec2D.add( R.mul( segj.p1 ), V );
                Vec2D j2 = Vec2D.add( R.mul( segj.p2 ), V ); // possible opt: precalc in array

                if( intersect( i1, i2, j1[j], j2[j] ) ) {
                    c.addElement( new Contact( segi.id, segj.id ) );
                    numContacts++;
                }
            }
        }
        */

        return numContacts;
    }

    static boolean boxTest4( Mat2D B, Vec2D T, Vec2D a, Vec2D b ) {

        Mat2D Bf = new Mat2D( Math.abs( B.xx ), Math.abs( B.xy ),
                              Math.abs( B.yx ), Math.abs( B.yy ) );


        // Check x bounds in a's space
        if( Math.abs(T.x) > (a.x + b.x * Bf.xx + b.y * Bf.xy) )
            return false;

        // Check y bounds in a's space
        if( Math.abs(T.y) > (a.y + b.x * Bf.yx + b.y * Bf.yy) )
            return false;

        // Check x bounds in b's space
        if( Math.abs(T.x*B.xx + T.y*B.yx) > (b.x + a.x * Bf.xx + a.y * Bf.yx) )
            return false;

        // Check y bounds in b's space
        if( Math.abs(T.x*B.xy + T.y*B.yy) > (b.y + a.x * Bf.xy + a.y * Bf.yy) )
            return false;

        return true;
    }


    /*
     * determines collision status between two box trees.
     */
    static int collide( Vector c, Box b1, Box b2, Mat2D B, Vec2D T ) {

        Vec2D a = new Vec2D( b1.dim );
        Vec2D b = new Vec2D( b2.dim );

        if( !boxTest4( B, T, a, b ) ) return 0; // disjoint boxes

        Mat2D Bp = new Mat2D();
        Vec2D Tp = new Vec2D();
        Vec2D tmp = new Vec2D();

        int numContacts = 0;

        boolean b1leaf = b1.isLeaf();
        boolean b2leaf = b2.isLeaf();

        if( b1.isLeaf() && b2leaf ) return segContact( c, b1, b2, B, T );
        if( b2leaf || (!b1leaf && b1.size() > b2.size()) ) {
            // Bp = b1->P->pR.T() * B
            // Tp = b1->P->pR.T() * (T - b1->P->pT)

            // Bp = Mat2D.mulTxM( b1.P.rotate, B );
            // Tp = b1.P.rotate.mulT( Vec2D.sub( T, b1.P.translate ) );

            Bp.mulTxM( b1.P.rotate, B );
            tmp.sub( T, b1.P.translate );
            Tp.mulT( b1.P.rotate, tmp );

            numContacts += collide( c, b1.P, b2, Bp, Tp);

            // Bp = b1->N->pR.T() * B
            // Tp = b1->N->pR.T() * (T - b1->N->pT)

            // Bp = Mat2D.mulTxM( b1.N.rotate, B );
            // Tp = b1.N.rotate.mulT( Vec2D.sub( T, b1.N.translate ) );

            Bp.mulTxM( b1.N.rotate, B );
            tmp.sub( T, b1.N.translate );
            Tp.mulT( b1.N.rotate, tmp );

            numContacts += collide( c, b1.N, b2, Bp, Tp);

        } else {
            // Bp = B * b2->P->pR
            // Tp = B * b2->P->pT + T

            // Bp = Mat2D.mul( B, b2.P.rotate );
            // Tp = Vec2D.add( B.mul( b2.P.translate ), T );

            Bp.mul( B, b2.P.rotate );
            tmp.mul( B, b2.P.translate );
            Tp.add( tmp, T );

            numContacts += collide( c, b1, b2.P, Bp, Tp);

            // Bp = B * b2->N->pR
            // Tp = B * b2->N->pT + T

            // Bp = Mat2D.mul( B, b2.N.rotate );
            // Tp = Vec2D.add( B.mul( b2.N.translate ), T );

            Bp.mul( B, b2.N.rotate );
            tmp.mul( B, b2.N.translate );
            Tp.add( tmp, T );

            numContacts += collide( c, b1, b2.N, Bp, Tp);
        }

        return numContacts;
    }

    public static int collide( Vector c,
                        Mat2D R1, Vec2D T1, Box b1,
                        Mat2D R2, Vec2D T2, Box b2 )
    {
        Mat2D tR1 = new Mat2D();
        Mat2D tR2 = new Mat2D();
        Mat2D R   = new Mat2D();
        Vec2D tT1 = new Vec2D();
        Vec2D tT2 = new Vec2D();
        Vec2D T   = new Vec2D();

        if( b1 == null || b2 == null ) return 0;

        // (R1,T1) and (R2,T2) are how the two triangle sets are positioned
        // in world space.  (tR1,tT1) and (tR2,tT2) are how the top level
        // boxes are positioned in world space

        tR1.mul( R1, b1.rotate );                        // tR1 = R1 * b1->pR;
        T.mul( R1, b1.translate ); tT1.add( T, T1 );    // tT1 = R1 * b1->pT + T1;

        tR2.mul( R2, b2.rotate );                        // tR2 = R2 * b2->pR;
        T.mul( R2, b2.translate ); tT2.add( T, T2 );    // tT2 = R2 * b2->pT + T2;

        // (R,T) is the placement of b2's top level box within
        // the coordinate system of b1's top level box.

        R.mulTxM( tR1, tR2 );            // R = tR1.T()*tR2;
        tT2.sub( tT2, tT1 );
        T.mulT( tR1, tT2 );                // T = tR1.T()*(tT2-tT1);

        // make the call

        return collide(c, b1, b2, R, T);
    }

    private static final void print( PrintStream out, int tabstop, String string ) {
        for( int i=0 ; i<tabstop ; i++ )
            out.print( " " );

        out.println( string );
    }

    private final int printBox( PrintStream out, int tab, int tabsize, int segno ) {
        int mysegno = segno;

        print( out, tab, "new Box( new Mat2D( " + rotate.xx + ", " + rotate.xy + ", " );
        print( out, tab, "                    " + rotate.yx + ", " + rotate.yy + " )," );
        print( out, tab, "         new Vec2D( " + translate.x + ", " + translate.y + " )," );
        print( out, tab, "         new Vec2D( " + dim.x + ", " + dim.y + " )," );
        print( out, tab, "         " + b_size + ", " + rad + "," );
        print( out, tab, "         " + ((segs == null) ? "null" : ("segments" + mysegno)) + "," );

        if( P != null ) segno = P.printBox( out, tab+tabsize, tabsize, segno+1 );
        else print( out, tab + tabsize, "null," );
        if( N != null ) segno = N.printBox( out, tab+tabsize, tabsize, segno+1 );
        else print( out, tab + tabsize, "null," );

        print( out, tab, "        )" + (tab == tabsize ? ";" : ",") );

        return segno;
    }

    private final int printSegments( PrintStream out, int tab, int tabsize, int segno ) {
        int mysegno = segno;
        int indent = tab + tabsize;

        if( segs != null ) {
            print( out, tab, "Segment segments" + mysegno + "[] = new Segment["+segs.size()+"];" );
            print( out, tab, "static {" );

            for( int i=0 ; i<segs.size() ; i++ ) {
                Segment seg = (Segment)segs.elementAt(i);
                print( out, indent, "segments" + mysegno + "["+i+"] = new Segment( " + seg.id + "," );
                print( out, indent+tabsize, "new Vec2D( " + seg.p1.x + ", " + seg.p1.y + " )," );
                print( out, indent+tabsize, "new Vec2D( " + seg.p2.x + ", " + seg.p2.y + " ) );" );
            }
            print( out, tab, "}" );
        }

        if( P != null ) segno = P.printSegments( out, tab, tabsize, segno+1 );
        if( N != null ) segno = N.printSegments( out, tab, tabsize, segno+1 );
        return segno;
    }


    public void printBox( PrintStream out, int tabstop ) {
        printSegments( out, tabstop, tabstop, 0 );
        printBox( out, tabstop, tabstop, 0 );
    }

}
