/*
 * FixedPt.java
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.rapid2d;
/**
 * <p>The FixedPt is a class for doing fixed point arithmetic.  It is NOT a class
 * for fixed point numbers.  The actual fixed point numbers are stored in
 * integers and add, subtracted, multiplied, etc... using the routines of the
 * fixed point.  This may lead to confusion between integers and fixed point
 * values, but um... deal with it.</p>
 *
<pre>
int a, b;

a = FixedPt.toFixed( 100 );
b = FixedPt.toFixed( 1.25 );
System.out.println( "100 * 1.25 = " + FixedPt.toString( FixedPt.mul( a, b ) ) );
</pre>
 *
 * @author Jeffrey Ichnowski
 * @version 1.0
 */
public final class FixedPt {
  public static final int SHIFT = 16;

  /** Fixed point for 1.  Equivalent to toFixed(1) */
  public static final int ONE = toFixed(1);
  /** Fixed point for 0.  Equivalent to toFixed(0) */
  public static final int ZERO = toFixed(0);
  /** Fixed point for 0.5.  Equivalent to toFixed(0.5) */
  public static final int HALF = divByInt( ONE, 2 );

  public static final int fractMask = ONE - 1;
  public static final int intMask   = ~fractMask;


  /** Truncates a fixed point value to an integer.
   * @param a fixed point
   * @return integer
   */
  public static final int trunc( int a ) { return a >> SHIFT; }
  /** Rounds a fixed point value to the nearest integer
   * @param a fixed point
   * @return integer
   */
  public static final int round( int a ) { return (a>0) ? ((a+HALF) >> SHIFT) : ((a-HALF) >> SHIFT); }

  /** Returns the fractional potion of a fixed point value.
   * @param a fixed point
   * @return fixed point
   */
  public static final int fract( int a ) { return a & fractMask; }

  /** Rounds a fixed point value to the largest integer less than or equal.
   * @param a fixed point
   * @return fixed point
   */
  public static final int ceil( int a ) { return (a & intMask) + ((fract(a) > 0) ? 1 : 0); }
  /** Rounds a fixed point value to the smallest integer greater than or equal.
   * @param a fixed point
   * @return fixed point
   */
  public static final int floor( int a ) { return a & intMask; }

  /** Returns a floating point representation of a fixed point value
   * @param a fixed point
   * @return value as a float
   */
  public static final float toFloat( int a ) { return (float)a / (float)ONE; }
  /** Returns a double representation of a fixed point value
   * @param a fixed point
   * @return the value as a double
   */
  public static final double toDouble( int a ) { return (double)a / ONE; }

  /** Converts an integer to a fixed point value.
   * @param a integer
   * @return fixed point
   */
  public static final int toFixed( int a )   { return a << SHIFT; }
  /** Converts a float to a fixed point value.
   * @param a float
   * @return fixed point
   */
  public static final int toFixed( float a ) { return (int)(a * ONE); }
  /** Converts a double to a fixed point value.
   * @param a double
   * @return fixed point
   */
  public static final int toFixed( double a ) { return (int)(a * ONE); }

  /** Adds two fixed point values.  Addition may also be done using the '+'
   * operator on the integers representing fixed points.
   * @param a fixed point
   * @param b fixed point
   * @return fixed point
   */
  public static final int add( int a, int b ) { return a + b; }
  /** Subtracts two fixed point values.  Subtraction may also be done with the
   * '-' operator.
   * @param a fixed point
   * @param b fixed point
   * @return fixed point
   */
  public static final int sub( int a, int b ) { return a - b; }
  /** Negates a fixed point value.  Negation may also be done with the unary
   * '-' operator.
   * @param a fixed point
   * @return fixed point
   */
  public static final int neg( int a )        { return -a; }
  /** Multiplies two fixed point values
   * @param a fixed point
   * @param b fixed point
   * @return fixed point
   */
  public static final int mul( int a, int b ) { return (int)(((long)a * (long)b) >> SHIFT); }
  /** Divides two fixed point values
   * @param a fixed point
   * @param b fixed point
   * @return fixed point
   */
  public static final int div( int n, int d ) { return (int)((((long)n) << SHIFT) / d); }
  /** Divides a fixed point value by an integer
   * @param a fixed point
   * @param b integer
   * @return fixed point
   */
  public static final int divByInt( int n, int di ) { return n / di; }

  // The following should be optimized for fixed point math (ie look up tables)
  public static final int abs( int a ) { return (a < 0) ? -a : a; }
  public static final int acos( int a ) { return toFixed(Math.acos(toDouble(a))); }
  public static final int asin( int a ) { return toFixed(Math.asin(toDouble(a))); }
  public static final int atan( int a ) { return toFixed(Math.atan(toDouble(a))); }
  public static final int atan2( int a, int b ) { return toFixed(Math.atan2(toDouble(a),toDouble(b))); }
  public static final int cos( int a ) { return toFixed(Math.cos(toDouble(a))); }
  public static final int exp( int a ) { return toFixed(Math.exp(toDouble(a))); }
  public static final int log( int a ) { return toFixed(Math.log(toDouble(a))); }
  public static final int max( int a, int b ) { return (a > b) ? a : b; }
  public static final int min( int a, int b ) { return (a < b) ? a : b; }
  public static final int pow( int a, int b ) { return toFixed(Math.pow(toDouble(a),toDouble(b))); }
  public static final int sin( int a ) { return toFixed(Math.sin(toDouble(a))); }
  public static final int tan( int a ) { return toFixed(Math.tan(toDouble(a))); }

  /** Returns the square root of an integer.  A negative parameter will cause
   * an ArithmeticException.
   *
   * @param a fixed point
   * @return the square root (fixed point)
   */
  public static final int sqrt( int a ) {
    return toFixed( Math.sqrt( toDouble( a ) ) );
  }
  /**
   * Returns a string of the form "1 23/256" where 1 is the integer part, 23
   * is the fractional part, and 256 is the denominator.
   *
   * @param a FixedPt int value
   * @return the string representation of the parameter
   */
  public static final String toString( int a ) {
    String s;

    if( a < 0 ) {
      a = neg( a );
      s = "-";
    } else
      s = "";

    return s + trunc( a ) + ' ' + (a & fractMask) + '/' + ONE;
  }

  /* Minimalist regression test:
  static final void doTest( FixedPt f, int a, char op, int b ) {
    int r = 0;

    switch( op ) {
      case '-': r = f.sub( a, b ); break;
      case '+': r = f.add( a, b ); break;
      case '*': r = f.mul( a, b ); break;
      case '/': r = f.div( a, b ); break;
      case '|': r = f.divByInt( a, f.trunc( b ) ); break;
    }

    System.out.println( f.toString(a) + " " + op + " " + f.toString(b) + " = " + f.toString(r) + "\t" +
                        f.toFloat(a)  + " " + op + " " + f.toFloat(b)  + " = " + f.toFloat(r) );

  }

  public static void main( String args[] ) {
    FixedPt f = new FixedPt( 8 );

    System.out.println( "ONE  = " + f.toString( f.ONE )  + " (" + f.ONE  + ")" );
    System.out.println( "HALF = " + f.toString( f.HALF ) + " (" + f.HALF + ")" );

    doTest( f, f.toFixed( 100.25 ), '+', f.toFixed( 125.75 ) );
    doTest( f, f.toFixed( 32 ), '-', f.toFixed( 16.3 ) );
    doTest( f, f.toFixed( 100 ), '*', f.toFixed( -1.25 ) );
    doTest( f, f.toFixed( 1000 ), '/', f.toFixed( -3 ) );
    doTest( f, f.toFixed( 100 ), '|', f.toFixed( -3 ) );
  }
  */
}
