/*
 * FxBox.java
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.rapid2d;

import java.util.Vector;
import java.io.PrintStream;

public final class FxBox {

    FxMat rotate;
    FxVec translate;

    FxBox P;
    FxBox N;
    FxBox U;

    FxSeg segs[];                   // segments

    FxVec dim;                      // half side length -- "radius"

    int b_size;                     // max( dim.x, dim.y )
    int rad;                        // Radius = sqrt(dim.x^2 + dim.y^2)

    // box to model space: x_m = R*x_b + T
    // model to box space: x_b = R.T()*(x_m - T)

    private boolean isLeaf() { return P == null && N == null; }
    private int     size()   { return b_size; }

    public FxBox() {
        U = P = N = null;
        rotate = new FxMat();
        translate = new FxVec( 0, 0 );
        dim = new FxVec( 1, 1 );
        b_size = FixedPt.toFixed( 1.0f );
        rad = FixedPt.toFixed( 1.732051f );
    }

    public FxBox( Box b ) {
        rotate = new FxMat( b.rotate );
        translate = new FxVec( b.translate );

        if( b.segs == null || b.segs.size() == 0 )
            segs = null;
        else {
            segs = new FxSeg[ b.segs.size() ];
            for( int i=0 ; i<segs.length ; i++ )
                segs[i] = new FxSeg( (Segment)b.segs.elementAt(i) );
        }

        dim = new FxVec( b.dim );

        b_size = FixedPt.toFixed( b.b_size );
        rad = FixedPt.toFixed( b.rad );

        U = null;
        if( b.N == null ) N = null;
        else {
            N = new FxBox( b.N );
            N.U = this;
        }

        if( b.P == null ) P = null;
        else {
            P = new FxBox( b.P );
            P.U = this;
        }
    }

    public void print( PrintStream out ) {
        out.println( "\t" + rotate.xx + ", " + rotate.xy + ", " + rotate.yx + ", " + rotate.yy + "," );
        out.println( "\t" + translate.x + ", " + translate.y + ", " + dim.x + ", " + dim.y + "," );
        out.println( "\t" + b_size + ", " + rad + ", " + ((segs == null) ? 0 : segs.length) + "," );
        if( segs != null ) for( int i=0 ; i<segs.length ; i++ ) {
            FxSeg s = segs[i];
            out.println( "\t" + s.id + ", " + s.p1.x + ", " + s.p1.y + ", " + s.p2.x + ", " + s.p2.y + "," );
        }
        out.println( "\t" + ((N == null?0:1) + (P == null?0:2)) + "," );
        if( N != null ) N.print( out );
        if( P != null ) P.print( out );
    }

    static boolean intersect( FxVec p1, FxVec p2, FxVec q1, FxVec q2 ) {
        int x1, y1, x2, y2;
        int c1, c2;

        // Quick reject...
        if( p1.x < p2.x ) {
            if( q1.x < p1.x && q2.x < p1.x ) return false;
            if( q1.x > p2.x && q2.x > p2.x ) return false;
        } else {
            if( q1.x < p2.x && q2.x < p2.x ) return false;
            if( q1.x > p1.x && q2.x > p1.x ) return false;
        }

        if( p1.y < p2.y ) {
            if( q1.y < p1.y && q2.y < p1.y ) return false;
            if( q1.y > p2.y && q2.y > p2.y ) return false;
        } else {
            if( q1.y < p2.y && q2.y < p2.y ) return false;
            if( q1.y > p1.y && q2.y > p1.y ) return false;
        }

        /*
         * Straddle test:
         *  Check to see if the end points on a line are on opposite sides
         * of the reference line.  Opposite signs on the cross products
         * indicates this, thus check to see if their product is negative.
         * Also, if one end point lies on reference line, it straddles that
         * line.  Thus if either cross product is 0, it straddles.
         *
         * Two line segments intersect if and only if they straddle each
         * other, provided that passed trivial rejection.
         */

        // q1q2 straddles p1p2?
        x1 = p2.x - p1.x; y1 = p2.y - p1.y;
        x2 = q1.x - p1.x; y2 = q1.y - p1.y;
        c1 = FixedPt.mul( x1, y2 ) - FixedPt.mul( x2, y1 );
        x2 = q2.x - p1.x; y2 = q2.y - p1.y;
        c2 = FixedPt.mul( x1, y2 ) - FixedPt.mul( x2, y1 );
        if( FixedPt.mul( c2, c1 ) > 0 ) return false;

        // p1p2 straddles q1q2?
        x1 = q2.x - q1.x; y1 = q2.y - q1.y;
        x2 = p1.x - q1.x; y2 = p1.y - q1.y;
        c1 = FixedPt.mul( x1, y2 ) - FixedPt.mul( x2, y1 );
        x2 = p2.x - q1.x; y2 = p2.y - q1.y;
        c2 = FixedPt.mul( x1, y2 ) - FixedPt.mul( x2, y1 );
        if( FixedPt.mul( c2, c1 ) > 0 ) return false;

        return true;
    }

    static int segContact( Vector c, FxBox b1, FxBox b2, FxMat R, FxVec V ) {
        int i,j;
        int numContacts = 0;
        int size1 = b1.segs.length;
        int size2 = b2.segs.length;

        FxVec j1 = new FxVec();
        FxVec j2 = new FxVec();
        FxVec tmp = new FxVec();

        for( j=0; j<size2 ; j++ ) {
            FxSeg segj = b2.segs[j];

            tmp.mul( R, segj.p1 );
            j1.add( tmp, V );

            tmp.mul( R, segj.p2 );
            j2.add( tmp, V );

            for( i=0; i<size1 ; i++ ) {
                FxSeg segi = b1.segs[i];

                if( intersect( segi.p1, segi.p2, j1, j2 ) ) {
                    c.addElement( new Contact( segi.id, segj.id ) );
                    numContacts++;
                }
            }
        }

        return numContacts;
    }

    static boolean boxTest4( FxMat B, FxVec T, FxVec a, FxVec b ) {

        FxMat Bf = new FxMat( Math.abs( B.xx ), Math.abs( B.xy ),
                              Math.abs( B.yx ), Math.abs( B.yy ) );


        // Check x bounds in a's space
        if( Math.abs(T.x) > (a.x + FixedPt.mul(b.x,Bf.xx) + FixedPt.mul(b.y,Bf.xy)) )
            return false;

        // Check y bounds in a's space
        if( Math.abs(T.y) > (a.y + FixedPt.mul(b.x,Bf.yx) + FixedPt.mul(b.y,Bf.yy)) )
            return false;

        // Check x bounds in b's space
        if( Math.abs(FixedPt.mul(T.x,B.xx) + FixedPt.mul(T.y,B.yx)) >
            (b.x + FixedPt.mul(a.x,Bf.xx) + FixedPt.mul(a.y,Bf.yx)) )
            return false;

        // Check y bounds in b's space
        if( Math.abs(FixedPt.mul(T.x,B.xy) + FixedPt.mul(T.y,B.yy)) >
            (b.y + FixedPt.mul(a.x,Bf.xy) + FixedPt.mul(a.y,Bf.yy)) )
            return false;

        return true;
    }


    /*
     * determines collision status between two box trees.
     */
    static int collide( Vector c, FxBox b1, FxBox b2, FxMat B, FxVec T ) {

        FxVec a = new FxVec( b1.dim );
        FxVec b = new FxVec( b2.dim );

        if( !boxTest4( B, T, a, b ) ) return 0; // disjoint boxes

        FxMat Bp = new FxMat();
        FxVec Tp = new FxVec();
        FxVec tmp = new FxVec();

        int numContacts = 0;

        boolean b1leaf = b1.isLeaf();
        boolean b2leaf = b2.isLeaf();

        if( b1leaf && b2leaf ) return segContact( c, b1, b2, B, T );
        if( b2leaf || (!b1leaf && b1.size() > b2.size()) ) {
            // Bp = b1->P->pR.T() * B
            // Tp = b1->P->pR.T() * (T - b1->P->pT)

            // Bp = Mat2D.mulTxM( b1.P.rotate, B );
            // Tp = b1.P.rotate.mulT( Vec2D.sub( T, b1.P.translate ) );

            Bp.mulTxM( b1.P.rotate, B );
            tmp.sub( T, b1.P.translate );
            Tp.mulT( b1.P.rotate, tmp );

            numContacts += collide( c, b1.P, b2, Bp, Tp);

            // Bp = b1->N->pR.T() * B
            // Tp = b1->N->pR.T() * (T - b1->N->pT)

            // Bp = Mat2D.mulTxM( b1.N.rotate, B );
            // Tp = b1.N.rotate.mulT( Vec2D.sub( T, b1.N.translate ) );

            Bp.mulTxM( b1.N.rotate, B );
            tmp.sub( T, b1.N.translate );
            Tp.mulT( b1.N.rotate, tmp );

            numContacts += collide( c, b1.N, b2, Bp, Tp);

        } else {
            // Bp = B * b2->P->pR
            // Tp = B * b2->P->pT + T

            // Bp = Mat2D.mul( B, b2.P.rotate );
            // Tp = Vec2D.add( B.mul( b2.P.translate ), T );

            Bp.mul( B, b2.P.rotate );
            tmp.mul( B, b2.P.translate );
            Tp.add( tmp, T );

            numContacts += collide( c, b1, b2.P, Bp, Tp);

            // Bp = B * b2->N->pR
            // Tp = B * b2->N->pT + T

            // Bp = Mat2D.mul( B, b2.N.rotate );
            // Tp = Vec2D.add( B.mul( b2.N.translate ), T );

            Bp.mul( B, b2.N.rotate );
            tmp.mul( B, b2.N.translate );
            Tp.add( tmp, T );

            numContacts += collide( c, b1, b2.N, Bp, Tp);
        }

        return numContacts;
    }

    static int collide( Vector c,
                        FxMat R1, FxVec T1, FxBox b1,
                        FxMat R2, FxVec T2, FxBox b2 )
    {
        FxMat tR1 = new FxMat();
        FxMat tR2 = new FxMat();
        FxMat R   = new FxMat();
        FxVec tT1 = new FxVec();
        FxVec tT2 = new FxVec();
        FxVec T   = new FxVec();

        if( b1 == null || b2 == null ) return 0;

        // (R1,T1) and (R2,T2) are how the two triangle sets are positioned
        // in world space.  (tR1,tT1) and (tR2,tT2) are how the top level
        // boxes are positioned in world space

        tR1.mul( R1, b1.rotate );                        // tR1 = R1 * b1->pR;
        T.mul( R1, b1.translate ); tT1.add( T, T1 );    // tT1 = R1 * b1->pT + T1;

        tR2.mul( R2, b2.rotate );                        // tR2 = R2 * b2->pR;
        T.mul( R2, b2.translate ); tT2.add( T, T2 );    // tT2 = R2 * b2->pT + T2;

        // (R,T) is the placement of b2's top level box within
        // the coordinate system of b1's top level box.

        R.mulTxM( tR1, tR2 );            // R = tR1.T()*tR2;
        tT2.sub( tT2, tT1 );
        T.mulT( tR1, tT2 );                // T = tR1.T()*(tT2-tT1);

        // make the call

        return collide(c, b1, b2, R, T);
    }

}
