/*
 * Mat2D.java
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.rapid2d;

public final class Mat2D {
	public float xx, xy;
	public float yx, yy;

	public Mat2D() { identity(); }
	public Mat2D( Mat2D m ) { set( m.xx, m.xy, m.yx, m.yy ); }
	public Mat2D( float a, float b, float c, float d ) { set( a,b,c,d ); }

	public final void identity() { set( 1,0, 0,1 ); }
	public final void set( float a, float b, float c, float d ) { xx=a; xy=b; yx=c; yy=d; }

	public final void transpose( Mat2D v ) { xx=v.xx; xy=v.yx; yx=v.xy; yy=v.yy; }

	public final void mul( Mat2D a, Mat2D b ) {
		xx = a.xx * b.xx + a.xy * b.yx;
		xy = a.xx * b.xy + a.xy * b.yy;
		yx = a.yx * b.xx + a.yy * b.yx;
		yy = a.yx * b.xy + a.yy * b.yy;
	}

	public final void mulTxM( Mat2D a, Mat2D b ) {
		xx = a.xx * b.xx + a.yx * b.yx;
		xy = a.xx * b.xy + a.yx * b.yy;
		yx = a.xy * b.xx + a.yy * b.yx;
		yy = a.xy * b.xy + a.yy * b.yy;
	}

	public final float comp( int r, int c ) {
		if( r==0 ) {
			if     ( c==0 ) return xx;
			else if( c==1 ) return xy;
		} else if( r == 1 ) {
			if	   ( c==0 ) return yx;
			else if( c==1 ) return yy;
		}

		throw new ArrayIndexOutOfBoundsException( "bad index (" + r + "," +c+ ")" );
	}

	/**
	 * Calculate the eigenvalues and eigenvectors of the matrix and
	 * return them in the passed matrix and vector.
	 *
	 * @param vout where the eigen vectors are put
	 * @param dout where the eigen values are put
	 */
	public final void eigen2( Mat2D vout, Vec2D dout ) {

		/*
		 * for the matrix:
	 	 *
		 *  | a b |
		 *  | c d |
		 *
		 * The two eigen values are:
		 *
		 * ( a + d +/- sqrt( (a-d)^2 + 4bc ) ) / 2
		 *
		 * The coresponding eigen vectors are:
		 *
		 * | t * ( a - d +/- sqrt( (a-d)^2 + 4bc ) ) / (2 * c) |
		 * |                     t                             |
		 *
		 *                     - or -
		 *
		 * |                     s                             |
		 * | s * ( d - a +/- sqrt( (a-d)^2 + 4bc ) ) / (2 * b) |
		 */

		float ad = (xx + yy);
		float root = (xx - yy);
		root *= root;
		root += 4 * xy * yx;

		if( root < 0.0 ) {
			System.err.println( "@Err: eigen - no real eigenvalues\n" );
			return;
		}

		root = (float)Math.sqrt( root );

		// Calculate the eigan values
		dout.x = (ad + root) * 0.5f;
		dout.y = (ad - root) * 0.5f;

		// Calculate the eigan vectors

		if( xy != 0.0 ) {        // avoid division by zero
			ad = yy - xx;
			vout.xx = (ad + root);
			vout.xy = (ad - root);
			vout.yy = vout.yx = 2.0f * xy;
		} else if( yx != 0.0 ) {
			ad = xx - yy;
			vout.yx = (ad + root);
			vout.yy = (ad - root);
			vout.xx = vout.xy = 2.0f * yx;
		} else {
			vout.identity();
			return;
		}

		// Normalized the eigen vectors.
		root = 1.0f / (float)Math.sqrt( vout.xx * vout.xx + vout.yx * vout.yx );
		vout.xx *= root;
		vout.yx *= root;

		root = 1.0f / (float)Math.sqrt( vout.xy * vout.xy + vout.yy * vout.yy );
		vout.xy *= root;
		vout.yy *= root;
	}

	public final void eigen( Mat2D v, Vec2D d )
	{
		int n = 2;
		int j,iq,ip,i;
		float tresh,theta,tau,t,sm,s,h,g,c;
		int nrot;
		Vec2D b = new Vec2D();
		Vec2D z = new Vec2D();
		Mat2D a = new Mat2D( this );

		v.identity();

		b.set( a.xx, a.yy );
		d.set( a.xx, a.yy );
		z.zero();

		nrot = 0;

		for(i=0; i<50; i++) {

			sm = Math.abs( a.xy );
			if( sm == 0.0 ) return;

			if (i < 3) tresh=0.2f*sm/(n*n);
			else tresh=0.0f;

			g = 100.0f*Math.abs( a.xy );

			if( i>3 && Math.abs(d.x)+g == Math.abs(d.x) &&
					   Math.abs(d.y)+g == Math.abs(d.y) )
				a.xy = 0.0f;
			else if( Math.abs(a.xy) > tresh ) {
				h = d.y - d.x;
				if( Math.abs(h)+g == Math.abs(h)) t= a.xy / h;
				else {
					theta=0.5f*h / a.xy;
					t=1.0f/(Math.abs(theta)+(float)Math.sqrt(1.0f+theta*theta));
					if( theta < 0.0f ) t = -t;
				}
				c=1.0f/(float)Math.sqrt(1+t*t);
				s=t*c;
				tau=s/(1.0f+c);
				h=t*a.xy;
				z.x -= h;
				z.y += h;
				d.x -= h;
				d.y += h;
				a.xy = 0.0f;

				g = v.xx;
				h = v.xy;
				v.xx = g-s*(h+g*tau);
				v.xy = h+s*(g-h*tau);

				g = v.yx;
				h = v.yy;
				v.yx = g-s*(h+g*tau);
				v.yy = h+s*(g-h*tau);
			}

			b.x += z.x;
			d.x = b.x;
			z.x = 0.0f;

			b.y += z.y;
			d.y = b.y;
			z.y = 0.0f;
		}

		System.err.println( "eigen: too many iterations in Jacobi transform." );

		return;
	}

	public final String toString() {
		return "[" + (float)xx + ", " + (float)xy + " / " + (float)yx + ", " + (float)yy + "]";
	}

}
