/*
 * DynamicsSystems.java
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.system;


/**
 * The DynamicsSystem interface is intended to provide a description of
 * a dynamical system sufficient for that system to be analysed by
 * transformation software, either the dynamics compiler of the periodic
 * system compiler.
 *
 * @author Stephen Chenney and Jeffrey Ichnowski
 * @version 1.0
 */
public interface DynamicsSystem {
    /**
     * Constant to indicate that a state variable is evaluated as
     * a closed form function of time, and hence requires no analysis.
     */
    public static final int CLOSED_FORM = 0;
    /**
     * Constant to indicate that a state variable is evaluated
     * using some interative method, such as numerical integration.
     */
    public static final int ITERATIVE = 1;
    /**
     * Constant to indicate that a state variable represents the first
     * derivative of the preceeding variable int he array with respect
     * to time.
     */
    public static final int FIRST_DERIVATIVE = 2;
    /**
     * Constant to indicate that a state variable represents the
     * second derivative of an earlier state variable with respect to
     * time. It is assumed the basic variable is that appearing 2 previous
     * in the state vector.
     */
    public static final int SECOND_DERIVATIVE = 3;

    /**
     * @return The number of state variables for the system.
     */
    public int numVariables();

    /**
     * Fields are the things that actually drive animations. It may
     * be that the state variables are best for describing state, but
     * don't correspond to geometric transformation. In such cases,
     * use fields. Note also that derivatives frequently appear as
     * state variables, but rarely as fields.
     *
     * @return the number of fields used to set transforms.
     */
    public int numFields();

    /**
     * Return the length of subsets with identical behaviour. This
     * number should be a divisor of the total number of variables.
     *
     * @return The length of indentically behaved subsets.
     */
    public int subsetLength();

    /**
     * Return the evaluation techniques for each variable.
     * ind will be less than subsetLength().
     *
     * @param ind The index of the variable in question.
     *
     * @return One of CLOSED_FORM, ITERATIVE, FIRST_DERIVATIVE
     *         or SECOND_DERIVATIVE
     */
    public int evaluationTechnique(int ind);

    /**
     * @return The period of any driving function, if one exists.
     *         0 if none exists.
     */
    public double period();

    /**
     * Return an array indicating which dimensions of the phase space
     * are cyclic. ind will be less than subsetLength(). By cyclic we mean
     * they wrap around modulo some number, such as an angle can be measured
     * modulo 2*PI. So any angular values should return true. Basically
     * this makes bounding easier.
     *
     * @param ind The index of the state variable in question.
     *
     * @return True if the variable can be measured modulo somthing.
     */
    public boolean isCyclic(int ind);

    /**
     * Get the bound for a given variable index.
     * ind will be less than subsetLength(). This method will only be called
     * for state variables that are cyclic, as given by isCyclic().
     *
     * @param ind The index of the variable in question.
     *
     * @return A 2d array, where the first element is the min and the
     *         second the max.
     */
    public double[] getBound(int ind);

    /**
     * Provide an initial condition in p. Returns true if this is a random
     * value. p will be the full set of parameters. This function should also
     * do burn in - ie the values returned are assumed to be typical for the
     * system. Finally, the values returned should be within their expected
     * bounds, particularly for cyclic variables.
     *
     * @param p Space for the initial condition. The method should fill this.
     *
     * @return True if the value is random, false if the same value will
     *         always be returned.
     */
    public boolean initialCondition(double[] p);

    /**
     * Evaluate the system from t_last, where it had state variables p_last,
     * to t_now and state variables p_now. p_last and p_now are the full
     * set of state variables. The value returned should be the percentage of
     * the maximum amount of work for a given frame. Returning 1.0 indicates
     * that no more work should be done for this system this frame, while
     * 0.1 indicates that 10 times more work could be done.
     *
     * @param t_last The time for the last set of state variables.
     * @param p_last The set of state variables at time t_last.
     * @param t_now The time to evaluate the state variables.
     * @param p_now The array to be filled with the new state variable values.
     *
     * @return The amount of work done, between 0 and 1.
     */
    public float evaluate(double t_last, double[] p_last,
			  double t_now, double[] p_now);

    /**
     * Given the state variables and time, fill in the field values.
     *
     * @param t The time corresponding to these values.
     * @param params The state variable values at time t.
     * @param fields The array of field values to be filled by the method.
     */
    public void evaluateFields(double t, double[] params, float[] fields);


    /**
     * Test the quality of an approximation. This routine is free to modify
     * the value before returning. The third argument is true if the
     * value can be discarded. In this case, the function should return true
     * if the value is usable, and false if it should be discarded.
     * p wil be the full set of state variables, and t is the evaluation time.
     *
     * @param t The time for the given state.
     * @param p The state variables. The method may modify these.
     * @param discard True if the method may discard the given state completely.
     *
     * @return True if the value is usable, False if it should be discarded.
     */
    public boolean checkState(double t, double[] p, boolean discard);

    /**
     * Return the time between buffered values that should be used. This is
     * really a measure of how quickly the system changes state, because
     * continuous values will be linearly interpolated between buffered
     * values. Also, the system will be asked to evaluate over periods
     * that are multiples of the buffer step.
     *
     * @return The interval to use between values in the state buffer.
     */
    public float bufferStep();

    /**
     * Return the number of time values that should be buffered. The
     * more buffered, the less chance the system will be without state.
     * But more buffered means much more memory used.
     *
     * @return The number of state entries in the buffer
     */
    public int bufferSize();


    /**
     * Return the resolutions of each variable. This defines the resolution
     * of the cell decomposition used for approximation. Higher resolutions
     * means better approximations, but much more memory used and bigger
     * final file sizes.
     *
     * @return An array giving the resolution of each state variable.
     */
    public double[] resolutions();


    /**
     * Return the threshold for the number of cells that should be tested
     * for the sampling threshold. N cells will be tested, where the sum of the
     * probability of the N cells is the smallest number greater than the
     * value returned by this function.
     *
     * @return The amount of probability density to test for the sampling
     *         threshold.
     */
    public double mixTestMax();

    /**
     * Return the difference between distributions required to measure
     * mixing in determining the sampling threshold. Higher values means
     * a weaker test, and smaller files and more efficient dynamics, but
     * with looser approximations. Lower values make the test harder.
     *
     * @return The minimum distance, between 0 and 1.
     */
    public double mixDiff();


    /**
     * Return the maximum allowable error in an approximation. Smaller
     * values give better results. The error of an approximation is
     * the sum of the squared error for each scaled sample. In other
     * samples are scaled between 0-1, the approximation is scaled identically,
     * then the difference is computed, squared and added to the total.
     *
     * You can always set this value to be very low and stop the approximation
     * learning by hand when you are happy with the quality.
     *
     * @return The maximum approximation error.
     */
    public double maxApproxError();

}


