/*
 * DynamicsVariable.java
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.system;


/**
 * The DynamicsVariable class is intended to provide a simple
 * standard interface between dynamical systems and VRML. Any
 * java classes used to animate geometry in the dmod modeller
 * must implement this interface.
 *
 * @author Stephen Chenney and Jeffery Icknowski
 * @version 1.0
 */
public interface DynamicsVariable {
    /**
     * @return The number of state variables in the system.
     */
    public int numVariables();

    /**
     * @return The number of fields used to set transforms.
     */
    public int numFields();

    /**
     * Return the time between buffered values that should be used. This is
     * really a measure of how quickly the system changes state, because
     * continuous values will be linearly interpolated between buffered
     * values. Also, the system will be asked to evaluate over periods
     * that are multiples of the buffer step.
     *
     * @return The interval to use between values in the state buffer.
     */
    public float bufferStep();

    /**
     * Return the number of time values that should be buffered. The
     * more buffered, the less chance the system will be without state.
     * But more buffered means much more memory used.
     *
     * @return The number of state entries in the buffer
     */
    public int bufferSize();

    /**
     * Provide an initial condition in p. Returns true if this is a random
     * value. p will be the full set of parameters. This function should also
     * do burn in - ie the values returned are assumed to be typical for the
     * system. Finally, the values returned should be within their expected
     * bounds, particularly for cyclic variables.
     *
     * @param p Space for the initial condition. The method should fill this.
     *
     * @return True if the value is random, false if the same value will
     *         always be returned.
     */
    public boolean initialCondition(double[] p);

    /**
     * Generate a state update, given the old state and time. In practice,
     * last_time and last_params will be the most recent time the
     * system was evaluated, corresponding to the latest buffer values.
     * Last_seen will be the last time the system was seen by the viewer.
     * It will be less than or equal to last_time. The function should
     * fill the new_params corresponding to new_time, and return the
     * amount of work done, a number between 0 and 1 where 0 indicates
     * nothing was done and 1 indicates enough work was done for one
     * frame of the animation.
     *
     * @param last_time The last time the system state was evaluated.
     * @param last_params The state variables evaluated at last_time.
     * @param last_seen The last time the system was seen by a viewer.
     * @param seen_params The state variables evaluated at last_seen.
     * @param new_time The new time to evaluate state.
     * @param new_params The method should fill this array with the new
     *        state variable values.
     *
     * @return A measure of the amount of work done, between 0 and 1.
     */
    public float newState(double last_time, double[] last_params,
                          double last_seen, double[] seen_params,
                          double new_time, double[] new_params);

    /**
     * Given the state variables and time, fill in the field values.
     *
     * @param t The time corresponding to these values.
     * @param params The state variable values at time t.
     * @param fields The array of field values to be filled by the method.
     */
    public void  evaluateFields(double t, double[] params, float[] fields);
}


