/*
 * Main.java
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.tovrml;

import java.io.*;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;

// import dynamics.model.parser.ParseException;
import berkeley.cs.dmc.model.parser.DynamicsParser;
import berkeley.cs.dmc.model.*;

import berkeley.cs.dmc.system.DynamicsVariable;


/**
 * An application class for transforming .dyn files, or dynamic models
 * into VRML and Java.
 *
 * @author Stephen Chenney and Jeffrey Ichnowski
 * @version 1.0
 */
public class Main {

    private static String[] timeStep_template = {
"  private void timeStep(MFFloat buffer, SFFloat lastTime,",
"                        SFFloat lastSeen, double bufStep, double phase,",
"                        DynamicsVariable var, double time, float[] fields,",
"                        int num_p, int num_b) {",
"    time += phase;",
"",
"    double last_time = (double)lastTime.getValue();",
"    double last_seen = (double)lastSeen.getValue();",
"",
"    if ( num_b == 1 )",
"    {",
"      for ( int i = 0 ; i < num_p ; i++ )",
"        old_params[i] = (double)buffer.get1Value(i);",
"      var.newState(last_time, old_params,",
"                   last_time, old_params,",
"                   time, new_params);",
"      for ( int i = 0 ; i < num_p ; i++ )",
"        buffer.set1Value(i, (float)new_params[i]);",
"      lastTime.setValue((float)time);",
"      var.evaluateFields(time, new_params, fields);",
"      return;",
"    }",
"",
"    int    low_index = (int)Math.floor(time / bufStep);",
"    double low_time = low_index * bufStep;",
"    int    high_index = ( low_index + 1 ) % num_b;",
"    int    last_index = (int)Math.floor((last_time + 0.001) / bufStep) % num_b;",
"    int    seen_index = (int)Math.floor((last_seen + 0.001) / bufStep) % num_b;",
"",
"    float work_done = 0.0f;",
"",
"    int base = 0;",
"",
"    low_index = low_index % num_b;",
"",
"    base = last_index * num_p;",
"    for ( int i = 0 ; i < num_p ; i++ )",
"      old_params[i] = (double)buffer.get1Value(base + i);",
"    base = seen_index * num_p;",
"    for ( int i = 0 ; i < num_p ; i++ )",
"      seen_params[i] = (double)buffer.get1Value(base + i);",
"",
"    if ( low_time > last_time + 0.00001 )",
"    {",
"      work_done = var.newState(last_time, old_params,",
"                               last_seen, seen_params,",
"                               low_time, new_params);",
"      for ( int i = 0 ; i < num_p ; i++ )",
"        old_params[i] = new_params[i];",
"      base = low_index * num_p;",
"      for ( int i = 0 ; i < num_p ; i++ )",
"        buffer.set1Value(base + i, (float)new_params[i]);",
"      last_time = low_time;",
"      last_index = low_index;",
"    }",
"    if ( low_time + bufStep > last_time + 0.00001 )",
"    {",
"      work_done += var.newState(low_time, old_params,",
"                                last_seen, seen_params,",
"                                low_time + bufStep, new_params);",
"      for ( int i = 0 ; i < num_p ; i++ )",
"        old_params[i] = new_params[i];",
"      base = high_index * num_p;",
"      for ( int i = 0 ; i < num_p ; i++ )",
"        buffer.set1Value(base + i, (float)new_params[i]);",
"      last_time = low_time + bufStep;",
"      last_index = high_index;",
"    }",
"",
"",
"    // Guaranteed to have low and high time buffer positions filled.",
"    int slots_used = last_index - low_index + 1;",
"    if ( slots_used < 1 )",
"      slots_used += num_b;",
"    base = 0;",
"    while (  work_done < 1.0f && slots_used < num_b )",
"    {",
"      work_done += var.newState(last_time, old_params,",
"                                last_seen, seen_params,",
"                                last_time + bufStep, new_params);",
"      last_index = ( last_index + 1 ) % num_b;",
"      base = last_index * num_p;",
"      for ( int i = 0 ; i < num_p ; i++ )",
"        old_params[i] = new_params[i];",
"      for ( int i = 0 ; i < num_p ; i++ )",
"        buffer.set1Value(base + i, (float)new_params[i]);",
"      last_time += bufStep;",
"      last_time = Math.floor(last_time / bufStep + 0.1) * bufStep;",
"      slots_used++;",
"    }",
"",
"    lastTime.setValue((float)last_time);",
"    lastSeen.setValue((float)(low_time + bufStep));",
"",
"    double fraction = ( time - low_time ) / bufStep;",
"    int base1 = high_index * num_p;",
"    base = low_index * num_p;",
"    for ( int i = 0 ; i < num_p ; i++ )",
"      new_params[i] = fraction * buffer.get1Value(base1 + i)",
"                    + ( 1.0 - fraction ) * buffer.get1Value(base + i);",
"    var.evaluateFields(time, new_params, fields);",
"  }"
    };

    private static final int        ROTATE = 1;
    private static final int        TRANSLATE = 2;
    private static final Integer    ROTATE_INT = new Integer(ROTATE);
    private static final Integer    TRANSLATE_INT = new Integer(TRANSLATE);

    private static void Tree_Hash_Protos(TreeNode tree, Hashtable hash)
    {
	if ( tree instanceof VRMLModelNode )
	{
	    Object  value = hash.get(tree.getName());

	    if ( value == null )
		hash.put(tree.getName(), ((VRMLModelNode)tree).getURL());
	    else
		if ( ! ((VRMLModelNode)tree).getURL().equals((String)value) )
		    System.err.println("Different URLs for the same VRML name: "
				     + tree.getName() + " at "
				     + (String)value + " and "
				     + ((VRMLModelNode)tree).getURL());
	}

	for ( int i = 0 ; i < tree.getChildCount() ; i++ )
	    Tree_Hash_Protos(tree.getChild(i), hash);
    }


    private static void Do_Indent(PrintWriter out, int num)
    {
	for ( int i = 0 ; i < num ; i++ )
	    out.print(" ");
    }


    private static double Get_Num(Object var)
    {
	if ( var instanceof Number )
	    return ((Number)var).doubleValue();
	else
	    return 0.0;
    }


    private static boolean Num_Constant(Object var)
    {
	return ( var instanceof Number );
    }


    private static void Rotate_To_VRML(RotateNode tree, PrintWriter out,
				       int indent)
    {
	if (  ! Num_Constant(tree.getX())
	   || ! Num_Constant(tree.getY())
	   || ! Num_Constant(tree.getZ())
	   || ! Num_Constant(tree.getAngle()) )
	{
	    Do_Indent(out, indent);
	    out.println("DEF " + tree.getName() + " Transform {");
	}
	else
	{
	    Do_Indent(out, indent);
	    out.println("Transform {");
	}

	Do_Indent(out, indent + 1);
	out.println("rotation " + Get_Num(tree.getX()) + " "
				+ Get_Num(tree.getY()) + " "
				+ Get_Num(tree.getZ()) + " "
				+ Get_Num(tree.getAngle()));
	Do_Indent(out, indent + 1);
	out.println("children [");

	for ( int i = 0 ; i < tree.getChildCount() ; i++ )
	    Node_To_VRML(tree.getChild(i), out, indent + 2);

	Do_Indent(out, indent + 1);
	out.println("]");
	Do_Indent(out, indent);
	out.println("}");
    }


    private static void Translate_To_VRML(TranslateNode tree, PrintWriter out,
					  int indent)
    {
	if (  ! Num_Constant(tree.getX())
	   || ! Num_Constant(tree.getY())
	   || ! Num_Constant(tree.getZ()) )
	{
	    Do_Indent(out, indent);
	    out.println("DEF " + tree.getName() + " Transform {");
	}
	else
	{
	    Do_Indent(out, indent);
	    out.println("Transform {");
	}

	Do_Indent(out, indent + 1);
	out.println("translation " + Get_Num(tree.getX())
		  + " " + Get_Num(tree.getY()) + " " + Get_Num(tree.getZ()));
	Do_Indent(out, indent + 1);
	out.println("children [");

	for ( int i = 0 ; i < tree.getChildCount() ; i++ )
	    Node_To_VRML(tree.getChild(i), out, indent + 2);

	Do_Indent(out, indent + 1);
	out.println("]");
	Do_Indent(out, indent);
	out.println("}");
    }


    private static void VRML_To_VRML(VRMLModelNode tree, PrintWriter out,
				     int indent)
    {
	Do_Indent(out, indent);
	out.println("Group {");
	Do_Indent(out, indent + 1);
	out.println("children [");

	Do_Indent(out, indent + 2);
	out.println(tree.getName() + " { }");

	for ( int i = 0 ; i < tree.getChildCount() ; i++ )
	    Node_To_VRML(tree.getChild(i), out, indent + 2);

	Do_Indent(out, indent + 1);
	out.println("]");
	Do_Indent(out, indent);
	out.println("}");
    }


    private static void Group_To_VRML(TreeNode tree, PrintWriter out,
				      int indent)
    {
	Do_Indent(out, indent);
	out.println("Group {");
	Do_Indent(out, indent + 1);
	out.println("children [");

	for ( int i = 0 ; i < tree.getChildCount() ; i++ )
	    Node_To_VRML(tree.getChild(i), out, indent + 2);

	Do_Indent(out, indent + 1);
	out.println("]");
	Do_Indent(out, indent);
	out.println("}");
    }


    private static void Bound_To_VRML(BoundNode bound, PrintWriter out,
				      int indent)
    {
	Do_Indent(out, indent);
	out.println("DEF " + bound.getName() + " VisibilitySensor {");
	Do_Indent(out, indent + 1);
	out.println("size " + Get_Num(bound.getX())
		  + " " + Get_Num(bound.getY()) + " " + Get_Num(bound.getZ()));
	Do_Indent(out, indent);
	out.println("}");

	for ( int i = 0 ; i < bound.getChildCount() ; i++ )
	    Node_To_VRML(bound.getChild(i), out, indent);
    }


    private static void Node_To_VRML(TreeNode tree, PrintWriter out, int indent)
    {
	if ( tree instanceof RotateNode )
	    Rotate_To_VRML((RotateNode)tree, out, indent);
	else if ( tree instanceof TranslateNode )
	    Translate_To_VRML((TranslateNode)tree, out, indent);
	else if ( tree instanceof VRMLModelNode )
	    VRML_To_VRML((VRMLModelNode)tree, out, indent);
	else if ( tree instanceof BoundNode )
	    Bound_To_VRML((BoundNode)tree, out, indent);
	else // Must be a group node.
	    Group_To_VRML(tree, out, indent);
    }


    private static void Tree_Visibility_Routes(TreeNode tree, String basename,
					       PrintWriter outfile)
    {
	if ( tree instanceof BoundNode )
	    outfile.println("ROUTE " + ((BoundNode)tree).getName()
			  + ".isActive TO " + basename + "_script."
			  + ((BoundNode)tree).getName());

	for ( int i = 0 ; i < tree.getChildCount() ; i++ )
	    Tree_Visibility_Routes(tree.getChild(i), basename, outfile);
    }


    private static void Tree_Visibility_Eventins(TreeNode tree, String basename,
					         PrintWriter outfile)
    {
	if ( tree instanceof BoundNode )
	    outfile.println("   eventIn SFBool " + ((BoundNode)tree).getName());

	for ( int i = 0 ; i < tree.getChildCount() ; i++ )
	    Tree_Visibility_Eventins(tree.getChild(i), basename, outfile);
    }


    private static void Tree_Visibility_Process(TreeNode tree, String basename,
					         PrintWriter outfile,
						 boolean is_root)
    {
	if ( tree instanceof BoundNode )
	{
	    BoundNode node = (BoundNode)tree;

	    outfile.println("    } else if ( \"" + node.getName()
			  + "\".equals( name ) ) {");
	    for ( int i = 0 ; i < node.getVarCount() ; i++ )
	    {
		VariableDefinition var = node.getVar(i);
		outfile.println("      " + var.getName() + "_visible = "
		    	+ "(boolean)((ConstSFBool)e.getValue()).getValue();");
                outfile.println("      System.out.println(\""
                              + var.getName() + "_visible = \" + "
                              + var.getName() + "_visible);");
	    }
	    if ( is_root )
	        outfile.println(
		     "      ((SFBool)getEventOut(\"visible\")).setValue("
                   + "(boolean)((ConstSFBool)e.getValue()).getValue());");
	}

	for ( int i = 0 ; i < tree.getChildCount() ; i++ )
	    Tree_Visibility_Process(tree.getChild(i), basename, outfile, false);
    }


    private static void Tree_To_VRML(TreeNode tree, String basename,
				     PrintWriter outfile, Hashtable variables,
				     Hashtable fields)
    {
	Enumeration hash_keys;
	String	    field_type;

	outfile.println("#VRML V2.0 utf8");
	outfile.println();

	outfile.println("PROTO " + basename + " [ eventIn SFFloat currentTime");
	outfile.println("    eventOut SFBool visible ] {");
	outfile.println();

	// Need to find all the vrml nodes, so they can be prototyped at the
	// start. Not strictly necessary, but easier to manage and cleaner.
	Hashtable   proto_hash = new Hashtable();
	Tree_Hash_Protos(tree, proto_hash);
	hash_keys = proto_hash.keys();
	while ( hash_keys.hasMoreElements() )
	{
	    String  key = (String)hash_keys.nextElement();
	    outfile.println("EXTERNPROTO " + key + " [] \""
			  + (String)proto_hash.get(key) + "\"");
	}
	outfile.println();

	outfile.println("Group {");
	outfile.println(" children [");

	// Dump the tree.
	Node_To_VRML(tree, outfile, 2);

	// Dump the script node.
	outfile.println("  DEF " + basename + "_script Script {");
	outfile.println("   url \"" + basename + ".class\"");
	outfile.println("   eventIn SFFloat currentTime IS currentTime");
	outfile.println("   eventOut SFBool visible IS visible");
	hash_keys = variables.keys();
	while ( hash_keys.hasMoreElements() )
	{
	    String  key = (String)hash_keys.nextElement();
	    VariableData var_data = (VariableData)variables.get(key);
	    outfile.print("   field MFFloat " + key + "_buffer [");
	    int num_vars = var_data.getType().getInstance().numVariables();
	    int num_buffered = var_data.getType().getInstance().bufferSize()
	                     * num_vars;
	    double[] init_vals = new double[num_vars];
	    var_data.getType().getInstance().initialCondition(init_vals);
	    for ( int i = 0 ; i < num_vars ; i++ )
		outfile.print(" " + init_vals[i]);
	    for ( int i = num_vars ; i < num_buffered ; i++ )
		outfile.print(" 0.0");
	    outfile.println(" ]");
	    outfile.println("   field SFFloat " + key + "_lastTime 0.0");
	    outfile.println("   field SFFloat " + key + "_lastSeen 0.0");
	}
	hash_keys = fields.keys();
	while ( hash_keys.hasMoreElements() )
	{
	    String  key = (String)hash_keys.nextElement();
	    field_type =
		( ((FieldData)fields.get(key)).getType() == ROTATE
		? "SFRotation" : "SFVec3f" );
	    outfile.println("   eventOut " + field_type + " " + key);
	}
	Tree_Visibility_Eventins(tree, basename, outfile);
	outfile.println("  }");

	outfile.println(" ]");
	outfile.println("}");
	outfile.println();

	// Dump the routes.

	// Visibility sensors to script.
	Tree_Visibility_Routes(tree, basename, outfile);

	// Script to everything else.
	hash_keys = fields.keys();
	while ( hash_keys.hasMoreElements() )
	{
	    String  key = (String)hash_keys.nextElement();
	    field_type =
		( ((FieldData)fields.get(key)).getType() == ROTATE
		? ".rotation" : ".translation" );
	    outfile.println("ROUTE " + basename + "_script." + key + " TO "
				     + key + field_type);
	}
	outfile.println("}");
    }


    private static String Evaluate_String(Object field)
    {
	if ( field instanceof Number )
	    return Float.toString(((Number)field).floatValue()) + "f";
	
	VariableField var = (VariableField)field;
	return var.getVar().getName() + "_elements[" + var.getIndex() + "]";
    }


    private static void Model_To_Java(DynamicsModel model, String basename,
				      PrintWriter outfile, Hashtable variables,
				      Hashtable fields)
    {
	Enumeration enum;

	outfile.println("import vrml.*;");
	outfile.println("import vrml.field.*;");
	outfile.println("import vrml.node.*;");
	outfile.println("import berkeley.cs.dmc.system.DynamicsVariable;");
	outfile.println();

	outfile.println("public class " + basename + " extends Script {");

	// Standard variables
	outfile.println("  double time = 0.0;");
	outfile.println("  double last_time = 0.0;");
	outfile.println("  boolean initialized = false;");

	int max_vars = 0;
	enum = variables.keys();
	while ( enum.hasMoreElements() )
	{
	    DynamicsVariable var = ((VariableData)variables.get(
	                           enum.nextElement())).getType().getInstance();
	    if ( var.numVariables() > max_vars )
		max_vars = var.numVariables();
	}
	outfile.println("  double[] old_params = new double["+max_vars+"];");
	outfile.println("  double[] seen_params = new double["+max_vars+"];");
	outfile.println("  double[] new_params = new double["+max_vars+"];");

	enum = variables.keys();
	while ( enum.hasMoreElements() )
	{
	    String name = (String)enum.nextElement();
	    VariableData var = (VariableData)variables.get(name);
	    outfile.println("  private " + var.getType().getName() + " "
	                  + name + " = null;");
	    outfile.println("  private MFFloat " + name + "_buffer = null;");
	    outfile.println("  private SFFloat " + name + "_lastTime = null;");
	    outfile.println("  private SFFloat " + name + "_lastSeen = null;");
	    outfile.println("  private float[] " + name
	                  + "_elements = new float["
			  + var.getType().getInstance().numFields() + "];");
	    if ( var.getBound() )
		outfile.println("  private boolean " + name+"_visible = true;");
	}

	// Dump all the eventout variables.
	enum = fields.keys();
	while ( enum.hasMoreElements() )
	{
	    String  key = (String)enum.nextElement();
	    String field_type =
		( ((FieldData)fields.get(key)).getType() == ROTATE
		? "SFRotation" : "SFVec3f" );
	    outfile.println("  private " + field_type + " "
	                  + key + "_event = null;");
	}

	// Output the initialization routine.
	outfile.println("  public void initialize() {");
	outfile.println("    if ( initialized ) return;");

	// Instantiate all the model variables.
	enum = variables.keys();
	while ( enum.hasMoreElements() )
	{
	    String name = (String)enum.nextElement();
	    VariableData var = (VariableData)variables.get(name);
	    outfile.println("    " + name + " = new "
	                  + var.getType().getName() + "();");
	    outfile.println("    " + name + "_buffer = (MFFloat)getField(\""
	                  + name + "_buffer\");");
	    outfile.println("    " + name + "_lastTime = (SFFloat)getField(\""
	                  + name + "_lastTime\");");
	    outfile.println("    " + name + "_lastSeen = (SFFloat)getField(\""
	                  + name + "_lastSeen\");");
	}

	// Get all the eventOuts
	enum = fields.keys();
	while ( enum.hasMoreElements() )
	{
	    String  key = (String)enum.nextElement();
	    String field_type =
		( ((FieldData)fields.get(key)).getType() == ROTATE
		? "SFRotation" : "SFVec3f" );
	    outfile.println("    " + key + "_event = (" + field_type
	                  + ")getEventOut(\"" + key + "\");");
	}



	// Close the init routine.
	outfile.println("    initialized = true;");
	outfile.println("  }");

	// Dump the event processing routine - processes timer events.
	outfile.println("  public void processEvent( Event e ) {");
	outfile.println("    String name = e.getName();");
	outfile.println("    if ( \"currentTime\".equals( name ) ) {");
	outfile.println("      time=((ConstSFFloat)e.getValue()).getValue();");
	Tree_Visibility_Process(model.getTree(), basename, outfile, true);
	outfile.println("    }");
	outfile.println("  }");

	for ( int i = 0 ; i < timeStep_template.length ; i++ )
	    outfile.println(timeStep_template[i]);

	// Dump the events processed routine - updates variables and sends
	// events.

        outfile.println("  public void eventsProcessed() {");
	outfile.println("    if ( ! initialized ) return;");
	outfile.println("    if ( time <= last_time )");
	outfile.println("        return;");
	outfile.println("    last_time = time;");

	// Update all the variables.
	enum = variables.keys();
	while ( enum.hasMoreElements() )
	{
	    String name = (String)enum.nextElement();
	    VariableData var = (VariableData)variables.get(name);
	    if ( var.getBound() )
	    {
		outfile.println("    if ( " + name + "_visible )");
		outfile.print("  ");
	    }
	    outfile.println("    timeStep(" + name + "_buffer, " + name
	                  + "_lastTime, " + name + "_lastSeen, "
			  + var.getType().getInstance().bufferStep() + "f, "
			  + var.getPhase() + "f, "
			  + name + ", last_time, " + name + "_elements, "
			  + var.getType().getInstance().numVariables() + ", "
			  + var.getType().getInstance().bufferSize() + ");");
	}

	// Set all the event outs.
	enum = fields.keys();
	while ( enum.hasMoreElements() )
	{
	    FieldData field = (FieldData)fields.get(enum.nextElement());
	    Vector    vars = field.getVariables();
	    boolean   need_vis = false;
	    for ( int i = 0 ; i < vars.size() && ! need_vis ; i++ )
		need_vis = ((VariableData)vars.elementAt(i)).getBound();
	    if ( need_vis )
	    {
		VariableData var_data;
		int i;
		for ( i = 0 ; i < vars.size() ; i++ )
		{
		    var_data = (VariableData)vars.elementAt(i);
		    if ( var_data.getBound() )
		    {
			outfile.print("    if ( " + var_data.getName()
			            + "_visible ");
			i++;
			break;
		    }
		}
		for ( ; i < vars.size() ; i++ )
		{
		    var_data = (VariableData)vars.elementAt(i);
		    if ( var_data.getBound() )
			outfile.print("&& " + var_data.getName()
			            + "_visible ");
		}
		outfile.println(")");
		outfile.print("  ");
	    }
	    if ( field.getType() == TRANSLATE )
	    {
		TranslateNode node = (TranslateNode)field.getNode();
		outfile.println("    " + node.getName() + "_event.setValue("
			      + Evaluate_String(node.getX()) + ", "
			      + Evaluate_String(node.getY()) + ", "
			      + Evaluate_String(node.getZ()) + ");");
	    }
	    else
	    {
		RotateNode node = (RotateNode)field.getNode();
		outfile.println("    " + node.getName() + "_event.setValue("
			      + Evaluate_String(node.getX()) + ", "
			      + Evaluate_String(node.getY()) + ", "
			      + Evaluate_String(node.getZ()) + ", "
			      + Evaluate_String(node.getAngle()) + ");");
	    }
	}

        outfile.println("  }");
	outfile.println("}");
	outfile.println();
    }


    private static void Store_Transforms(TreeNode tree, Hashtable variables,
                                         Hashtable fields)
    {
	if ( tree instanceof BoundNode )
	{
	    BoundNode node = (BoundNode)tree;
	    for ( int i = 0 ; i < node.getVarCount() ; i++ )
	    {
		VariableDefinition var = node.getVar(i);
		VariableData var_data =
				    (VariableData)variables.get(var.getName());
		if ( var_data == null )
		{
		    System.err.println("Variable '" + var.getName()
		                     + "' not found in bound '" + node.getName()
				     + "'");
		    System.exit(1);
		}
		var_data.setBound(true);
	    }
	}

	if ( tree instanceof TranslateNode )
	{
	    TranslateNode node = (TranslateNode)tree;
	    if (  ! Num_Constant(node.getX())
	       || ! Num_Constant(node.getY())
	       || ! Num_Constant(node.getZ()) )
	    {
		if ( fields.containsKey(node.getName()) )
		{
		    System.err.println("Node '" + node.getName()
		                     + "' redefined");
		    System.exit(1);
		}
		FieldData field_data =
				new FieldData(node.getName(), TRANSLATE, node);
		Vector all_vars = field_data.getVariables();

		if (  ! Num_Constant(node.getX()) )
		{
		    VariableData var_data =
			(VariableData)variables.get(
			    ((VariableField)node.getX()).getVar().getName());
		    if ( var_data == null )
		    {
			System.err.println("Variable '"
			    + ((VariableField)node.getX()).getVar().getName()
			    + "' not defined");
			System.exit(1);
		    }
		    if ( ! all_vars.contains(var_data) )
			field_data.addVariable(var_data);
		}
		if (  ! Num_Constant(node.getY()) )
		{
		    VariableData var_data =
			(VariableData)variables.get(
			    ((VariableField)node.getY()).getVar().getName());
		    if ( var_data == null )
		    {
			System.err.println("Variable '"
			    + ((VariableField)node.getY()).getVar().getName()
			    + "' not defined");
			System.exit(1);
		    }
		    if ( ! all_vars.contains(var_data) )
			field_data.addVariable(var_data);
		}
		if (  ! Num_Constant(node.getZ()) )
		{
		    VariableData var_data =
			(VariableData)variables.get(
			    ((VariableField)node.getZ()).getVar().getName());
		    if ( var_data == null )
		    {
			System.err.println("Variable '"
			    + ((VariableField)node.getZ()).getVar().getName()
			    + "' not defined");
			System.exit(1);
		    }
		    if ( ! all_vars.contains(var_data) )
			field_data.addVariable(var_data);
		}

		fields.put(node.getName(), field_data);
	    }
	}

	if ( tree instanceof RotateNode )
	{
	    RotateNode node = (RotateNode)tree;
	    if (  ! Num_Constant(node.getX())
	       || ! Num_Constant(node.getY())
	       || ! Num_Constant(node.getZ())
	       || ! Num_Constant(node.getAngle()) )
	    {
		if ( fields.containsKey(node.getName()) )
		{
		    System.err.println("Node '" + node.getName()
		                     + "' redefined");
		    System.exit(1);
		}
		FieldData field_data =
				new FieldData(node.getName(), ROTATE, node);
		Vector all_vars = field_data.getVariables();

		if (  ! Num_Constant(node.getX()) )
		{
		    VariableData var_data =
			(VariableData)variables.get(
			    ((VariableField)node.getX()).getVar().getName());
		    if ( var_data == null )
		    {
			System.err.println("Variable '"
			    + ((VariableField)node.getX()).getVar().getName()
			    + "' not defined");
			System.exit(1);
		    }
		    if ( ! all_vars.contains(var_data) )
			field_data.addVariable(var_data);
		}
		if (  ! Num_Constant(node.getY()) )
		{
		    VariableData var_data =
			(VariableData)variables.get(
			    ((VariableField)node.getY()).getVar().getName());
		    if ( var_data == null )
		    {
			System.err.println("Variable '"
			    + ((VariableField)node.getY()).getVar().getName()
			    + "' not defined");
			System.exit(1);
		    }
		    if ( ! all_vars.contains(var_data) )
			field_data.addVariable(var_data);
		}
		if (  ! Num_Constant(node.getZ()) )
		{
		    VariableData var_data =
			(VariableData)variables.get(
			    ((VariableField)node.getZ()).getVar().getName());
		    if ( var_data == null )
		    {
			System.err.println("Variable '"
			    + ((VariableField)node.getZ()).getVar().getName()
			    + "' not defined");
			System.exit(1);
		    }
		    if ( ! all_vars.contains(var_data) )
			field_data.addVariable(var_data);
		}
		if (  ! Num_Constant(node.getAngle()) )
		{
		    VariableData var_data =
			(VariableData)variables.get(
			   ((VariableField)node.getAngle()).getVar().getName());
		    if ( var_data == null )
		    {
			System.err.println("Variable '"
			  + ((VariableField)node.getAngle()).getVar().getName()
			  + "' not defined");
			System.exit(1);
		    }
		    if ( ! all_vars.contains(var_data) )
			field_data.addVariable(var_data);
		}

		fields.put(node.getName(), field_data);
	    }
	}

	for ( int i = 0 ; i < tree.getChildCount() ; i++ )
	    Store_Transforms(tree.getChild(i), variables, fields);
    }


    private static void Store_Symbols(DynamicsModel model, Hashtable classes,
                                     Hashtable variables, Hashtable fields)
    {
	Enumeration      enum = null;
	
	enum = model.getVars().keys();
	while ( enum.hasMoreElements() ) {
	    VariableDefinition var =
	        (VariableDefinition)model.getVars().get(enum.nextElement());

	    if ( variables.containsKey(var.getName()) )
	    {
		System.err.println("Variable '" + var.getName()
		                 + "' redefined");
		System.exit(1);
	    }

	    ClassData class_data = (ClassData)classes.get(var.getType());
	    if ( class_data == null )
	    {
		class_data = new ClassData(var.getType());
		classes.put(var.getType(), class_data);
	    }

	    VariableData var_data = new VariableData(var.getName(), class_data,
	                                             var.getPhase());
	    variables.put(var.getName(), var_data);
	}

	Store_Transforms(model.getTree(), variables, fields);
    }


    /**
     * Given a model data structure object, create a VRML file
     * describing the geometry and transformations, and a Java
     * file that animates it.
     *
     * @param basename The root file name for the generated files.
     * @param model The object describing the model.
     */
    public static void dump(String basename, DynamicsModel model)
	throws IOException
    {
	PrintWriter	wrlfile = null;
	PrintWriter	javafile = null;
	Hashtable       classes = new Hashtable(100);
	Hashtable       variables = new Hashtable(100);
	Hashtable       fields = new Hashtable(100);

	wrlfile = new PrintWriter(new FileWriter(basename + ".wrl"));
	javafile = new PrintWriter(new FileWriter(basename + ".java"));

	System.out.println("Building symbol table...");
	Store_Symbols(model, classes, variables, fields);
	System.out.println("Done");

	System.out.println("Writing VRML...");
	Tree_To_VRML(model.getTree(), basename, wrlfile, variables, fields);
	System.out.println("Done");

	System.out.println("Writing Java...");
	Model_To_Java(model, basename, javafile, variables, fields);
	System.out.println("Done");

	wrlfile.close();
	javafile.close();
    }


    /**
     * The application main function. This app should be invoked as
     *
     * <PRE>
     * java berkeley.cs.dmc.tovrml.Main <ModelFile>
     * </PRE>
     *
     * ModelFile should be a .dyn file containing a model
     * description. The extension will be stripped and the base file
     * name used as the root for the VRML and Java files produced by the
     * program.
     *
     * See the documentation for a description of the file format.
     *
     * This routine loads the file, creating a model data structure, and
     * calls dump(). See dump for details on the files generated.
     */
    public static void main( String args[] ) {

	if ( args.length < 1 )
	{
	    System.err.println("Expected file base name as argument");
	    System.exit(1);
	}

	try {
	    InputStream	    infile = new FileInputStream(args[0] + ".dyn");
	    DynamicsParser  parser = new DynamicsParser(infile);
	    System.out.println("Parsing...");
	    DynamicsModel   model = parser.parse();
	    System.out.println("Done");
	    dump(args[0], model);
	    infile.close();
	} catch (Exception e) {
	    System.err.println(e);
	    System.exit(1);
	}

	System.exit(0);
    }

}

