/*
 * HTMLParser.java
 * Copyright (C) 1998 Stephen Chenney and Jeffrey Ichnowski
 *
 * This program is free software.  You can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package berkeley.cs.dmc.vrml.parser.spec;

import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.PrintStream;
import java.io.Reader;
import java.util.Stack;
import java.util.StringTokenizer;
import java.util.Vector;

public class HTMLParser {
    private static final int TOKEN_ERROR = -2;
    private static final int TOKEN_EOF = -1;
    private static final int TOKEN_PCDATA = 0;
    private static final int TOKEN_PCWHITE = 8;
    private static final int TOKEN_START_TAG = 1;
    private static final int TOKEN_END_TAG = 2;
    private static final int TOKEN_TAG = 3;
    private static final int TOKEN_TAG_CLOSE = 4;
    private static final int TOKEN_ATTNAME = 5;
    private static final int TOKEN_ATTVAL = 6;
    private static final int TOKEN_EQUALS = 7;

    private static final String TOKEN_NAMES[] = { // add two to token.
	"ERROR", "EOF", "PCDATA", "START_TAG", "END_TAG",
	"TAG", "TAG_CLOSE", "ATTNAME", "ATTVAL", "EQUALS"
    };

    private static final int BUFFER_SIZE = 0x100;

    Reader in;
    char buffer[] = new char[BUFFER_SIZE];
    int  bufread = 0;
    int  buffill = 0;
    int  buflen = 0;
    boolean eof = false;

    String token_string;

    public HTMLParser( Reader in ) {
	this.in = in;
    }

    //
    // Stream handling
    // 

    boolean crlf = false;
    int line = 1;

    boolean peekcharavail = false;
    int peekchar;

    private final int read()  throws IOException {
	int ch;
	if( peekcharavail ) {
	    peekcharavail = false;
	    ch = peekchar;
	} else {
	    ch = in.read();

	    switch(ch) {
	    case '\r':
		if( peek() == '\n' )
		    consume();
		// fall through
	    case '\n':
		line++;
		return '\n';
	    }
	}

	// System.out.println( "consumed: "+ ((ch < 32) ? ("\\"+ch) : (""+(char)ch)) );
	return ch;
    }
    private final void pushback( int ch ) {
	peekchar = ch;
	peekcharavail = true;
	// System.out.println( "pushed: "+((ch < 32) ? ("\\"+ch) : (""+(char)ch)) );
    }
    private final int peek() throws IOException {
	if( !peekcharavail ) {
	    peekchar = in.read();
	    peekcharavail = true;
	}
	return peekchar;
    }
    private final void consume() throws IOException {
	if( peekcharavail )
	    peekcharavail = false;
	else
	    in.read();
    }

    //
    // HTML Helpers
    //
    private static final String HTMLtoPlain( String str ) {
	int s = str.indexOf( '&' );
	if( s == -1 ) return str;
	int e = str.indexOf( ';', s+1 );
	if( e == -1 ) return str;

	String start = str.substring( 0, s );
	String mid = str.substring( s+1, e );
	String end = HTMLtoPlain( str.substring( e+1 ) );

	try {
	    if( mid.charAt(0) == '#' )
		return start + (char)Integer.parseInt( mid.substring(1) ) + end;
	    if( "quot".equals(mid) )
		return start + '"' + end;
	    else if( "lt".equals(mid) )
		return start + '<' + end;
	    else if( "gt".equals(mid) )
		return start + '>' + end;
	    else
		return start + mid + end;
	} catch( Exception ex ) {
	    return str;
	}
    }

    //
    // Tokenizing
    // 

    private final boolean isSpace( char ch ) {
	return Character.isSpaceChar(ch) || ch == '\n';
    }

    int nextToken() throws IOException {
	while(true) {
	    int ch = read();
	    StringBuffer str = null;

	    switch( ch ) {
	    case -1:
		return TOKEN_EOF;
	    case ' ':
	    case '\t':
	    case '\n':
		str = new StringBuffer();
		do {
		    str.append( (char)ch );
		    ch = read();
		} while( (ch == ' ') || (ch == '\t') || (ch == '\n') );
		pushback( ch );
		token_string = HTMLtoPlain( new String(str) );
		return TOKEN_PCWHITE;
	    case '<':
		if( peek() == '/' ) {
		    consume();
		    return TOKEN_END_TAG;
		} else if( peek() == '!' ) {
		    consume();
		    while( read() != '>' );
		} else {
		    return TOKEN_START_TAG;
		}
		break;
	    default:
		str = new StringBuffer();
		do {
		    str.append( (char)ch);
		    ch = read();
		} while( (ch != -1) && (ch != '\t') && (ch != '\n') && (ch != ' ') && (ch != '<') );
		pushback(ch);
		token_string = HTMLtoPlain( new String(str) );
		return TOKEN_PCDATA;
	    }
	}
	/*
	while( true ) {
	    do {
		if( (ch = read()) == -1 )
		    return TOKEN_EOF;
	    } while( isSpace((char)ch) );

	    if( ch == '<' ) {
		if( peek() == '/' ) {
		    consume();
		    return TOKEN_END_TAG;
		} else if( peek() == '!' ) {
		    consume();
		    while( read() != '>' );
		} else {
		    return TOKEN_START_TAG;
		}
	    } else {
		StringBuffer str = new StringBuffer();

		do {
		    str.append((char)ch);
		    ch = read();
		} while( !isSpace((char)ch) && ch != '<' && ch != -1 );
		pushback(ch);

		token_string = new String(str);
		return TOKEN_PCDATA;
	    }
	}
	*/
    }

    int nextTagToken() throws IOException {
	StringBuffer str = new StringBuffer();
	int i = 1;
	int ch = read();

	if( ch == -1 )
	    return TOKEN_EOF;
	if( !Character.isLetter((char)ch) )
	    return TOKEN_ERROR;

	do {
	    str.append((char)ch);
	    ch = read();
	} while( ch != -1 && Character.isLetterOrDigit((char)ch) );
	pushback(ch);

	token_string = HTMLtoPlain( new String(str) );
	return TOKEN_TAG;
    }

    int nextAttListToken() throws IOException {
	int ch;

	while( true ) {
	    do {
		if( (ch = read()) == -1 )
		    return TOKEN_EOF;
	    } while( isSpace((char)ch) );
	    
	    if( ch == '=' ) {
		return TOKEN_EQUALS;
	    } else if( ch == '>' ) {
		return TOKEN_TAG_CLOSE;
	    } else if( ch == '-' && peek() == '-' ) { // comment
		consume();
		do {
		    do {
			if( (ch = read()) == -1 )
			    return TOKEN_EOF;
		    } while( ch != '-' );
		    
		    if( (ch = read()) == -1 )
			return TOKEN_EOF;
		    
		} while( ch != '-' );
	    } else if( Character.isLetter((char)ch) || ch == '_' || ch == '-' || ch == '.' ) {
		StringBuffer str = new StringBuffer();
		int i = 0;
		do {
		    str.append((char)ch);
		    ch = read();
		} while( ch != -1 && (Character.isLetterOrDigit((char)ch) ||
				      ch == '_' || ch == '-' || ch == ',') );
		pushback(ch);
		token_string = HTMLtoPlain( new String(str) );
		return TOKEN_ATTNAME;
	    } else {
		return TOKEN_ERROR;
	    }
	}
    }

    int nextAttValToken() throws IOException {
	int ch;
	do {
	    if( (ch = read()) == -1 )
		return TOKEN_EOF;
	} while( isSpace((char)ch) );

	StringBuffer str = new StringBuffer();

	if( ch == '\'' || ch == '\"' ) {
	    int delim = ch;
	    ch = read();
	    while( ch != delim && ch != -1 ) {
		str.append((char)ch);
		ch = read();
	    }
	} else {
	    do {
		str.append((char)ch);
		ch = read();
	    } while( !isSpace((char)ch) && ch != '>' && ch != -1 );
	    pushback(ch);
	}

	token_string = HTMLtoPlain( new String(str) );
	return TOKEN_ATTVAL;
    }
    
    //
    // Error recovery
    //
    
    void closeAttribute() throws IOException {
	int ch;
	do {
	    ch = read();
	} while( ch != '>' && ch != -1 );
    }

    //
    // Parsing
    //


    /*
    Stack stack = new Stack();
    StringBuffer pcdata = null;

    void prt( String str ) {
	int i = stack.size();
	while( i-- != 0 )
	    System.out.print( " ");
	System.out.println(str);
    }

    void openTag( String name ) {
	if( stack.size() > 0 && name.equals(stack.peek()) )
	    closeTag( name );
	else if( pcdata != null ) {
	    prt( new String(pcdata) );
	    pcdata = null;
	}

	prt( "<"+name+">" );
	stack.push( name );
	pcdata = null;
    }

    void attribute( String name, String value ) {
	prt( name+"="+value );
    }

    void closeTag( String name ) {
	if( pcdata != null ) {
	    prt( new String(pcdata) );
	    pcdata = null;
	}

	if( stack.search(name) != -1 ) {
	    while( !stack.pop().equals(name) );
	}

	prt( "</"+name+">" );
    }

    void pcdata( String str ) {
	if( pcdata == null )
	    pcdata = new StringBuffer( str );
	else
	    pcdata.append( ' ' ).append( str );
    }
    */


    private final void state( int s ) {
	// System.out.println( "State "+state+" -> "+s );
	state = s;
    }

    String nodename;
    int state = 0;
    String current = "";
    Stack stack = new Stack();
    void openTag( String name ) {
	if( stack.size() > 0 && name.equals(stack.peek()) )
	    closeTag( name );

	stack.push( current = name );

	switch( state ) {
	case 0:
	    if( "a".equals( current ) ) {
		nodename = "";
		state( 1 );
	    }
	    break;
	case 3:
	    state( "pre".equals( current ) ? 4 : 0 );
	    break;
	case 4:
	    state( "b".equals( current ) ? 5 : 0 );
	    break;
	}
    }
    void attribute( String name, String value ) {
	switch( state ) {
	case 1:
	    if( "name".equals(name) ) {
		nodename = value;
		state( 2 );
	    }
	    break;
	}
    }
    void closeTag( String name ) {
	if( stack.search(name) != -1 )
	    while( !stack.pop().equals(name) );

	current = (stack.size() > 0) ? (String)stack.peek() : "";

	switch( state ) {
	case 2:
	    state( "a".equals(name) ? 3 : 0 );
	    break;
	}
    }

    String fieldCat; // eventIn, exposedField, ...
    String fieldType; // MFNode, SFString, ...
    String fieldName;
    String fieldDefault;

    /*
    Vector fieldCats = new Vector();
    Vector fieldTypes = new Vector();
    Vector fieldNames = new Vector();
    Vector fieldDefaults = new Vector();
    */
    /*
    void writeField( String cat, String type, String name, String def ) {
	String out;

	// if( astfile == null ) return;
	// try {

	if( cat.endsWith( "ield" ) ) { // ie. field and exposedField
	    String fd;

	    if( def == null ||
		"\"\"".equals(def) ||
		"NULL".equals(def) )
		fd = "";
	    else if( def.equals( "FALSE" ) )
		fd = "false";
	    else if( def.equals( "TRUE" ) )
		fd = "true";
	    else {
		StringTokenizer st = new StringTokenizer( def, " ,\t\n[]" );
		StringBuffer tmp = new StringBuffer();
		if( st.hasMoreTokens() ) {
		    String tok = st.nextToken();
		    if( tok.indexOf('.') != -1 && !type.endsWith( "Time" ) )
			tok += "f";
		    
		    tmp.append(tok);
		    
		    while( st.hasMoreTokens() ) {
			tok = st.nextToken();
			if( tok.indexOf('.') != -1 && !type.endsWith( "Time" ) )
			    tok += "f";
			
			tmp.append( ", " ).append( tok );
		    }
		}
		fd = new String(tmp);
	    }

	    astfile.println( "    public AST"+type+" "+name+";" ); // "= new AST"+type+"(" );
	    nodefile.print("    public "+type+" "+name+" = new "+type+"(" );

	    if( type.charAt(0) != 'M' ) {
		if( !"SFImage".equals(type) ) {
		    // astfile.print( fd );
		    nodefile.print( fd );
		}

		out = "); // "+cat;
		// astfile.println( out ); 
		nodefile.println( out );
		
	    } else { // MField
		out = "); // "+cat;
		// astfile.println( out );
		nodefile.println( out );

		if( fd.length() > 0 ) {
		    int len;
		    if( "MFRotation".equals(type) )
			len = 4;
		    else if( "MFVec3f".equals(type) ||
			"MFColor".equals(type) )
			len = 3;
		    else if( "MFVec2f".equals(type) )
			len = 2;
		    else
			len = 1;

		    StringTokenizer st = new StringTokenizer( fd, " ," );
		    if( st.hasMoreTokens() ) {
			out = "    {";
			// astfile.println( out );
			nodefile.println( out );
			do {
			    out = "        "+name+".addValue(";
			    // astfile.print( out );
			    nodefile.print( out );

			    out = st.nextToken();
			    // astfile.print( out );
			    nodefile.print( out );
			    for( int c=len ; --c>0 && st.hasMoreTokens() ; ) {
				out = ", "+st.nextToken();
				// astfile.print( out );
				nodefile.print( out );
			    }
			    out = ");";
			    // astfile.println( out );
			    nodefile.println( out );
			} while( st.hasMoreTokens() );
			out = "    }";
			// astfile.println( out);
			nodefile.println( out );
		    }
		}
	    }
	} else { // not a field or exposed field
	    astfile.println( out="    // "+cat+" "+type+" "+name+" "+def );
	    nodefile.println( out );
	}
	
	    
	// } catch( IOException e ) {
	// System.out.println( "WARNING! Error writing "+nodename+".java" );
	// }
    }
    */
    void pcwhite( String str ) {
	switch( state ) {
	case 10:
	    if( str.indexOf('\n') != -1 ) {
		// System.out.println("  "+fieldCat+" "+fieldType+" "+fieldName );
		out.println( "\t"+fieldCat+" "+fieldType+" "+fieldName );
		state( 7 );

		// fieldCats.addElement( fieldCat );
		// fieldNames.addElement( fieldName );
		// fieldTypes.addElement( fieldType );
		// fieldDefaults.addElement("");
		// writeField();
	    }
	    break;
	case 11:
	    if( str.indexOf('\n') != -1 ) {
		int i = fieldDefault.indexOf( '#' );
		if( i != -1 )
		    fieldDefault = fieldDefault.substring(0,i);
		fieldDefault = fieldDefault.trim();
		// System.out.println("  "+fieldCat+" "+fieldType+" "+fieldName+" <"+fieldDefault+">" );
		out.println( "\t"+fieldCat+" "+fieldType+" "+fieldName+" "+fieldDefault );
		state( 7 );

		// fieldCats.addElement( fieldCat );
		// fieldNames.addElement( fieldName );
		// fieldTypes.addElement( fieldType );
		// fieldDefaults.addElement( fieldDefault );
		// writeField();
	    } else
		fieldDefault += str;
	    break;
	}
    }

    // PrintWriter astfile;
    // PrintWriter nodefile;
    PrintStream out = System.out;

    void pcdata( String str ) {
	switch( state ) {
	case 5:
	    state( str.equals( nodename ) ? 6 : 0 );
	    break;
	case 6:
	    state( "{".equals( str ) && !"Script".equals( nodename ) ? 7 : 0 );
	    if( state == 7 ) {
		/*
		astfile = null;
		fieldNames.setSize(0);
		fieldTypes.setSize(0);
		fieldCats.setSize(0);
		fieldDefaults.setSize(0);
		try {
		    astfile = new PrintWriter( new FileWriter( "../ast/node/AST"+nodename+".java" ) );
		    astfile.println( "package berkeley.cs.dmc.vrml.parser.ast.node;" );
		    astfile.println();
		    astfile.println( "import berkeley.cs.dmc.vrml.parser.ast.ASTNode;" );
		    astfile.println( "import berkeley.cs.dmc.vrml.parser.ast.field.*;" );
		    astfile.println( "import berkeley.cs.dmc.vrml.parser.*;" );
		    astfile.println();
		    astfile.println( "public class AST"+nodename+" extends ASTNode {" );

		    nodefile = new PrintWriter( new FileWriter( "../node/"+nodename+".java" ) );
		    nodefile.println( "package berkeley.cs.dmc.vrml.parser.node;" );
		    nodefile.println();
		    nodefile.println( "import berkeley.cs.dmc.vrml.node.Node;" );
		    nodefile.println( "import berkeley.cs.dmc.vrml.field.*;" );
		    nodefile.println();
		    nodefile.println( "public class "+nodename+" extends Node {" );

		} catch( IOException e ) {
		    System.out.println( "IO Error: "+e );
		}
		System.out.println( nodename + " {" );
		*/

		out.println( nodename+" {" );
	    }
	    break;
	case 7:
	    if( "}".equals( str ) ) {
		/*
		if( astfile != null ) {
		    try {
			int i;

			for( i=0 ; i<fieldNames.size() ; i++ )
			    writeField( (String)fieldCats.elementAt(i),
					(String)fieldTypes.elementAt(i),
					(String)fieldNames.elementAt(i),
					(String)fieldDefaults.elementAt(i) );
					
			nodefile.println();
			nodefile.println( "    public String getType() { return \""+nodename+"\"; }" );
			if( fieldNames.size() > 0 ) {
			    astfile.println();
			    astfile.println( "    public void parseField( VRMLParser parser, Token field )" );
			    astfile.println( "    throws ParseException {" );
			    astfile.println( "        String name=field.image;" );
			    astfile.println();
			    astfile.print( "       " );
			    for( i=0 ; i<fieldNames.size() ; i++ ) {
				String cat = (String)fieldCats.elementAt(i);
				if( cat.endsWith("ield") ) { // i.e. field or exposedField
				    String name = (String)fieldNames.elementAt(i);
				    String type = (String)fieldTypes.elementAt(i);
				    astfile.println( " if( \""+name+"\".equals(name) )" );
				    astfile.println( "            "+name+" = parser."+type.toLowerCase()+"();" );
				    astfile.print( "        else" );
				}
			    }
			    astfile.println();
			    astfile.println( "            super.parseField( parser, field );" );
			    astfile.println( "    }" );
			}

			astfile.println();
			astfile.println( "    public void dump( String indent, java.io.PrintWriter out ) {" );
			astfile.println( "        String indent2=indent+INDENT_LEVEL;" );
			astfile.println();
			astfile.println( "        out.println( indent+\""+nodename+" {\" );" );
			for( i=0 ; i<fieldNames.size() ; i++ ) {
			    String cat = (String)fieldCats.elementAt(i);
			    if( cat.endsWith("ield") ) {
				String name = (String)fieldNames.elementAt(i);
				String type = (String)fieldTypes.elementAt(i);
				astfile.println( "        if( "+name+" != null ) {" );
				astfile.println( "            out.println( indent2+\""+name+"\" );" );
				astfile.println( "            "+name+".dump( indent2, out );" );
				astfile.println( "        }" );
			    }
			}
			astfile.println( "        out.println( indent+\"}\" );" );
			astfile.println( "    }" );
			

			astfile.println( "}" );

			nodefile.println( "}" );
			// } catch( IOException e ) {
			// System.out.println( "File IO Exception " + e.getMessage() );
		    } finally {
		        astfile.close();
			nodefile.close();
		    }
		}
		*/
		out.println("}");
		// System.out.println( "}" );
		state( 0 );
	    } else {
		fieldCat = str;
		state( 8 );
	    }
	    break;
	case 8:
	    fieldType = str;
	    state( 9 );
	    break;
	case 9:
	    fieldName = str;
	    state( 10 );
	    break;
	case 10:
	    fieldDefault = str;
	    state( 11 );
	    break;
	case 11:
	    fieldDefault += str;
	    break;
	}
    }

    public synchronized void parse() throws IOException {
	int token = -3;

    loop:
	while( true ) {
	    token = nextToken();
	    // System.out.println( "TOKEN: " + TOKEN_NAMES[token+2] );
	    switch( token ) {
	    case TOKEN_EOF: break loop;
	    case TOKEN_PCDATA:
		pcdata( token_string );
		break;
	    case TOKEN_PCWHITE:
		pcwhite( token_string );
		break;
	    case TOKEN_START_TAG:
		parseAttribute();
		break;
	    case TOKEN_END_TAG:
		parseEndTag();
		break;
	    default:
		throw new InternalError();
	    }
	}
    }

    void parseAttribute() throws IOException {
	int token;

	token = nextTagToken();
	// System.out.println( "TOKEN: " + TOKEN_NAMES[token+2] );

	switch( token ) {
	case TOKEN_EOF: return;
	case TOKEN_ERROR:
	    System.err.println( "error at line "+line+", expected tag name" );
	    closeAttribute();
	    return;
	case TOKEN_TAG:
	    openTag( token_string.toLowerCase() );
	    break;
	default:
	    throw new InternalError();
	}

        loop:
	while(true) {
	    String name = "";
	    token = nextAttListToken();
	    // System.out.println( "TOKEN: " + TOKEN_NAMES[token+2] );

	    switch( token ) {
	    case TOKEN_EOF: return;
	    case TOKEN_ERROR:
		System.err.println( "error at line "+line+", expected attribute name" );
		closeAttribute();
		return;
	    case TOKEN_TAG_CLOSE:
		return;
	    case TOKEN_EQUALS:
		System.err.println( "'=' w/o attribute" );
		closeAttribute();
		break;
	    case TOKEN_ATTNAME:
		name = token_string;
		break;
	    default:
		throw new InternalError();
	    }

	    token = nextAttListToken();
  	    // System.out.println( "TOKEN: " + TOKEN_NAMES[token+2] );

	    switch( token ) {
	    case TOKEN_EOF: return;
	    case TOKEN_ERROR:
	    case TOKEN_ATTNAME:
		System.err.println( "error at line "+line+", expected '='" );
		closeAttribute();
		return;
	    case TOKEN_TAG_CLOSE:
		System.err.println( "error at line "+line+", attribute w/o value" );
		closeAttribute();
		return;
	    case TOKEN_EQUALS: break;
	    default:
		throw new InternalError();
	    }


	token = nextAttValToken();
	// System.out.println( "TOKEN: " + TOKEN_NAMES[token+2] );

	switch( token ) {
	    case TOKEN_EOF: return;
	    case TOKEN_ATTVAL:
		attribute( name.toLowerCase(), token_string );
		break;
	    default:
		throw new InternalError();
	    }
	}
    }

    void parseEndTag() throws IOException {
	int token;

	token = nextTagToken();
	// System.out.println( "TOKEN: " + TOKEN_NAMES[token+2] );

	switch( token ) {
	case TOKEN_EOF: return;
	case TOKEN_ERROR:
	    System.err.println( "error at line "+line+", expected close tag name" );
	    closeAttribute();
	    return;
	case TOKEN_TAG:
	    // System.out.println( "</"+token_string+">" );
	    closeTag( token_string.toLowerCase() );
	    closeAttribute();
	    break;
	default:
	    throw new InternalError();
	}
    }

    public static void main( String args[] ) {
	System.err.println( "Reading from STDIN..." );
	HTMLParser parser = new HTMLParser( new java.io.InputStreamReader( System.in ) );
	try {
	    parser.parse();
	} catch( IOException e ) {
	    e.printStackTrace();
	}
    }
}
